# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php//"
# 1 "<built-in>"
# 1 "<command-line>"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
# 21 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 1
# 33 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_version.h" 1
# 34 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 1
# 51 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_config.h" 1
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 1
# 2423 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h"
# 1 "/usr/include/stdlib.h" 1 3 4
# 25 "/usr/include/stdlib.h" 3 4
# 1 "/usr/include/features.h" 1 3 4
# 361 "/usr/include/features.h" 3 4
# 1 "/usr/include/sys/cdefs.h" 1 3 4
# 365 "/usr/include/sys/cdefs.h" 3 4
# 1 "/usr/include/bits/wordsize.h" 1 3 4
# 366 "/usr/include/sys/cdefs.h" 2 3 4
# 362 "/usr/include/features.h" 2 3 4
# 385 "/usr/include/features.h" 3 4
# 1 "/usr/include/gnu/stubs.h" 1 3 4



# 1 "/usr/include/bits/wordsize.h" 1 3 4
# 5 "/usr/include/gnu/stubs.h" 2 3 4


# 1 "/usr/include/gnu/stubs-32.h" 1 3 4
# 8 "/usr/include/gnu/stubs.h" 2 3 4
# 386 "/usr/include/features.h" 2 3 4
# 26 "/usr/include/stdlib.h" 2 3 4







# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 211 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 3 4
typedef unsigned int size_t;
# 323 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 3 4
typedef long int wchar_t;
# 34 "/usr/include/stdlib.h" 2 3 4








# 1 "/usr/include/bits/waitflags.h" 1 3 4
# 43 "/usr/include/stdlib.h" 2 3 4
# 1 "/usr/include/bits/waitstatus.h" 1 3 4
# 65 "/usr/include/bits/waitstatus.h" 3 4
# 1 "/usr/include/endian.h" 1 3 4
# 37 "/usr/include/endian.h" 3 4
# 1 "/usr/include/bits/endian.h" 1 3 4
# 38 "/usr/include/endian.h" 2 3 4
# 61 "/usr/include/endian.h" 3 4
# 1 "/usr/include/bits/byteswap.h" 1 3 4
# 62 "/usr/include/endian.h" 2 3 4
# 66 "/usr/include/bits/waitstatus.h" 2 3 4

union wait
  {
    int w_status;
    struct
      {

 unsigned int __w_termsig:7;
 unsigned int __w_coredump:1;
 unsigned int __w_retcode:8;
 unsigned int:16;







      } __wait_terminated;
    struct
      {

 unsigned int __w_stopval:8;
 unsigned int __w_stopsig:8;
 unsigned int:16;






      } __wait_stopped;
  };
# 44 "/usr/include/stdlib.h" 2 3 4
# 68 "/usr/include/stdlib.h" 3 4
typedef union
  {
    union wait *__uptr;
    int *__iptr;
  } __WAIT_STATUS __attribute__ ((__transparent_union__));
# 96 "/usr/include/stdlib.h" 3 4


typedef struct
  {
    int quot;
    int rem;
  } div_t;



typedef struct
  {
    long int quot;
    long int rem;
  } ldiv_t;







__extension__ typedef struct
  {
    long long int quot;
    long long int rem;
  } lldiv_t;


# 140 "/usr/include/stdlib.h" 3 4
extern size_t __ctype_get_mb_cur_max (void) __attribute__ ((__nothrow__)) ;




extern double atof (__const char *__nptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1))) ;

extern int atoi (__const char *__nptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1))) ;

extern long int atol (__const char *__nptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1))) ;





__extension__ extern long long int atoll (__const char *__nptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1))) ;





extern double strtod (__const char *__restrict __nptr,
        char **__restrict __endptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





extern float strtof (__const char *__restrict __nptr,
       char **__restrict __endptr) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern long double strtold (__const char *__restrict __nptr,
       char **__restrict __endptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





extern long int strtol (__const char *__restrict __nptr,
   char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern unsigned long int strtoul (__const char *__restrict __nptr,
      char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;




__extension__
extern long long int strtoq (__const char *__restrict __nptr,
        char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

__extension__
extern unsigned long long int strtouq (__const char *__restrict __nptr,
           char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





__extension__
extern long long int strtoll (__const char *__restrict __nptr,
         char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

__extension__
extern unsigned long long int strtoull (__const char *__restrict __nptr,
     char **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

# 236 "/usr/include/stdlib.h" 3 4
# 1 "/usr/include/xlocale.h" 1 3 4
# 28 "/usr/include/xlocale.h" 3 4
typedef struct __locale_struct
{

  struct __locale_data *__locales[13];


  const unsigned short int *__ctype_b;
  const int *__ctype_tolower;
  const int *__ctype_toupper;


  const char *__names[13];
} *__locale_t;


typedef __locale_t locale_t;
# 237 "/usr/include/stdlib.h" 2 3 4



extern long int strtol_l (__const char *__restrict __nptr,
     char **__restrict __endptr, int __base,
     __locale_t __loc) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 4))) ;

extern unsigned long int strtoul_l (__const char *__restrict __nptr,
        char **__restrict __endptr,
        int __base, __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 4))) ;

__extension__
extern long long int strtoll_l (__const char *__restrict __nptr,
    char **__restrict __endptr, int __base,
    __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 4))) ;

__extension__
extern unsigned long long int strtoull_l (__const char *__restrict __nptr,
       char **__restrict __endptr,
       int __base, __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 4))) ;

extern double strtod_l (__const char *__restrict __nptr,
   char **__restrict __endptr, __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 3))) ;

extern float strtof_l (__const char *__restrict __nptr,
         char **__restrict __endptr, __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 3))) ;

extern long double strtold_l (__const char *__restrict __nptr,
         char **__restrict __endptr,
         __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 3))) ;





extern __inline double
__attribute__ ((__nothrow__)) atof (__const char *__nptr)
{
  return strtod (__nptr, (char **) ((void *)0));
}
extern __inline int
__attribute__ ((__nothrow__)) atoi (__const char *__nptr)
{
  return (int) strtol (__nptr, (char **) ((void *)0), 10);
}
extern __inline long int
__attribute__ ((__nothrow__)) atol (__const char *__nptr)
{
  return strtol (__nptr, (char **) ((void *)0), 10);
}




__extension__ extern __inline long long int
__attribute__ ((__nothrow__)) atoll (__const char *__nptr)
{
  return strtoll (__nptr, (char **) ((void *)0), 10);
}

# 311 "/usr/include/stdlib.h" 3 4
extern char *l64a (long int __n) __attribute__ ((__nothrow__)) ;


extern long int a64l (__const char *__s)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1))) ;




# 1 "/usr/include/sys/types.h" 1 3 4
# 28 "/usr/include/sys/types.h" 3 4


# 1 "/usr/include/bits/types.h" 1 3 4
# 28 "/usr/include/bits/types.h" 3 4
# 1 "/usr/include/bits/wordsize.h" 1 3 4
# 29 "/usr/include/bits/types.h" 2 3 4


typedef unsigned char __u_char;
typedef unsigned short int __u_short;
typedef unsigned int __u_int;
typedef unsigned long int __u_long;


typedef signed char __int8_t;
typedef unsigned char __uint8_t;
typedef signed short int __int16_t;
typedef unsigned short int __uint16_t;
typedef signed int __int32_t;
typedef unsigned int __uint32_t;




__extension__ typedef signed long long int __int64_t;
__extension__ typedef unsigned long long int __uint64_t;







__extension__ typedef long long int __quad_t;
__extension__ typedef unsigned long long int __u_quad_t;
# 131 "/usr/include/bits/types.h" 3 4
# 1 "/usr/include/bits/typesizes.h" 1 3 4
# 132 "/usr/include/bits/types.h" 2 3 4


__extension__ typedef __u_quad_t __dev_t;
__extension__ typedef unsigned int __uid_t;
__extension__ typedef unsigned int __gid_t;
__extension__ typedef unsigned long int __ino_t;
__extension__ typedef __u_quad_t __ino64_t;
__extension__ typedef unsigned int __mode_t;
__extension__ typedef unsigned int __nlink_t;
__extension__ typedef long int __off_t;
__extension__ typedef __quad_t __off64_t;
__extension__ typedef int __pid_t;
__extension__ typedef struct { int __val[2]; } __fsid_t;
__extension__ typedef long int __clock_t;
__extension__ typedef unsigned long int __rlim_t;
__extension__ typedef __u_quad_t __rlim64_t;
__extension__ typedef unsigned int __id_t;
__extension__ typedef long int __time_t;
__extension__ typedef unsigned int __useconds_t;
__extension__ typedef long int __suseconds_t;

__extension__ typedef int __daddr_t;
__extension__ typedef long int __swblk_t;
__extension__ typedef int __key_t;


__extension__ typedef int __clockid_t;


__extension__ typedef void * __timer_t;


__extension__ typedef long int __blksize_t;




__extension__ typedef long int __blkcnt_t;
__extension__ typedef __quad_t __blkcnt64_t;


__extension__ typedef unsigned long int __fsblkcnt_t;
__extension__ typedef __u_quad_t __fsblkcnt64_t;


__extension__ typedef unsigned long int __fsfilcnt_t;
__extension__ typedef __u_quad_t __fsfilcnt64_t;

__extension__ typedef int __ssize_t;



typedef __off64_t __loff_t;
typedef __quad_t *__qaddr_t;
typedef char *__caddr_t;


__extension__ typedef int __intptr_t;


__extension__ typedef unsigned int __socklen_t;
# 31 "/usr/include/sys/types.h" 2 3 4



typedef __u_char u_char;
typedef __u_short u_short;
typedef __u_int u_int;
typedef __u_long u_long;
typedef __quad_t quad_t;
typedef __u_quad_t u_quad_t;
typedef __fsid_t fsid_t;




typedef __loff_t loff_t;



typedef __ino_t ino_t;






typedef __ino64_t ino64_t;




typedef __dev_t dev_t;




typedef __gid_t gid_t;




typedef __mode_t mode_t;




typedef __nlink_t nlink_t;




typedef __uid_t uid_t;





typedef __off_t off_t;






typedef __off64_t off64_t;




typedef __pid_t pid_t;





typedef __id_t id_t;




typedef __ssize_t ssize_t;





typedef __daddr_t daddr_t;
typedef __caddr_t caddr_t;





typedef __key_t key_t;
# 133 "/usr/include/sys/types.h" 3 4
# 1 "/usr/include/time.h" 1 3 4
# 58 "/usr/include/time.h" 3 4


typedef __clock_t clock_t;



# 74 "/usr/include/time.h" 3 4


typedef __time_t time_t;



# 92 "/usr/include/time.h" 3 4
typedef __clockid_t clockid_t;
# 104 "/usr/include/time.h" 3 4
typedef __timer_t timer_t;
# 134 "/usr/include/sys/types.h" 2 3 4



typedef __useconds_t useconds_t;



typedef __suseconds_t suseconds_t;





# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 148 "/usr/include/sys/types.h" 2 3 4



typedef unsigned long int ulong;
typedef unsigned short int ushort;
typedef unsigned int uint;
# 195 "/usr/include/sys/types.h" 3 4
typedef int int8_t __attribute__ ((__mode__ (__QI__)));
typedef int int16_t __attribute__ ((__mode__ (__HI__)));
typedef int int32_t __attribute__ ((__mode__ (__SI__)));
typedef int int64_t __attribute__ ((__mode__ (__DI__)));


typedef unsigned int u_int8_t __attribute__ ((__mode__ (__QI__)));
typedef unsigned int u_int16_t __attribute__ ((__mode__ (__HI__)));
typedef unsigned int u_int32_t __attribute__ ((__mode__ (__SI__)));
typedef unsigned int u_int64_t __attribute__ ((__mode__ (__DI__)));

typedef int register_t __attribute__ ((__mode__ (__word__)));
# 220 "/usr/include/sys/types.h" 3 4
# 1 "/usr/include/sys/select.h" 1 3 4
# 31 "/usr/include/sys/select.h" 3 4
# 1 "/usr/include/bits/select.h" 1 3 4
# 32 "/usr/include/sys/select.h" 2 3 4


# 1 "/usr/include/bits/sigset.h" 1 3 4
# 24 "/usr/include/bits/sigset.h" 3 4
typedef int __sig_atomic_t;




typedef struct
  {
    unsigned long int __val[(1024 / (8 * sizeof (unsigned long int)))];
  } __sigset_t;
# 35 "/usr/include/sys/select.h" 2 3 4



typedef __sigset_t sigset_t;





# 1 "/usr/include/time.h" 1 3 4
# 120 "/usr/include/time.h" 3 4
struct timespec
  {
    __time_t tv_sec;
    long int tv_nsec;
  };
# 45 "/usr/include/sys/select.h" 2 3 4

# 1 "/usr/include/bits/time.h" 1 3 4
# 75 "/usr/include/bits/time.h" 3 4
struct timeval
  {
    __time_t tv_sec;
    __suseconds_t tv_usec;
  };
# 47 "/usr/include/sys/select.h" 2 3 4
# 55 "/usr/include/sys/select.h" 3 4
typedef long int __fd_mask;
# 67 "/usr/include/sys/select.h" 3 4
typedef struct
  {



    __fd_mask fds_bits[1024 / (8 * (int) sizeof (__fd_mask))];





  } fd_set;






typedef __fd_mask fd_mask;
# 99 "/usr/include/sys/select.h" 3 4

# 109 "/usr/include/sys/select.h" 3 4
extern int select (int __nfds, fd_set *__restrict __readfds,
     fd_set *__restrict __writefds,
     fd_set *__restrict __exceptfds,
     struct timeval *__restrict __timeout);
# 121 "/usr/include/sys/select.h" 3 4
extern int pselect (int __nfds, fd_set *__restrict __readfds,
      fd_set *__restrict __writefds,
      fd_set *__restrict __exceptfds,
      const struct timespec *__restrict __timeout,
      const __sigset_t *__restrict __sigmask);



# 221 "/usr/include/sys/types.h" 2 3 4


# 1 "/usr/include/sys/sysmacros.h" 1 3 4
# 30 "/usr/include/sys/sysmacros.h" 3 4
__extension__
extern unsigned int gnu_dev_major (unsigned long long int __dev)
     __attribute__ ((__nothrow__));
__extension__
extern unsigned int gnu_dev_minor (unsigned long long int __dev)
     __attribute__ ((__nothrow__));
__extension__
extern unsigned long long int gnu_dev_makedev (unsigned int __major,
            unsigned int __minor)
     __attribute__ ((__nothrow__));


__extension__ extern __inline unsigned int
__attribute__ ((__nothrow__)) gnu_dev_major (unsigned long long int __dev)
{
  return ((__dev >> 8) & 0xfff) | ((unsigned int) (__dev >> 32) & ~0xfff);
}

__extension__ extern __inline unsigned int
__attribute__ ((__nothrow__)) gnu_dev_minor (unsigned long long int __dev)
{
  return (__dev & 0xff) | ((unsigned int) (__dev >> 12) & ~0xff);
}

__extension__ extern __inline unsigned long long int
__attribute__ ((__nothrow__)) gnu_dev_makedev (unsigned int __major, unsigned int __minor)
{
  return ((__minor & 0xff) | ((__major & 0xfff) << 8)
   | (((unsigned long long int) (__minor & ~0xff)) << 12)
   | (((unsigned long long int) (__major & ~0xfff)) << 32));
}
# 224 "/usr/include/sys/types.h" 2 3 4





typedef __blksize_t blksize_t;






typedef __blkcnt_t blkcnt_t;



typedef __fsblkcnt_t fsblkcnt_t;



typedef __fsfilcnt_t fsfilcnt_t;
# 263 "/usr/include/sys/types.h" 3 4
typedef __blkcnt64_t blkcnt64_t;
typedef __fsblkcnt64_t fsblkcnt64_t;
typedef __fsfilcnt64_t fsfilcnt64_t;





# 1 "/usr/include/bits/pthreadtypes.h" 1 3 4
# 36 "/usr/include/bits/pthreadtypes.h" 3 4
typedef unsigned long int pthread_t;


typedef union
{
  char __size[36];
  long int __align;
} pthread_attr_t;


typedef struct __pthread_internal_slist
{
  struct __pthread_internal_slist *__next;
} __pthread_slist_t;




typedef union
{
  struct __pthread_mutex_s
  {
    int __lock;
    unsigned int __count;
    int __owner;


    int __kind;
    unsigned int __nusers;
    __extension__ union
    {
      int __spins;
      __pthread_slist_t __list;
    };
  } __data;
  char __size[24];
  long int __align;
} pthread_mutex_t;

typedef union
{
  char __size[4];
  long int __align;
} pthread_mutexattr_t;




typedef union
{
  struct
  {
    int __lock;
    unsigned int __futex;
    __extension__ unsigned long long int __total_seq;
    __extension__ unsigned long long int __wakeup_seq;
    __extension__ unsigned long long int __woken_seq;
    void *__mutex;
    unsigned int __nwaiters;
    unsigned int __broadcast_seq;
  } __data;
  char __size[48];
  __extension__ long long int __align;
} pthread_cond_t;

typedef union
{
  char __size[4];
  long int __align;
} pthread_condattr_t;



typedef unsigned int pthread_key_t;



typedef int pthread_once_t;





typedef union
{
  struct
  {
    int __lock;
    unsigned int __nr_readers;
    unsigned int __readers_wakeup;
    unsigned int __writer_wakeup;
    unsigned int __nr_readers_queued;
    unsigned int __nr_writers_queued;


    unsigned char __flags;
    unsigned char __shared;
    unsigned char __pad1;
    unsigned char __pad2;
    int __writer;
  } __data;
  char __size[32];
  long int __align;
} pthread_rwlock_t;

typedef union
{
  char __size[8];
  long int __align;
} pthread_rwlockattr_t;





typedef volatile int pthread_spinlock_t;




typedef union
{
  char __size[20];
  long int __align;
} pthread_barrier_t;

typedef union
{
  char __size[4];
  int __align;
} pthread_barrierattr_t;
# 272 "/usr/include/sys/types.h" 2 3 4



# 321 "/usr/include/stdlib.h" 2 3 4






extern long int random (void) __attribute__ ((__nothrow__));


extern void srandom (unsigned int __seed) __attribute__ ((__nothrow__));





extern char *initstate (unsigned int __seed, char *__statebuf,
   size_t __statelen) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));



extern char *setstate (char *__statebuf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));







struct random_data
  {
    int32_t *fptr;
    int32_t *rptr;
    int32_t *state;
    int rand_type;
    int rand_deg;
    int rand_sep;
    int32_t *end_ptr;
  };

extern int random_r (struct random_data *__restrict __buf,
       int32_t *__restrict __result) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));

extern int srandom_r (unsigned int __seed, struct random_data *__buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));

extern int initstate_r (unsigned int __seed, char *__restrict __statebuf,
   size_t __statelen,
   struct random_data *__restrict __buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4)));

extern int setstate_r (char *__restrict __statebuf,
         struct random_data *__restrict __buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));






extern int rand (void) __attribute__ ((__nothrow__));

extern void srand (unsigned int __seed) __attribute__ ((__nothrow__));




extern int rand_r (unsigned int *__seed) __attribute__ ((__nothrow__));







extern double drand48 (void) __attribute__ ((__nothrow__));
extern double erand48 (unsigned short int __xsubi[3]) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern long int lrand48 (void) __attribute__ ((__nothrow__));
extern long int nrand48 (unsigned short int __xsubi[3])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern long int mrand48 (void) __attribute__ ((__nothrow__));
extern long int jrand48 (unsigned short int __xsubi[3])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern void srand48 (long int __seedval) __attribute__ ((__nothrow__));
extern unsigned short int *seed48 (unsigned short int __seed16v[3])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
extern void lcong48 (unsigned short int __param[7]) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





struct drand48_data
  {
    unsigned short int __x[3];
    unsigned short int __old_x[3];
    unsigned short int __c;
    unsigned short int __init;
    unsigned long long int __a;
  };


extern int drand48_r (struct drand48_data *__restrict __buffer,
        double *__restrict __result) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern int erand48_r (unsigned short int __xsubi[3],
        struct drand48_data *__restrict __buffer,
        double *__restrict __result) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern int lrand48_r (struct drand48_data *__restrict __buffer,
        long int *__restrict __result)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern int nrand48_r (unsigned short int __xsubi[3],
        struct drand48_data *__restrict __buffer,
        long int *__restrict __result)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern int mrand48_r (struct drand48_data *__restrict __buffer,
        long int *__restrict __result)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern int jrand48_r (unsigned short int __xsubi[3],
        struct drand48_data *__restrict __buffer,
        long int *__restrict __result)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern int srand48_r (long int __seedval, struct drand48_data *__buffer)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));

extern int seed48_r (unsigned short int __seed16v[3],
       struct drand48_data *__buffer) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));

extern int lcong48_r (unsigned short int __param[7],
        struct drand48_data *__buffer)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));









extern void *malloc (size_t __size) __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) ;

extern void *calloc (size_t __nmemb, size_t __size)
     __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) ;










extern void *realloc (void *__ptr, size_t __size)
     __attribute__ ((__nothrow__)) __attribute__ ((__warn_unused_result__));

extern void free (void *__ptr) __attribute__ ((__nothrow__));




extern void cfree (void *__ptr) __attribute__ ((__nothrow__));



# 1 "/usr/include/alloca.h" 1 3 4
# 25 "/usr/include/alloca.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 26 "/usr/include/alloca.h" 2 3 4







extern void *alloca (size_t __size) __attribute__ ((__nothrow__));






# 498 "/usr/include/stdlib.h" 2 3 4





extern void *valloc (size_t __size) __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) ;




extern int posix_memalign (void **__memptr, size_t __alignment, size_t __size)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;




extern void abort (void) __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));



extern int atexit (void (*__func) (void)) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 528 "/usr/include/stdlib.h" 3 4
extern int at_quick_exit (void (*__func) (void)) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));







extern int on_exit (void (*__func) (int __status, void *__arg), void *__arg)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern void exit (int __status) __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));







extern void quick_exit (int __status) __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));







extern void _Exit (int __status) __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));






extern char *getenv (__const char *__name) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;




extern char *__secure_getenv (__const char *__name)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





extern int putenv (char *__string) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int setenv (__const char *__name, __const char *__value, int __replace)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));


extern int unsetenv (__const char *__name) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern int clearenv (void) __attribute__ ((__nothrow__));
# 606 "/usr/include/stdlib.h" 3 4
extern char *mktemp (char *__template) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
# 620 "/usr/include/stdlib.h" 3 4
extern int mkstemp (char *__template) __attribute__ ((__nonnull__ (1))) ;
# 630 "/usr/include/stdlib.h" 3 4
extern int mkstemp64 (char *__template) __attribute__ ((__nonnull__ (1))) ;
# 642 "/usr/include/stdlib.h" 3 4
extern int mkstemps (char *__template, int __suffixlen) __attribute__ ((__nonnull__ (1))) ;
# 652 "/usr/include/stdlib.h" 3 4
extern int mkstemps64 (char *__template, int __suffixlen)
     __attribute__ ((__nonnull__ (1))) ;
# 663 "/usr/include/stdlib.h" 3 4
extern char *mkdtemp (char *__template) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
# 674 "/usr/include/stdlib.h" 3 4
extern int mkostemp (char *__template, int __flags) __attribute__ ((__nonnull__ (1))) ;
# 684 "/usr/include/stdlib.h" 3 4
extern int mkostemp64 (char *__template, int __flags) __attribute__ ((__nonnull__ (1))) ;
# 694 "/usr/include/stdlib.h" 3 4
extern int mkostemps (char *__template, int __suffixlen, int __flags)
     __attribute__ ((__nonnull__ (1))) ;
# 706 "/usr/include/stdlib.h" 3 4
extern int mkostemps64 (char *__template, int __suffixlen, int __flags)
     __attribute__ ((__nonnull__ (1))) ;









extern int system (__const char *__command) ;






extern char *canonicalize_file_name (__const char *__name)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
# 734 "/usr/include/stdlib.h" 3 4
extern char *realpath (__const char *__restrict __name,
         char *__restrict __resolved) __attribute__ ((__nothrow__)) ;






typedef int (*__compar_fn_t) (__const void *, __const void *);


typedef __compar_fn_t comparison_fn_t;



typedef int (*__compar_d_fn_t) (__const void *, __const void *, void *);





extern void *bsearch (__const void *__key, __const void *__base,
        size_t __nmemb, size_t __size, __compar_fn_t __compar)
     __attribute__ ((__nonnull__ (1, 2, 5))) ;



extern void qsort (void *__base, size_t __nmemb, size_t __size,
     __compar_fn_t __compar) __attribute__ ((__nonnull__ (1, 4)));

extern void qsort_r (void *__base, size_t __nmemb, size_t __size,
       __compar_d_fn_t __compar, void *__arg)
  __attribute__ ((__nonnull__ (1, 4)));




extern int abs (int __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;
extern long int labs (long int __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;



__extension__ extern long long int llabs (long long int __x)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;







extern div_t div (int __numer, int __denom)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;
extern ldiv_t ldiv (long int __numer, long int __denom)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;




__extension__ extern lldiv_t lldiv (long long int __numer,
        long long int __denom)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__)) ;

# 808 "/usr/include/stdlib.h" 3 4
extern char *ecvt (double __value, int __ndigit, int *__restrict __decpt,
     int *__restrict __sign) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4))) ;




extern char *fcvt (double __value, int __ndigit, int *__restrict __decpt,
     int *__restrict __sign) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4))) ;




extern char *gcvt (double __value, int __ndigit, char *__buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3))) ;




extern char *qecvt (long double __value, int __ndigit,
      int *__restrict __decpt, int *__restrict __sign)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4))) ;
extern char *qfcvt (long double __value, int __ndigit,
      int *__restrict __decpt, int *__restrict __sign)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4))) ;
extern char *qgcvt (long double __value, int __ndigit, char *__buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3))) ;




extern int ecvt_r (double __value, int __ndigit, int *__restrict __decpt,
     int *__restrict __sign, char *__restrict __buf,
     size_t __len) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4, 5)));
extern int fcvt_r (double __value, int __ndigit, int *__restrict __decpt,
     int *__restrict __sign, char *__restrict __buf,
     size_t __len) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4, 5)));

extern int qecvt_r (long double __value, int __ndigit,
      int *__restrict __decpt, int *__restrict __sign,
      char *__restrict __buf, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4, 5)));
extern int qfcvt_r (long double __value, int __ndigit,
      int *__restrict __decpt, int *__restrict __sign,
      char *__restrict __buf, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4, 5)));







extern int mblen (__const char *__s, size_t __n) __attribute__ ((__nothrow__)) ;


extern int mbtowc (wchar_t *__restrict __pwc,
     __const char *__restrict __s, size_t __n) __attribute__ ((__nothrow__)) ;


extern int wctomb (char *__s, wchar_t __wchar) __attribute__ ((__nothrow__)) ;



extern size_t mbstowcs (wchar_t *__restrict __pwcs,
   __const char *__restrict __s, size_t __n) __attribute__ ((__nothrow__));

extern size_t wcstombs (char *__restrict __s,
   __const wchar_t *__restrict __pwcs, size_t __n)
     __attribute__ ((__nothrow__));








extern int rpmatch (__const char *__response) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
# 896 "/usr/include/stdlib.h" 3 4
extern int getsubopt (char **__restrict __optionp,
        char *__const *__restrict __tokens,
        char **__restrict __valuep)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2, 3))) ;





extern void setkey (__const char *__key) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));







extern int posix_openpt (int __oflag) ;







extern int grantpt (int __fd) __attribute__ ((__nothrow__));



extern int unlockpt (int __fd) __attribute__ ((__nothrow__));




extern char *ptsname (int __fd) __attribute__ ((__nothrow__)) ;






extern int ptsname_r (int __fd, char *__buf, size_t __buflen)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));


extern int getpt (void);






extern int getloadavg (double __loadavg[], int __nelem)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 964 "/usr/include/stdlib.h" 3 4

# 2424 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 2
# 2439 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h"
# 1 "/usr/include/string.h" 1 3 4
# 29 "/usr/include/string.h" 3 4





# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 35 "/usr/include/string.h" 2 3 4









extern void *memcpy (void *__restrict __dest,
       __const void *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern void *memmove (void *__dest, __const void *__src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));






extern void *memccpy (void *__restrict __dest, __const void *__restrict __src,
        int __c, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));





extern void *memset (void *__s, int __c, size_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int memcmp (__const void *__s1, __const void *__s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 95 "/usr/include/string.h" 3 4
extern void *memchr (__const void *__s, int __c, size_t __n)
      __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));


# 109 "/usr/include/string.h" 3 4
extern void *rawmemchr (__const void *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));
# 120 "/usr/include/string.h" 3 4
extern void *memrchr (__const void *__s, int __c, size_t __n)
      __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));






extern char *strcpy (char *__restrict __dest, __const char *__restrict __src)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));

extern char *strncpy (char *__restrict __dest,
        __const char *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern char *strcat (char *__restrict __dest, __const char *__restrict __src)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));

extern char *strncat (char *__restrict __dest, __const char *__restrict __src,
        size_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern int strcmp (__const char *__s1, __const char *__s2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));

extern int strncmp (__const char *__s1, __const char *__s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));


extern int strcoll (__const char *__s1, __const char *__s2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));

extern size_t strxfrm (char *__restrict __dest,
         __const char *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));

# 165 "/usr/include/string.h" 3 4
extern int strcoll_l (__const char *__s1, __const char *__s2, __locale_t __l)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2, 3)));

extern size_t strxfrm_l (char *__dest, __const char *__src, size_t __n,
    __locale_t __l) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4)));





extern char *strdup (__const char *__s)
     __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) __attribute__ ((__nonnull__ (1)));






extern char *strndup (__const char *__string, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) __attribute__ ((__nonnull__ (1)));
# 210 "/usr/include/string.h" 3 4

# 235 "/usr/include/string.h" 3 4
extern char *strchr (__const char *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));
# 262 "/usr/include/string.h" 3 4
extern char *strrchr (__const char *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));


# 276 "/usr/include/string.h" 3 4
extern char *strchrnul (__const char *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));






extern size_t strcspn (__const char *__s, __const char *__reject)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));


extern size_t strspn (__const char *__s, __const char *__accept)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 314 "/usr/include/string.h" 3 4
extern char *strpbrk (__const char *__s, __const char *__accept)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 342 "/usr/include/string.h" 3 4
extern char *strstr (__const char *__haystack, __const char *__needle)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));




extern char *strtok (char *__restrict __s, __const char *__restrict __delim)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));




extern char *__strtok_r (char *__restrict __s,
    __const char *__restrict __delim,
    char **__restrict __save_ptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));

extern char *strtok_r (char *__restrict __s, __const char *__restrict __delim,
         char **__restrict __save_ptr)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
# 373 "/usr/include/string.h" 3 4
extern char *strcasestr (__const char *__haystack, __const char *__needle)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));







extern void *memmem (__const void *__haystack, size_t __haystacklen,
       __const void *__needle, size_t __needlelen)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 3)));



extern void *__mempcpy (void *__restrict __dest,
   __const void *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern void *mempcpy (void *__restrict __dest,
        __const void *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));





extern size_t strlen (__const char *__s)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));





extern size_t strnlen (__const char *__string, size_t __maxlen)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));





extern char *strerror (int __errnum) __attribute__ ((__nothrow__));

# 438 "/usr/include/string.h" 3 4
extern char *strerror_r (int __errnum, char *__buf, size_t __buflen)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));





extern char *strerror_l (int __errnum, __locale_t __l) __attribute__ ((__nothrow__));





extern void __bzero (void *__s, size_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern void bcopy (__const void *__src, void *__dest, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));


extern void bzero (void *__s, size_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int bcmp (__const void *__s1, __const void *__s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 489 "/usr/include/string.h" 3 4
extern char *index (__const char *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));
# 517 "/usr/include/string.h" 3 4
extern char *rindex (__const char *__s, int __c)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1)));




extern int ffs (int __i) __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern int ffsl (long int __l) __attribute__ ((__nothrow__)) __attribute__ ((__const__));

__extension__ extern int ffsll (long long int __ll)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern int strcasecmp (__const char *__s1, __const char *__s2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));


extern int strncasecmp (__const char *__s1, __const char *__s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));





extern int strcasecmp_l (__const char *__s1, __const char *__s2,
    __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2, 3)));

extern int strncasecmp_l (__const char *__s1, __const char *__s2,
     size_t __n, __locale_t __loc)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2, 4)));





extern char *strsep (char **__restrict __stringp,
       __const char *__restrict __delim)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));




extern char *strsignal (int __sig) __attribute__ ((__nothrow__));


extern char *__stpcpy (char *__restrict __dest, __const char *__restrict __src)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern char *stpcpy (char *__restrict __dest, __const char *__restrict __src)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern char *__stpncpy (char *__restrict __dest,
   __const char *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern char *stpncpy (char *__restrict __dest,
        __const char *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));




extern int strverscmp (__const char *__s1, __const char *__s2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));


extern char *strfry (char *__string) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern void *memfrob (void *__s, size_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 606 "/usr/include/string.h" 3 4
extern char *basename (__const char *__filename) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 634 "/usr/include/string.h" 3 4
# 1 "/usr/include/bits/string.h" 1 3 4
# 635 "/usr/include/string.h" 2 3 4


# 1 "/usr/include/bits/string2.h" 1 3 4
# 394 "/usr/include/bits/string2.h" 3 4
extern void *__rawmemchr (const void *__s, int __c);
# 969 "/usr/include/bits/string2.h" 3 4
extern __inline size_t __strcspn_c1 (__const char *__s, int __reject);
extern __inline size_t
__strcspn_c1 (__const char *__s, int __reject)
{
  register size_t __result = 0;
  while (__s[__result] != '\0' && __s[__result] != __reject)
    ++__result;
  return __result;
}

extern __inline size_t __strcspn_c2 (__const char *__s, int __reject1,
         int __reject2);
extern __inline size_t
__strcspn_c2 (__const char *__s, int __reject1, int __reject2)
{
  register size_t __result = 0;
  while (__s[__result] != '\0' && __s[__result] != __reject1
  && __s[__result] != __reject2)
    ++__result;
  return __result;
}

extern __inline size_t __strcspn_c3 (__const char *__s, int __reject1,
         int __reject2, int __reject3);
extern __inline size_t
__strcspn_c3 (__const char *__s, int __reject1, int __reject2,
       int __reject3)
{
  register size_t __result = 0;
  while (__s[__result] != '\0' && __s[__result] != __reject1
  && __s[__result] != __reject2 && __s[__result] != __reject3)
    ++__result;
  return __result;
}
# 1045 "/usr/include/bits/string2.h" 3 4
extern __inline size_t __strspn_c1 (__const char *__s, int __accept);
extern __inline size_t
__strspn_c1 (__const char *__s, int __accept)
{
  register size_t __result = 0;

  while (__s[__result] == __accept)
    ++__result;
  return __result;
}

extern __inline size_t __strspn_c2 (__const char *__s, int __accept1,
        int __accept2);
extern __inline size_t
__strspn_c2 (__const char *__s, int __accept1, int __accept2)
{
  register size_t __result = 0;

  while (__s[__result] == __accept1 || __s[__result] == __accept2)
    ++__result;
  return __result;
}

extern __inline size_t __strspn_c3 (__const char *__s, int __accept1,
        int __accept2, int __accept3);
extern __inline size_t
__strspn_c3 (__const char *__s, int __accept1, int __accept2, int __accept3)
{
  register size_t __result = 0;

  while (__s[__result] == __accept1 || __s[__result] == __accept2
  || __s[__result] == __accept3)
    ++__result;
  return __result;
}
# 1121 "/usr/include/bits/string2.h" 3 4
extern __inline char *__strpbrk_c2 (__const char *__s, int __accept1,
         int __accept2);
extern __inline char *
__strpbrk_c2 (__const char *__s, int __accept1, int __accept2)
{

  while (*__s != '\0' && *__s != __accept1 && *__s != __accept2)
    ++__s;
  return *__s == '\0' ? ((void *)0) : (char *) (size_t) __s;
}

extern __inline char *__strpbrk_c3 (__const char *__s, int __accept1,
         int __accept2, int __accept3);
extern __inline char *
__strpbrk_c3 (__const char *__s, int __accept1, int __accept2,
       int __accept3)
{

  while (*__s != '\0' && *__s != __accept1 && *__s != __accept2
  && *__s != __accept3)
    ++__s;
  return *__s == '\0' ? ((void *)0) : (char *) (size_t) __s;
}
# 1172 "/usr/include/bits/string2.h" 3 4
extern __inline char *__strtok_r_1c (char *__s, char __sep, char **__nextp);
extern __inline char *
__strtok_r_1c (char *__s, char __sep, char **__nextp)
{
  char *__result;
  if (__s == ((void *)0))
    __s = *__nextp;
  while (*__s == __sep)
    ++__s;
  __result = ((void *)0);
  if (*__s != '\0')
    {
      __result = __s++;
      while (*__s != '\0')
 if (*__s++ == __sep)
   {
     __s[-1] = '\0';
     break;
   }
    }
  *__nextp = __s;
  return __result;
}
# 1204 "/usr/include/bits/string2.h" 3 4
extern char *__strsep_g (char **__stringp, __const char *__delim);
# 1222 "/usr/include/bits/string2.h" 3 4
extern __inline char *__strsep_1c (char **__s, char __reject);
extern __inline char *
__strsep_1c (char **__s, char __reject)
{
  register char *__retval = *__s;
  if (__retval != ((void *)0) && (*__s = (__extension__ (__builtin_constant_p (__reject) && !__builtin_constant_p (__retval) && (__reject) == '\0' ? (char *) __rawmemchr (__retval, __reject) : __builtin_strchr (__retval, __reject)))) != ((void *)0))
    *(*__s)++ = '\0';
  return __retval;
}

extern __inline char *__strsep_2c (char **__s, char __reject1, char __reject2);
extern __inline char *
__strsep_2c (char **__s, char __reject1, char __reject2)
{
  register char *__retval = *__s;
  if (__retval != ((void *)0))
    {
      register char *__cp = __retval;
      while (1)
 {
   if (*__cp == '\0')
     {
       __cp = ((void *)0);
   break;
     }
   if (*__cp == __reject1 || *__cp == __reject2)
     {
       *__cp++ = '\0';
       break;
     }
   ++__cp;
 }
      *__s = __cp;
    }
  return __retval;
}

extern __inline char *__strsep_3c (char **__s, char __reject1, char __reject2,
       char __reject3);
extern __inline char *
__strsep_3c (char **__s, char __reject1, char __reject2, char __reject3)
{
  register char *__retval = *__s;
  if (__retval != ((void *)0))
    {
      register char *__cp = __retval;
      while (1)
 {
   if (*__cp == '\0')
     {
       __cp = ((void *)0);
   break;
     }
   if (*__cp == __reject1 || *__cp == __reject2 || *__cp == __reject3)
     {
       *__cp++ = '\0';
       break;
     }
   ++__cp;
 }
      *__s = __cp;
    }
  return __retval;
}
# 1303 "/usr/include/bits/string2.h" 3 4
extern char *__strdup (__const char *__string) __attribute__ ((__nothrow__)) __attribute__ ((__malloc__));
# 1322 "/usr/include/bits/string2.h" 3 4
extern char *__strndup (__const char *__string, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__malloc__));
# 638 "/usr/include/string.h" 2 3 4
# 646 "/usr/include/string.h" 3 4

# 2440 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 2
# 2450 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h"
# 1 "/usr/include/math.h" 1 3 4
# 30 "/usr/include/math.h" 3 4




# 1 "/usr/include/bits/huge_val.h" 1 3 4
# 35 "/usr/include/math.h" 2 3 4

# 1 "/usr/include/bits/huge_valf.h" 1 3 4
# 37 "/usr/include/math.h" 2 3 4
# 1 "/usr/include/bits/huge_vall.h" 1 3 4
# 38 "/usr/include/math.h" 2 3 4


# 1 "/usr/include/bits/inf.h" 1 3 4
# 41 "/usr/include/math.h" 2 3 4


# 1 "/usr/include/bits/nan.h" 1 3 4
# 44 "/usr/include/math.h" 2 3 4



# 1 "/usr/include/bits/mathdef.h" 1 3 4
# 36 "/usr/include/bits/mathdef.h" 3 4
typedef long double float_t;

typedef long double double_t;
# 48 "/usr/include/math.h" 2 3 4
# 71 "/usr/include/math.h" 3 4
# 1 "/usr/include/bits/mathcalls.h" 1 3 4
# 53 "/usr/include/bits/mathcalls.h" 3 4


extern double acos (double __x) __attribute__ ((__nothrow__)); extern double __acos (double __x) __attribute__ ((__nothrow__));

extern double asin (double __x) __attribute__ ((__nothrow__)); extern double __asin (double __x) __attribute__ ((__nothrow__));

extern double atan (double __x) __attribute__ ((__nothrow__)); extern double __atan (double __x) __attribute__ ((__nothrow__));

extern double atan2 (double __y, double __x) __attribute__ ((__nothrow__)); extern double __atan2 (double __y, double __x) __attribute__ ((__nothrow__));


extern double cos (double __x) __attribute__ ((__nothrow__)); extern double __cos (double __x) __attribute__ ((__nothrow__));

extern double sin (double __x) __attribute__ ((__nothrow__)); extern double __sin (double __x) __attribute__ ((__nothrow__));

extern double tan (double __x) __attribute__ ((__nothrow__)); extern double __tan (double __x) __attribute__ ((__nothrow__));




extern double cosh (double __x) __attribute__ ((__nothrow__)); extern double __cosh (double __x) __attribute__ ((__nothrow__));

extern double sinh (double __x) __attribute__ ((__nothrow__)); extern double __sinh (double __x) __attribute__ ((__nothrow__));

extern double tanh (double __x) __attribute__ ((__nothrow__)); extern double __tanh (double __x) __attribute__ ((__nothrow__));




extern void sincos (double __x, double *__sinx, double *__cosx) __attribute__ ((__nothrow__)); extern void __sincos (double __x, double *__sinx, double *__cosx) __attribute__ ((__nothrow__))
                                                           ;





extern double acosh (double __x) __attribute__ ((__nothrow__)); extern double __acosh (double __x) __attribute__ ((__nothrow__));

extern double asinh (double __x) __attribute__ ((__nothrow__)); extern double __asinh (double __x) __attribute__ ((__nothrow__));

extern double atanh (double __x) __attribute__ ((__nothrow__)); extern double __atanh (double __x) __attribute__ ((__nothrow__));







extern double exp (double __x) __attribute__ ((__nothrow__)); extern double __exp (double __x) __attribute__ ((__nothrow__));


extern double frexp (double __x, int *__exponent) __attribute__ ((__nothrow__)); extern double __frexp (double __x, int *__exponent) __attribute__ ((__nothrow__));


extern double ldexp (double __x, int __exponent) __attribute__ ((__nothrow__)); extern double __ldexp (double __x, int __exponent) __attribute__ ((__nothrow__));


extern double log (double __x) __attribute__ ((__nothrow__)); extern double __log (double __x) __attribute__ ((__nothrow__));


extern double log10 (double __x) __attribute__ ((__nothrow__)); extern double __log10 (double __x) __attribute__ ((__nothrow__));


extern double modf (double __x, double *__iptr) __attribute__ ((__nothrow__)); extern double __modf (double __x, double *__iptr) __attribute__ ((__nothrow__));




extern double exp10 (double __x) __attribute__ ((__nothrow__)); extern double __exp10 (double __x) __attribute__ ((__nothrow__));

extern double pow10 (double __x) __attribute__ ((__nothrow__)); extern double __pow10 (double __x) __attribute__ ((__nothrow__));





extern double expm1 (double __x) __attribute__ ((__nothrow__)); extern double __expm1 (double __x) __attribute__ ((__nothrow__));


extern double log1p (double __x) __attribute__ ((__nothrow__)); extern double __log1p (double __x) __attribute__ ((__nothrow__));


extern double logb (double __x) __attribute__ ((__nothrow__)); extern double __logb (double __x) __attribute__ ((__nothrow__));






extern double exp2 (double __x) __attribute__ ((__nothrow__)); extern double __exp2 (double __x) __attribute__ ((__nothrow__));


extern double log2 (double __x) __attribute__ ((__nothrow__)); extern double __log2 (double __x) __attribute__ ((__nothrow__));








extern double pow (double __x, double __y) __attribute__ ((__nothrow__)); extern double __pow (double __x, double __y) __attribute__ ((__nothrow__));


extern double sqrt (double __x) __attribute__ ((__nothrow__)); extern double __sqrt (double __x) __attribute__ ((__nothrow__));





extern double hypot (double __x, double __y) __attribute__ ((__nothrow__)); extern double __hypot (double __x, double __y) __attribute__ ((__nothrow__));






extern double cbrt (double __x) __attribute__ ((__nothrow__)); extern double __cbrt (double __x) __attribute__ ((__nothrow__));








extern double ceil (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __ceil (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern double fabs (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __fabs (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern double floor (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __floor (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern double fmod (double __x, double __y) __attribute__ ((__nothrow__)); extern double __fmod (double __x, double __y) __attribute__ ((__nothrow__));




extern int __isinf (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int __finite (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int isinf (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int finite (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern double drem (double __x, double __y) __attribute__ ((__nothrow__)); extern double __drem (double __x, double __y) __attribute__ ((__nothrow__));



extern double significand (double __x) __attribute__ ((__nothrow__)); extern double __significand (double __x) __attribute__ ((__nothrow__));





extern double copysign (double __x, double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __copysign (double __x, double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));






extern double nan (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __nan (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int __isnan (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern int isnan (double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern double j0 (double) __attribute__ ((__nothrow__)); extern double __j0 (double) __attribute__ ((__nothrow__));
extern double j1 (double) __attribute__ ((__nothrow__)); extern double __j1 (double) __attribute__ ((__nothrow__));
extern double jn (int, double) __attribute__ ((__nothrow__)); extern double __jn (int, double) __attribute__ ((__nothrow__));
extern double y0 (double) __attribute__ ((__nothrow__)); extern double __y0 (double) __attribute__ ((__nothrow__));
extern double y1 (double) __attribute__ ((__nothrow__)); extern double __y1 (double) __attribute__ ((__nothrow__));
extern double yn (int, double) __attribute__ ((__nothrow__)); extern double __yn (int, double) __attribute__ ((__nothrow__));






extern double erf (double) __attribute__ ((__nothrow__)); extern double __erf (double) __attribute__ ((__nothrow__));
extern double erfc (double) __attribute__ ((__nothrow__)); extern double __erfc (double) __attribute__ ((__nothrow__));
extern double lgamma (double) __attribute__ ((__nothrow__)); extern double __lgamma (double) __attribute__ ((__nothrow__));






extern double tgamma (double) __attribute__ ((__nothrow__)); extern double __tgamma (double) __attribute__ ((__nothrow__));





extern double gamma (double) __attribute__ ((__nothrow__)); extern double __gamma (double) __attribute__ ((__nothrow__));






extern double lgamma_r (double, int *__signgamp) __attribute__ ((__nothrow__)); extern double __lgamma_r (double, int *__signgamp) __attribute__ ((__nothrow__));







extern double rint (double __x) __attribute__ ((__nothrow__)); extern double __rint (double __x) __attribute__ ((__nothrow__));


extern double nextafter (double __x, double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __nextafter (double __x, double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));

extern double nexttoward (double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __nexttoward (double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern double remainder (double __x, double __y) __attribute__ ((__nothrow__)); extern double __remainder (double __x, double __y) __attribute__ ((__nothrow__));



extern double scalbn (double __x, int __n) __attribute__ ((__nothrow__)); extern double __scalbn (double __x, int __n) __attribute__ ((__nothrow__));



extern int ilogb (double __x) __attribute__ ((__nothrow__)); extern int __ilogb (double __x) __attribute__ ((__nothrow__));




extern double scalbln (double __x, long int __n) __attribute__ ((__nothrow__)); extern double __scalbln (double __x, long int __n) __attribute__ ((__nothrow__));



extern double nearbyint (double __x) __attribute__ ((__nothrow__)); extern double __nearbyint (double __x) __attribute__ ((__nothrow__));



extern double round (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __round (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern double trunc (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern double __trunc (double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern double remquo (double __x, double __y, int *__quo) __attribute__ ((__nothrow__)); extern double __remquo (double __x, double __y, int *__quo) __attribute__ ((__nothrow__));






extern long int lrint (double __x) __attribute__ ((__nothrow__)); extern long int __lrint (double __x) __attribute__ ((__nothrow__));
extern long long int llrint (double __x) __attribute__ ((__nothrow__)); extern long long int __llrint (double __x) __attribute__ ((__nothrow__));



extern long int lround (double __x) __attribute__ ((__nothrow__)); extern long int __lround (double __x) __attribute__ ((__nothrow__));
extern long long int llround (double __x) __attribute__ ((__nothrow__)); extern long long int __llround (double __x) __attribute__ ((__nothrow__));



extern double fdim (double __x, double __y) __attribute__ ((__nothrow__)); extern double __fdim (double __x, double __y) __attribute__ ((__nothrow__));


extern double fmax (double __x, double __y) __attribute__ ((__nothrow__)); extern double __fmax (double __x, double __y) __attribute__ ((__nothrow__));


extern double fmin (double __x, double __y) __attribute__ ((__nothrow__)); extern double __fmin (double __x, double __y) __attribute__ ((__nothrow__));



extern int __fpclassify (double __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));


extern int __signbit (double __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));



extern double fma (double __x, double __y, double __z) __attribute__ ((__nothrow__)); extern double __fma (double __x, double __y, double __z) __attribute__ ((__nothrow__));








extern double scalb (double __x, double __n) __attribute__ ((__nothrow__)); extern double __scalb (double __x, double __n) __attribute__ ((__nothrow__));
# 72 "/usr/include/math.h" 2 3 4
# 94 "/usr/include/math.h" 3 4
# 1 "/usr/include/bits/mathcalls.h" 1 3 4
# 53 "/usr/include/bits/mathcalls.h" 3 4


extern float acosf (float __x) __attribute__ ((__nothrow__)); extern float __acosf (float __x) __attribute__ ((__nothrow__));

extern float asinf (float __x) __attribute__ ((__nothrow__)); extern float __asinf (float __x) __attribute__ ((__nothrow__));

extern float atanf (float __x) __attribute__ ((__nothrow__)); extern float __atanf (float __x) __attribute__ ((__nothrow__));

extern float atan2f (float __y, float __x) __attribute__ ((__nothrow__)); extern float __atan2f (float __y, float __x) __attribute__ ((__nothrow__));


extern float cosf (float __x) __attribute__ ((__nothrow__)); extern float __cosf (float __x) __attribute__ ((__nothrow__));

extern float sinf (float __x) __attribute__ ((__nothrow__)); extern float __sinf (float __x) __attribute__ ((__nothrow__));

extern float tanf (float __x) __attribute__ ((__nothrow__)); extern float __tanf (float __x) __attribute__ ((__nothrow__));




extern float coshf (float __x) __attribute__ ((__nothrow__)); extern float __coshf (float __x) __attribute__ ((__nothrow__));

extern float sinhf (float __x) __attribute__ ((__nothrow__)); extern float __sinhf (float __x) __attribute__ ((__nothrow__));

extern float tanhf (float __x) __attribute__ ((__nothrow__)); extern float __tanhf (float __x) __attribute__ ((__nothrow__));




extern void
 sincosf
# 82 "/usr/include/bits/mathcalls.h" 3 4
 (float __x, float *__sinx, float *__cosx) __attribute__ ((__nothrow__)); extern void
 __sincosf
# 82 "/usr/include/bits/mathcalls.h" 3 4
 (float __x, float *__sinx, float *__cosx) __attribute__ ((__nothrow__))
                                                           ;





extern float acoshf (float __x) __attribute__ ((__nothrow__)); extern float __acoshf (float __x) __attribute__ ((__nothrow__));

extern float asinhf (float __x) __attribute__ ((__nothrow__)); extern float __asinhf (float __x) __attribute__ ((__nothrow__));

extern float atanhf (float __x) __attribute__ ((__nothrow__)); extern float __atanhf (float __x) __attribute__ ((__nothrow__));







extern float expf (float __x) __attribute__ ((__nothrow__)); extern float __expf (float __x) __attribute__ ((__nothrow__));


extern float frexpf (float __x, int *__exponent) __attribute__ ((__nothrow__)); extern float __frexpf (float __x, int *__exponent) __attribute__ ((__nothrow__));


extern float ldexpf (float __x, int __exponent) __attribute__ ((__nothrow__)); extern float __ldexpf (float __x, int __exponent) __attribute__ ((__nothrow__));


extern float logf (float __x) __attribute__ ((__nothrow__)); extern float __logf (float __x) __attribute__ ((__nothrow__));


extern float log10f (float __x) __attribute__ ((__nothrow__)); extern float __log10f (float __x) __attribute__ ((__nothrow__));


extern float modff (float __x, float *__iptr) __attribute__ ((__nothrow__)); extern float __modff (float __x, float *__iptr) __attribute__ ((__nothrow__));




extern float exp10f (float __x) __attribute__ ((__nothrow__)); extern float __exp10f (float __x) __attribute__ ((__nothrow__));

extern float pow10f (float __x) __attribute__ ((__nothrow__)); extern float __pow10f (float __x) __attribute__ ((__nothrow__));





extern float expm1f (float __x) __attribute__ ((__nothrow__)); extern float __expm1f (float __x) __attribute__ ((__nothrow__));


extern float log1pf (float __x) __attribute__ ((__nothrow__)); extern float __log1pf (float __x) __attribute__ ((__nothrow__));


extern float logbf (float __x) __attribute__ ((__nothrow__)); extern float __logbf (float __x) __attribute__ ((__nothrow__));






extern float exp2f (float __x) __attribute__ ((__nothrow__)); extern float __exp2f (float __x) __attribute__ ((__nothrow__));


extern float log2f (float __x) __attribute__ ((__nothrow__)); extern float __log2f (float __x) __attribute__ ((__nothrow__));








extern float powf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __powf (float __x, float __y) __attribute__ ((__nothrow__));


extern float sqrtf (float __x) __attribute__ ((__nothrow__)); extern float __sqrtf (float __x) __attribute__ ((__nothrow__));





extern float hypotf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __hypotf (float __x, float __y) __attribute__ ((__nothrow__));






extern float cbrtf (float __x) __attribute__ ((__nothrow__)); extern float __cbrtf (float __x) __attribute__ ((__nothrow__));








extern float ceilf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __ceilf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern float fabsf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __fabsf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern float floorf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __floorf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern float fmodf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __fmodf (float __x, float __y) __attribute__ ((__nothrow__));




extern int __isinff (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int __finitef (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int isinff (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int finitef (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern float dremf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __dremf (float __x, float __y) __attribute__ ((__nothrow__));



extern float significandf (float __x) __attribute__ ((__nothrow__)); extern float __significandf (float __x) __attribute__ ((__nothrow__));





extern float copysignf (float __x, float __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __copysignf (float __x, float __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));






extern float nanf (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __nanf (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int __isnanf (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern int isnanf (float __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern float j0f (float) __attribute__ ((__nothrow__)); extern float __j0f (float) __attribute__ ((__nothrow__));
extern float j1f (float) __attribute__ ((__nothrow__)); extern float __j1f (float) __attribute__ ((__nothrow__));
extern float jnf (int, float) __attribute__ ((__nothrow__)); extern float __jnf (int, float) __attribute__ ((__nothrow__));
extern float y0f (float) __attribute__ ((__nothrow__)); extern float __y0f (float) __attribute__ ((__nothrow__));
extern float y1f (float) __attribute__ ((__nothrow__)); extern float __y1f (float) __attribute__ ((__nothrow__));
extern float ynf (int, float) __attribute__ ((__nothrow__)); extern float __ynf (int, float) __attribute__ ((__nothrow__));






extern float erff (float) __attribute__ ((__nothrow__)); extern float __erff (float) __attribute__ ((__nothrow__));
extern float erfcf (float) __attribute__ ((__nothrow__)); extern float __erfcf (float) __attribute__ ((__nothrow__));
extern float lgammaf (float) __attribute__ ((__nothrow__)); extern float __lgammaf (float) __attribute__ ((__nothrow__));






extern float tgammaf (float) __attribute__ ((__nothrow__)); extern float __tgammaf (float) __attribute__ ((__nothrow__));





extern float gammaf (float) __attribute__ ((__nothrow__)); extern float __gammaf (float) __attribute__ ((__nothrow__));






extern float lgammaf_r (float, int *__signgamp) __attribute__ ((__nothrow__)); extern float __lgammaf_r (float, int *__signgamp) __attribute__ ((__nothrow__));







extern float rintf (float __x) __attribute__ ((__nothrow__)); extern float __rintf (float __x) __attribute__ ((__nothrow__));


extern float nextafterf (float __x, float __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __nextafterf (float __x, float __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));

extern float nexttowardf (float __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __nexttowardf (float __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern float remainderf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __remainderf (float __x, float __y) __attribute__ ((__nothrow__));



extern float scalbnf (float __x, int __n) __attribute__ ((__nothrow__)); extern float __scalbnf (float __x, int __n) __attribute__ ((__nothrow__));



extern int ilogbf (float __x) __attribute__ ((__nothrow__)); extern int __ilogbf (float __x) __attribute__ ((__nothrow__));




extern float scalblnf (float __x, long int __n) __attribute__ ((__nothrow__)); extern float __scalblnf (float __x, long int __n) __attribute__ ((__nothrow__));



extern float nearbyintf (float __x) __attribute__ ((__nothrow__)); extern float __nearbyintf (float __x) __attribute__ ((__nothrow__));



extern float roundf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __roundf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern float truncf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern float __truncf (float __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern float remquof (float __x, float __y, int *__quo) __attribute__ ((__nothrow__)); extern float __remquof (float __x, float __y, int *__quo) __attribute__ ((__nothrow__));






extern long int lrintf (float __x) __attribute__ ((__nothrow__)); extern long int __lrintf (float __x) __attribute__ ((__nothrow__));
extern long long int llrintf (float __x) __attribute__ ((__nothrow__)); extern long long int __llrintf (float __x) __attribute__ ((__nothrow__));



extern long int lroundf (float __x) __attribute__ ((__nothrow__)); extern long int __lroundf (float __x) __attribute__ ((__nothrow__));
extern long long int llroundf (float __x) __attribute__ ((__nothrow__)); extern long long int __llroundf (float __x) __attribute__ ((__nothrow__));



extern float fdimf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __fdimf (float __x, float __y) __attribute__ ((__nothrow__));


extern float fmaxf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __fmaxf (float __x, float __y) __attribute__ ((__nothrow__));


extern float fminf (float __x, float __y) __attribute__ ((__nothrow__)); extern float __fminf (float __x, float __y) __attribute__ ((__nothrow__));



extern int __fpclassifyf (float __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));


extern int __signbitf (float __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));



extern float fmaf (float __x, float __y, float __z) __attribute__ ((__nothrow__)); extern float __fmaf (float __x, float __y, float __z) __attribute__ ((__nothrow__));








extern float scalbf (float __x, float __n) __attribute__ ((__nothrow__)); extern float __scalbf (float __x, float __n) __attribute__ ((__nothrow__));
# 95 "/usr/include/math.h" 2 3 4
# 141 "/usr/include/math.h" 3 4
# 1 "/usr/include/bits/mathcalls.h" 1 3 4
# 53 "/usr/include/bits/mathcalls.h" 3 4


extern long double acosl (long double __x) __attribute__ ((__nothrow__)); extern long double __acosl (long double __x) __attribute__ ((__nothrow__));

extern long double asinl (long double __x) __attribute__ ((__nothrow__)); extern long double __asinl (long double __x) __attribute__ ((__nothrow__));

extern long double atanl (long double __x) __attribute__ ((__nothrow__)); extern long double __atanl (long double __x) __attribute__ ((__nothrow__));

extern long double atan2l (long double __y, long double __x) __attribute__ ((__nothrow__)); extern long double __atan2l (long double __y, long double __x) __attribute__ ((__nothrow__));


extern long double cosl (long double __x) __attribute__ ((__nothrow__)); extern long double __cosl (long double __x) __attribute__ ((__nothrow__));

extern long double sinl (long double __x) __attribute__ ((__nothrow__)); extern long double __sinl (long double __x) __attribute__ ((__nothrow__));

extern long double tanl (long double __x) __attribute__ ((__nothrow__)); extern long double __tanl (long double __x) __attribute__ ((__nothrow__));




extern long double coshl (long double __x) __attribute__ ((__nothrow__)); extern long double __coshl (long double __x) __attribute__ ((__nothrow__));

extern long double sinhl (long double __x) __attribute__ ((__nothrow__)); extern long double __sinhl (long double __x) __attribute__ ((__nothrow__));

extern long double tanhl (long double __x) __attribute__ ((__nothrow__)); extern long double __tanhl (long double __x) __attribute__ ((__nothrow__));




extern void
 sincosl
# 82 "/usr/include/bits/mathcalls.h" 3 4
 (long double __x, long double *__sinx, long double *__cosx) __attribute__ ((__nothrow__)); extern void
 __sincosl
# 82 "/usr/include/bits/mathcalls.h" 3 4
 (long double __x, long double *__sinx, long double *__cosx) __attribute__ ((__nothrow__))
                                                           ;





extern long double acoshl (long double __x) __attribute__ ((__nothrow__)); extern long double __acoshl (long double __x) __attribute__ ((__nothrow__));

extern long double asinhl (long double __x) __attribute__ ((__nothrow__)); extern long double __asinhl (long double __x) __attribute__ ((__nothrow__));

extern long double atanhl (long double __x) __attribute__ ((__nothrow__)); extern long double __atanhl (long double __x) __attribute__ ((__nothrow__));







extern long double expl (long double __x) __attribute__ ((__nothrow__)); extern long double __expl (long double __x) __attribute__ ((__nothrow__));


extern long double frexpl (long double __x, int *__exponent) __attribute__ ((__nothrow__)); extern long double __frexpl (long double __x, int *__exponent) __attribute__ ((__nothrow__));


extern long double ldexpl (long double __x, int __exponent) __attribute__ ((__nothrow__)); extern long double __ldexpl (long double __x, int __exponent) __attribute__ ((__nothrow__));


extern long double logl (long double __x) __attribute__ ((__nothrow__)); extern long double __logl (long double __x) __attribute__ ((__nothrow__));


extern long double log10l (long double __x) __attribute__ ((__nothrow__)); extern long double __log10l (long double __x) __attribute__ ((__nothrow__));


extern long double modfl (long double __x, long double *__iptr) __attribute__ ((__nothrow__)); extern long double __modfl (long double __x, long double *__iptr) __attribute__ ((__nothrow__));




extern long double exp10l (long double __x) __attribute__ ((__nothrow__)); extern long double __exp10l (long double __x) __attribute__ ((__nothrow__));

extern long double pow10l (long double __x) __attribute__ ((__nothrow__)); extern long double __pow10l (long double __x) __attribute__ ((__nothrow__));





extern long double expm1l (long double __x) __attribute__ ((__nothrow__)); extern long double __expm1l (long double __x) __attribute__ ((__nothrow__));


extern long double log1pl (long double __x) __attribute__ ((__nothrow__)); extern long double __log1pl (long double __x) __attribute__ ((__nothrow__));


extern long double logbl (long double __x) __attribute__ ((__nothrow__)); extern long double __logbl (long double __x) __attribute__ ((__nothrow__));






extern long double exp2l (long double __x) __attribute__ ((__nothrow__)); extern long double __exp2l (long double __x) __attribute__ ((__nothrow__));


extern long double log2l (long double __x) __attribute__ ((__nothrow__)); extern long double __log2l (long double __x) __attribute__ ((__nothrow__));








extern long double powl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __powl (long double __x, long double __y) __attribute__ ((__nothrow__));


extern long double sqrtl (long double __x) __attribute__ ((__nothrow__)); extern long double __sqrtl (long double __x) __attribute__ ((__nothrow__));





extern long double hypotl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __hypotl (long double __x, long double __y) __attribute__ ((__nothrow__));






extern long double cbrtl (long double __x) __attribute__ ((__nothrow__)); extern long double __cbrtl (long double __x) __attribute__ ((__nothrow__));








extern long double ceill (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __ceill (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern long double fabsl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __fabsl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern long double floorl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __floorl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern long double fmodl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __fmodl (long double __x, long double __y) __attribute__ ((__nothrow__));




extern int __isinfl (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int __finitel (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int isinfl (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern int finitel (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern long double dreml (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __dreml (long double __x, long double __y) __attribute__ ((__nothrow__));



extern long double significandl (long double __x) __attribute__ ((__nothrow__)); extern long double __significandl (long double __x) __attribute__ ((__nothrow__));





extern long double copysignl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __copysignl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));






extern long double nanl (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __nanl (__const char *__tagb) __attribute__ ((__nothrow__)) __attribute__ ((__const__));





extern int __isnanl (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern int isnanl (long double __value) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern long double j0l (long double) __attribute__ ((__nothrow__)); extern long double __j0l (long double) __attribute__ ((__nothrow__));
extern long double j1l (long double) __attribute__ ((__nothrow__)); extern long double __j1l (long double) __attribute__ ((__nothrow__));
extern long double jnl (int, long double) __attribute__ ((__nothrow__)); extern long double __jnl (int, long double) __attribute__ ((__nothrow__));
extern long double y0l (long double) __attribute__ ((__nothrow__)); extern long double __y0l (long double) __attribute__ ((__nothrow__));
extern long double y1l (long double) __attribute__ ((__nothrow__)); extern long double __y1l (long double) __attribute__ ((__nothrow__));
extern long double ynl (int, long double) __attribute__ ((__nothrow__)); extern long double __ynl (int, long double) __attribute__ ((__nothrow__));






extern long double erfl (long double) __attribute__ ((__nothrow__)); extern long double __erfl (long double) __attribute__ ((__nothrow__));
extern long double erfcl (long double) __attribute__ ((__nothrow__)); extern long double __erfcl (long double) __attribute__ ((__nothrow__));
extern long double lgammal (long double) __attribute__ ((__nothrow__)); extern long double __lgammal (long double) __attribute__ ((__nothrow__));






extern long double tgammal (long double) __attribute__ ((__nothrow__)); extern long double __tgammal (long double) __attribute__ ((__nothrow__));





extern long double gammal (long double) __attribute__ ((__nothrow__)); extern long double __gammal (long double) __attribute__ ((__nothrow__));






extern long double lgammal_r (long double, int *__signgamp) __attribute__ ((__nothrow__)); extern long double __lgammal_r (long double, int *__signgamp) __attribute__ ((__nothrow__));







extern long double rintl (long double __x) __attribute__ ((__nothrow__)); extern long double __rintl (long double __x) __attribute__ ((__nothrow__));


extern long double nextafterl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __nextafterl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));

extern long double nexttowardl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __nexttowardl (long double __x, long double __y) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern long double remainderl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __remainderl (long double __x, long double __y) __attribute__ ((__nothrow__));



extern long double scalbnl (long double __x, int __n) __attribute__ ((__nothrow__)); extern long double __scalbnl (long double __x, int __n) __attribute__ ((__nothrow__));



extern int ilogbl (long double __x) __attribute__ ((__nothrow__)); extern int __ilogbl (long double __x) __attribute__ ((__nothrow__));




extern long double scalblnl (long double __x, long int __n) __attribute__ ((__nothrow__)); extern long double __scalblnl (long double __x, long int __n) __attribute__ ((__nothrow__));



extern long double nearbyintl (long double __x) __attribute__ ((__nothrow__)); extern long double __nearbyintl (long double __x) __attribute__ ((__nothrow__));



extern long double roundl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __roundl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));



extern long double truncl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__)); extern long double __truncl (long double __x) __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern long double remquol (long double __x, long double __y, int *__quo) __attribute__ ((__nothrow__)); extern long double __remquol (long double __x, long double __y, int *__quo) __attribute__ ((__nothrow__));






extern long int lrintl (long double __x) __attribute__ ((__nothrow__)); extern long int __lrintl (long double __x) __attribute__ ((__nothrow__));
extern long long int llrintl (long double __x) __attribute__ ((__nothrow__)); extern long long int __llrintl (long double __x) __attribute__ ((__nothrow__));



extern long int lroundl (long double __x) __attribute__ ((__nothrow__)); extern long int __lroundl (long double __x) __attribute__ ((__nothrow__));
extern long long int llroundl (long double __x) __attribute__ ((__nothrow__)); extern long long int __llroundl (long double __x) __attribute__ ((__nothrow__));



extern long double fdiml (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __fdiml (long double __x, long double __y) __attribute__ ((__nothrow__));


extern long double fmaxl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __fmaxl (long double __x, long double __y) __attribute__ ((__nothrow__));


extern long double fminl (long double __x, long double __y) __attribute__ ((__nothrow__)); extern long double __fminl (long double __x, long double __y) __attribute__ ((__nothrow__));



extern int __fpclassifyl (long double __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));


extern int __signbitl (long double __value) __attribute__ ((__nothrow__))
     __attribute__ ((__const__));



extern long double fmal (long double __x, long double __y, long double __z) __attribute__ ((__nothrow__)); extern long double __fmal (long double __x, long double __y, long double __z) __attribute__ ((__nothrow__));








extern long double scalbl (long double __x, long double __n) __attribute__ ((__nothrow__)); extern long double __scalbl (long double __x, long double __n) __attribute__ ((__nothrow__));
# 142 "/usr/include/math.h" 2 3 4
# 157 "/usr/include/math.h" 3 4
extern int signgam;
# 198 "/usr/include/math.h" 3 4
enum
  {
    FP_NAN,

    FP_INFINITE,

    FP_ZERO,

    FP_SUBNORMAL,

    FP_NORMAL

  };
# 291 "/usr/include/math.h" 3 4
typedef enum
{
  _IEEE_ = -1,
  _SVID_,
  _XOPEN_,
  _POSIX_,
  _ISOC_
} _LIB_VERSION_TYPE;




extern _LIB_VERSION_TYPE _LIB_VERSION;
# 316 "/usr/include/math.h" 3 4
struct exception

  {
    int type;
    char *name;
    double arg1;
    double arg2;
    double retval;
  };




extern int matherr (struct exception *__exc);
# 416 "/usr/include/math.h" 3 4
# 1 "/usr/include/bits/mathinline.h" 1 3 4
# 127 "/usr/include/bits/mathinline.h" 3 4
extern __inline int
__attribute__ ((__nothrow__)) __signbitf (float __x)
{
  __extension__ union { float __f; int __i; } __u = { __f: __x };
  return __u.__i < 0;
}
extern __inline int
__attribute__ ((__nothrow__)) __signbit (double __x)
{
  __extension__ union { double __d; int __i[2]; } __u = { __d: __x };
  return __u.__i[1] < 0;
}
extern __inline int
__attribute__ ((__nothrow__)) __signbitl (long double __x)
{
  __extension__ union { long double __l; int __i[3]; } __u = { __l: __x };
  return (__u.__i[2] & 0x8000) != 0;
}
# 295 "/usr/include/bits/mathinline.h" 3 4
extern __inline double __sgn (double) __attribute__ ((__nothrow__)); extern __inline double __attribute__ ((__nothrow__)) __sgn (double __x) { return __x == 0.0 ? 0.0 : (__x > 0.0 ? 1.0 : -1.0); } extern __inline float __sgnf (float) __attribute__ ((__nothrow__)); extern __inline float __attribute__ ((__nothrow__)) __sgnf (float __x) { return __x == 0.0 ? 0.0 : (__x > 0.0 ? 1.0 : -1.0); } extern __inline long double __sgnl (long double) __attribute__ ((__nothrow__)); extern __inline long double __attribute__ ((__nothrow__)) __sgnl (long double __x) { return __x == 0.0 ? 0.0 : (__x > 0.0 ? 1.0 : -1.0); }
# 437 "/usr/include/bits/mathinline.h" 3 4
extern __inline long double __attribute__ ((__nothrow__)) __atan2l (long double __y, long double __x) { return __builtin_atan2l (__y, __x); }
# 477 "/usr/include/bits/mathinline.h" 3 4
extern __inline double __attribute__ ((__nothrow__)) fabs (double __x) { return __builtin_fabs (__x); }

extern __inline float __attribute__ ((__nothrow__)) fabsf (float __x) { return __builtin_fabsf (__x); }
extern __inline long double __attribute__ ((__nothrow__)) fabsl (long double __x) { return __builtin_fabsl (__x); }

extern __inline long double __attribute__ ((__nothrow__)) __fabsl (long double __x) { return __builtin_fabsl (__x); }
# 510 "/usr/include/bits/mathinline.h" 3 4
extern __inline long double __sgn1l (long double) __attribute__ ((__nothrow__)); extern __inline long double __attribute__ ((__nothrow__)) __sgn1l (long double __x) { __extension__ union { long double __xld; unsigned int __xi[3]; } __n = { __xld: __x }; __n.__xi[2] = (__n.__xi[2] & 0x8000) | 0x3fff; __n.__xi[1] = 0x80000000; __n.__xi[0] = 0; return __n.__xld; }
# 534 "/usr/include/bits/mathinline.h" 3 4
extern __inline double __attribute__ ((__nothrow__)) floor (double __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0400, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; } extern __inline float __attribute__ ((__nothrow__)) floorf (float __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0400, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; } extern __inline long double __attribute__ ((__nothrow__)) floorl (long double __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0400, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; }
# 552 "/usr/include/bits/mathinline.h" 3 4
extern __inline double __attribute__ ((__nothrow__)) ceil (double __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0800, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; } extern __inline float __attribute__ ((__nothrow__)) ceilf (float __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0800, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; } extern __inline long double __attribute__ ((__nothrow__)) ceill (long double __x) { register long double __value; register int __ignore; unsigned short int __cw; unsigned short int __cwtmp; __asm __volatile ("fnstcw %3\n\t" "movzwl %3, %1\n\t" "andl $0xf3ff, %1\n\t" "orl $0x0800, %1\n\t" "movw %w1, %2\n\t" "fldcw %2\n\t" "frndint\n\t" "fldcw %3" : "=t" (__value), "=&q" (__ignore), "=m" (__cwtmp), "=m" (__cw) : "0" (__x)); return __value; }
# 671 "/usr/include/bits/mathinline.h" 3 4
extern __inline long int
__attribute__ ((__nothrow__)) lrintf (float __x)
{
  long int __lrintres; __asm__ __volatile__ ("fistpl %0" : "=m" (__lrintres) : "t" (__x) : "st"); return __lrintres;
}
extern __inline long int
__attribute__ ((__nothrow__)) lrint (double __x)
{
  long int __lrintres; __asm__ __volatile__ ("fistpl %0" : "=m" (__lrintres) : "t" (__x) : "st"); return __lrintres;
}
extern __inline long int
__attribute__ ((__nothrow__)) lrintl (long double __x)
{
  long int __lrintres; __asm__ __volatile__ ("fistpl %0" : "=m" (__lrintres) : "t" (__x) : "st"); return __lrintres;
}
# 694 "/usr/include/bits/mathinline.h" 3 4
extern __inline long long int
__attribute__ ((__nothrow__)) llrintf (float __x)
{
  long long int __llrintres; __asm__ __volatile__ ("fistpll %0" : "=m" (__llrintres) : "t" (__x) : "st"); return __llrintres;
}
extern __inline long long int
__attribute__ ((__nothrow__)) llrint (double __x)
{
  long long int __llrintres; __asm__ __volatile__ ("fistpll %0" : "=m" (__llrintres) : "t" (__x) : "st"); return __llrintres;
}
extern __inline long long int
__attribute__ ((__nothrow__)) llrintl (long double __x)
{
  long long int __llrintres; __asm__ __volatile__ ("fistpll %0" : "=m" (__llrintres) : "t" (__x) : "st"); return __llrintres;
}
# 731 "/usr/include/bits/mathinline.h" 3 4
extern __inline int
__attribute__ ((__nothrow__)) __finite (double __x)
{
  return (__extension__
   (((((union { double __d; int __i[2]; }) {__d: __x}).__i[1]
      | 0x800fffffu) + 1) >> 31));
}
# 417 "/usr/include/math.h" 2 3 4
# 472 "/usr/include/math.h" 3 4

# 2451 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 2
# 2 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_config.h" 2
# 52 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 68 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
# 1 "/usr/include/stdio.h" 1 3 4
# 30 "/usr/include/stdio.h" 3 4




# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 35 "/usr/include/stdio.h" 2 3 4
# 45 "/usr/include/stdio.h" 3 4
struct _IO_FILE;



typedef struct _IO_FILE FILE;





# 65 "/usr/include/stdio.h" 3 4
typedef struct _IO_FILE __FILE;
# 75 "/usr/include/stdio.h" 3 4
# 1 "/usr/include/libio.h" 1 3 4
# 32 "/usr/include/libio.h" 3 4
# 1 "/usr/include/_G_config.h" 1 3 4
# 15 "/usr/include/_G_config.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 16 "/usr/include/_G_config.h" 2 3 4




# 1 "/usr/include/wchar.h" 1 3 4
# 83 "/usr/include/wchar.h" 3 4
typedef struct
{
  int __count;
  union
  {

    unsigned int __wch;



    char __wchb[4];
  } __value;
} __mbstate_t;
# 21 "/usr/include/_G_config.h" 2 3 4

typedef struct
{
  __off_t __pos;
  __mbstate_t __state;
} _G_fpos_t;
typedef struct
{
  __off64_t __pos;
  __mbstate_t __state;
} _G_fpos64_t;
# 53 "/usr/include/_G_config.h" 3 4
typedef int _G_int16_t __attribute__ ((__mode__ (__HI__)));
typedef int _G_int32_t __attribute__ ((__mode__ (__SI__)));
typedef unsigned int _G_uint16_t __attribute__ ((__mode__ (__HI__)));
typedef unsigned int _G_uint32_t __attribute__ ((__mode__ (__SI__)));
# 33 "/usr/include/libio.h" 2 3 4
# 53 "/usr/include/libio.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stdarg.h" 1 3 4
# 40 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stdarg.h" 3 4
typedef __builtin_va_list __gnuc_va_list;
# 54 "/usr/include/libio.h" 2 3 4
# 170 "/usr/include/libio.h" 3 4
struct _IO_jump_t; struct _IO_FILE;
# 180 "/usr/include/libio.h" 3 4
typedef void _IO_lock_t;





struct _IO_marker {
  struct _IO_marker *_next;
  struct _IO_FILE *_sbuf;



  int _pos;
# 203 "/usr/include/libio.h" 3 4
};


enum __codecvt_result
{
  __codecvt_ok,
  __codecvt_partial,
  __codecvt_error,
  __codecvt_noconv
};
# 271 "/usr/include/libio.h" 3 4
struct _IO_FILE {
  int _flags;




  char* _IO_read_ptr;
  char* _IO_read_end;
  char* _IO_read_base;
  char* _IO_write_base;
  char* _IO_write_ptr;
  char* _IO_write_end;
  char* _IO_buf_base;
  char* _IO_buf_end;

  char *_IO_save_base;
  char *_IO_backup_base;
  char *_IO_save_end;

  struct _IO_marker *_markers;

  struct _IO_FILE *_chain;

  int _fileno;



  int _flags2;

  __off_t _old_offset;



  unsigned short _cur_column;
  signed char _vtable_offset;
  char _shortbuf[1];



  _IO_lock_t *_lock;
# 319 "/usr/include/libio.h" 3 4
  __off64_t _offset;
# 328 "/usr/include/libio.h" 3 4
  void *__pad1;
  void *__pad2;
  void *__pad3;
  void *__pad4;
  size_t __pad5;

  int _mode;

  char _unused2[15 * sizeof (int) - 4 * sizeof (void *) - sizeof (size_t)];

};


typedef struct _IO_FILE _IO_FILE;


struct _IO_FILE_plus;

extern struct _IO_FILE_plus _IO_2_1_stdin_;
extern struct _IO_FILE_plus _IO_2_1_stdout_;
extern struct _IO_FILE_plus _IO_2_1_stderr_;
# 364 "/usr/include/libio.h" 3 4
typedef __ssize_t __io_read_fn (void *__cookie, char *__buf, size_t __nbytes);







typedef __ssize_t __io_write_fn (void *__cookie, __const char *__buf,
     size_t __n);







typedef int __io_seek_fn (void *__cookie, __off64_t *__pos, int __w);


typedef int __io_close_fn (void *__cookie);




typedef __io_read_fn cookie_read_function_t;
typedef __io_write_fn cookie_write_function_t;
typedef __io_seek_fn cookie_seek_function_t;
typedef __io_close_fn cookie_close_function_t;


typedef struct
{
  __io_read_fn *read;
  __io_write_fn *write;
  __io_seek_fn *seek;
  __io_close_fn *close;
} _IO_cookie_io_functions_t;
typedef _IO_cookie_io_functions_t cookie_io_functions_t;

struct _IO_cookie_file;


extern void _IO_cookie_init (struct _IO_cookie_file *__cfile, int __read_write,
        void *__cookie, _IO_cookie_io_functions_t __fns);







extern int __underflow (_IO_FILE *);
extern int __uflow (_IO_FILE *);
extern int __overflow (_IO_FILE *, int);
# 460 "/usr/include/libio.h" 3 4
extern int _IO_getc (_IO_FILE *__fp);
extern int _IO_putc (int __c, _IO_FILE *__fp);
extern int _IO_feof (_IO_FILE *__fp) __attribute__ ((__nothrow__));
extern int _IO_ferror (_IO_FILE *__fp) __attribute__ ((__nothrow__));

extern int _IO_peekc_locked (_IO_FILE *__fp);





extern void _IO_flockfile (_IO_FILE *) __attribute__ ((__nothrow__));
extern void _IO_funlockfile (_IO_FILE *) __attribute__ ((__nothrow__));
extern int _IO_ftrylockfile (_IO_FILE *) __attribute__ ((__nothrow__));
# 490 "/usr/include/libio.h" 3 4
extern int _IO_vfscanf (_IO_FILE * __restrict, const char * __restrict,
   __gnuc_va_list, int *__restrict);
extern int _IO_vfprintf (_IO_FILE *__restrict, const char *__restrict,
    __gnuc_va_list);
extern __ssize_t _IO_padn (_IO_FILE *, int, __ssize_t);
extern size_t _IO_sgetn (_IO_FILE *, void *, size_t);

extern __off64_t _IO_seekoff (_IO_FILE *, __off64_t, int, int);
extern __off64_t _IO_seekpos (_IO_FILE *, __off64_t, int);

extern void _IO_free_backup_area (_IO_FILE *) __attribute__ ((__nothrow__));
# 76 "/usr/include/stdio.h" 2 3 4




typedef __gnuc_va_list va_list;
# 109 "/usr/include/stdio.h" 3 4


typedef _G_fpos_t fpos_t;





typedef _G_fpos64_t fpos64_t;
# 161 "/usr/include/stdio.h" 3 4
# 1 "/usr/include/bits/stdio_lim.h" 1 3 4
# 162 "/usr/include/stdio.h" 2 3 4



extern struct _IO_FILE *stdin;
extern struct _IO_FILE *stdout;
extern struct _IO_FILE *stderr;









extern int remove (__const char *__filename) __attribute__ ((__nothrow__));

extern int rename (__const char *__old, __const char *__new) __attribute__ ((__nothrow__));




extern int renameat (int __oldfd, __const char *__old, int __newfd,
       __const char *__new) __attribute__ ((__nothrow__));








extern FILE *tmpfile (void) ;
# 204 "/usr/include/stdio.h" 3 4
extern FILE *tmpfile64 (void) ;



extern char *tmpnam (char *__s) __attribute__ ((__nothrow__)) ;





extern char *tmpnam_r (char *__s) __attribute__ ((__nothrow__)) ;
# 226 "/usr/include/stdio.h" 3 4
extern char *tempnam (__const char *__dir, __const char *__pfx)
     __attribute__ ((__nothrow__)) __attribute__ ((__malloc__)) ;








extern int fclose (FILE *__stream);




extern int fflush (FILE *__stream);

# 251 "/usr/include/stdio.h" 3 4
extern int fflush_unlocked (FILE *__stream);
# 261 "/usr/include/stdio.h" 3 4
extern int fcloseall (void);









extern FILE *fopen (__const char *__restrict __filename,
      __const char *__restrict __modes) ;




extern FILE *freopen (__const char *__restrict __filename,
        __const char *__restrict __modes,
        FILE *__restrict __stream) ;
# 294 "/usr/include/stdio.h" 3 4


extern FILE *fopen64 (__const char *__restrict __filename,
        __const char *__restrict __modes) ;
extern FILE *freopen64 (__const char *__restrict __filename,
   __const char *__restrict __modes,
   FILE *__restrict __stream) ;




extern FILE *fdopen (int __fd, __const char *__modes) __attribute__ ((__nothrow__)) ;





extern FILE *fopencookie (void *__restrict __magic_cookie,
     __const char *__restrict __modes,
     _IO_cookie_io_functions_t __io_funcs) __attribute__ ((__nothrow__)) ;




extern FILE *fmemopen (void *__s, size_t __len, __const char *__modes)
  __attribute__ ((__nothrow__)) ;




extern FILE *open_memstream (char **__bufloc, size_t *__sizeloc) __attribute__ ((__nothrow__)) ;






extern void setbuf (FILE *__restrict __stream, char *__restrict __buf) __attribute__ ((__nothrow__));



extern int setvbuf (FILE *__restrict __stream, char *__restrict __buf,
      int __modes, size_t __n) __attribute__ ((__nothrow__));





extern void setbuffer (FILE *__restrict __stream, char *__restrict __buf,
         size_t __size) __attribute__ ((__nothrow__));


extern void setlinebuf (FILE *__stream) __attribute__ ((__nothrow__));








extern int fprintf (FILE *__restrict __stream,
      __const char *__restrict __format, ...);




extern int printf (__const char *__restrict __format, ...);

extern int sprintf (char *__restrict __s,
      __const char *__restrict __format, ...) __attribute__ ((__nothrow__));





extern int vfprintf (FILE *__restrict __s, __const char *__restrict __format,
       __gnuc_va_list __arg);




extern int vprintf (__const char *__restrict __format, __gnuc_va_list __arg);

extern int vsprintf (char *__restrict __s, __const char *__restrict __format,
       __gnuc_va_list __arg) __attribute__ ((__nothrow__));





extern int snprintf (char *__restrict __s, size_t __maxlen,
       __const char *__restrict __format, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 3, 4)));

extern int vsnprintf (char *__restrict __s, size_t __maxlen,
        __const char *__restrict __format, __gnuc_va_list __arg)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 3, 0)));






extern int vasprintf (char **__restrict __ptr, __const char *__restrict __f,
        __gnuc_va_list __arg)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 2, 0))) ;
extern int __asprintf (char **__restrict __ptr,
         __const char *__restrict __fmt, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 2, 3))) ;
extern int asprintf (char **__restrict __ptr,
       __const char *__restrict __fmt, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 2, 3))) ;
# 416 "/usr/include/stdio.h" 3 4
extern int vdprintf (int __fd, __const char *__restrict __fmt,
       __gnuc_va_list __arg)
     __attribute__ ((__format__ (__printf__, 2, 0)));
extern int dprintf (int __fd, __const char *__restrict __fmt, ...)
     __attribute__ ((__format__ (__printf__, 2, 3)));








extern int fscanf (FILE *__restrict __stream,
     __const char *__restrict __format, ...) ;




extern int scanf (__const char *__restrict __format, ...) ;

extern int sscanf (__const char *__restrict __s,
     __const char *__restrict __format, ...) __attribute__ ((__nothrow__));
# 467 "/usr/include/stdio.h" 3 4








extern int vfscanf (FILE *__restrict __s, __const char *__restrict __format,
      __gnuc_va_list __arg)
     __attribute__ ((__format__ (__scanf__, 2, 0))) ;





extern int vscanf (__const char *__restrict __format, __gnuc_va_list __arg)
     __attribute__ ((__format__ (__scanf__, 1, 0))) ;


extern int vsscanf (__const char *__restrict __s,
      __const char *__restrict __format, __gnuc_va_list __arg)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__scanf__, 2, 0)));
# 526 "/usr/include/stdio.h" 3 4









extern int fgetc (FILE *__stream);
extern int getc (FILE *__stream);





extern int getchar (void);

# 554 "/usr/include/stdio.h" 3 4
extern int getc_unlocked (FILE *__stream);
extern int getchar_unlocked (void);
# 565 "/usr/include/stdio.h" 3 4
extern int fgetc_unlocked (FILE *__stream);











extern int fputc (int __c, FILE *__stream);
extern int putc (int __c, FILE *__stream);





extern int putchar (int __c);

# 598 "/usr/include/stdio.h" 3 4
extern int fputc_unlocked (int __c, FILE *__stream);







extern int putc_unlocked (int __c, FILE *__stream);
extern int putchar_unlocked (int __c);






extern int getw (FILE *__stream);


extern int putw (int __w, FILE *__stream);








extern char *fgets (char *__restrict __s, int __n, FILE *__restrict __stream)
     ;






extern char *gets (char *__s) ;

# 644 "/usr/include/stdio.h" 3 4
extern char *fgets_unlocked (char *__restrict __s, int __n,
        FILE *__restrict __stream) ;
# 660 "/usr/include/stdio.h" 3 4
extern __ssize_t __getdelim (char **__restrict __lineptr,
          size_t *__restrict __n, int __delimiter,
          FILE *__restrict __stream) ;
extern __ssize_t getdelim (char **__restrict __lineptr,
        size_t *__restrict __n, int __delimiter,
        FILE *__restrict __stream) ;







extern __ssize_t getline (char **__restrict __lineptr,
       size_t *__restrict __n,
       FILE *__restrict __stream) ;








extern int fputs (__const char *__restrict __s, FILE *__restrict __stream);





extern int puts (__const char *__s);






extern int ungetc (int __c, FILE *__stream);






extern size_t fread (void *__restrict __ptr, size_t __size,
       size_t __n, FILE *__restrict __stream) ;




extern size_t fwrite (__const void *__restrict __ptr, size_t __size,
        size_t __n, FILE *__restrict __s) ;

# 721 "/usr/include/stdio.h" 3 4
extern int fputs_unlocked (__const char *__restrict __s,
      FILE *__restrict __stream);
# 732 "/usr/include/stdio.h" 3 4
extern size_t fread_unlocked (void *__restrict __ptr, size_t __size,
         size_t __n, FILE *__restrict __stream) ;
extern size_t fwrite_unlocked (__const void *__restrict __ptr, size_t __size,
          size_t __n, FILE *__restrict __stream) ;








extern int fseek (FILE *__stream, long int __off, int __whence);




extern long int ftell (FILE *__stream) ;




extern void rewind (FILE *__stream);

# 768 "/usr/include/stdio.h" 3 4
extern int fseeko (FILE *__stream, __off_t __off, int __whence);




extern __off_t ftello (FILE *__stream) ;
# 787 "/usr/include/stdio.h" 3 4






extern int fgetpos (FILE *__restrict __stream, fpos_t *__restrict __pos);




extern int fsetpos (FILE *__stream, __const fpos_t *__pos);
# 810 "/usr/include/stdio.h" 3 4



extern int fseeko64 (FILE *__stream, __off64_t __off, int __whence);
extern __off64_t ftello64 (FILE *__stream) ;
extern int fgetpos64 (FILE *__restrict __stream, fpos64_t *__restrict __pos);
extern int fsetpos64 (FILE *__stream, __const fpos64_t *__pos);




extern void clearerr (FILE *__stream) __attribute__ ((__nothrow__));

extern int feof (FILE *__stream) __attribute__ ((__nothrow__)) ;

extern int ferror (FILE *__stream) __attribute__ ((__nothrow__)) ;




extern void clearerr_unlocked (FILE *__stream) __attribute__ ((__nothrow__));
extern int feof_unlocked (FILE *__stream) __attribute__ ((__nothrow__)) ;
extern int ferror_unlocked (FILE *__stream) __attribute__ ((__nothrow__)) ;








extern void perror (__const char *__s);






# 1 "/usr/include/bits/sys_errlist.h" 1 3 4
# 27 "/usr/include/bits/sys_errlist.h" 3 4
extern int sys_nerr;
extern __const char *__const sys_errlist[];


extern int _sys_nerr;
extern __const char *__const _sys_errlist[];
# 849 "/usr/include/stdio.h" 2 3 4




extern int fileno (FILE *__stream) __attribute__ ((__nothrow__)) ;




extern int fileno_unlocked (FILE *__stream) __attribute__ ((__nothrow__)) ;
# 868 "/usr/include/stdio.h" 3 4
extern FILE *popen (__const char *__command, __const char *__modes) ;





extern int pclose (FILE *__stream);





extern char *ctermid (char *__s) __attribute__ ((__nothrow__));





extern char *cuserid (char *__s);




struct obstack;


extern int obstack_printf (struct obstack *__restrict __obstack,
      __const char *__restrict __format, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 2, 3)));
extern int obstack_vprintf (struct obstack *__restrict __obstack,
       __const char *__restrict __format,
       __gnuc_va_list __args)
     __attribute__ ((__nothrow__)) __attribute__ ((__format__ (__printf__, 2, 0)));







extern void flockfile (FILE *__stream) __attribute__ ((__nothrow__));



extern int ftrylockfile (FILE *__stream) __attribute__ ((__nothrow__)) ;


extern void funlockfile (FILE *__stream) __attribute__ ((__nothrow__));
# 929 "/usr/include/stdio.h" 3 4
# 1 "/usr/include/bits/stdio.h" 1 3 4
# 36 "/usr/include/bits/stdio.h" 3 4
extern __inline int
vprintf (__const char *__restrict __fmt, __gnuc_va_list __arg)
{
  return vfprintf (stdout, __fmt, __arg);
}



extern __inline int
getchar (void)
{
  return _IO_getc (stdin);
}




extern __inline int
fgetc_unlocked (FILE *__fp)
{
  return (__builtin_expect (((__fp)->_IO_read_ptr >= (__fp)->_IO_read_end), 0) ? __uflow (__fp) : *(unsigned char *) (__fp)->_IO_read_ptr++);
}





extern __inline int
getc_unlocked (FILE *__fp)
{
  return (__builtin_expect (((__fp)->_IO_read_ptr >= (__fp)->_IO_read_end), 0) ? __uflow (__fp) : *(unsigned char *) (__fp)->_IO_read_ptr++);
}


extern __inline int
getchar_unlocked (void)
{
  return (__builtin_expect (((stdin)->_IO_read_ptr >= (stdin)->_IO_read_end), 0) ? __uflow (stdin) : *(unsigned char *) (stdin)->_IO_read_ptr++);
}




extern __inline int
putchar (int __c)
{
  return _IO_putc (__c, stdout);
}




extern __inline int
fputc_unlocked (int __c, FILE *__stream)
{
  return (__builtin_expect (((__stream)->_IO_write_ptr >= (__stream)->_IO_write_end), 0) ? __overflow (__stream, (unsigned char) (__c)) : (unsigned char) (*(__stream)->_IO_write_ptr++ = (__c)));
}





extern __inline int
putc_unlocked (int __c, FILE *__stream)
{
  return (__builtin_expect (((__stream)->_IO_write_ptr >= (__stream)->_IO_write_end), 0) ? __overflow (__stream, (unsigned char) (__c)) : (unsigned char) (*(__stream)->_IO_write_ptr++ = (__c)));
}


extern __inline int
putchar_unlocked (int __c)
{
  return (__builtin_expect (((stdout)->_IO_write_ptr >= (stdout)->_IO_write_end), 0) ? __overflow (stdout, (unsigned char) (__c)) : (unsigned char) (*(stdout)->_IO_write_ptr++ = (__c)));
}





extern __inline __ssize_t
getline (char **__lineptr, size_t *__n, FILE *__stream)
{
  return __getdelim (__lineptr, __n, '\n', __stream);
}





extern __inline int
__attribute__ ((__nothrow__)) feof_unlocked (FILE *__stream)
{
  return (((__stream)->_flags & 0x10) != 0);
}


extern __inline int
__attribute__ ((__nothrow__)) ferror_unlocked (FILE *__stream)
{
  return (((__stream)->_flags & 0x20) != 0);
}
# 930 "/usr/include/stdio.h" 2 3 4
# 938 "/usr/include/stdio.h" 3 4

# 69 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2






# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stdarg.h" 1 3 4
# 76 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2



# 1 "/usr/include/dlfcn.h" 1 3 4
# 25 "/usr/include/dlfcn.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 26 "/usr/include/dlfcn.h" 2 3 4


# 1 "/usr/include/bits/dlfcn.h" 1 3 4
# 58 "/usr/include/bits/dlfcn.h" 3 4



extern void _dl_mcount_wrapper_check (void *__selfpc) __attribute__ ((__nothrow__));


# 29 "/usr/include/dlfcn.h" 2 3 4
# 45 "/usr/include/dlfcn.h" 3 4
typedef long int Lmid_t;











extern void *dlopen (__const char *__file, int __mode) __attribute__ ((__nothrow__));



extern int dlclose (void *__handle) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern void *dlsym (void *__restrict __handle,
      __const char *__restrict __name) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));



extern void *dlmopen (Lmid_t __nsid, __const char *__file, int __mode) __attribute__ ((__nothrow__));



extern void *dlvsym (void *__restrict __handle,
       __const char *__restrict __name,
       __const char *__restrict __version)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));





extern char *dlerror (void) __attribute__ ((__nothrow__));





typedef struct
{
  __const char *dli_fname;
  void *dli_fbase;
  __const char *dli_sname;
  void *dli_saddr;
} Dl_info;



extern int dladdr (__const void *__address, Dl_info *__info)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));


extern int dladdr1 (__const void *__address, Dl_info *__info,
      void **__extra_info, int __flags) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));




enum
  {

    RTLD_DL_SYMENT = 1,


    RTLD_DL_LINKMAP = 2
  };







extern int dlinfo (void *__restrict __handle,
     int __request, void *__restrict __arg)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 3)));


enum
  {

    RTLD_DI_LMID = 1,



    RTLD_DI_LINKMAP = 2,

    RTLD_DI_CONFIGADDR = 3,






    RTLD_DI_SERINFO = 4,
    RTLD_DI_SERINFOSIZE = 5,



    RTLD_DI_ORIGIN = 6,

    RTLD_DI_PROFILENAME = 7,
    RTLD_DI_PROFILEOUT = 8,




    RTLD_DI_TLS_MODID = 9,





    RTLD_DI_TLS_DATA = 10,

    RTLD_DI_MAX = 10
  };




typedef struct
{
  char *dls_name;
  unsigned int dls_flags;
} Dl_serpath;



typedef struct
{
  size_t dls_size;
  unsigned int dls_cnt;
  Dl_serpath dls_serpath[1];
} Dl_serinfo;




# 80 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 236 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_errors.h" 1
# 237 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/../TSRM/TSRM.h" 1
# 20 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/../TSRM/TSRM.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config.h" 1
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 1
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config.h" 2
# 21 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/../TSRM/TSRM.h" 2
# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/../TSRM/TSRM.h"
typedef long tsrm_intptr_t;
typedef unsigned long tsrm_uintptr_t;
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 1
# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h" 2
# 44 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
typedef struct _zend_leak_info {
 void *addr;
 size_t size;
 const char *filename;
 uint lineno;
 const char *orig_filename;
 uint orig_lineno;
} zend_leak_info;



__attribute__ ((visibility("default"))) char *zend_strndup(const char *s, unsigned int length) __attribute__ ((__malloc__));

__attribute__ ((visibility("default"))) void *_emalloc(size_t size ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) void *_safe_emalloc(size_t nmemb, size_t size, size_t offset ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) void *_safe_malloc(size_t nmemb, size_t size, size_t offset) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) void _efree(void *ptr );
__attribute__ ((visibility("default"))) void *_ecalloc(size_t nmemb, size_t size ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) void *_erealloc(void *ptr, size_t size, int allow_failure );
__attribute__ ((visibility("default"))) void *_safe_erealloc(void *ptr, size_t nmemb, size_t size, size_t offset );
__attribute__ ((visibility("default"))) void *_safe_realloc(void *ptr, size_t nmemb, size_t size, size_t offset);
__attribute__ ((visibility("default"))) char *_estrdup(const char *s ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) char *_estrndup(const char *s, unsigned int length ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) size_t _zend_mem_block_size(void *ptr );
# 93 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
inline static void * __zend_malloc(size_t len)
{
 void *tmp = malloc(len);
 if (tmp) {
  return tmp;
 }
 fprintf(stderr, "Out of memory\n");
 exit(1);
}

inline static void * __zend_calloc(size_t nmemb, size_t len)
{
 void *tmp = _safe_malloc(nmemb, len, 0);
 memset(tmp, 0, nmemb * len);
 return tmp;
}

inline static void * __zend_realloc(void *p, size_t len)
{
 p = realloc(p, len);
 if (p) {
  return p;
 }
 fprintf(stderr, "Out of memory\n");
 exit(1);
}
# 142 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
__attribute__ ((visibility("default"))) int zend_set_memory_limit(size_t memory_limit);

__attribute__ ((visibility("default"))) void start_memory_manager(void);
__attribute__ ((visibility("default"))) void shutdown_memory_manager(int silent, int full_shutdown );
__attribute__ ((visibility("default"))) int is_zend_mm(void);
# 159 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
__attribute__ ((visibility("default"))) size_t zend_memory_usage(int real_usage );
__attribute__ ((visibility("default"))) size_t zend_memory_peak_usage(int real_usage );


# 191 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
typedef struct _zend_mm_heap zend_mm_heap;

__attribute__ ((visibility("default"))) zend_mm_heap *zend_mm_startup(void);
__attribute__ ((visibility("default"))) void zend_mm_shutdown(zend_mm_heap *heap, int full_shutdown, int silent );
__attribute__ ((visibility("default"))) void *_zend_mm_alloc(zend_mm_heap *heap, size_t size ) __attribute__ ((__malloc__));
__attribute__ ((visibility("default"))) void _zend_mm_free(zend_mm_heap *heap, void *p );
__attribute__ ((visibility("default"))) void *_zend_mm_realloc(zend_mm_heap *heap, void *p, size_t size );
__attribute__ ((visibility("default"))) size_t _zend_mm_block_size(zend_mm_heap *heap, void *p );
# 211 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h"
typedef struct _zend_mm_storage zend_mm_storage;

typedef struct _zend_mm_segment {
 size_t size;
 struct _zend_mm_segment *next_segment;
} zend_mm_segment;

typedef struct _zend_mm_mem_handlers {
 const char *name;
 zend_mm_storage* (*init)(void *params);
 void (*dtor)(zend_mm_storage *storage);
 void (*compact)(zend_mm_storage *storage);
 zend_mm_segment* (*_alloc)(zend_mm_storage *storage, size_t size);
 zend_mm_segment* (*_realloc)(zend_mm_storage *storage, zend_mm_segment *ptr, size_t size);
 void (*_free)(zend_mm_storage *storage, zend_mm_segment *ptr);
} zend_mm_mem_handlers;

struct _zend_mm_storage {
 const zend_mm_mem_handlers *handlers;
 void *data;
};

__attribute__ ((visibility("default"))) zend_mm_heap *zend_mm_startup_ex(const zend_mm_mem_handlers *handlers, size_t block_size, size_t reserve_size, int internal, void *params);
__attribute__ ((visibility("default"))) zend_mm_heap *zend_mm_set_heap(zend_mm_heap *new_heap );
__attribute__ ((visibility("default"))) zend_mm_storage *zend_mm_get_storage(zend_mm_heap *heap);

__attribute__ ((visibility("default"))) void zend_mm_set_custom_handlers(zend_mm_heap *heap,
                                          void* (*_malloc)(size_t),
                                          void (*_free)(void*),
                                          void* (*_realloc)(void*, size_t));
# 238 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_types.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_types.h"
typedef unsigned char zend_bool;
typedef unsigned char zend_uchar;
typedef unsigned int zend_uint;
typedef unsigned long zend_ulong;
typedef unsigned short zend_ushort;






typedef long long int zend_long64;
typedef unsigned long long int zend_ulong64;
# 49 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_types.h"
typedef long zend_intptr_t;
typedef unsigned long zend_uintptr_t;


typedef unsigned int zend_object_handle;
typedef struct _zend_object_handlers zend_object_handlers;

typedef struct _zend_object_value {
 zend_object_handle handle;
 const zend_object_handlers *handlers;
} zend_object_value;
# 240 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_string.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_string.h"
__attribute__ ((visibility("default"))) extern const char *(*zend_new_interned_string)(const char *str, int len, int free_src );
__attribute__ ((visibility("default"))) extern void (*zend_interned_strings_snapshot)(void);
__attribute__ ((visibility("default"))) extern void (*zend_interned_strings_restore)(void);

void zend_interned_strings_init(void);
void zend_interned_strings_dtor(void);
# 241 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2


# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 11 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/syslimits.h" 1 3 4






# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 122 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 3 4
# 1 "/usr/include/limits.h" 1 3 4
# 145 "/usr/include/limits.h" 3 4
# 1 "/usr/include/bits/posix1_lim.h" 1 3 4
# 157 "/usr/include/bits/posix1_lim.h" 3 4
# 1 "/usr/include/bits/local_lim.h" 1 3 4
# 39 "/usr/include/bits/local_lim.h" 3 4
# 1 "/usr/include/linux/limits.h" 1 3 4
# 40 "/usr/include/bits/local_lim.h" 2 3 4
# 158 "/usr/include/bits/posix1_lim.h" 2 3 4
# 146 "/usr/include/limits.h" 2 3 4



# 1 "/usr/include/bits/posix2_lim.h" 1 3 4
# 150 "/usr/include/limits.h" 2 3 4



# 1 "/usr/include/bits/xopen_lim.h" 1 3 4
# 34 "/usr/include/bits/xopen_lim.h" 3 4
# 1 "/usr/include/bits/stdio_lim.h" 1 3 4
# 35 "/usr/include/bits/xopen_lim.h" 2 3 4
# 154 "/usr/include/limits.h" 2 3 4
# 123 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 2 3 4
# 8 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/syslimits.h" 2 3 4
# 12 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 2 3 4
# 244 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 256 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
static const char long_min_digits[] = "2147483648";
# 271 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h" 1
# 45 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h"
typedef ulong (*hash_func_t)(const char *arKey, uint nKeyLength);
typedef int (*compare_func_t)(const void *, const void * );
typedef void (*sort_func_t)(void *, size_t, register size_t, compare_func_t );
typedef void (*dtor_func_t)(void *pDest);
typedef void (*copy_ctor_func_t)(void *pElement);
typedef void (*copy_ctor_param_func_t)(void *pElement, void *pParam);

struct _hashtable;

typedef struct bucket {
 ulong h;
 uint nKeyLength;
 void *pData;
 void *pDataPtr;
 struct bucket *pListNext;
 struct bucket *pListLast;
 struct bucket *pNext;
 struct bucket *pLast;
 const char *arKey;
} Bucket;

typedef struct _hashtable {
 uint nTableSize;
 uint nTableMask;
 uint nNumOfElements;
 ulong nNextFreeElement;
 Bucket *pInternalPointer;
 Bucket *pListHead;
 Bucket *pListTail;
 Bucket **arBuckets;
 dtor_func_t pDestructor;
 zend_bool persistent;
 unsigned char nApplyCount;
 zend_bool bApplyProtection;



} HashTable;


typedef struct _zend_hash_key {
 const char *arKey;
 uint nKeyLength;
 ulong h;
} zend_hash_key;


typedef zend_bool (*merge_checker_func_t)(HashTable *target_ht, void *source_data, zend_hash_key *hash_key, void *pParam);

typedef Bucket* HashPosition;




__attribute__ ((visibility("default"))) int _zend_hash_init(HashTable *ht, uint nSize, hash_func_t pHashFunction, dtor_func_t pDestructor, zend_bool persistent );
__attribute__ ((visibility("default"))) int _zend_hash_init_ex(HashTable *ht, uint nSize, hash_func_t pHashFunction, dtor_func_t pDestructor, zend_bool persistent, zend_bool bApplyProtection );
__attribute__ ((visibility("default"))) void zend_hash_destroy(HashTable *ht);
__attribute__ ((visibility("default"))) void zend_hash_clean(HashTable *ht);




__attribute__ ((visibility("default"))) int _zend_hash_add_or_update(HashTable *ht, const char *arKey, uint nKeyLength, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int _zend_hash_quick_add_or_update(HashTable *ht, const char *arKey, uint nKeyLength, ulong h, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int _zend_hash_index_update_or_next_insert(HashTable *ht, ulong h, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int zend_hash_add_empty_element(HashTable *ht, const char *arKey, uint nKeyLength);






typedef int (*apply_func_t)(void *pDest );
typedef int (*apply_func_arg_t)(void *pDest, void *argument );
typedef int (*apply_func_args_t)(void *pDest , int num_args, va_list args, zend_hash_key *hash_key);

__attribute__ ((visibility("default"))) void zend_hash_graceful_destroy(HashTable *ht);
__attribute__ ((visibility("default"))) void zend_hash_graceful_reverse_destroy(HashTable *ht);
__attribute__ ((visibility("default"))) void zend_hash_apply(HashTable *ht, apply_func_t apply_func );
__attribute__ ((visibility("default"))) void zend_hash_apply_with_argument(HashTable *ht, apply_func_arg_t apply_func, void * );
__attribute__ ((visibility("default"))) void zend_hash_apply_with_arguments(HashTable *ht , apply_func_args_t apply_func, int, ...);







__attribute__ ((visibility("default"))) void zend_hash_reverse_apply(HashTable *ht, apply_func_t apply_func );



__attribute__ ((visibility("default"))) int zend_hash_del_key_or_index(HashTable *ht, const char *arKey, uint nKeyLength, ulong h, int flag);







__attribute__ ((visibility("default"))) ulong zend_get_hash_value(const char *arKey, uint nKeyLength);


__attribute__ ((visibility("default"))) int zend_hash_find(const HashTable *ht, const char *arKey, uint nKeyLength, void **pData);
__attribute__ ((visibility("default"))) int zend_hash_quick_find(const HashTable *ht, const char *arKey, uint nKeyLength, ulong h, void **pData);
__attribute__ ((visibility("default"))) int zend_hash_index_find(const HashTable *ht, ulong h, void **pData);


__attribute__ ((visibility("default"))) int zend_hash_exists(const HashTable *ht, const char *arKey, uint nKeyLength);
__attribute__ ((visibility("default"))) int zend_hash_quick_exists(const HashTable *ht, const char *arKey, uint nKeyLength, ulong h);
__attribute__ ((visibility("default"))) int zend_hash_index_exists(const HashTable *ht, ulong h);
__attribute__ ((visibility("default"))) ulong zend_hash_next_free_element(const HashTable *ht);





__attribute__ ((visibility("default"))) int zend_hash_move_forward_ex(HashTable *ht, HashPosition *pos);
__attribute__ ((visibility("default"))) int zend_hash_move_backwards_ex(HashTable *ht, HashPosition *pos);
__attribute__ ((visibility("default"))) int zend_hash_get_current_key_ex(const HashTable *ht, char **str_index, uint *str_length, ulong *num_index, zend_bool duplicate, HashPosition *pos);
__attribute__ ((visibility("default"))) int zend_hash_get_current_key_type_ex(HashTable *ht, HashPosition *pos);
__attribute__ ((visibility("default"))) int zend_hash_get_current_data_ex(HashTable *ht, void **pData, HashPosition *pos);
__attribute__ ((visibility("default"))) void zend_hash_internal_pointer_reset_ex(HashTable *ht, HashPosition *pos);
__attribute__ ((visibility("default"))) void zend_hash_internal_pointer_end_ex(HashTable *ht, HashPosition *pos);
__attribute__ ((visibility("default"))) int zend_hash_update_current_key_ex(HashTable *ht, int key_type, const char *str_index, uint str_length, ulong num_index, int mode, HashPosition *pos);

typedef struct _HashPointer {
 HashPosition pos;
 ulong h;
} HashPointer;

__attribute__ ((visibility("default"))) int zend_hash_get_pointer(const HashTable *ht, HashPointer *ptr);
__attribute__ ((visibility("default"))) int zend_hash_set_pointer(HashTable *ht, const HashPointer *ptr);
# 214 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h"
__attribute__ ((visibility("default"))) void zend_hash_copy(HashTable *target, HashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size);
__attribute__ ((visibility("default"))) void _zend_hash_merge(HashTable *target, HashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size, int overwrite );
__attribute__ ((visibility("default"))) void zend_hash_merge_ex(HashTable *target, HashTable *source, copy_ctor_func_t pCopyConstructor, uint size, merge_checker_func_t pMergeSource, void *pParam);
__attribute__ ((visibility("default"))) int zend_hash_sort(HashTable *ht, sort_func_t sort_func, compare_func_t compare_func, int renumber );
__attribute__ ((visibility("default"))) int zend_hash_compare(HashTable *ht1, HashTable *ht2, compare_func_t compar, zend_bool ordered );
__attribute__ ((visibility("default"))) int zend_hash_minmax(const HashTable *ht, compare_func_t compar, int flag, void **pData );




__attribute__ ((visibility("default"))) int zend_hash_num_elements(const HashTable *ht);

__attribute__ ((visibility("default"))) int zend_hash_rehash(HashTable *ht);
# 261 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h"
static inline ulong zend_inline_hash_func(const char *arKey, uint nKeyLength)
{
 register ulong hash = 5381;


 for (; nKeyLength >= 8; nKeyLength -= 8) {
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
  hash = ((hash << 5) + hash) + *arKey++;
 }
 switch (nKeyLength) {
  case 7: hash = ((hash << 5) + hash) + *arKey++;
  case 6: hash = ((hash << 5) + hash) + *arKey++;
  case 5: hash = ((hash << 5) + hash) + *arKey++;
  case 4: hash = ((hash << 5) + hash) + *arKey++;
  case 3: hash = ((hash << 5) + hash) + *arKey++;
  case 2: hash = ((hash << 5) + hash) + *arKey++;
  case 1: hash = ((hash << 5) + hash) + *arKey++; break;
  case 0: break;

 }
 return hash;
}


__attribute__ ((visibility("default"))) ulong zend_hash_func(const char *arKey, uint nKeyLength);








# 348 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h"
static inline int zend_symtable_update(HashTable *ht, const char *arKey, uint nKeyLength, void *pData, uint nDataSize, void **pDest)
 {
 do { ulong idx; do { register const char *tmp = arKey; if (*tmp == '-') { tmp++; } if (*tmp >= '0' && *tmp <= '9') { const char *end = arKey + nKeyLength - 1; if ((*end != '\0') || (*tmp == '0' && nKeyLength > 2) || (end - tmp > 11 - 1) || (4 == 4 && end - tmp == 11 - 1 && *tmp > '2')) { break; } idx = (*tmp - '0'); while (++tmp != end && *tmp >= '0' && *tmp <= '9') { idx = (idx * 10) + (*tmp - '0'); } if (tmp == end) { if (*arKey == '-') { if (idx-1 > 2147483647L) { break; } idx = (ulong)(-(long)idx); } else if (idx > 2147483647L) { break; } return _zend_hash_index_update_or_next_insert(ht, idx, pData, nDataSize, pDest, (1<<0) ); } } } while (0); } while (0);
 return _zend_hash_add_or_update(ht, arKey, nKeyLength, pData, nDataSize, pDest, (1<<0) );
}


static inline int zend_symtable_del(HashTable *ht, const char *arKey, uint nKeyLength)
{
 do { ulong idx; do { register const char *tmp = arKey; if (*tmp == '-') { tmp++; } if (*tmp >= '0' && *tmp <= '9') { const char *end = arKey + nKeyLength - 1; if ((*end != '\0') || (*tmp == '0' && nKeyLength > 2) || (end - tmp > 11 - 1) || (4 == 4 && end - tmp == 11 - 1 && *tmp > '2')) { break; } idx = (*tmp - '0'); while (++tmp != end && *tmp >= '0' && *tmp <= '9') { idx = (idx * 10) + (*tmp - '0'); } if (tmp == end) { if (*arKey == '-') { if (idx-1 > 2147483647L) { break; } idx = (ulong)(-(long)idx); } else if (idx > 2147483647L) { break; } return zend_hash_del_key_or_index(ht, ((void *)0), 0, idx, 1); } } } while (0); } while (0);
 return zend_hash_del_key_or_index(ht, arKey, nKeyLength, 0, 0);
}


static inline int zend_symtable_find(HashTable *ht, const char *arKey, uint nKeyLength, void **pData)
{
 do { ulong idx; do { register const char *tmp = arKey; if (*tmp == '-') { tmp++; } if (*tmp >= '0' && *tmp <= '9') { const char *end = arKey + nKeyLength - 1; if ((*end != '\0') || (*tmp == '0' && nKeyLength > 2) || (end - tmp > 11 - 1) || (4 == 4 && end - tmp == 11 - 1 && *tmp > '2')) { break; } idx = (*tmp - '0'); while (++tmp != end && *tmp >= '0' && *tmp <= '9') { idx = (idx * 10) + (*tmp - '0'); } if (tmp == end) { if (*arKey == '-') { if (idx-1 > 2147483647L) { break; } idx = (ulong)(-(long)idx); } else if (idx > 2147483647L) { break; } return zend_hash_index_find(ht, idx, pData); } } } while (0); } while (0);
 return zend_hash_find(ht, arKey, nKeyLength, pData);
}


static inline int zend_symtable_exists(HashTable *ht, const char *arKey, uint nKeyLength)
{
 do { ulong idx; do { register const char *tmp = arKey; if (*tmp == '-') { tmp++; } if (*tmp >= '0' && *tmp <= '9') { const char *end = arKey + nKeyLength - 1; if ((*end != '\0') || (*tmp == '0' && nKeyLength > 2) || (end - tmp > 11 - 1) || (4 == 4 && end - tmp == 11 - 1 && *tmp > '2')) { break; } idx = (*tmp - '0'); while (++tmp != end && *tmp >= '0' && *tmp <= '9') { idx = (idx * 10) + (*tmp - '0'); } if (tmp == end) { if (*arKey == '-') { if (idx-1 > 2147483647L) { break; } idx = (ulong)(-(long)idx); } else if (idx > 2147483647L) { break; } return zend_hash_index_exists(ht, idx); } } } while (0); } while (0);
 return zend_hash_exists(ht, arKey, nKeyLength);
}

static inline int zend_symtable_update_current_key_ex(HashTable *ht, const char *arKey, uint nKeyLength, int mode, HashPosition *pos)
{
 do { ulong idx; do { register const char *tmp = arKey; if (*tmp == '-') { tmp++; } if (*tmp >= '0' && *tmp <= '9') { const char *end = arKey + nKeyLength - 1; if ((*end != '\0') || (*tmp == '0' && nKeyLength > 2) || (end - tmp > 11 - 1) || (4 == 4 && end - tmp == 11 - 1 && *tmp > '2')) { break; } idx = (*tmp - '0'); while (++tmp != end && *tmp >= '0' && *tmp <= '9') { idx = (idx * 10) + (*tmp - '0'); } if (tmp == end) { if (*arKey == '-') { if (idx-1 > 2147483647L) { break; } idx = (ulong)(-(long)idx); } else if (idx > 2147483647L) { break; } return zend_hash_update_current_key_ex(ht, 2, ((void *)0), 0, idx, mode, pos); } } } while (0); } while (0);
 return zend_hash_update_current_key_ex(ht, 1, arKey, nKeyLength, 0, mode, pos);
}
# 272 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ts_hash.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ts_hash.h"
typedef struct _zend_ts_hashtable {
 HashTable hash;
 zend_uint reader;




} TsHashTable;






__attribute__ ((visibility("default"))) int _zend_ts_hash_init(TsHashTable *ht, uint nSize, hash_func_t pHashFunction, dtor_func_t pDestructor, zend_bool persistent );
__attribute__ ((visibility("default"))) int _zend_ts_hash_init_ex(TsHashTable *ht, uint nSize, hash_func_t pHashFunction, dtor_func_t pDestructor, zend_bool persistent, zend_bool bApplyProtection );
__attribute__ ((visibility("default"))) void zend_ts_hash_destroy(TsHashTable *ht);
__attribute__ ((visibility("default"))) void zend_ts_hash_clean(TsHashTable *ht);
# 52 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ts_hash.h"
__attribute__ ((visibility("default"))) int _zend_ts_hash_add_or_update(TsHashTable *ht, char *arKey, uint nKeyLength, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int _zend_ts_hash_quick_add_or_update(TsHashTable *ht, char *arKey, uint nKeyLength, ulong h, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int _zend_ts_hash_index_update_or_next_insert(TsHashTable *ht, ulong h, void *pData, uint nDataSize, void **pDest, int flag );





__attribute__ ((visibility("default"))) int zend_ts_hash_add_empty_element(TsHashTable *ht, char *arKey, uint nKeyLength);

__attribute__ ((visibility("default"))) void zend_ts_hash_graceful_destroy(TsHashTable *ht);
__attribute__ ((visibility("default"))) void zend_ts_hash_apply(TsHashTable *ht, apply_func_t apply_func );
__attribute__ ((visibility("default"))) void zend_ts_hash_apply_with_argument(TsHashTable *ht, apply_func_arg_t apply_func, void * );
__attribute__ ((visibility("default"))) void zend_ts_hash_apply_with_arguments(TsHashTable *ht , apply_func_args_t apply_func, int, ...);

__attribute__ ((visibility("default"))) void zend_ts_hash_reverse_apply(TsHashTable *ht, apply_func_t apply_func );



__attribute__ ((visibility("default"))) int zend_ts_hash_del_key_or_index(TsHashTable *ht, char *arKey, uint nKeyLength, ulong h, int flag);





__attribute__ ((visibility("default"))) ulong zend_ts_get_hash_value(TsHashTable *ht, char *arKey, uint nKeyLength);


__attribute__ ((visibility("default"))) int zend_ts_hash_find(TsHashTable *ht, char *arKey, uint nKeyLength, void **pData);
__attribute__ ((visibility("default"))) int zend_ts_hash_quick_find(TsHashTable *ht, char *arKey, uint nKeyLength, ulong h, void **pData);
__attribute__ ((visibility("default"))) int zend_ts_hash_index_find(TsHashTable *ht, ulong h, void **pData);


__attribute__ ((visibility("default"))) int zend_ts_hash_exists(TsHashTable *ht, char *arKey, uint nKeyLength);
__attribute__ ((visibility("default"))) int zend_ts_hash_index_exists(TsHashTable *ht, ulong h);


__attribute__ ((visibility("default"))) void zend_ts_hash_copy(TsHashTable *target, TsHashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size);
__attribute__ ((visibility("default"))) void zend_ts_hash_copy_to_hash(HashTable *target, TsHashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size);
__attribute__ ((visibility("default"))) void zend_ts_hash_merge(TsHashTable *target, TsHashTable *source, copy_ctor_func_t pCopyConstructor, void *tmp, uint size, int overwrite);
__attribute__ ((visibility("default"))) void zend_ts_hash_merge_ex(TsHashTable *target, TsHashTable *source, copy_ctor_func_t pCopyConstructor, uint size, merge_checker_func_t pMergeSource, void *pParam);
__attribute__ ((visibility("default"))) int zend_ts_hash_sort(TsHashTable *ht, sort_func_t sort_func, compare_func_t compare_func, int renumber );
__attribute__ ((visibility("default"))) int zend_ts_hash_compare(TsHashTable *ht1, TsHashTable *ht2, compare_func_t compar, zend_bool ordered );
__attribute__ ((visibility("default"))) int zend_ts_hash_minmax(TsHashTable *ht, compare_func_t compar, int flag, void **pData );

__attribute__ ((visibility("default"))) int zend_ts_hash_num_elements(TsHashTable *ht);

__attribute__ ((visibility("default"))) int zend_ts_hash_rehash(TsHashTable *ht);

__attribute__ ((visibility("default"))) ulong zend_ts_hash_func(char *arKey, uint nKeyLength);








# 273 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_llist.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_llist.h"
typedef struct _zend_llist_element {
 struct _zend_llist_element *next;
 struct _zend_llist_element *prev;
 char data[1];
} zend_llist_element;

typedef void (*llist_dtor_func_t)(void *);
typedef int (*llist_compare_func_t)(const zend_llist_element **, const zend_llist_element ** );
typedef void (*llist_apply_with_args_func_t)(void *data, int num_args, va_list args );
typedef void (*llist_apply_with_arg_func_t)(void *data, void *arg );
typedef void (*llist_apply_func_t)(void * );

typedef struct _zend_llist {
 zend_llist_element *head;
 zend_llist_element *tail;
 size_t count;
 size_t size;
 llist_dtor_func_t dtor;
 unsigned char persistent;
 zend_llist_element *traverse_ptr;
} zend_llist;

typedef zend_llist_element* zend_llist_position;


__attribute__ ((visibility("default"))) void zend_llist_init(zend_llist *l, size_t size, llist_dtor_func_t dtor, unsigned char persistent);
__attribute__ ((visibility("default"))) void zend_llist_add_element(zend_llist *l, void *element);
__attribute__ ((visibility("default"))) void zend_llist_prepend_element(zend_llist *l, void *element);
__attribute__ ((visibility("default"))) void zend_llist_del_element(zend_llist *l, void *element, int (*compare)(void *element1, void *element2));
__attribute__ ((visibility("default"))) void zend_llist_destroy(zend_llist *l);
__attribute__ ((visibility("default"))) void zend_llist_clean(zend_llist *l);
__attribute__ ((visibility("default"))) void *zend_llist_remove_tail(zend_llist *l);
__attribute__ ((visibility("default"))) void zend_llist_copy(zend_llist *dst, zend_llist *src);
__attribute__ ((visibility("default"))) void zend_llist_apply(zend_llist *l, llist_apply_func_t func );
__attribute__ ((visibility("default"))) void zend_llist_apply_with_del(zend_llist *l, int (*func)(void *data));
__attribute__ ((visibility("default"))) void zend_llist_apply_with_argument(zend_llist *l, llist_apply_with_arg_func_t func, void *arg );
__attribute__ ((visibility("default"))) void zend_llist_apply_with_arguments(zend_llist *l, llist_apply_with_args_func_t func , int num_args, ...);
__attribute__ ((visibility("default"))) int zend_llist_count(zend_llist *l);
__attribute__ ((visibility("default"))) void zend_llist_sort(zend_llist *l, llist_compare_func_t comp_func );


__attribute__ ((visibility("default"))) void *zend_llist_get_first_ex(zend_llist *l, zend_llist_position *pos);
__attribute__ ((visibility("default"))) void *zend_llist_get_last_ex(zend_llist *l, zend_llist_position *pos);
__attribute__ ((visibility("default"))) void *zend_llist_get_next_ex(zend_llist *l, zend_llist_position *pos);
__attribute__ ((visibility("default"))) void *zend_llist_get_prev_ex(zend_llist *l, zend_llist_position *pos);







# 274 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2





void zend_error_noreturn(int type, const char *format, ...) __attribute__ ((noreturn));







typedef struct _zval_struct zval;
typedef struct _zend_class_entry zend_class_entry;

typedef struct _zend_guard {
 zend_bool in_get;
 zend_bool in_set;
 zend_bool in_unset;
 zend_bool in_isset;
 zend_bool dummy;
} zend_guard;

typedef struct _zend_object {
 zend_class_entry *ce;
 HashTable *properties;
 zval **properties_table;
 HashTable *guards;
} zend_object;

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_object_handlers.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_object_handlers.h"
union _zend_function;
struct _zend_property_info;
struct _zend_literal;






typedef zval *(*zend_object_read_property_t)(zval *object, zval *member, int type, const struct _zend_literal *key );


typedef zval *(*zend_object_read_dimension_t)(zval *object, zval *offset, int type );
# 46 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_object_handlers.h"
typedef void (*zend_object_write_property_t)(zval *object, zval *member, zval *value, const struct _zend_literal *key );


typedef void (*zend_object_write_dimension_t)(zval *object, zval *offset, zval *value );



typedef zval **(*zend_object_get_property_ptr_ptr_t)(zval *object, zval *member, const struct _zend_literal *key );



typedef void (*zend_object_set_t)(zval **object, zval *value );




typedef zval* (*zend_object_get_t)(zval *object );







typedef int (*zend_object_has_property_t)(zval *object, zval *member, int has_set_exists, const struct _zend_literal *key );


typedef int (*zend_object_has_dimension_t)(zval *object, zval *member, int check_empty );


typedef void (*zend_object_unset_property_t)(zval *object, zval *member, const struct _zend_literal *key );


typedef void (*zend_object_unset_dimension_t)(zval *object, zval *offset );


typedef HashTable *(*zend_object_get_properties_t)(zval *object );

typedef HashTable *(*zend_object_get_debug_info_t)(zval *object, int *is_temp );





typedef int (*zend_object_call_method_t)(const char *method, int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
typedef union _zend_function *(*zend_object_get_method_t)(zval **object_ptr, char *method, int method_len, const struct _zend_literal *key );
typedef union _zend_function *(*zend_object_get_constructor_t)(zval *object );


typedef void (*zend_object_add_ref_t)(zval *object );
typedef void (*zend_object_del_ref_t)(zval *object );
typedef void (*zend_object_delete_obj_t)(zval *object );
typedef zend_object_value (*zend_object_clone_obj_t)(zval *object );

typedef zend_class_entry *(*zend_object_get_class_entry_t)(const zval *object );
typedef int (*zend_object_get_class_name_t)(const zval *object, const char **class_name, zend_uint *class_name_len, int parent );
typedef int (*zend_object_compare_t)(zval *object1, zval *object2 );



typedef int (*zend_object_cast_t)(zval *readobj, zval *retval, int type );



typedef int (*zend_object_count_elements_t)(zval *object, long *count );

typedef int (*zend_object_get_closure_t)(zval *obj, zend_class_entry **ce_ptr, union _zend_function **fptr_ptr, zval **zobj_ptr );

typedef HashTable *(*zend_object_get_gc_t)(zval *object, zval ***table, int *n );

struct _zend_object_handlers {

 zend_object_add_ref_t add_ref;
 zend_object_del_ref_t del_ref;
 zend_object_clone_obj_t clone_obj;

 zend_object_read_property_t read_property;
 zend_object_write_property_t write_property;
 zend_object_read_dimension_t read_dimension;
 zend_object_write_dimension_t write_dimension;
 zend_object_get_property_ptr_ptr_t get_property_ptr_ptr;
 zend_object_get_t get;
 zend_object_set_t set;
 zend_object_has_property_t has_property;
 zend_object_unset_property_t unset_property;
 zend_object_has_dimension_t has_dimension;
 zend_object_unset_dimension_t unset_dimension;
 zend_object_get_properties_t get_properties;
 zend_object_get_method_t get_method;
 zend_object_call_method_t call_method;
 zend_object_get_constructor_t get_constructor;
 zend_object_get_class_entry_t get_class_entry;
 zend_object_get_class_name_t get_class_name;
 zend_object_compare_t compare_objects;
 zend_object_cast_t cast_object;
 zend_object_count_elements_t count_elements;
 zend_object_get_debug_info_t get_debug_info;
 zend_object_get_closure_t get_closure;
 zend_object_get_gc_t get_gc;
};

extern __attribute__ ((visibility("default"))) zend_object_handlers std_object_handlers;


__attribute__ ((visibility("default"))) union _zend_function *zend_std_get_static_method(zend_class_entry *ce, const char *function_name_strval, int function_name_strlen, const struct _zend_literal *key );
__attribute__ ((visibility("default"))) zval **zend_std_get_static_property(zend_class_entry *ce, const char *property_name, int property_name_len, zend_bool silent, const struct _zend_literal *key );
__attribute__ ((visibility("default"))) zend_bool zend_std_unset_static_property(zend_class_entry *ce, const char *property_name, int property_name_len, const struct _zend_literal *key );
__attribute__ ((visibility("default"))) union _zend_function *zend_std_get_constructor(zval *object );
__attribute__ ((visibility("default"))) struct _zend_property_info *zend_get_property_info(zend_class_entry *ce, zval *member, int silent );
__attribute__ ((visibility("default"))) HashTable *zend_std_get_properties(zval *object );
__attribute__ ((visibility("default"))) HashTable *zend_std_get_debug_info(zval *object, int *is_temp );
__attribute__ ((visibility("default"))) int zend_std_cast_object_tostring(zval *readobj, zval *writeobj, int type );
__attribute__ ((visibility("default"))) void zend_std_write_property(zval *object, zval *member, zval *value, const struct _zend_literal *key );
__attribute__ ((visibility("default"))) void rebuild_object_properties(zend_object *zobj);





__attribute__ ((visibility("default"))) int zend_check_private(union _zend_function *fbc, zend_class_entry *ce, char *function_name_strval, int function_name_strlen );

__attribute__ ((visibility("default"))) int zend_check_protected(zend_class_entry *ce, zend_class_entry *scope);

__attribute__ ((visibility("default"))) int zend_check_property_access(zend_object *zobj, const char *prop_info_name, int prop_info_name_len );

__attribute__ ((visibility("default"))) void zend_std_call_user_call(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

# 306 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2

typedef union _zvalue_value {
 long lval;
 double dval;
 struct {
  char *val;
  int len;
 } str;
 HashTable *ht;
 zend_object_value obj;
} zvalue_value;

struct _zval_struct {

 zvalue_value value;
 zend_uint refcount__gc;
 zend_uchar type;
 zend_uchar is_ref__gc;
};
# 378 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
static inline __attribute__((always_inline)) zend_uint zval_refcount_p(zval* pz) {
 return pz->refcount__gc;
}

static inline __attribute__((always_inline)) zend_uint zval_set_refcount_p(zval* pz, zend_uint rc) {
 return pz->refcount__gc = rc;
}

static inline __attribute__((always_inline)) zend_uint zval_addref_p(zval* pz) {
 return ++pz->refcount__gc;
}

static inline __attribute__((always_inline)) zend_uint zval_delref_p(zval* pz) {
 return --pz->refcount__gc;
}

static inline __attribute__((always_inline)) zend_bool zval_isref_p(zval* pz) {
 return pz->is_ref__gc;
}

static inline __attribute__((always_inline)) zend_bool zval_set_isref_p(zval* pz) {
 return pz->is_ref__gc = 1;
}

static inline __attribute__((always_inline)) zend_bool zval_unset_isref_p(zval* pz) {
 return pz->is_ref__gc = 0;
}

static inline __attribute__((always_inline)) zend_bool zval_set_isref_to_p(zval* pz, zend_bool isref) {
 return pz->is_ref__gc = isref;
}





union _zend_function;

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_iterators.h" 1
# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_iterators.h"
typedef struct _zend_object_iterator zend_object_iterator;

typedef struct _zend_object_iterator_funcs {

 void (*dtor)(zend_object_iterator *iter );


 int (*valid)(zend_object_iterator *iter );


 void (*get_current_data)(zend_object_iterator *iter, zval ***data );


 int (*get_current_key)(zend_object_iterator *iter, char **str_key, uint *str_key_len, ulong *int_key );


 void (*move_forward)(zend_object_iterator *iter );


 void (*rewind)(zend_object_iterator *iter );


 void (*invalidate_current)(zend_object_iterator *iter );
} zend_object_iterator_funcs;

struct _zend_object_iterator {
 void *data;
 zend_object_iterator_funcs *funcs;
 ulong index;
};

typedef struct _zend_class_iterator_funcs {
 zend_object_iterator_funcs *funcs;
 union _zend_function *zf_new_iterator;
 union _zend_function *zf_valid;
 union _zend_function *zf_current;
 union _zend_function *zf_key;
 union _zend_function *zf_next;
 union _zend_function *zf_rewind;
} zend_class_iterator_funcs;

enum zend_object_iterator_kind {
 ZEND_ITER_INVALID,
 ZEND_ITER_PLAIN_ARRAY,
 ZEND_ITER_PLAIN_OBJECT,
 ZEND_ITER_OBJECT
};



__attribute__ ((visibility("default"))) enum zend_object_iterator_kind zend_iterator_unwrap(zval *array_ptr, zend_object_iterator **iter );


__attribute__ ((visibility("default"))) zval *zend_iterator_wrap(zend_object_iterator *iter );

__attribute__ ((visibility("default"))) void zend_register_iterator_wrapper(void);

# 417 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2

struct _zend_serialize_data;
struct _zend_unserialize_data;

typedef struct _zend_serialize_data zend_serialize_data;
typedef struct _zend_unserialize_data zend_unserialize_data;

struct _zend_trait_method_reference {
 const char* method_name;
 unsigned int mname_len;

 zend_class_entry *ce;

 const char* class_name;
 unsigned int cname_len;
};
typedef struct _zend_trait_method_reference zend_trait_method_reference;

struct _zend_trait_precedence {
 zend_trait_method_reference *trait_method;

 zend_class_entry** exclude_from_classes;

 union _zend_function* function;
};
typedef struct _zend_trait_precedence zend_trait_precedence;

struct _zend_trait_alias {
 zend_trait_method_reference *trait_method;




 const char* alias;
 unsigned int alias_len;




 zend_uint modifiers;

 union _zend_function* function;
};
typedef struct _zend_trait_alias zend_trait_alias;

struct _zend_class_entry {
 char type;
 const char *name;
 zend_uint name_length;
 struct _zend_class_entry *parent;
 int refcount;
 zend_uint ce_flags;

 HashTable function_table;
 HashTable properties_info;
 zval **default_properties_table;
 zval **default_static_members_table;
 zval **static_members_table;
 HashTable constants_table;
 int default_properties_count;
 int default_static_members_count;

 union _zend_function *constructor;
 union _zend_function *destructor;
 union _zend_function *clone;
 union _zend_function *__get;
 union _zend_function *__set;
 union _zend_function *__unset;
 union _zend_function *__isset;
 union _zend_function *__call;
 union _zend_function *__callstatic;
 union _zend_function *__tostring;
 union _zend_function *serialize_func;
 union _zend_function *unserialize_func;

 zend_class_iterator_funcs iterator_funcs;


 zend_object_value (*create_object)(zend_class_entry *class_type );
 zend_object_iterator *(*get_iterator)(zend_class_entry *ce, zval *object, int by_ref );
 int (*interface_gets_implemented)(zend_class_entry *iface, zend_class_entry *class_type );
 union _zend_function *(*get_static_method)(zend_class_entry *ce, char* method, int method_len );


 int (*serialize)(zval *object, unsigned char **buffer, zend_uint *buf_len, zend_serialize_data *data );
 int (*unserialize)(zval **object, zend_class_entry *ce, const unsigned char *buf, zend_uint buf_len, zend_unserialize_data *data );

 zend_class_entry **interfaces;
 zend_uint num_interfaces;

 zend_class_entry **traits;
 zend_uint num_traits;
 zend_trait_alias **trait_aliases;
 zend_trait_precedence **trait_precedences;

 union {
  struct {
   const char *filename;
   zend_uint line_start;
   zend_uint line_end;
   const char *doc_comment;
   zend_uint doc_comment_len;
  } user;
  struct {
   const struct _zend_function_entry *builtin_functions;
   struct _zend_module_entry *module;
  } internal;
 } info;
};

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_stream.h" 1
# 30 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_stream.h"
typedef size_t (*zend_stream_fsizer_t)(void* handle );
typedef size_t (*zend_stream_reader_t)(void* handle, char *buf, size_t len );
typedef void (*zend_stream_closer_t)(void* handle );



typedef enum {
 ZEND_HANDLE_FILENAME,
 ZEND_HANDLE_FD,
 ZEND_HANDLE_FP,
 ZEND_HANDLE_STREAM,
 ZEND_HANDLE_MAPPED
} zend_stream_type;

typedef struct _zend_mmap {
 size_t len;
 size_t pos;
 void *map;
 char *buf;
 void *old_handle;
 zend_stream_closer_t old_closer;
} zend_mmap;

typedef struct _zend_stream {
 void *handle;
 int isatty;
 zend_mmap mmap;
 zend_stream_reader_t reader;
 zend_stream_fsizer_t fsizer;
 zend_stream_closer_t closer;
} zend_stream;

typedef struct _zend_file_handle {
 zend_stream_type type;
 const char *filename;
 char *opened_path;
 union {
  int fd;
  FILE *fp;
  zend_stream stream;
 } handle;
 zend_bool free_filename;
} zend_file_handle;


__attribute__ ((visibility("default"))) int zend_stream_open(const char *filename, zend_file_handle *handle );
__attribute__ ((visibility("default"))) int zend_stream_fixup(zend_file_handle *file_handle, char **buf, size_t *len );
__attribute__ ((visibility("default"))) void zend_file_handle_dtor(zend_file_handle *fh );
__attribute__ ((visibility("default"))) int zend_compare_file_handles(zend_file_handle *fh1, zend_file_handle *fh2);

# 528 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
typedef struct _zend_utility_functions {
 void (*error_function)(int type, const char *error_filename, const uint error_lineno, const char *format, va_list args) __attribute__ ((format(printf, 4, 0)));
 int (*printf_function)(const char *format, ...) __attribute__ ((format(printf, 1, 2)));
 int (*write_function)(const char *str, uint str_length);
 FILE *(*fopen_function)(const char *filename, char **opened_path );
 void (*message_handler)(long message, const void *data );




 int (*get_configuration_directive)(const char *name, uint name_length, zval *contents);
 void (*ticks_function)(int ticks);
 void (*on_timeout)(int seconds );
 int (*stream_open_function)(const char *filename, zend_file_handle *handle );
 int (*vspprintf_function)(char **pbuf, size_t max_len, const char *format, va_list ap);
 char *(*getenv_function)(char *name, size_t name_len );
 char *(*resolve_path_function)(const char *filename, int filename_len );
} zend_utility_functions;

typedef struct _zend_utility_values {
 char *import_use_extension;
 uint import_use_extension_length;
 zend_bool html_errors;
} zend_utility_values;

typedef int (*zend_write_func_t)(const char *str, uint str_length);
# 593 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
int zend_startup(zend_utility_functions *utility_functions, char **extensions );
void zend_shutdown(void);
void zend_register_standard_ini_entries(void);
void zend_post_startup(void);
void zend_set_utility_values(zend_utility_values *utility_values);


__attribute__ ((visibility("default"))) void _zend_bailout(char *filename, uint lineno);

# 631 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"

__attribute__ ((visibility("default"))) char *get_zend_version(void);
__attribute__ ((visibility("default"))) void zend_make_printable_zval(zval *expr, zval *expr_copy, int *use_copy);
__attribute__ ((visibility("default"))) int zend_print_zval(zval *expr, int indent);
__attribute__ ((visibility("default"))) int zend_print_zval_ex(zend_write_func_t write_func, zval *expr, int indent);
__attribute__ ((visibility("default"))) void zend_print_zval_r(zval *expr, int indent );
__attribute__ ((visibility("default"))) void zend_print_flat_zval_r(zval *expr );
__attribute__ ((visibility("default"))) void zend_print_zval_r_ex(zend_write_func_t write_func, zval *expr, int indent );
__attribute__ ((visibility("default"))) void zend_output_debug_string(zend_bool trigger_break, const char *format, ...) __attribute__ ((format(printf, 2, 3)));


void zend_activate(void);
void zend_deactivate(void);
void zend_call_destructors(void);
void zend_activate_modules(void);
void zend_deactivate_modules(void);
void zend_post_deactivate_modules(void);








__attribute__ ((visibility("default"))) void free_estring(char **str_p);

# 676 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"

extern __attribute__ ((visibility("default"))) int (*zend_printf)(const char *format, ...) __attribute__ ((format(printf, 1, 2)));
extern __attribute__ ((visibility("default"))) zend_write_func_t zend_write;
extern __attribute__ ((visibility("default"))) FILE *(*zend_fopen)(const char *filename, char **opened_path );




extern __attribute__ ((visibility("default"))) void (*zend_ticks_function)(int ticks);
extern __attribute__ ((visibility("default"))) void (*zend_error_cb)(int type, const char *error_filename, const uint error_lineno, const char *format, va_list args) __attribute__ ((format(printf, 4, 0)));
extern __attribute__ ((visibility("default"))) void (*zend_on_timeout)(int seconds );
extern __attribute__ ((visibility("default"))) int (*zend_stream_open_function)(const char *filename, zend_file_handle *handle );
extern int (*zend_vspprintf)(char **pbuf, size_t max_len, const char *format, va_list ap);
extern __attribute__ ((visibility("default"))) char *(*zend_getenv)(char *name, size_t name_len );
extern __attribute__ ((visibility("default"))) char *(*zend_resolve_path)(const char *filename, int filename_len );

__attribute__ ((visibility("default"))) void zend_error(int type, const char *format, ...) __attribute__ ((format(printf, 2, 3)));

void zenderror(const char *error);



extern __attribute__ ((visibility("default"))) zend_class_entry *zend_standard_class_def;
extern __attribute__ ((visibility("default"))) zend_utility_values zend_uv;
extern __attribute__ ((visibility("default"))) zval zval_used_for_init;









# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_signal.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_signal.h"
# 1 "/usr/include/signal.h" 1 3 4
# 31 "/usr/include/signal.h" 3 4


# 1 "/usr/include/bits/sigset.h" 1 3 4
# 104 "/usr/include/bits/sigset.h" 3 4
extern int __sigismember (__const __sigset_t *, int);
extern int __sigaddset (__sigset_t *, int);
extern int __sigdelset (__sigset_t *, int);
# 118 "/usr/include/bits/sigset.h" 3 4
extern __inline int __sigismember (__const __sigset_t *__set, int __sig) { unsigned long int __mask = (((unsigned long int) 1) << (((__sig) - 1) % (8 * sizeof (unsigned long int)))); unsigned long int __word = (((__sig) - 1) / (8 * sizeof (unsigned long int))); return (__set->__val[__word] & __mask) ? 1 : 0; }
extern __inline int __sigaddset ( __sigset_t *__set, int __sig) { unsigned long int __mask = (((unsigned long int) 1) << (((__sig) - 1) % (8 * sizeof (unsigned long int)))); unsigned long int __word = (((__sig) - 1) / (8 * sizeof (unsigned long int))); return ((__set->__val[__word] |= __mask), 0); }
extern __inline int __sigdelset ( __sigset_t *__set, int __sig) { unsigned long int __mask = (((unsigned long int) 1) << (((__sig) - 1) % (8 * sizeof (unsigned long int)))); unsigned long int __word = (((__sig) - 1) / (8 * sizeof (unsigned long int))); return ((__set->__val[__word] &= ~__mask), 0); }
# 34 "/usr/include/signal.h" 2 3 4







typedef __sig_atomic_t sig_atomic_t;

# 58 "/usr/include/signal.h" 3 4
# 1 "/usr/include/bits/signum.h" 1 3 4
# 59 "/usr/include/signal.h" 2 3 4
# 76 "/usr/include/signal.h" 3 4
# 1 "/usr/include/time.h" 1 3 4
# 77 "/usr/include/signal.h" 2 3 4


# 1 "/usr/include/bits/siginfo.h" 1 3 4
# 25 "/usr/include/bits/siginfo.h" 3 4
# 1 "/usr/include/bits/wordsize.h" 1 3 4
# 26 "/usr/include/bits/siginfo.h" 2 3 4







typedef union sigval
  {
    int sival_int;
    void *sival_ptr;
  } sigval_t;
# 51 "/usr/include/bits/siginfo.h" 3 4
typedef struct siginfo
  {
    int si_signo;
    int si_errno;

    int si_code;

    union
      {
 int _pad[((128 / sizeof (int)) - 3)];


 struct
   {
     __pid_t si_pid;
     __uid_t si_uid;
   } _kill;


 struct
   {
     int si_tid;
     int si_overrun;
     sigval_t si_sigval;
   } _timer;


 struct
   {
     __pid_t si_pid;
     __uid_t si_uid;
     sigval_t si_sigval;
   } _rt;


 struct
   {
     __pid_t si_pid;
     __uid_t si_uid;
     int si_status;
     __clock_t si_utime;
     __clock_t si_stime;
   } _sigchld;


 struct
   {
     void *si_addr;
   } _sigfault;


 struct
   {
     long int si_band;
     int si_fd;
   } _sigpoll;
      } _sifields;
  } siginfo_t;
# 129 "/usr/include/bits/siginfo.h" 3 4
enum
{
  SI_ASYNCNL = -60,

  SI_TKILL = -6,

  SI_SIGIO,

  SI_ASYNCIO,

  SI_MESGQ,

  SI_TIMER,

  SI_QUEUE,

  SI_USER,

  SI_KERNEL = 0x80

};



enum
{
  ILL_ILLOPC = 1,

  ILL_ILLOPN,

  ILL_ILLADR,

  ILL_ILLTRP,

  ILL_PRVOPC,

  ILL_PRVREG,

  ILL_COPROC,

  ILL_BADSTK

};


enum
{
  FPE_INTDIV = 1,

  FPE_INTOVF,

  FPE_FLTDIV,

  FPE_FLTOVF,

  FPE_FLTUND,

  FPE_FLTRES,

  FPE_FLTINV,

  FPE_FLTSUB

};


enum
{
  SEGV_MAPERR = 1,

  SEGV_ACCERR

};


enum
{
  BUS_ADRALN = 1,

  BUS_ADRERR,

  BUS_OBJERR

};


enum
{
  TRAP_BRKPT = 1,

  TRAP_TRACE

};


enum
{
  CLD_EXITED = 1,

  CLD_KILLED,

  CLD_DUMPED,

  CLD_TRAPPED,

  CLD_STOPPED,

  CLD_CONTINUED

};


enum
{
  POLL_IN = 1,

  POLL_OUT,

  POLL_MSG,

  POLL_ERR,

  POLL_PRI,

  POLL_HUP

};
# 273 "/usr/include/bits/siginfo.h" 3 4
typedef struct sigevent
  {
    sigval_t sigev_value;
    int sigev_signo;
    int sigev_notify;

    union
      {
 int _pad[((64 / sizeof (int)) - 3)];



 __pid_t _tid;

 struct
   {
     void (*_function) (sigval_t);
     void *_attribute;
   } _sigev_thread;
      } _sigev_un;
  } sigevent_t;






enum
{
  SIGEV_SIGNAL = 0,

  SIGEV_NONE,

  SIGEV_THREAD,


  SIGEV_THREAD_ID = 4

};
# 80 "/usr/include/signal.h" 2 3 4




typedef void (*__sighandler_t) (int);




extern __sighandler_t __sysv_signal (int __sig, __sighandler_t __handler)
     __attribute__ ((__nothrow__));

extern __sighandler_t sysv_signal (int __sig, __sighandler_t __handler)
     __attribute__ ((__nothrow__));







extern __sighandler_t signal (int __sig, __sighandler_t __handler)
     __attribute__ ((__nothrow__));
# 113 "/usr/include/signal.h" 3 4





extern __sighandler_t bsd_signal (int __sig, __sighandler_t __handler)
     __attribute__ ((__nothrow__));






extern int kill (__pid_t __pid, int __sig) __attribute__ ((__nothrow__));






extern int killpg (__pid_t __pgrp, int __sig) __attribute__ ((__nothrow__));




extern int raise (int __sig) __attribute__ ((__nothrow__));




extern __sighandler_t ssignal (int __sig, __sighandler_t __handler)
     __attribute__ ((__nothrow__));
extern int gsignal (int __sig) __attribute__ ((__nothrow__));




extern void psignal (int __sig, __const char *__s);




extern void psiginfo (__const siginfo_t *__pinfo, __const char *__s);
# 168 "/usr/include/signal.h" 3 4
extern int __sigpause (int __sig_or_mask, int __is_sig);
# 177 "/usr/include/signal.h" 3 4
extern int sigpause (int __sig) __asm__ ("__xpg_sigpause");
# 196 "/usr/include/signal.h" 3 4
extern int sigblock (int __mask) __attribute__ ((__nothrow__)) __attribute__ ((__deprecated__));


extern int sigsetmask (int __mask) __attribute__ ((__nothrow__)) __attribute__ ((__deprecated__));


extern int siggetmask (void) __attribute__ ((__nothrow__)) __attribute__ ((__deprecated__));
# 211 "/usr/include/signal.h" 3 4
typedef __sighandler_t sighandler_t;




typedef __sighandler_t sig_t;





extern int sigemptyset (sigset_t *__set) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int sigfillset (sigset_t *__set) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int sigaddset (sigset_t *__set, int __signo) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int sigdelset (sigset_t *__set, int __signo) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int sigismember (__const sigset_t *__set, int __signo)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern int sigisemptyset (__const sigset_t *__set) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int sigandset (sigset_t *__set, __const sigset_t *__left,
        __const sigset_t *__right) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2, 3)));


extern int sigorset (sigset_t *__set, __const sigset_t *__left,
       __const sigset_t *__right) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2, 3)));




# 1 "/usr/include/bits/sigaction.h" 1 3 4
# 25 "/usr/include/bits/sigaction.h" 3 4
struct sigaction
  {


    union
      {

 __sighandler_t sa_handler;

 void (*sa_sigaction) (int, siginfo_t *, void *);
      }
    __sigaction_handler;







    __sigset_t sa_mask;


    int sa_flags;


    void (*sa_restorer) (void);
  };
# 253 "/usr/include/signal.h" 2 3 4


extern int sigprocmask (int __how, __const sigset_t *__restrict __set,
   sigset_t *__restrict __oset) __attribute__ ((__nothrow__));






extern int sigsuspend (__const sigset_t *__set) __attribute__ ((__nonnull__ (1)));


extern int sigaction (int __sig, __const struct sigaction *__restrict __act,
        struct sigaction *__restrict __oact) __attribute__ ((__nothrow__));


extern int sigpending (sigset_t *__set) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern int sigwait (__const sigset_t *__restrict __set, int *__restrict __sig)
     __attribute__ ((__nonnull__ (1, 2)));






extern int sigwaitinfo (__const sigset_t *__restrict __set,
   siginfo_t *__restrict __info) __attribute__ ((__nonnull__ (1)));






extern int sigtimedwait (__const sigset_t *__restrict __set,
    siginfo_t *__restrict __info,
    __const struct timespec *__restrict __timeout)
     __attribute__ ((__nonnull__ (1)));



extern int sigqueue (__pid_t __pid, int __sig, __const union sigval __val)
     __attribute__ ((__nothrow__));
# 310 "/usr/include/signal.h" 3 4
extern __const char *__const _sys_siglist[65];
extern __const char *__const sys_siglist[65];


struct sigvec
  {
    __sighandler_t sv_handler;
    int sv_mask;

    int sv_flags;

  };
# 334 "/usr/include/signal.h" 3 4
extern int sigvec (int __sig, __const struct sigvec *__vec,
     struct sigvec *__ovec) __attribute__ ((__nothrow__));



# 1 "/usr/include/bits/sigcontext.h" 1 3 4
# 28 "/usr/include/bits/sigcontext.h" 3 4
# 1 "/usr/include/asm/sigcontext.h" 1 3 4




# 1 "/usr/include/linux/types.h" 1 3 4



# 1 "/usr/include/asm/types.h" 1 3 4





# 1 "/usr/include/asm-generic/types.h" 1 3 4






# 1 "/usr/include/asm-generic/int-ll64.h" 1 3 4
# 11 "/usr/include/asm-generic/int-ll64.h" 3 4
# 1 "/usr/include/asm/bitsperlong.h" 1 3 4
# 10 "/usr/include/asm/bitsperlong.h" 3 4
# 1 "/usr/include/asm-generic/bitsperlong.h" 1 3 4
# 11 "/usr/include/asm/bitsperlong.h" 2 3 4
# 12 "/usr/include/asm-generic/int-ll64.h" 2 3 4







typedef __signed__ char __s8;
typedef unsigned char __u8;

typedef __signed__ short __s16;
typedef unsigned short __u16;

typedef __signed__ int __s32;
typedef unsigned int __u32;


__extension__ typedef __signed__ long long __s64;
__extension__ typedef unsigned long long __u64;
# 8 "/usr/include/asm-generic/types.h" 2 3 4



typedef unsigned short umode_t;
# 7 "/usr/include/asm/types.h" 2 3 4
# 5 "/usr/include/linux/types.h" 2 3 4



# 1 "/usr/include/linux/posix_types.h" 1 3 4



# 1 "/usr/include/linux/stddef.h" 1 3 4
# 5 "/usr/include/linux/posix_types.h" 2 3 4
# 36 "/usr/include/linux/posix_types.h" 3 4
typedef struct {
 unsigned long fds_bits [(1024/(8 * sizeof(unsigned long)))];
} __kernel_fd_set;


typedef void (*__kernel_sighandler_t)(int);


typedef int __kernel_key_t;
typedef int __kernel_mqd_t;

# 1 "/usr/include/asm/posix_types.h" 1 3 4

# 1 "/usr/include/asm/posix_types_32.h" 1 3 4
# 10 "/usr/include/asm/posix_types_32.h" 3 4
typedef unsigned long __kernel_ino_t;
typedef unsigned short __kernel_mode_t;
typedef unsigned short __kernel_nlink_t;
typedef long __kernel_off_t;
typedef int __kernel_pid_t;
typedef unsigned short __kernel_ipc_pid_t;
typedef unsigned short __kernel_uid_t;
typedef unsigned short __kernel_gid_t;
typedef unsigned int __kernel_size_t;
typedef int __kernel_ssize_t;
typedef int __kernel_ptrdiff_t;
typedef long __kernel_time_t;
typedef long __kernel_suseconds_t;
typedef long __kernel_clock_t;
typedef int __kernel_timer_t;
typedef int __kernel_clockid_t;
typedef int __kernel_daddr_t;
typedef char * __kernel_caddr_t;
typedef unsigned short __kernel_uid16_t;
typedef unsigned short __kernel_gid16_t;
typedef unsigned int __kernel_uid32_t;
typedef unsigned int __kernel_gid32_t;

typedef unsigned short __kernel_old_uid_t;
typedef unsigned short __kernel_old_gid_t;
typedef unsigned short __kernel_old_dev_t;


typedef long long __kernel_loff_t;


typedef struct {
 int val[2];
} __kernel_fsid_t;
# 3 "/usr/include/asm/posix_types.h" 2 3 4
# 48 "/usr/include/linux/posix_types.h" 2 3 4
# 9 "/usr/include/linux/types.h" 2 3 4
# 27 "/usr/include/linux/types.h" 3 4
typedef __u16 __le16;
typedef __u16 __be16;
typedef __u32 __le32;
typedef __u32 __be32;
typedef __u64 __le64;
typedef __u64 __be64;

typedef __u16 __sum16;
typedef __u32 __wsum;
# 6 "/usr/include/asm/sigcontext.h" 2 3 4
# 23 "/usr/include/asm/sigcontext.h" 3 4
struct _fpx_sw_bytes {
 __u32 magic1;
 __u32 extended_size;


 __u64 xstate_bv;




 __u32 xstate_size;




 __u32 padding[7];
};
# 56 "/usr/include/asm/sigcontext.h" 3 4
struct _fpreg {
 unsigned short significand[4];
 unsigned short exponent;
};

struct _fpxreg {
 unsigned short significand[4];
 unsigned short exponent;
 unsigned short padding[3];
};

struct _xmmreg {
 unsigned long element[4];
};

struct _fpstate {

 unsigned long cw;
 unsigned long sw;
 unsigned long tag;
 unsigned long ipoff;
 unsigned long cssel;
 unsigned long dataoff;
 unsigned long datasel;
 struct _fpreg _st[8];
 unsigned short status;
 unsigned short magic;


 unsigned long _fxsr_env[6];
 unsigned long mxcsr;
 unsigned long reserved;
 struct _fpxreg _fxsr_st[8];
 struct _xmmreg _xmm[8];
 unsigned long padding1[44];

 union {
  unsigned long padding2[12];
  struct _fpx_sw_bytes sw_reserved;

 };
};






struct sigcontext {
 unsigned short gs, __gsh;
 unsigned short fs, __fsh;
 unsigned short es, __esh;
 unsigned short ds, __dsh;
 unsigned long edi;
 unsigned long esi;
 unsigned long ebp;
 unsigned long esp;
 unsigned long ebx;
 unsigned long edx;
 unsigned long ecx;
 unsigned long eax;
 unsigned long trapno;
 unsigned long err;
 unsigned long eip;
 unsigned short cs, __csh;
 unsigned long eflags;
 unsigned long esp_at_signal;
 unsigned short ss, __ssh;
 struct _fpstate *fpstate;
 unsigned long oldmask;
 unsigned long cr2;
};
# 190 "/usr/include/asm/sigcontext.h" 3 4
struct _xsave_hdr {
 __u64 xstate_bv;
 __u64 reserved1[2];
 __u64 reserved2[5];
};

struct _ymmh_state {

 __u32 ymmh_space[64];
};







struct _xstate {
 struct _fpstate fpstate;
 struct _xsave_hdr xstate_hdr;
 struct _ymmh_state ymmh;

};
# 29 "/usr/include/bits/sigcontext.h" 2 3 4
# 340 "/usr/include/signal.h" 2 3 4


extern int sigreturn (struct sigcontext *__scp) __attribute__ ((__nothrow__));






# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 350 "/usr/include/signal.h" 2 3 4




extern int siginterrupt (int __sig, int __interrupt) __attribute__ ((__nothrow__));

# 1 "/usr/include/bits/sigstack.h" 1 3 4
# 26 "/usr/include/bits/sigstack.h" 3 4
struct sigstack
  {
    void *ss_sp;
    int ss_onstack;
  };



enum
{
  SS_ONSTACK = 1,

  SS_DISABLE

};
# 50 "/usr/include/bits/sigstack.h" 3 4
typedef struct sigaltstack
  {
    void *ss_sp;
    int ss_flags;
    size_t ss_size;
  } stack_t;
# 357 "/usr/include/signal.h" 2 3 4


# 1 "/usr/include/sys/ucontext.h" 1 3 4
# 23 "/usr/include/sys/ucontext.h" 3 4
# 1 "/usr/include/signal.h" 1 3 4
# 24 "/usr/include/sys/ucontext.h" 2 3 4



# 1 "/usr/include/bits/sigcontext.h" 1 3 4
# 28 "/usr/include/sys/ucontext.h" 2 3 4



typedef int greg_t;





typedef greg_t gregset_t[19];



enum
{
  REG_GS = 0,

  REG_FS,

  REG_ES,

  REG_DS,

  REG_EDI,

  REG_ESI,

  REG_EBP,

  REG_ESP,

  REG_EBX,

  REG_EDX,

  REG_ECX,

  REG_EAX,

  REG_TRAPNO,

  REG_ERR,

  REG_EIP,

  REG_CS,

  REG_EFL,

  REG_UESP,

  REG_SS

};



struct _libc_fpreg
{
  unsigned short int significand[4];
  unsigned short int exponent;
};

struct _libc_fpstate
{
  unsigned long int cw;
  unsigned long int sw;
  unsigned long int tag;
  unsigned long int ipoff;
  unsigned long int cssel;
  unsigned long int dataoff;
  unsigned long int datasel;
  struct _libc_fpreg _st[8];
  unsigned long int status;
};


typedef struct _libc_fpstate *fpregset_t;


typedef struct
  {
    gregset_t gregs;


    fpregset_t fpregs;
    unsigned long int oldmask;
    unsigned long int cr2;
  } mcontext_t;


typedef struct ucontext
  {
    unsigned long int uc_flags;
    struct ucontext *uc_link;
    stack_t uc_stack;
    mcontext_t uc_mcontext;
    __sigset_t uc_sigmask;
    struct _libc_fpstate __fpregs_mem;
  } ucontext_t;
# 360 "/usr/include/signal.h" 2 3 4





extern int sigstack (struct sigstack *__ss, struct sigstack *__oss)
     __attribute__ ((__nothrow__)) __attribute__ ((__deprecated__));



extern int sigaltstack (__const struct sigaltstack *__restrict __ss,
   struct sigaltstack *__restrict __oss) __attribute__ ((__nothrow__));







extern int sighold (int __sig) __attribute__ ((__nothrow__));


extern int sigrelse (int __sig) __attribute__ ((__nothrow__));


extern int sigignore (int __sig) __attribute__ ((__nothrow__));


extern __sighandler_t sigset (int __sig, __sighandler_t __disp) __attribute__ ((__nothrow__));






# 1 "/usr/include/bits/sigthread.h" 1 3 4
# 31 "/usr/include/bits/sigthread.h" 3 4
extern int pthread_sigmask (int __how,
       __const __sigset_t *__restrict __newmask,
       __sigset_t *__restrict __oldmask)__attribute__ ((__nothrow__));


extern int pthread_kill (pthread_t __threadid, int __signo) __attribute__ ((__nothrow__));



extern int pthread_sigqueue (pthread_t __threadid, int __signo,
        const union sigval __value) __attribute__ ((__nothrow__));
# 396 "/usr/include/signal.h" 2 3 4






extern int __libc_current_sigrtmin (void) __attribute__ ((__nothrow__));

extern int __libc_current_sigrtmax (void) __attribute__ ((__nothrow__));




# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_signal.h" 2
# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_signal.h"
typedef struct _zend_signal_entry_t {
 int flags;
 void* handler;
} zend_signal_entry_t;

typedef struct _zend_signal_t {
 int signo;
 siginfo_t *siginfo;
 void* context;
} zend_signal_t;

typedef struct _zend_signal_queue_t {
 zend_signal_t zend_signal;
 struct _zend_signal_queue_t *next;
} zend_signal_queue_t;


typedef struct _zend_signal_globals_t {
 int depth;
 int blocked;
 int running;
 int active;
 int initialized;
 zend_bool check;
 zend_signal_entry_t handlers[65];
 zend_signal_queue_t pstorage[32], *phead, *ptail, *pavail;
} zend_signal_globals_t;
# 74 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_signal.h"
extern __attribute__ ((visibility("default"))) zend_signal_globals_t zend_signal_globals;






void zend_signal_handler_defer(int signo, siginfo_t *siginfo, void *context);
__attribute__ ((visibility("default"))) void zend_signal_handler_unblock();
void zend_signal_activate(void);
void zend_signal_deactivate(void);
void zend_signal_startup();
void zend_signal_shutdown(void);
__attribute__ ((visibility("default"))) int zend_signal(int signo, void (*handler)(int) );
__attribute__ ((visibility("default"))) int zend_sigaction(int signo, const struct sigaction *act, struct sigaction *oldact );
# 711 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2






__attribute__ ((visibility("default"))) void zend_message_dispatcher(long message, const void *data );

__attribute__ ((visibility("default"))) int zend_get_configuration_directive(const char *name, uint name_length, zval *contents);

# 829 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_gc.h" 1
# 81 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_gc.h"
typedef struct _gc_root_buffer {
 struct _gc_root_buffer *prev;
 struct _gc_root_buffer *next;
 zend_object_handle handle;
 union {
  zval *pz;
  const zend_object_handlers *handlers;
 } u;
} gc_root_buffer;

typedef struct _zval_gc_info {
 zval z;
 union {
  gc_root_buffer *buffered;
  struct _zval_gc_info *next;
 } u;
} zval_gc_info;

typedef struct _zend_gc_globals {
 zend_bool gc_enabled;
 zend_bool gc_active;

 gc_root_buffer *buf;
 gc_root_buffer roots;
 gc_root_buffer *unused;
 gc_root_buffer *first_unused;
 gc_root_buffer *last_unused;

 zval_gc_info *zval_to_free;
 zval_gc_info *free_list;
 zval_gc_info *next_to_free;

 zend_uint gc_runs;
 zend_uint collected;
# 129 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_gc.h"
} zend_gc_globals;
# 138 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_gc.h"
extern __attribute__ ((visibility("default"))) zend_gc_globals gc_globals;



__attribute__ ((visibility("default"))) int gc_collect_cycles(void);
__attribute__ ((visibility("default"))) void gc_zval_possible_root(zval *zv );
__attribute__ ((visibility("default"))) void gc_zobj_possible_root(zval *zv );
__attribute__ ((visibility("default"))) void gc_remove_zval_from_buffer(zval *zv );
__attribute__ ((visibility("default"))) void gc_globals_ctor(void);
__attribute__ ((visibility("default"))) void gc_globals_dtor(void);
__attribute__ ((visibility("default"))) void gc_init(void);
__attribute__ ((visibility("default"))) void gc_reset(void);

# 180 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_gc.h"
static inline __attribute__((always_inline)) void gc_zval_check_possible_root(zval *z )
{
 if (z->type == 4 || z->type == 5) {
  gc_zval_possible_root(z );
 }
}

static inline __attribute__((always_inline)) void gc_remove_from_buffer(gc_root_buffer *root )
{
 root->next->prev = root->prev;
 root->prev->next = root->next;
 root->prev = (gc_globals.unused);
 (gc_globals.unused) = root;
 ;
}
# 830 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
# 1 "/usr/include/errno.h" 1 3 4
# 32 "/usr/include/errno.h" 3 4




# 1 "/usr/include/bits/errno.h" 1 3 4
# 25 "/usr/include/bits/errno.h" 3 4
# 1 "/usr/include/linux/errno.h" 1 3 4



# 1 "/usr/include/asm/errno.h" 1 3 4
# 1 "/usr/include/asm-generic/errno.h" 1 3 4



# 1 "/usr/include/asm-generic/errno-base.h" 1 3 4
# 5 "/usr/include/asm-generic/errno.h" 2 3 4
# 1 "/usr/include/asm/errno.h" 2 3 4
# 5 "/usr/include/linux/errno.h" 2 3 4
# 26 "/usr/include/bits/errno.h" 2 3 4
# 47 "/usr/include/bits/errno.h" 3 4
extern int *__errno_location (void) __attribute__ ((__nothrow__)) __attribute__ ((__const__));
# 37 "/usr/include/errno.h" 2 3 4
# 55 "/usr/include/errno.h" 3 4
extern char *program_invocation_name, *program_invocation_short_name;




# 69 "/usr/include/errno.h" 3 4
typedef int error_t;
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 2

# 1 "/usr/include/assert.h" 1 3 4
# 66 "/usr/include/assert.h" 3 4



extern void __assert_fail (__const char *__assertion, __const char *__file,
      unsigned int __line, __const char *__function)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));


extern void __assert_perror_fail (int __errnum, __const char *__file,
      unsigned int __line,
      __const char *__function)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));




extern void __assert (const char *__assertion, const char *__file, int __line)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));



# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 2





# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_strtod.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_strtod.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_strtod.h" 2


__attribute__ ((visibility("default"))) void zend_freedtoa(char *s);
__attribute__ ((visibility("default"))) char * zend_dtoa(double _d, int mode, int ndigits, int *decpt, int *sign, char **rve);
__attribute__ ((visibility("default"))) double zend_strtod(const char *s00, const char **se);
__attribute__ ((visibility("default"))) double zend_hex_strtod(const char *str, const char **endptr);
__attribute__ ((visibility("default"))) double zend_oct_strtod(const char *str, const char **endptr);
__attribute__ ((visibility("default"))) double zend_bin_strtod(const char *str, const char **endptr);
__attribute__ ((visibility("default"))) int zend_startup_strtod(void);
__attribute__ ((visibility("default"))) int zend_shutdown_strtod(void);

# 34 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_multiply.h" 1
# 35 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 2








__attribute__ ((visibility("default"))) int add_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int sub_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int mul_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int div_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int mod_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int boolean_xor_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int boolean_not_function(zval *result, zval *op1 );
__attribute__ ((visibility("default"))) int bitwise_not_function(zval *result, zval *op1 );
__attribute__ ((visibility("default"))) int bitwise_or_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int bitwise_and_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int bitwise_xor_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int shift_left_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int shift_right_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int concat_function(zval *result, zval *op1, zval *op2 );

__attribute__ ((visibility("default"))) int is_equal_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int is_identical_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int is_not_identical_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int is_not_equal_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int is_smaller_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int is_smaller_or_equal_function(zval *result, zval *op1, zval *op2 );

__attribute__ ((visibility("default"))) zend_bool instanceof_function_ex(const zend_class_entry *instance_ce, const zend_class_entry *ce, zend_bool interfaces_only );
__attribute__ ((visibility("default"))) zend_bool instanceof_function(const zend_class_entry *instance_ce, const zend_class_entry *ce );





static inline __attribute__((always_inline)) long zend_dval_to_lval(double d)
{
 if (d > 2147483647L || d < (-2147483647L - 1L)) {
  return (long)(unsigned long)(zend_long64) d;
 }
 return (long) d;
}
# 105 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
static inline zend_uchar is_numeric_string(const char *str, int length, long *lval, double *dval, int allow_errors)
{
 const char *ptr;
 int base = 10, digits = 0, dp_or_e = 0;
 double local_dval;
 zend_uchar type;

 if (!length) {
  return 0;
 }



 while (*str == ' ' || *str == '\t' || *str == '\n' || *str == '\r' || *str == '\v' || *str == '\f') {
  str++;
  length--;
 }
 ptr = str;

 if (*ptr == '-' || *ptr == '+') {
  ptr++;
 }

 if (((*ptr) >= '0' && (*ptr) <= '9')) {


  if (length > 2 && *str == '0' && (str[1] == 'x' || str[1] == 'X')) {
   base = 16;
   ptr += 2;
  }


  while (*ptr == '0') {
   ptr++;
  }




  for (type = 1; !(digits >= 11 && (dval || allow_errors == 1)); digits++, ptr++) {
check_digits:
   if (((*ptr) >= '0' && (*ptr) <= '9') || (base == 16 && (((*ptr) >= 'A' && (*ptr) <= 'F') || ((*ptr) >= 'a' && (*ptr) <= 'f')))) {
    continue;
   } else if (base == 10) {
    if (*ptr == '.' && dp_or_e < 1) {
     goto process_double;
    } else if ((*ptr == 'e' || *ptr == 'E') && dp_or_e < 2) {
     const char *e = ptr + 1;

     if (*e == '-' || *e == '+') {
      ptr = e++;
     }
     if (((*e) >= '0' && (*e) <= '9')) {
      goto process_double;
     }
    }
   }

   break;
  }

  if (base == 10) {
   if (digits >= 11) {
    dp_or_e = -1;
    goto process_double;
   }
  } else if (!(digits < 4 * 2 || (digits == 4 * 2 && ptr[-digits] <= '7'))) {
   if (dval) {
    local_dval = zend_hex_strtod(str, &ptr);
   }
   type = 2;
  }
 } else if (*ptr == '.' && ((ptr[1]) >= '0' && (ptr[1]) <= '9')) {
process_double:
  type = 2;



  if (dval) {
   local_dval = zend_strtod(str, &ptr);
  } else if (allow_errors != 1 && dp_or_e != -1) {
   dp_or_e = (*ptr++ == '.') ? 1 : 2;
   goto check_digits;
  }
 } else {
  return 0;
 }

 if (ptr != str + length) {
  if (!allow_errors) {
   return 0;
  }
  if (allow_errors == -1) {
   zend_error((1<<3L), "A non well formed numeric value encountered");
  }
 }

 if (type == 1) {
  if (digits == 11 - 1) {
   int cmp = __extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p (&ptr[-digits]) && __builtin_constant_p (long_min_digits) && (__s1_len = strlen (&ptr[-digits]), __s2_len = strlen (long_min_digits), (!((size_t)(const void *)((&ptr[-digits]) + 1) - (size_t)(const void *)(&ptr[-digits]) == 1) || __s1_len >= 4) && (!((size_t)(const void *)((long_min_digits) + 1) - (size_t)(const void *)(long_min_digits) == 1) || __s2_len >= 4)) ? __builtin_strcmp (&ptr[-digits], long_min_digits) : (__builtin_constant_p (&ptr[-digits]) && ((size_t)(const void *)((&ptr[-digits]) + 1) - (size_t)(const void *)(&ptr[-digits]) == 1) && (__s1_len = strlen (&ptr[-digits]), __s1_len < 4) ? (__builtin_constant_p (long_min_digits) && ((size_t)(const void *)((long_min_digits) + 1) - (size_t)(const void *)(long_min_digits) == 1) ? __builtin_strcmp (&ptr[-digits], long_min_digits) : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) (long_min_digits); register int __result = (((__const unsigned char *) (__const char *) (&ptr[-digits]))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) (&ptr[-digits]))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) (&ptr[-digits]))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) (&ptr[-digits]))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p (long_min_digits) && ((size_t)(const void *)((long_min_digits) + 1) - (size_t)(const void *)(long_min_digits) == 1) && (__s2_len = strlen (long_min_digits), __s2_len < 4) ? (__builtin_constant_p (&ptr[-digits]) && ((size_t)(const void *)((&ptr[-digits]) + 1) - (size_t)(const void *)(&ptr[-digits]) == 1) ? __builtin_strcmp (&ptr[-digits], long_min_digits) : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) (&ptr[-digits]); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) (long_min_digits))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) (long_min_digits))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) (long_min_digits))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) (long_min_digits))[3]); } } __result; }))) : __builtin_strcmp (&ptr[-digits], long_min_digits)))); });

   if (!(cmp < 0 || (cmp == 0 && *str == '-'))) {
    if (dval) {
     *dval = zend_strtod(str, ((void *)0));
    }

    return 2;
   }
  }

  if (lval) {
   *lval = strtol(str, ((void *)0), base);
  }

  return 1;
 } else {
  if (dval) {
   *dval = local_dval;
  }

  return 2;
 }
}

static inline char *
zend_memnstr(char *haystack, char *needle, int needle_len, char *end)
{
 char *p = haystack;
 char ne = needle[needle_len-1];

 if (needle_len == 1) {
  return (char *)memchr(p, *needle, (end-p));
 }

 if (needle_len > end-haystack) {
  return ((void *)0);
 }

 end -= needle_len;

 while (p <= end) {
  if ((p = (char *)memchr(p, *needle, (end-p+1))) && ne == p[needle_len-1]) {
   if (!memcmp(needle, p, needle_len-1)) {
    return p;
   }
  }

  if (p == ((void *)0)) {
   return ((void *)0);
  }

  p++;
 }

 return ((void *)0);
}

static inline const void *zend_memrchr(const void *s, int c, size_t n)
{
 register const unsigned char *e;

 if (n <= 0) {
  return ((void *)0);
 }

 for (e = (const unsigned char *)s + n - 1; e >= (const unsigned char *)s; e--) {
  if (*e == (const unsigned char)c) {
   return (const void *)e;
  }
 }

 return ((void *)0);
}


__attribute__ ((visibility("default"))) int increment_function(zval *op1);
__attribute__ ((visibility("default"))) int decrement_function(zval *op2);

__attribute__ ((visibility("default"))) void convert_scalar_to_number(zval *op );
__attribute__ ((visibility("default"))) void _convert_to_string(zval *op );
__attribute__ ((visibility("default"))) void convert_to_long(zval *op);
__attribute__ ((visibility("default"))) void convert_to_double(zval *op);
__attribute__ ((visibility("default"))) void convert_to_long_base(zval *op, int base);
__attribute__ ((visibility("default"))) void convert_to_null(zval *op);
__attribute__ ((visibility("default"))) void convert_to_boolean(zval *op);
__attribute__ ((visibility("default"))) void convert_to_array(zval *op);
__attribute__ ((visibility("default"))) void convert_to_object(zval *op);
__attribute__ ((visibility("default"))) void multi_convert_to_long_ex(int argc, ...);
__attribute__ ((visibility("default"))) void multi_convert_to_double_ex(int argc, ...);
__attribute__ ((visibility("default"))) void multi_convert_to_string_ex(int argc, ...);
__attribute__ ((visibility("default"))) int add_char_to_string(zval *result, const zval *op1, const zval *op2);
__attribute__ ((visibility("default"))) int add_string_to_string(zval *result, const zval *op1, const zval *op2);


__attribute__ ((visibility("default"))) double zend_string_to_double(const char *number, zend_uint length);

__attribute__ ((visibility("default"))) int zval_is_true(zval *op);
__attribute__ ((visibility("default"))) int compare_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int numeric_compare_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int string_compare_function_ex(zval *result, zval *op1, zval *op2, zend_bool case_insensitive );
__attribute__ ((visibility("default"))) int string_compare_function(zval *result, zval *op1, zval *op2 );
__attribute__ ((visibility("default"))) int string_case_compare_function(zval *result, zval *op1, zval *op2 );

__attribute__ ((visibility("default"))) int string_locale_compare_function(zval *result, zval *op1, zval *op2 );


__attribute__ ((visibility("default"))) void zend_str_tolower(char *str, unsigned int length);
__attribute__ ((visibility("default"))) char *zend_str_tolower_copy(char *dest, const char *source, unsigned int length);
__attribute__ ((visibility("default"))) char *zend_str_tolower_dup(const char *source, unsigned int length);

__attribute__ ((visibility("default"))) int zend_binary_zval_strcmp(zval *s1, zval *s2);
__attribute__ ((visibility("default"))) int zend_binary_zval_strncmp(zval *s1, zval *s2, zval *s3);
__attribute__ ((visibility("default"))) int zend_binary_zval_strcasecmp(zval *s1, zval *s2);
__attribute__ ((visibility("default"))) int zend_binary_zval_strncasecmp(zval *s1, zval *s2, zval *s3);
__attribute__ ((visibility("default"))) int zend_binary_strcmp(const char *s1, uint len1, const char *s2, uint len2);
__attribute__ ((visibility("default"))) int zend_binary_strncmp(const char *s1, uint len1, const char *s2, uint len2, uint length);
__attribute__ ((visibility("default"))) int zend_binary_strcasecmp(const char *s1, uint len1, const char *s2, uint len2);
__attribute__ ((visibility("default"))) int zend_binary_strncasecmp(const char *s1, uint len1, const char *s2, uint len2, uint length);

__attribute__ ((visibility("default"))) void zendi_smart_strcmp(zval *result, zval *s1, zval *s2);
__attribute__ ((visibility("default"))) void zend_compare_symbol_tables(zval *result, HashTable *ht1, HashTable *ht2 );
__attribute__ ((visibility("default"))) void zend_compare_arrays(zval *result, zval *a1, zval *a2 );
__attribute__ ((visibility("default"))) void zend_compare_objects(zval *result, zval *o1, zval *o2 );

__attribute__ ((visibility("default"))) int zend_atoi(const char *str, int str_len);
__attribute__ ((visibility("default"))) long zend_atol(const char *str, int str_len);

__attribute__ ((visibility("default"))) void zend_locale_sprintf_double(zval *op );

# 456 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
static inline __attribute__((always_inline)) int fast_increment_function(zval *op1)
{
 if (__builtin_expect((*op1).type == 1, 1)) {

  __asm__(
   "incl (%0)\n\t"
   "jno  0f\n\t"
   "movl $0x0, (%0)\n\t"
   "movl $0x41e00000, 0x4(%0)\n\t"
   "movb $0x2,0xc(%0)\n"
   "0:"
   :
   : "r"(op1));
# 488 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
  return 0;
 }
 return increment_function(op1);
}

static inline __attribute__((always_inline)) int fast_decrement_function(zval *op1)
{
 if (__builtin_expect((*op1).type == 1, 1)) {

  __asm__(
   "decl (%0)\n\t"
   "jno  0f\n\t"
   "movl $0x00200000, (%0)\n\t"
   "movl $0xc1e00000, 0x4(%0)\n\t"
   "movb $0x2,0xc(%0)\n"
   "0:"
   :
   : "r"(op1));
# 525 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
  return 0;
 }
 return decrement_function(op1);
}

static inline __attribute__((always_inline)) int fast_add_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {

  __asm__(
   "movl	(%1), %%eax\n\t"
   "addl   (%2), %%eax\n\t"
   "jo     0f\n\t"
   "movl   %%eax, (%0)\n\t"
   "movb   $0x1,0xc(%0)\n\t"
   "jmp    1f\n"
   "0:\n\t"
   "fildl	(%1)\n\t"
   "fildl	(%2)\n\t"
   "faddp	%%st, %%st(1)\n\t"
   "movb   $0x2,0xc(%0)\n\t"
   "fstpl	(%0)\n"
   "1:"
   :
   : "r"(result),
     "r"(op1),
     "r"(op2)
   : "eax");
# 585 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
   return 0;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = ((double)(*op1).value.lval) + (*op2).value.dval;
   (*result).type = 2;
   return 0;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = (*op1).value.dval + (*op2).value.dval;
   (*result).type = 2;
   return 0;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   (*result).value.dval = (*op1).value.dval + ((double)(*op2).value.lval);
   (*result).type = 2;
   return 0;
  }
 }
 return add_function(result, op1, op2 );
}

static inline __attribute__((always_inline)) int fast_sub_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {

  __asm__(
   "movl	(%1), %%eax\n\t"
   "subl   (%2), %%eax\n\t"
   "jo     0f\n\t"
   "movl   %%eax, (%0)\n\t"
   "movb   $0x1,0xc(%0)\n\t"
   "jmp    1f\n"
   "0:\n\t"
   "fildl	(%2)\n\t"
   "fildl	(%1)\n\t"



   "fsubp	%%st, %%st(1)\n\t"

   "movb   $0x2,0xc(%0)\n\t"
   "fstpl	(%0)\n"
   "1:"
   :
   : "r"(result),
     "r"(op1),
     "r"(op2)
   : "eax");
# 668 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h"
   return 0;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = ((double)(*op1).value.lval) - (*op2).value.dval;
   (*result).type = 2;
   return 0;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = (*op1).value.dval - (*op2).value.dval;
   (*result).type = 2;
   return 0;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   (*result).value.dval = (*op1).value.dval - ((double)(*op2).value.lval);
   (*result).type = 2;
   return 0;
  }
 }
 return sub_function(result, op1, op2 );
}

static inline __attribute__((always_inline)) int fast_mul_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   long overflow;

   do { long __tmpvar; __asm__ ("imul %3,%0\n" "adc $0,%1" : "=r"(__tmpvar),"=r"(overflow) : "0"((*op1).value.lval), "r"((*op2).value.lval), "1"(0)); if (overflow) ((*result).value.dval) = (double) ((*op1).value.lval) * (double) ((*op2).value.lval); else ((*result).value.lval) = __tmpvar; } while (0);
   (*result).type = overflow ? 2 : 1;
   return 0;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = ((double)(*op1).value.lval) * (*op2).value.dval;
   (*result).type = 2;
   return 0;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   (*result).value.dval = (*op1).value.dval * (*op2).value.dval;
   (*result).type = 2;
   return 0;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   (*result).value.dval = (*op1).value.dval * ((double)(*op2).value.lval);
   (*result).type = 2;
   return 0;
  }
 }
 return mul_function(result, op1, op2 );
}

static inline __attribute__((always_inline)) int fast_div_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1) && 0) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   if (__builtin_expect((*op2).value.lval == 0, 0)) {
    zend_error((1<<1L), "Division by zero");
    (*result).value.lval = 0;
    (*result).type = 3;
    return -1;
   } else if (__builtin_expect((*op2).value.lval == -1 && (*op1).value.lval == (-2147483647L - 1L), 0)) {

    (*result).value.dval = (double) (-2147483647L - 1L) / -1;
    (*result).type = 2;
   } else if (__builtin_expect((*op1).value.lval % (*op2).value.lval == 0, 1)) {

    (*result).value.lval = (*op1).value.lval / (*op2).value.lval;
    (*result).type = 1;
   } else {
    (*result).value.dval = ((double) (*op1).value.lval) / ((double)(*op2).value.lval);
    (*result).type = 2;
   }
   return 0;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   if (__builtin_expect((*op2).value.dval == 0, 0)) {
    zend_error((1<<1L), "Division by zero");
    (*result).value.lval = 0;
    (*result).type = 3;
    return -1;
   }
   (*result).value.dval = ((double)(*op1).value.lval) / (*op2).value.dval;
   (*result).type = 2;
   return 0;
  }
 } else if (__builtin_expect((*op1).type == 2, 1) && 0) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   if (__builtin_expect((*op2).value.dval == 0, 0)) {
    zend_error((1<<1L), "Division by zero");
    (*result).value.lval = 0;
    (*result).type = 3;
    return -1;
   }
   (*result).value.dval = (*op1).value.dval / (*op2).value.dval;
   (*result).type = 2;
   return 0;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   if (__builtin_expect((*op2).value.lval == 0, 0)) {
    zend_error((1<<1L), "Division by zero");
    (*result).value.lval = 0;
    (*result).type = 3;
    return -1;
   }
   (*result).value.dval = (*op1).value.dval / ((double)(*op2).value.lval);
   (*result).type = 2;
   return 0;
  }
 }
 return div_function(result, op1, op2 );
}

static inline __attribute__((always_inline)) int fast_mod_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   if (__builtin_expect((*op2).value.lval == 0, 0)) {
    zend_error((1<<1L), "Division by zero");
    (*result).value.lval = 0;
    (*result).type = 3;
    return -1;
   } else if (__builtin_expect((*op2).value.lval == -1, 0)) {

    (*result).value.lval = 0;
    (*result).type = 1;
    return 0;
   }
   (*result).value.lval = (*op1).value.lval % (*op2).value.lval;
   (*result).type = 1;
   return 0;
  }
 }
 return mod_function(result, op1, op2 );
}

static inline __attribute__((always_inline)) int fast_equal_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.lval == (*op2).value.lval;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   return ((double)(*op1).value.lval) == (*op2).value.dval;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   return (*op1).value.dval == (*op2).value.dval;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.dval == ((double)(*op2).value.lval);
  }
 }
 compare_function(result, op1, op2 );
 return (*result).value.lval == 0;
}

static inline __attribute__((always_inline)) int fast_not_equal_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.lval != (*op2).value.lval;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   return ((double)(*op1).value.lval) != (*op2).value.dval;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   return (*op1).value.dval != (*op2).value.dval;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.dval != ((double)(*op2).value.lval);
  }
 }
 compare_function(result, op1, op2 );
 return (*result).value.lval != 0;
}

static inline __attribute__((always_inline)) int fast_is_smaller_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.lval < (*op2).value.lval;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   return ((double)(*op1).value.lval) < (*op2).value.dval;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   return (*op1).value.dval < (*op2).value.dval;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.dval < ((double)(*op2).value.lval);
  }
 }
 compare_function(result, op1, op2 );
 return (*result).value.lval < 0;
}

static inline __attribute__((always_inline)) int fast_is_smaller_or_equal_function(zval *result, zval *op1, zval *op2 )
{
 if (__builtin_expect((*op1).type == 1, 1)) {
  if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.lval <= (*op2).value.lval;
  } else if (__builtin_expect((*op2).type == 2, 1)) {
   return ((double)(*op1).value.lval) <= (*op2).value.dval;
  }
 } else if (__builtin_expect((*op1).type == 2, 1)) {
  if (__builtin_expect((*op2).type == 2, 1)) {
   return (*op1).value.dval <= (*op2).value.dval;
  } else if (__builtin_expect((*op2).type == 1, 1)) {
   return (*op1).value.dval <= ((double)(*op2).value.lval);
  }
 }
 compare_function(result, op1, op2 );
 return (*result).value.lval <= 0;
}
# 831 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_variables.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_variables.h"


__attribute__ ((visibility("default"))) void _zval_dtor_func(zval *zvalue );

static inline __attribute__((always_inline)) void _zval_dtor(zval *zvalue )
{
 if (zvalue->type <= 3) {
  return;
 }
 _zval_dtor_func(zvalue );
}

__attribute__ ((visibility("default"))) void _zval_copy_ctor_func(zval *zvalue );

static inline __attribute__((always_inline)) void _zval_copy_ctor(zval *zvalue )
{
 if (zvalue->type <= 3) {
  return;
 }
 _zval_copy_ctor_func(zvalue );
}

__attribute__ ((visibility("default"))) int zval_copy_static_var(zval **p , int num_args, va_list args, zend_hash_key *key);

__attribute__ ((visibility("default"))) int zend_print_variable(zval *var);
__attribute__ ((visibility("default"))) void _zval_ptr_dtor(zval **zval_ptr );
__attribute__ ((visibility("default"))) void _zval_internal_dtor(zval *zvalue );
__attribute__ ((visibility("default"))) void _zval_internal_ptr_dtor(zval **zvalue );
__attribute__ ((visibility("default"))) void _zval_dtor_wrapper(zval *zvalue);
# 78 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_variables.h"
__attribute__ ((visibility("default"))) void zval_add_ref(zval **p);


# 832 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend.h" 2

typedef enum {
 EH_NORMAL = 0,
 EH_SUPPRESS,
 EH_THROW
} zend_error_handling_t;

typedef struct {
 zend_error_handling_t handling;
 zend_class_entry *exception;
 zval *user_handler;
} zend_error_handling;

__attribute__ ((visibility("default"))) void zend_save_error_handling(zend_error_handling *current );
__attribute__ ((visibility("default"))) void zend_replace_error_handling(zend_error_handling_t error_handling, zend_class_entry *exception_class, zend_error_handling *current );
__attribute__ ((visibility("default"))) void zend_restore_error_handling(zend_error_handling *saved );
# 35 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_qsort.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_qsort.h"

__attribute__ ((visibility("default"))) void zend_qsort(void *base, size_t nmemb, size_t siz, compare_func_t compare );

# 36 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_compat.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_compat.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_config.h" 1
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_compat.h" 2
# 37 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h" 1
# 53 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h"
typedef struct _zend_op_array zend_op_array;
typedef struct _zend_op zend_op;

typedef struct _zend_compiler_context {
 zend_uint opcodes_size;
 int vars_size;
 int literals_size;
 int current_brk_cont;
 int backpatch_count;
 HashTable *labels;
} zend_compiler_context;

typedef struct _zend_literal {
 zval constant;
 zend_ulong hash_value;
 zend_uint cache_slot;
} zend_literal;




typedef union _znode_op {
 zend_uint constant;
 zend_uint var;
 zend_uint num;
 zend_ulong hash;
 zend_uint opline_num;
 zend_op *jmp_addr;
 zval *zv;
 zend_literal *literal;
  void *ptr;
} znode_op;

typedef struct _znode {
 int op_type;
 union {
  znode_op op;
  zval constant;
  zend_op_array *op_array;
 } u;
 zend_uint EA;
} znode;

typedef struct _zend_execute_data zend_execute_data;




typedef int (*user_opcode_handler_t) (zend_execute_data *execute_data );
typedef int (__attribute__((fastcall)) *opcode_handler_t) (zend_execute_data *execute_data );

extern __attribute__ ((visibility("default"))) opcode_handler_t *zend_opcode_handlers;

struct _zend_op {
 opcode_handler_t handler;
 znode_op op1;
 znode_op op2;
 znode_op result;
 ulong extended_value;
 uint lineno;
 zend_uchar opcode;
 zend_uchar op1_type;
 zend_uchar op2_type;
 zend_uchar result_type;
};


typedef struct _zend_brk_cont_element {
 int start;
 int cont;
 int brk;
 int parent;
} zend_brk_cont_element;

typedef struct _zend_label {
 int brk_cont;
 zend_uint opline_num;
} zend_label;

typedef struct _zend_try_catch_element {
 zend_uint try_op;
 zend_uint catch_op;
} zend_try_catch_element;
# 210 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h"
char *zend_visibility_string(zend_uint fn_flags);


typedef struct _zend_property_info {
 zend_uint flags;
 const char *name;
 int name_length;
 ulong h;
 int offset;
 const char *doc_comment;
 int doc_comment_len;
 zend_class_entry *ce;
} zend_property_info;


typedef struct _zend_arg_info {
 const char *name;
 zend_uint name_len;
 const char *class_name;
 zend_uint class_name_len;
 zend_uchar type_hint;
 zend_bool allow_null;
 zend_bool pass_by_reference;
} zend_arg_info;





typedef struct _zend_internal_function_info {
 const char *_name;
 zend_uint _name_len;
 const char *_class_name;
 zend_uint required_num_args;
 zend_uchar _type_hint;
 zend_bool return_reference;
 zend_bool pass_rest_by_reference;
} zend_internal_function_info;

typedef struct _zend_compiled_variable {
 const char *name;
 int name_len;
 ulong hash_value;
} zend_compiled_variable;

struct _zend_op_array {

 zend_uchar type;
 const char *function_name;
 zend_class_entry *scope;
 zend_uint fn_flags;
 union _zend_function *prototype;
 zend_uint num_args;
 zend_uint required_num_args;
 zend_arg_info *arg_info;


 zend_uint *refcount;

 zend_op *opcodes;
 zend_uint last;

 zend_compiled_variable *vars;
 int last_var;

 zend_uint T;

 zend_brk_cont_element *brk_cont_array;
 int last_brk_cont;

 zend_try_catch_element *try_catch_array;
 int last_try_catch;


 HashTable *static_variables;

 zend_uint this_var;

 const char *filename;
 zend_uint line_start;
 zend_uint line_end;
 const char *doc_comment;
 zend_uint doc_comment_len;
 zend_uint early_binding;

 zend_literal *literals;
 int last_literal;

 void **run_time_cache;
 int last_cache_slot;

 void *reserved[4];
};





typedef struct _zend_internal_function {

 zend_uchar type;
 const char * function_name;
 zend_class_entry *scope;
 zend_uint fn_flags;
 union _zend_function *prototype;
 zend_uint num_args;
 zend_uint required_num_args;
 zend_arg_info *arg_info;


 void (*handler)(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
 struct _zend_module_entry *module;
} zend_internal_function;



typedef union _zend_function {
 zend_uchar type;

 struct {
  zend_uchar type;
  const char *function_name;
  zend_class_entry *scope;
  zend_uint fn_flags;
  union _zend_function *prototype;
  zend_uint num_args;
  zend_uint required_num_args;
  zend_arg_info *arg_info;
 } common;

 zend_op_array op_array;
 zend_internal_function internal_function;
} zend_function;


typedef struct _zend_function_state {
 zend_function *function;
 void **arguments;
} zend_function_state;


typedef struct _zend_switch_entry {
 znode cond;
 int default_case;
 int control_var;
} zend_switch_entry;


typedef struct _list_llist_element {
 znode var;
 zend_llist dimensions;
 znode value;
} list_llist_element;

union _temp_variable;

struct _zend_execute_data {
 struct _zend_op *opline;
 zend_function_state function_state;
 zend_function *fbc;
 zend_class_entry *called_scope;
 zend_op_array *op_array;
 zval *object;
 union _temp_variable *Ts;
 zval ***CVs;
 HashTable *symbol_table;
 struct _zend_execute_data *prev_execute_data;
 zval *old_error_reporting;
 zend_bool nested;
 zval **original_return_value;
 zend_class_entry *current_scope;
 zend_class_entry *current_called_scope;
 zval *current_this;
 zval *current_object;
};
# 397 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h"
# 1 "/usr/include/setjmp.h" 1 3 4
# 28 "/usr/include/setjmp.h" 3 4


# 1 "/usr/include/bits/setjmp.h" 1 3 4
# 29 "/usr/include/bits/setjmp.h" 3 4
typedef int __jmp_buf[6];
# 31 "/usr/include/setjmp.h" 2 3 4
# 1 "/usr/include/bits/sigset.h" 1 3 4
# 32 "/usr/include/setjmp.h" 2 3 4



struct __jmp_buf_tag
  {




    __jmp_buf __jmpbuf;
    int __mask_was_saved;
    __sigset_t __saved_mask;
  };




typedef struct __jmp_buf_tag jmp_buf[1];



extern int setjmp (jmp_buf __env) __attribute__ ((__nothrow__));






extern int __sigsetjmp (struct __jmp_buf_tag __env[1], int __savemask) __attribute__ ((__nothrow__));




extern int _setjmp (struct __jmp_buf_tag __env[1]) __attribute__ ((__nothrow__));
# 78 "/usr/include/setjmp.h" 3 4




extern void longjmp (struct __jmp_buf_tag __env[1], int __val)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));







extern void _longjmp (struct __jmp_buf_tag __env[1], int __val)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));







typedef struct __jmp_buf_tag sigjmp_buf[1];
# 110 "/usr/include/setjmp.h" 3 4
extern void siglongjmp (sigjmp_buf __env, int __val)
     __attribute__ ((__nothrow__)) __attribute__ ((__noreturn__));
# 120 "/usr/include/setjmp.h" 3 4

# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals_macros.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals_macros.h"
typedef struct _zend_compiler_globals zend_compiler_globals;
typedef struct _zend_executor_globals zend_executor_globals;
typedef struct _zend_php_scanner_globals zend_php_scanner_globals;
typedef struct _zend_ini_scanner_globals zend_ini_scanner_globals;









extern __attribute__ ((visibility("default"))) struct _zend_compiler_globals compiler_globals;
int zendparse(void);
# 48 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals_macros.h"
extern __attribute__ ((visibility("default"))) zend_executor_globals executor_globals;
# 57 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals_macros.h"
extern __attribute__ ((visibility("default"))) zend_php_scanner_globals language_scanner_globals;
# 67 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals_macros.h"
extern __attribute__ ((visibility("default"))) zend_ini_scanner_globals ini_scanner_globals;



# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_stack.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_stack.h"
typedef struct _zend_stack {
 int top, max;
 void **elements;
} zend_stack;





__attribute__ ((visibility("default"))) int zend_stack_init(zend_stack *stack);
__attribute__ ((visibility("default"))) int zend_stack_push(zend_stack *stack, const void *element, int size);
__attribute__ ((visibility("default"))) int zend_stack_top(const zend_stack *stack, void **element);
__attribute__ ((visibility("default"))) int zend_stack_del_top(zend_stack *stack);
__attribute__ ((visibility("default"))) int zend_stack_int_top(const zend_stack *stack);
__attribute__ ((visibility("default"))) int zend_stack_is_empty(const zend_stack *stack);
__attribute__ ((visibility("default"))) int zend_stack_destroy(zend_stack *stack);
__attribute__ ((visibility("default"))) void **zend_stack_base(const zend_stack *stack);
__attribute__ ((visibility("default"))) int zend_stack_count(const zend_stack *stack);
__attribute__ ((visibility("default"))) void zend_stack_apply(zend_stack *stack, int type, int (*apply_function)(void *element));
__attribute__ ((visibility("default"))) void zend_stack_apply_with_argument(zend_stack *stack, int type, int (*apply_function)(void *element, void *arg), void *arg);

# 31 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ptr_stack.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ptr_stack.h"
typedef struct _zend_ptr_stack {
 int top, max;
 void **elements;
 void **top_element;
 zend_bool persistent;
} zend_ptr_stack;





__attribute__ ((visibility("default"))) void zend_ptr_stack_init(zend_ptr_stack *stack);
__attribute__ ((visibility("default"))) void zend_ptr_stack_init_ex(zend_ptr_stack *stack, zend_bool persistent);
__attribute__ ((visibility("default"))) void zend_ptr_stack_n_push(zend_ptr_stack *stack, int count, ...);
__attribute__ ((visibility("default"))) void zend_ptr_stack_n_pop(zend_ptr_stack *stack, int count, ...);
__attribute__ ((visibility("default"))) void zend_ptr_stack_destroy(zend_ptr_stack *stack);
__attribute__ ((visibility("default"))) void zend_ptr_stack_apply(zend_ptr_stack *stack, void (*func)(void *));
__attribute__ ((visibility("default"))) void zend_ptr_stack_clean(zend_ptr_stack *stack, void (*func)(void *), zend_bool free_elements);
__attribute__ ((visibility("default"))) int zend_ptr_stack_num_elements(zend_ptr_stack *stack);

# 58 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ptr_stack.h"
static inline __attribute__((always_inline)) void zend_ptr_stack_3_push(zend_ptr_stack *stack, void *a, void *b, void *c)
{


 if (stack->top+3 > stack->max) { do { stack->max += 64; } while (stack->top+3 > stack->max); stack->elements = (void **) ((stack->persistent)?__zend_realloc((stack->elements), ((sizeof(void *) * (stack->max)))):_erealloc(((stack->elements)), (((sizeof(void *) * (stack->max)))), 0 )); stack->top_element = stack->elements+stack->top; }

 stack->top += 3;
 *(stack->top_element++) = a;
 *(stack->top_element++) = b;
 *(stack->top_element++) = c;


}

static inline __attribute__((always_inline)) void zend_ptr_stack_2_push(zend_ptr_stack *stack, void *a, void *b)
{


 if (stack->top+2 > stack->max) { do { stack->max += 64; } while (stack->top+2 > stack->max); stack->elements = (void **) ((stack->persistent)?__zend_realloc((stack->elements), ((sizeof(void *) * (stack->max)))):_erealloc(((stack->elements)), (((sizeof(void *) * (stack->max)))), 0 )); stack->top_element = stack->elements+stack->top; }

 stack->top += 2;
 *(stack->top_element++) = a;
 *(stack->top_element++) = b;


}

static inline __attribute__((always_inline)) void zend_ptr_stack_3_pop(zend_ptr_stack *stack, void **a, void **b, void **c)
{
 *a = *(--stack->top_element);
 *b = *(--stack->top_element);
 *c = *(--stack->top_element);
 stack->top -= 3;
}

static inline __attribute__((always_inline)) void zend_ptr_stack_2_pop(zend_ptr_stack *stack, void **a, void **b)
{
 *a = *(--stack->top_element);
 *b = *(--stack->top_element);
 stack->top -= 2;
}

static inline __attribute__((always_inline)) void zend_ptr_stack_push(zend_ptr_stack *stack, void *ptr)
{
 if (stack->top+1 > stack->max) { do { stack->max += 64; } while (stack->top+1 > stack->max); stack->elements = (void **) ((stack->persistent)?__zend_realloc((stack->elements), ((sizeof(void *) * (stack->max)))):_erealloc(((stack->elements)), (((sizeof(void *) * (stack->max)))), 0 )); stack->top_element = stack->elements+stack->top; }

 stack->top++;
 *(stack->top_element++) = ptr;
}

static inline __attribute__((always_inline)) void *zend_ptr_stack_pop(zend_ptr_stack *stack)
{
 stack->top--;
 return *(--stack->top_element);
}
# 32 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2


# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_objects.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_objects.h"

__attribute__ ((visibility("default"))) void zend_object_std_init(zend_object *object, zend_class_entry *ce );
__attribute__ ((visibility("default"))) void zend_object_std_dtor(zend_object *object );
__attribute__ ((visibility("default"))) zend_object_value zend_objects_new(zend_object **object, zend_class_entry *class_type );
__attribute__ ((visibility("default"))) void zend_objects_destroy_object(zend_object *object, zend_object_handle handle );
__attribute__ ((visibility("default"))) zend_object *zend_objects_get_address(const zval *object );
__attribute__ ((visibility("default"))) void zend_objects_clone_members(zend_object *new_object, zend_object_value new_obj_val, zend_object *old_object, zend_object_handle handle );
__attribute__ ((visibility("default"))) zend_object_value zend_objects_clone_obj(zval *object );
__attribute__ ((visibility("default"))) void zend_objects_free_object_storage(zend_object *object );

# 35 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_objects_API.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_objects_API.h"
typedef void (*zend_objects_store_dtor_t)(void *object, zend_object_handle handle );
typedef void (*zend_objects_free_object_storage_t)(void *object );
typedef void (*zend_objects_store_clone_t)(void *object, void **object_clone );

typedef struct _zend_object_store_bucket {
 zend_bool destructor_called;
 zend_bool valid;
 union _store_bucket {
  struct _store_object {
   void *object;
   zend_objects_store_dtor_t dtor;
   zend_objects_free_object_storage_t free_storage;
   zend_objects_store_clone_t clone;
   const zend_object_handlers *handlers;
   zend_uint refcount;
   gc_root_buffer *buffered;
  } obj;
  struct {
   int next;
  } free_list;
 } bucket;
} zend_object_store_bucket;

typedef struct _zend_objects_store {
 zend_object_store_bucket *object_buckets;
 zend_uint top;
 zend_uint size;
 int free_list_head;
} zend_objects_store;



__attribute__ ((visibility("default"))) void zend_objects_store_init(zend_objects_store *objects, zend_uint init_size);
__attribute__ ((visibility("default"))) void zend_objects_store_call_destructors(zend_objects_store *objects );
__attribute__ ((visibility("default"))) void zend_objects_store_mark_destructed(zend_objects_store *objects );
__attribute__ ((visibility("default"))) void zend_objects_store_destroy(zend_objects_store *objects);


__attribute__ ((visibility("default"))) zend_object_handle zend_objects_store_put(void *object, zend_objects_store_dtor_t dtor, zend_objects_free_object_storage_t storage, zend_objects_store_clone_t clone );

__attribute__ ((visibility("default"))) void zend_objects_store_add_ref(zval *object );
__attribute__ ((visibility("default"))) void zend_objects_store_del_ref(zval *object );
__attribute__ ((visibility("default"))) void zend_objects_store_add_ref_by_handle(zend_object_handle handle );
__attribute__ ((visibility("default"))) void zend_objects_store_del_ref_by_handle_ex(zend_object_handle handle, const zend_object_handlers *handlers );
static inline __attribute__((always_inline)) void zend_objects_store_del_ref_by_handle(zend_object_handle handle ) {
 zend_objects_store_del_ref_by_handle_ex(handle, ((void *)0) );
}
__attribute__ ((visibility("default"))) zend_uint zend_objects_store_get_refcount(zval *object );
__attribute__ ((visibility("default"))) zend_object_value zend_objects_store_clone_obj(zval *object );
__attribute__ ((visibility("default"))) void *zend_object_store_get_object(const zval *object );
__attribute__ ((visibility("default"))) void *zend_object_store_get_object_by_handle(zend_object_handle handle );

__attribute__ ((visibility("default"))) void zend_object_store_set_object(zval *zobject, void *object );
__attribute__ ((visibility("default"))) void zend_object_store_ctor_failed(zval *zobject );

__attribute__ ((visibility("default"))) void zend_objects_store_free_object_storage(zend_objects_store *objects );



__attribute__ ((visibility("default"))) zval *zend_object_create_proxy(zval *object, zval *member );

__attribute__ ((visibility("default"))) zend_object_handlers *zend_get_std_object_handlers(void);

# 36 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h" 1
# 37 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_float.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_float.h"
extern __attribute__ ((visibility("default"))) void zend_init_fpu(void);
extern __attribute__ ((visibility("default"))) void zend_shutdown_fpu(void);
extern __attribute__ ((visibility("default"))) void zend_ensure_fpu_mode(void);
# 212 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_float.h"
# 1 "/usr/include/fpu_control.h" 1 3 4
# 89 "/usr/include/fpu_control.h" 3 4
typedef unsigned int fpu_control_t __attribute__ ((__mode__ (__HI__)));
# 100 "/usr/include/fpu_control.h" 3 4
extern fpu_control_t __fpu_control;
# 213 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_float.h" 2
# 38 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_multibyte.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_multibyte.h"
typedef struct _zend_encoding zend_encoding;

typedef size_t (*zend_encoding_filter)(unsigned char **str, size_t *str_length, const unsigned char *buf, size_t length );

typedef const zend_encoding* (*zend_encoding_fetcher)(const char *encoding_name );
typedef const char* (*zend_encoding_name_getter)(const zend_encoding *encoding);
typedef int (*zend_encoding_lexer_compatibility_checker)(const zend_encoding *encoding);
typedef const zend_encoding *(*zend_encoding_detector)(const unsigned char *string, size_t length, const zend_encoding **list, size_t list_size );
typedef size_t (*zend_encoding_converter)(unsigned char **to, size_t *to_length, const unsigned char *from, size_t from_length, const zend_encoding *encoding_to, const zend_encoding *encoding_from );
typedef int (*zend_encoding_list_parser)(const char *encoding_list, size_t encoding_list_len, const zend_encoding ***return_list, size_t *return_size, int persistent );
typedef const zend_encoding *(*zend_encoding_internal_encoding_getter)(void);
typedef int (*zend_encoding_internal_encoding_setter)(const zend_encoding *encoding );

typedef struct _zend_multibyte_functions {
    const char *provider_name;
    zend_encoding_fetcher encoding_fetcher;
    zend_encoding_name_getter encoding_name_getter;
    zend_encoding_lexer_compatibility_checker lexer_compatibility_checker;
    zend_encoding_detector encoding_detector;
    zend_encoding_converter encoding_converter;
    zend_encoding_list_parser encoding_list_parser;
    zend_encoding_internal_encoding_getter internal_encoding_getter;
    zend_encoding_internal_encoding_setter internal_encoding_setter;
} zend_multibyte_functions;






__attribute__ ((visibility("default"))) extern const zend_encoding *zend_multibyte_encoding_utf32be;
__attribute__ ((visibility("default"))) extern const zend_encoding *zend_multibyte_encoding_utf32le;
__attribute__ ((visibility("default"))) extern const zend_encoding *zend_multibyte_encoding_utf16be;
__attribute__ ((visibility("default"))) extern const zend_encoding *zend_multibyte_encoding_utf16le;
__attribute__ ((visibility("default"))) extern const zend_encoding *zend_multibyte_encoding_utf8;


__attribute__ ((visibility("default"))) int zend_multibyte_set_functions(const zend_multibyte_functions *functions );
__attribute__ ((visibility("default"))) const zend_multibyte_functions *zend_multibyte_get_functions(void);

__attribute__ ((visibility("default"))) const zend_encoding *zend_multibyte_fetch_encoding(const char *name );
__attribute__ ((visibility("default"))) const char *zend_multibyte_get_encoding_name(const zend_encoding *encoding);
__attribute__ ((visibility("default"))) int zend_multibyte_check_lexer_compatibility(const zend_encoding *encoding);
__attribute__ ((visibility("default"))) const zend_encoding *zend_multibyte_encoding_detector(const unsigned char *string, size_t length, const zend_encoding **list, size_t list_size );
__attribute__ ((visibility("default"))) size_t zend_multibyte_encoding_converter(unsigned char **to, size_t *to_length, const unsigned char *from, size_t from_length, const zend_encoding *encoding_to, const zend_encoding *encoding_from );
__attribute__ ((visibility("default"))) int zend_multibyte_parse_encoding_list(const char *encoding_list, size_t encoding_list_len, const zend_encoding ***return_list, size_t *return_size, int persistent );

__attribute__ ((visibility("default"))) const zend_encoding *zend_multibyte_get_internal_encoding(void);
__attribute__ ((visibility("default"))) const zend_encoding *zend_multibyte_get_script_encoding(void);
__attribute__ ((visibility("default"))) int zend_multibyte_set_script_encoding(const zend_encoding **encoding_list, size_t encoding_list_size );
__attribute__ ((visibility("default"))) int zend_multibyte_set_internal_encoding(const zend_encoding *encoding );
__attribute__ ((visibility("default"))) int zend_multibyte_set_script_encoding_by_string(const char *new_value, size_t new_value_length );


# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 55 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h" 1
# 56 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 2
# 64 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h"
typedef struct _zend_declarables {
 zval ticks;
} zend_declarables;

typedef struct _zend_vm_stack *zend_vm_stack;
typedef struct _zend_ini_entry zend_ini_entry;


struct _zend_compiler_globals {
 zend_stack bp_stack;
 zend_stack switch_cond_stack;
 zend_stack foreach_copy_stack;
 zend_stack object_stack;
 zend_stack declare_stack;

 zend_class_entry *active_class_entry;


 zend_llist list_llist;
 zend_llist dimension_llist;
 zend_stack list_stack;

 zend_stack function_call_stack;

 char *compiled_filename;

 int zend_lineno;

 char *heredoc;
 int heredoc_len;

 zend_op_array *active_op_array;

 HashTable *function_table;
 HashTable *class_table;

 HashTable filenames_table;

 HashTable *auto_globals;

 zend_bool parse_error;
 zend_bool in_compilation;
 zend_bool short_tags;
 zend_bool asp_tags;

 zend_declarables declarables;

 zend_bool unclean_shutdown;

 zend_bool ini_parser_unbuffered_errors;

 zend_llist open_files;

 long catch_begin;

 struct _zend_ini_parser_param *ini_parser_param;

 int interactive;

 zend_uint start_lineno;
 zend_bool increment_lineno;

 znode implementing_class;

 zend_uint access_type;

 char *doc_comment;
 zend_uint doc_comment_len;

 zend_uint compiler_options;

 zval *current_namespace;
 HashTable *current_import;
 zend_bool in_namespace;
 zend_bool has_bracketed_namespaces;

 zend_compiler_context context;
 zend_stack context_stack;


 char *interned_strings_start;
 char *interned_strings_end;
 char *interned_strings_top;
 char *interned_strings_snapshot_top;

 HashTable interned_strings;

 const zend_encoding **script_encoding_list;
 size_t script_encoding_list_size;
 zend_bool multibyte;
 zend_bool detect_unicode;
 zend_bool encoding_declared;





};


struct _zend_executor_globals {
 zval **return_value_ptr_ptr;

 zval uninitialized_zval;
 zval *uninitialized_zval_ptr;

 zval error_zval;
 zval *error_zval_ptr;

 zend_ptr_stack arg_types_stack;


 HashTable *symtable_cache[32];
 HashTable **symtable_cache_limit;
 HashTable **symtable_cache_ptr;

 zend_op **opline_ptr;

 HashTable *active_symbol_table;
 HashTable symbol_table;

 HashTable included_files;

 jmp_buf *bailout;

 int error_reporting;
 int orig_error_reporting;
 int exit_status;

 zend_op_array *active_op_array;

 HashTable *function_table;
 HashTable *class_table;
 HashTable *zend_constants;

 zend_class_entry *scope;
 zend_class_entry *called_scope;

 zval *This;

 long precision;

 int ticks_count;

 zend_bool in_execution;
 HashTable *in_autoload;
 zend_function *autoload_func;
 zend_bool full_tables_cleanup;


 zend_bool no_extensions;






 HashTable regular_list;
 HashTable persistent_list;

 zend_vm_stack argument_stack;

 int user_error_handler_error_reporting;
 zval *user_error_handler;
 zval *user_exception_handler;
 zend_stack user_error_handlers_error_reporting;
 zend_ptr_stack user_error_handlers;
 zend_ptr_stack user_exception_handlers;

 zend_error_handling_t error_handling;
 zend_class_entry *exception_class;


 int timeout_seconds;

 int lambda_count;

 HashTable *ini_directives;
 HashTable *modified_ini_directives;
 zend_ini_entry *error_reporting_ini_entry;

 zend_objects_store objects_store;
 zval *exception, *prev_exception;
 zend_op *opline_before_exception;
 zend_op exception_op[3];

 struct _zend_execute_data *current_execute_data;

 struct _zend_module_entry *current_module;

 zend_property_info std_property_info;

 zend_bool active;

 zend_op *start_op;

 void *saved_fpu_cw_ptr;

 fpu_control_t saved_fpu_cw;


 void *reserved[4];
};

struct _zend_ini_scanner_globals {
 zend_file_handle *yy_in;
 zend_file_handle *yy_out;

 unsigned int yy_leng;
 unsigned char *yy_start;
 unsigned char *yy_text;
 unsigned char *yy_cursor;
 unsigned char *yy_marker;
 unsigned char *yy_limit;
 int yy_state;
 zend_stack state_stack;

 char *filename;
 int lineno;


 int scanner_mode;
};

struct _zend_php_scanner_globals {
 zend_file_handle *yy_in;
 zend_file_handle *yy_out;

 unsigned int yy_leng;
 unsigned char *yy_start;
 unsigned char *yy_text;
 unsigned char *yy_cursor;
 unsigned char *yy_marker;
 unsigned char *yy_limit;
 int yy_state;
 zend_stack state_stack;


 unsigned char *script_org;
 size_t script_org_size;


 unsigned char *script_filtered;
 size_t script_filtered_size;


 zend_encoding_filter input_filter;
 zend_encoding_filter output_filter;
 const zend_encoding *script_encoding;
};
# 398 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h" 2



void init_compiler(void);
void shutdown_compiler(void);
void zend_init_compiler_data_structures(void);
void zend_init_compiler_context(void);

extern __attribute__ ((visibility("default"))) zend_op_array *(*zend_compile_file)(zend_file_handle *file_handle, int type );
extern __attribute__ ((visibility("default"))) zend_op_array *(*zend_compile_string)(zval *source_string, char *filename );

__attribute__ ((visibility("default"))) int lex_scan(zval *zendlval );
void startup_scanner(void);
void shutdown_scanner(void);

__attribute__ ((visibility("default"))) char *zend_set_compiled_filename(const char *new_compiled_filename );
__attribute__ ((visibility("default"))) void zend_restore_compiled_filename(char *original_compiled_filename );
__attribute__ ((visibility("default"))) char *zend_get_compiled_filename(void);
__attribute__ ((visibility("default"))) int zend_get_compiled_lineno(void);
__attribute__ ((visibility("default"))) size_t zend_get_scanned_file_offset(void);

void zend_resolve_non_class_name(znode *element_name, zend_bool check_namespace );
void zend_resolve_class_name(znode *class_name, ulong fetch_type, int check_ns_name );
__attribute__ ((visibility("default"))) const char* zend_get_compiled_variable_name(const zend_op_array *op_array, zend_uint var, int* name_len);
# 430 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h"
void zend_do_binary_op(zend_uchar op, znode *result, const znode *op1, const znode *op2 );
void zend_do_unary_op(zend_uchar op, znode *result, const znode *op1 );
void zend_do_binary_assign_op(zend_uchar op, znode *result, const znode *op1, const znode *op2 );
void zend_do_assign(znode *result, znode *variable, znode *value );
void zend_do_assign_ref(znode *result, const znode *lvar, const znode *rvar );
void fetch_simple_variable(znode *result, znode *varname, int bp );
void fetch_simple_variable_ex(znode *result, znode *varname, int bp, zend_uchar op );
void zend_do_indirect_references(znode *result, const znode *num_references, znode *variable );
void zend_do_fetch_static_variable(znode *varname, const znode *static_assignment, int fetch_type );
void zend_do_fetch_global_variable(znode *varname, const znode *static_assignment, int fetch_type );

void fetch_array_begin(znode *result, znode *varname, znode *first_dim );
void fetch_array_dim(znode *result, const znode *parent, const znode *dim );
void fetch_string_offset(znode *result, const znode *parent, const znode *offset );
void zend_do_fetch_static_member(znode *result, znode *class_znode );
void zend_do_print(znode *result, const znode *arg );
void zend_do_echo(const znode *arg );
typedef int (*unary_op_type)(zval *, zval * );
typedef int (*binary_op_type)(zval *, zval *, zval * );
__attribute__ ((visibility("default"))) unary_op_type get_unary_op(int opcode);
__attribute__ ((visibility("default"))) binary_op_type get_binary_op(int opcode);

void zend_do_while_cond(const znode *expr, znode *close_bracket_token );
void zend_do_while_end(const znode *while_token, const znode *close_bracket_token );
void zend_do_do_while_begin(void);
void zend_do_do_while_end(const znode *do_token, const znode *expr_open_bracket, const znode *expr );


void zend_do_if_cond(const znode *cond, znode *closing_bracket_token );
void zend_do_if_after_statement(const znode *closing_bracket_token, unsigned char initialize );
void zend_do_if_end(void);

void zend_do_for_cond(const znode *expr, znode *second_semicolon_token );
void zend_do_for_before_statement(const znode *cond_start, const znode *second_semicolon_token );
void zend_do_for_end(const znode *second_semicolon_token );

void zend_do_pre_incdec(znode *result, const znode *op1, zend_uchar op );
void zend_do_post_incdec(znode *result, const znode *op1, zend_uchar op );

void zend_do_begin_variable_parse(void);
void zend_do_end_variable_parse(znode *variable, int type, int arg_offset );

void zend_check_writable_variable(const znode *variable);

void zend_do_free(znode *op1 );

void zend_do_add_string(znode *result, const znode *op1, znode *op2 );
void zend_do_add_variable(znode *result, const znode *op1, const znode *op2 );

int zend_do_verify_access_types(const znode *current_access_type, const znode *new_modifier);
void zend_do_begin_function_declaration(znode *function_token, znode *function_name, int is_method, int return_reference, znode *fn_flags_znode );
void zend_do_end_function_declaration(const znode *function_token );
void zend_do_receive_arg(zend_uchar op, znode *varname, const znode *offset, const znode *initialization, znode *class_type, zend_bool pass_by_reference );
int zend_do_begin_function_call(znode *function_name, zend_bool check_namespace );
void zend_do_begin_method_call(znode *left_bracket );
void zend_do_clone(znode *result, const znode *expr );
void zend_do_begin_dynamic_function_call(znode *function_name, int prefix_len );
void zend_do_fetch_class(znode *result, znode *class_name );
void zend_do_build_full_name(znode *result, znode *prefix, znode *name, int is_class_member );
int zend_do_begin_class_member_function_call(znode *class_name, znode *method_name );
void zend_do_end_function_call(znode *function_name, znode *result, const znode *argument_list, int is_method, int is_dynamic_fcall );
void zend_do_return(znode *expr, int do_end_vparse );
void zend_do_handle_exception(void);

void zend_do_begin_lambda_function_declaration(znode *result, znode *function_token, int return_reference, int is_static );
void zend_do_fetch_lexical_variable(znode *varname, zend_bool is_ref );

void zend_do_try(znode *try_token );
void zend_do_begin_catch(znode *try_token, znode *catch_class, znode *catch_var, znode *first_catch );
void zend_do_end_catch(const znode *try_token );
void zend_do_throw(const znode *expr );

__attribute__ ((visibility("default"))) int do_bind_function(const zend_op_array *op_array, zend_op *opline, HashTable *function_table, zend_bool compile_time);
__attribute__ ((visibility("default"))) zend_class_entry *do_bind_class(const zend_op_array *op_array, const zend_op *opline, HashTable *class_table, zend_bool compile_time );
__attribute__ ((visibility("default"))) zend_class_entry *do_bind_inherited_class(const zend_op_array *op_array, const zend_op *opline, HashTable *class_table, zend_class_entry *parent_ce, zend_bool compile_time );
__attribute__ ((visibility("default"))) void zend_do_inherit_interfaces(zend_class_entry *ce, const zend_class_entry *iface );
__attribute__ ((visibility("default"))) void zend_do_implement_interface(zend_class_entry *ce, zend_class_entry *iface );
void zend_do_implements_interface(znode *interface_znode );


void zend_add_trait_precedence(znode *precedence_znode );
void zend_add_trait_alias(znode *alias_znode );


void zend_do_implements_trait(znode *interface_znode );
__attribute__ ((visibility("default"))) void zend_do_implement_trait(zend_class_entry *ce, zend_class_entry *trait );
__attribute__ ((visibility("default"))) void zend_do_bind_traits(zend_class_entry *ce );
void zend_prepare_trait_precedence(znode *result, znode *method_reference, znode *trait_list );
void zend_prepare_reference(znode *result, znode *class_name, znode *method_name );
void zend_prepare_trait_alias(znode *result, znode *method_reference, znode *modifiers, znode *alias );

void init_trait_alias_list(znode* result, const znode* trait_alias );
void add_trait_alias(znode* result, const znode* trait_alias );
void init_trait_alias(znode* result, const znode* method_name, const znode* alias, const znode* modifiers );

__attribute__ ((visibility("default"))) void zend_do_inheritance(zend_class_entry *ce, zend_class_entry *parent_ce );
void zend_do_early_binding(void);
__attribute__ ((visibility("default"))) void zend_do_delayed_early_binding(const zend_op_array *op_array );

void zend_do_pass_param(znode *param, zend_uchar op, int offset );


void zend_do_boolean_or_begin(znode *expr1, znode *op_token );
void zend_do_boolean_or_end(znode *result, const znode *expr1, const znode *expr2, znode *op_token );
void zend_do_boolean_and_begin(znode *expr1, znode *op_token );
void zend_do_boolean_and_end(znode *result, const znode *expr1, const znode *expr2, const znode *op_token );

void zend_do_brk_cont(zend_uchar op, const znode *expr );

void zend_do_switch_cond(const znode *cond );
void zend_do_switch_end(const znode *case_list );
void zend_do_case_before_statement(const znode *case_list, znode *case_token, const znode *case_expr );
void zend_do_case_after_statement(znode *result, const znode *case_token );
void zend_do_default_before_statement(const znode *case_list, znode *default_token );

void zend_do_begin_class_declaration(const znode *class_token, znode *class_name, const znode *parent_class_name );
void zend_do_end_class_declaration(const znode *class_token, const znode *parent_token );
void zend_do_declare_property(const znode *var_name, const znode *value, zend_uint access_type );
void zend_do_declare_class_constant(znode *var_name, const znode *value );

void zend_do_fetch_property(znode *result, znode *object, const znode *property );

void zend_do_halt_compiler_register(void);

void zend_do_push_object(const znode *object );
void zend_do_pop_object(znode *object );


void zend_do_begin_new_object(znode *new_token, znode *class_type );
void zend_do_end_new_object(znode *result, const znode *new_token, const znode *argument_list );

void zend_do_fetch_constant(znode *result, znode *constant_container, znode *constant_name, int mode, zend_bool check_namespace );

void zend_do_shell_exec(znode *result, const znode *cmd );

void zend_do_init_array(znode *result, const znode *expr, const znode *offset, zend_bool is_ref );
void zend_do_add_array_element(znode *result, const znode *expr, const znode *offset, zend_bool is_ref );
void zend_do_add_static_array_element(znode *result, znode *offset, const znode *expr);
void zend_do_list_init(void);
void zend_do_list_end(znode *result, znode *expr );
void zend_do_add_list_element(const znode *element );
void zend_do_new_list_begin(void);
void zend_do_new_list_end(void);


void zend_init_list(void *result, void *item );
void zend_add_to_list(void *result, void *item );


void zend_do_cast(znode *result, const znode *expr, int type );
void zend_do_include_or_eval(int type, znode *result, const znode *op1 );

void zend_do_unset(const znode *variable );
void zend_do_isset_or_isempty(int type, znode *result, znode *variable );

void zend_do_instanceof(znode *result, const znode *expr, const znode *class_znode, int type );

void zend_do_foreach_begin(znode *foreach_token, znode *open_brackets_token, znode *array, znode *as_token, int variable );
void zend_do_foreach_cont(znode *foreach_token, const znode *open_brackets_token, const znode *as_token, znode *value, znode *key );
void zend_do_foreach_end(const znode *foreach_token, const znode *as_token );

void zend_do_declare_begin(void);
void zend_do_declare_stmt(znode *var, znode *val );
void zend_do_declare_end(const znode *declare_token );

void zend_do_exit(znode *result, const znode *message );

void zend_do_begin_silence(znode *strudel_token );
void zend_do_end_silence(const znode *strudel_token );

void zend_do_jmp_set(const znode *value, znode *jmp_token, znode *colon_token );
void zend_do_jmp_set_else(znode *result, const znode *false_value, const znode *jmp_token, const znode *colon_token );

void zend_do_begin_qm_op(const znode *cond, znode *qm_token );
void zend_do_qm_true(const znode *true_value, znode *qm_token, znode *colon_token );
void zend_do_qm_false(znode *result, const znode *false_value, const znode *qm_token, const znode *colon_token );

void zend_do_extended_info(void);
void zend_do_extended_fcall_begin(void);
void zend_do_extended_fcall_end(void);

void zend_do_ticks(void);

void zend_do_abstract_method(const znode *function_name, znode *modifiers, const znode *body );

void zend_do_declare_constant(znode *name, znode *value );
void zend_do_build_namespace_name(znode *result, znode *prefix, znode *name );
void zend_do_begin_namespace(const znode *name, zend_bool with_brackets );
void zend_do_end_namespace(void);
void zend_verify_namespace(void);
void zend_do_use(znode *name, znode *new_name, int is_global );
void zend_do_end_compilation(void);

void zend_do_label(znode *label );
void zend_do_goto(const znode *label );
void zend_resolve_goto_label(zend_op_array *op_array, zend_op *opline, int pass2 );
void zend_release_labels(void);

__attribute__ ((visibility("default"))) void function_add_ref(zend_function *function);






__attribute__ ((visibility("default"))) zend_op_array *compile_file(zend_file_handle *file_handle, int type );
__attribute__ ((visibility("default"))) zend_op_array *compile_string(zval *source_string, char *filename );
__attribute__ ((visibility("default"))) zend_op_array *compile_filename(int type, zval *filename );
__attribute__ ((visibility("default"))) int zend_execute_scripts(int type , zval **retval, int file_count, ...);
__attribute__ ((visibility("default"))) int open_file_for_scanning(zend_file_handle *file_handle );
__attribute__ ((visibility("default"))) void init_op_array(zend_op_array *op_array, zend_uchar type, int initial_ops_size );
__attribute__ ((visibility("default"))) void destroy_op_array(zend_op_array *op_array );
__attribute__ ((visibility("default"))) void zend_destroy_file_handle(zend_file_handle *file_handle );
__attribute__ ((visibility("default"))) int zend_cleanup_class_data(zend_class_entry **pce );
__attribute__ ((visibility("default"))) int zend_cleanup_user_class_data(zend_class_entry **pce );
__attribute__ ((visibility("default"))) void zend_cleanup_internal_class_data(zend_class_entry *ce );
__attribute__ ((visibility("default"))) void zend_cleanup_internal_classes(void);
__attribute__ ((visibility("default"))) int zend_cleanup_function_data(zend_function *function );
__attribute__ ((visibility("default"))) int zend_cleanup_function_data_full(zend_function *function );

__attribute__ ((visibility("default"))) void destroy_zend_function(zend_function *function );
__attribute__ ((visibility("default"))) void zend_function_dtor(zend_function *function);
__attribute__ ((visibility("default"))) void destroy_zend_class(zend_class_entry **pce);
void zend_class_add_ref(zend_class_entry **ce);

__attribute__ ((visibility("default"))) void zend_mangle_property_name(char **dest, int *dest_length, const char *src1, int src1_length, const char *src2, int src2_length, int internal);
__attribute__ ((visibility("default"))) int zend_unmangle_property_name(const char *mangled_property, int mangled_property_len, const char **class_name, const char **prop_name);




zend_op *get_next_op(zend_op_array *op_array );
void init_op(zend_op *op );
int get_next_op_number(zend_op_array *op_array);
int print_class(zend_class_entry *class_entry );
void print_op_array(zend_op_array *op_array, int optimizations);
__attribute__ ((visibility("default"))) int pass_two(zend_op_array *op_array );
zend_brk_cont_element *get_next_brk_cont_element(zend_op_array *op_array);
void zend_do_first_catch(znode *open_parentheses );
void zend_initialize_try_catch_element(const znode *try_token );
void zend_do_mark_last_catch(const znode *first_catch, const znode *last_additional_catch );
__attribute__ ((visibility("default"))) zend_bool zend_is_compiling(void);
__attribute__ ((visibility("default"))) char *zend_make_compiled_string_description(const char *name );
__attribute__ ((visibility("default"))) void zend_initialize_class_data(zend_class_entry *ce, zend_bool nullify_handlers );
int zend_get_class_fetch_type(const char *class_name, uint class_name_len);

typedef zend_bool (*zend_auto_global_callback)(const char *name, uint name_len );
typedef struct _zend_auto_global {
 const char *name;
 uint name_len;
 zend_auto_global_callback auto_global_callback;
 zend_bool jit;
 zend_bool armed;
} zend_auto_global;

__attribute__ ((visibility("default"))) int zend_register_auto_global(const char *name, uint name_len, zend_bool jit, zend_auto_global_callback auto_global_callback );
__attribute__ ((visibility("default"))) void zend_activate_auto_globals(void);
__attribute__ ((visibility("default"))) zend_bool zend_is_auto_global(const char *name, uint name_len );
__attribute__ ((visibility("default"))) zend_bool zend_is_auto_global_quick(const char *name, uint name_len, ulong hashval );
__attribute__ ((visibility("default"))) size_t zend_dirname(char *path, size_t len);

int zendlex(znode *zendlval );

int zend_add_literal(zend_op_array *op_array, const zval *zv );



# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_vm_opcodes.h" 1
# 698 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h" 2
# 820 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_compile.h"

# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_build.h" 1
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h" 2
# 69 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h"
struct _zend_ini_entry;
typedef struct _zend_module_entry zend_module_entry;
typedef struct _zend_module_dep zend_module_dep;

struct _zend_module_entry {
 unsigned short size;
 unsigned int zend_api;
 unsigned char zend_debug;
 unsigned char zts;
 const struct _zend_ini_entry *ini_entry;
 const struct _zend_module_dep *deps;
 const char *name;
 const struct _zend_function_entry *functions;
 int (*module_startup_func)(int type, int module_number );
 int (*module_shutdown_func)(int type, int module_number );
 int (*request_startup_func)(int type, int module_number );
 int (*request_shutdown_func)(int type, int module_number );
 void (*info_func)(zend_module_entry *zend_module );
 const char *version;
 size_t globals_size;



 void* globals_ptr;

 void (*globals_ctor)(void *global );
 void (*globals_dtor)(void *global );
 int (*post_deactivate_func)(void);
 int module_started;
 unsigned char type;
 void *handle;
 int module_number;
 const char *build_id;
};
# 118 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_modules.h"
struct _zend_module_dep {
 const char *name;
 const char *rel;
 const char *version;
 unsigned char type;
};

extern __attribute__ ((visibility("default"))) HashTable module_registry;

void module_destructor(zend_module_entry *module);
int module_registry_cleanup(zend_module_entry *module );
int module_registry_request_startup(zend_module_entry *module );
int module_registry_unload_temp(const zend_module_entry *module );
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_list.h" 1
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_list.h"





typedef struct _zend_rsrc_list_entry {
 void *ptr;
 int type;
 int refcount;
} zend_rsrc_list_entry;

typedef void (*rsrc_dtor_func_t)(zend_rsrc_list_entry *rsrc );


typedef struct _zend_rsrc_list_dtors_entry {

 void (*list_dtor)(void *);
 void (*plist_dtor)(void *);


 rsrc_dtor_func_t list_dtor_ex;
 rsrc_dtor_func_t plist_dtor_ex;

 const char *type_name;

 int module_number;
 int resource_id;
 unsigned char type;
} zend_rsrc_list_dtors_entry;



__attribute__ ((visibility("default"))) int zend_register_list_destructors(void (*ld)(void *), void (*pld)(void *), int module_number);
__attribute__ ((visibility("default"))) int zend_register_list_destructors_ex(rsrc_dtor_func_t ld, rsrc_dtor_func_t pld, const char *type_name, int module_number);

void list_entry_destructor(void *ptr);
void plist_entry_destructor(void *ptr);

void zend_clean_module_rsrc_dtors(int module_number );
int zend_init_rsrc_list(void);
int zend_init_rsrc_plist(void);
void zend_destroy_rsrc_list(HashTable *ht );
int zend_init_rsrc_list_dtors(void);
void zend_destroy_rsrc_list_dtors(void);

__attribute__ ((visibility("default"))) int zend_list_insert(void *ptr, int type );
__attribute__ ((visibility("default"))) int _zend_list_addref(int id );
__attribute__ ((visibility("default"))) int _zend_list_delete(int id );
__attribute__ ((visibility("default"))) void *_zend_list_find(int id, int *type );





__attribute__ ((visibility("default"))) int zend_register_resource(zval *rsrc_result, void *rsrc_pointer, int rsrc_type );
__attribute__ ((visibility("default"))) void *zend_fetch_resource(zval **passed_id , int default_id, const char *resource_type_name, int *found_resource_type, int num_resource_types, ...);

__attribute__ ((visibility("default"))) const char *zend_rsrc_list_get_rsrc_type(int resource );
__attribute__ ((visibility("default"))) int zend_fetch_list_dtor_id(char *type_name);

extern __attribute__ ((visibility("default"))) int le_index_ptr;
# 116 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_list.h"

# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h" 2


# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h" 1
# 30 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h"
typedef union _temp_variable {
 zval tmp_var;
 struct {
  zval **ptr_ptr;
  zval *ptr;
  zend_bool fcall_returned_reference;
 } var;
 struct {
  zval **ptr_ptr;
  zval *str;
  zend_uint offset;
 } str_offset;
 struct {
  zval **ptr_ptr;
  zval *ptr;
  HashPointer fe_pos;
 } fe;
 zend_class_entry *class_entry;
} temp_variable;



__attribute__ ((visibility("default"))) extern void (*zend_execute)(zend_op_array *op_array );
__attribute__ ((visibility("default"))) extern void (*zend_execute_internal)(zend_execute_data *execute_data_ptr, int return_value_used );

void init_executor(void);
void shutdown_executor(void);
void shutdown_destructors(void);
__attribute__ ((visibility("default"))) void execute(zend_op_array *op_array );
__attribute__ ((visibility("default"))) void execute_internal(zend_execute_data *execute_data_ptr, int return_value_used );
__attribute__ ((visibility("default"))) int zend_is_true(zval *op);

static inline __attribute__((always_inline)) void safe_free_zval_ptr_rel(zval *p )
{
 ;

 if (p!=(executor_globals.uninitialized_zval_ptr)) {
  do { if (((gc_root_buffer*)(((zend_uintptr_t)(((zval_gc_info*)p)->u.buffered)) & ~0x03))) { gc_remove_zval_from_buffer(p ); }; _efree((p) ); } while (0);
 }
}
__attribute__ ((visibility("default"))) int zend_lookup_class(const char *name, int name_length, zend_class_entry ***ce );
__attribute__ ((visibility("default"))) int zend_lookup_class_ex(const char *name, int name_length, const zend_literal *key, int use_autoload, zend_class_entry ***ce );
__attribute__ ((visibility("default"))) int zend_eval_string(char *str, zval *retval_ptr, char *string_name );
__attribute__ ((visibility("default"))) int zend_eval_stringl(char *str, int str_len, zval *retval_ptr, char *string_name );
__attribute__ ((visibility("default"))) int zend_eval_string_ex(char *str, zval *retval_ptr, char *string_name, int handle_exceptions );
__attribute__ ((visibility("default"))) int zend_eval_stringl_ex(char *str, int str_len, zval *retval_ptr, char *string_name, int handle_exceptions );

__attribute__ ((visibility("default"))) char * zend_verify_arg_class_kind(const zend_arg_info *cur_arg_info, ulong fetch_type, const char **class_name, zend_class_entry **pce );
__attribute__ ((visibility("default"))) int zend_verify_arg_error(int error_type, const zend_function *zf, zend_uint arg_num, const char *need_msg, const char *need_kind, const char *given_msg, const char *given_kind );

static inline __attribute__((always_inline)) void i_zval_ptr_dtor(zval *zval_ptr )
{
 if (!zval_delref_p(zval_ptr)) {
  ;

  if (zval_ptr != &(executor_globals.uninitialized_zval)) {
   if (((gc_root_buffer*)(((zend_uintptr_t)(((zval_gc_info*)zval_ptr)->u.buffered)) & ~0x03))) { gc_remove_zval_from_buffer(zval_ptr ); };
   _zval_dtor((zval_ptr) );
   _efree((zval_ptr) );
  }
 } else {
  ;

  if (zval_refcount_p(zval_ptr) == 1) {
   zval_unset_isref_p(zval_ptr);
  }

  gc_zval_check_possible_root((zval_ptr) );
 }
}

static inline __attribute__((always_inline)) int i_zend_is_true(zval *op)
{
 int result;

 switch ((*op).type) {
  case 0:
   result = 0;
   break;
  case 1:
  case 3:
  case 7:
   result = ((*op).value.lval?1:0);
   break;
  case 2:
   result = ((*op).value.dval ? 1 : 0);
   break;
  case 6:
   if ((*op).value.str.len == 0
    || ((*op).value.str.len==1 && (*op).value.str.val[0]=='0')) {
    result = 0;
   } else {
    result = 1;
   }
   break;
  case 4:
   result = (zend_hash_num_elements((*op).value.ht)?1:0);
   break;
  case 5:
   if(((*op).type == 5 && (((*op)).value.obj.handlers->get_class_entry != ((void *)0)))) {
    ;

    if ((*op).value.obj.handlers->cast_object) {
     zval tmp;
     if ((*op).value.obj.handlers->cast_object(op, &tmp, 3 ) == 0) {
      result = (tmp).value.lval;
      break;
     }
    } else if ((*op).value.obj.handlers->get) {
     zval *tmp = (*op).value.obj.handlers->get(op );
     if((*tmp).type != 5) {

      convert_to_boolean(tmp);
      result = (*tmp).value.lval;
      _zval_ptr_dtor((&tmp) );
      break;
     }
    }
   }
   result = 1;
   break;
  default:
   result = 0;
   break;
 }
 return result;
}

__attribute__ ((visibility("default"))) int zval_update_constant(zval **pp, void *arg );
__attribute__ ((visibility("default"))) int zval_update_constant_inline_change(zval **pp, void *arg );
__attribute__ ((visibility("default"))) int zval_update_constant_no_inline_change(zval **pp, void *arg );
__attribute__ ((visibility("default"))) int zval_update_constant_ex(zval **pp, void *arg, zend_class_entry *scope );




struct _zend_vm_stack {
 void **top;
 void **end;
 zend_vm_stack prev;
};
# 183 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h"
static inline __attribute__((always_inline)) zend_vm_stack zend_vm_stack_new_page(int count) {
 zend_vm_stack page = (zend_vm_stack)_emalloc(((((sizeof(*page)) + 4 - 1) & ~(4 -1)) + sizeof(void*) * count) );

 page->top = ((void**)(((char*)(page)) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1))));
 page->end = page->top + count;
 page->prev = ((void *)0);
 return page;
}

static inline __attribute__((always_inline)) void zend_vm_stack_init(void)
{
 (executor_globals.argument_stack) = zend_vm_stack_new_page(((16 * 1024) - 16));
}

static inline __attribute__((always_inline)) void zend_vm_stack_destroy(void)
{
 zend_vm_stack stack = (executor_globals.argument_stack);

 while (stack != ((void *)0)) {
  zend_vm_stack p = stack->prev;
  _efree((stack) );
  stack = p;
 }
}

static inline __attribute__((always_inline)) void zend_vm_stack_extend(int count )
{
 zend_vm_stack p = zend_vm_stack_new_page(count >= ((16 * 1024) - 16) ? count : ((16 * 1024) - 16));
 p->prev = (executor_globals.argument_stack);
 (executor_globals.argument_stack) = p;
}

static inline __attribute__((always_inline)) void **zend_vm_stack_top(void)
{
 return (executor_globals.argument_stack)->top;
}

static inline __attribute__((always_inline)) void zend_vm_stack_push(void *ptr )
{
 do { if (__builtin_expect((1) > (executor_globals.argument_stack)->end - (executor_globals.argument_stack)->top, 0)) { zend_vm_stack_extend((1) ); } } while (0);
 *((executor_globals.argument_stack)->top++) = ptr;
}

static inline __attribute__((always_inline)) void zend_vm_stack_push_nocheck(void *ptr )
{
 *((executor_globals.argument_stack)->top++) = ptr;
}

static inline __attribute__((always_inline)) void *zend_vm_stack_pop(void)
{
 void *el = *(--(executor_globals.argument_stack)->top);

 if (__builtin_expect((executor_globals.argument_stack)->top == ((void**)(((char*)((executor_globals.argument_stack))) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))), 0)) {
  zend_vm_stack p = (executor_globals.argument_stack);
  (executor_globals.argument_stack) = p->prev;
  _efree((p) );
  }
 return el;
}

static inline __attribute__((always_inline)) void *zend_vm_stack_alloc(size_t size )
{
 void *ret;

 size = (size + (sizeof(void*) - 1)) / sizeof(void*);


 if (4 > sizeof(void*)) {
  int extra = (4 - ((zend_uintptr_t)(executor_globals.argument_stack)->top & (4 - 1))) / sizeof(void*);

  if (__builtin_expect(size + extra + (((sizeof(void*)) + 4 - 1) & ~(4 -1)) / sizeof(void*) > (zend_uintptr_t)((executor_globals.argument_stack)->end - (executor_globals.argument_stack)->top), 0)
                                                                          ) {
   zend_vm_stack_extend(size );
  } else {
   void **old_top = (executor_globals.argument_stack)->top;

   (executor_globals.argument_stack)->top += extra;

   *(executor_globals.argument_stack)->top = (void*)old_top;
   (executor_globals.argument_stack)->top += (((sizeof(void*)) + 4 - 1) & ~(4 -1)) / sizeof(void*);
  }
 } else {
  do { if (__builtin_expect(((int)size) > (executor_globals.argument_stack)->end - (executor_globals.argument_stack)->top, 0)) { zend_vm_stack_extend(((int)size) ); } } while (0);
 }
 ret = (void*)(executor_globals.argument_stack)->top;
 (executor_globals.argument_stack)->top += size;
 return ret;
}

static inline __attribute__((always_inline)) void zend_vm_stack_free_int(void *ptr )
{
 if (__builtin_expect(((void**)(((char*)((executor_globals.argument_stack))) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))) == (void**)ptr, 0)) {
  zend_vm_stack p = (executor_globals.argument_stack);

  (executor_globals.argument_stack) = p->prev;
  _efree((p) );
 } else {
  (executor_globals.argument_stack)->top = (void**)ptr;
 }
}

static inline __attribute__((always_inline)) void zend_vm_stack_free(void *ptr )
{
 if (__builtin_expect(((void**)(((char*)((executor_globals.argument_stack))) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))) == (void**)ptr, 0)) {
  zend_vm_stack p = (executor_globals.argument_stack);

  (executor_globals.argument_stack) = p->prev;
  _efree((p) );
 } else {

  if (4 > sizeof(void*)) {
   ptr = (void*)(((char*)ptr) - (((sizeof(void*)) + 4 - 1) & ~(4 -1)));
   (executor_globals.argument_stack)->top = *(void***)ptr;
  } else {
   (executor_globals.argument_stack)->top = (void**)ptr;
  }
 }
}

static inline __attribute__((always_inline)) void** zend_vm_stack_push_args(int count )
{

 if (__builtin_expect((executor_globals.argument_stack)->top - ((void**)(((char*)((executor_globals.argument_stack))) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))) < count, 0) ||
  __builtin_expect((executor_globals.argument_stack)->top == (executor_globals.argument_stack)->end, 0)) {
  zend_vm_stack p = (executor_globals.argument_stack);

  zend_vm_stack_extend(count + 1 );

  (executor_globals.argument_stack)->top += count;
  *((executor_globals.argument_stack)->top) = (void*)(zend_uintptr_t)count;
  while (count-- > 0) {
   void *data = *(--p->top);

   if (__builtin_expect(p->top == ((void**)(((char*)(p)) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))), 0)) {
    zend_vm_stack r = p;

    (executor_globals.argument_stack)->prev = p->prev;
    p = p->prev;
    _efree((r) );
   }
   *(((void**)(((char*)((executor_globals.argument_stack))) + (((sizeof(struct _zend_vm_stack)) + 4 - 1) & ~(4 -1)))) + count) = data;
  }
  return (executor_globals.argument_stack)->top++;
 }
 *((executor_globals.argument_stack)->top) = (void*)(zend_uintptr_t)count;
 return (executor_globals.argument_stack)->top++;
}

static inline __attribute__((always_inline)) void zend_vm_stack_clear_multiple(void)
{
 void **p = (executor_globals.argument_stack)->top - 1;
 int delete_count = (int)(zend_uintptr_t) *p;

 while (--delete_count>=0) {
  zval *q = *(zval **)(--p);
  *p = ((void *)0);
  i_zval_ptr_dtor(q );
 }
 zend_vm_stack_free_int(p );
}

static inline __attribute__((always_inline)) zval** zend_vm_stack_get_arg(int requested_arg )
{
 void **p = (executor_globals.current_execute_data)->prev_execute_data->function_state.arguments;
 int arg_count = (int)(zend_uintptr_t) *p;

 if (__builtin_expect(requested_arg > arg_count, 0)) {
  return ((void *)0);
 }
 return (zval**)p - arg_count + requested_arg - 1;
}

static inline __attribute__((always_inline)) void zend_arg_types_stack_2_pop(zend_ptr_stack *stack, zval **object, zend_function **fbc)
{
 void *a, *b;

 zend_ptr_stack_2_pop(stack, &a, &b);

 *object = (zval *) a;
 *fbc = (zend_function *) b;
}

static inline __attribute__((always_inline)) void zend_arg_types_stack_3_pop(zend_ptr_stack *stack, zend_class_entry **called_scope, zval **object, zend_function **fbc)
{
 void *a, *b, *c;

 zend_ptr_stack_3_pop(stack, &a, &b, &c);

 *called_scope = (zend_class_entry *) a;
 *object = (zval *) b;
 *fbc = (zend_function *) c;
}

void execute_new_code(void);



__attribute__ ((visibility("default"))) const char *get_active_class_name(const char **space );
__attribute__ ((visibility("default"))) const char *get_active_function_name(void);
__attribute__ ((visibility("default"))) const char *zend_get_executed_filename(void);
__attribute__ ((visibility("default"))) uint zend_get_executed_lineno(void);
__attribute__ ((visibility("default"))) zend_bool zend_is_executing(void);

__attribute__ ((visibility("default"))) void zend_set_timeout(long seconds, int reset_signals);
__attribute__ ((visibility("default"))) void zend_unset_timeout(void);
__attribute__ ((visibility("default"))) void zend_timeout(int dummy);
__attribute__ ((visibility("default"))) zend_class_entry *zend_fetch_class(const char *class_name, uint class_name_len, int fetch_type );
__attribute__ ((visibility("default"))) zend_class_entry *zend_fetch_class_by_name(const char *class_name, uint class_name_len, const zend_literal *key, int fetch_type );
void zend_verify_abstract_class(zend_class_entry *ce );
# 410 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h"
__attribute__ ((visibility("default"))) zval** zend_get_compiled_variable_value(const zend_execute_data *execute_data_ptr, zend_uint var);
# 420 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h"
__attribute__ ((visibility("default"))) int zend_set_user_opcode_handler(zend_uchar opcode, user_opcode_handler_t handler);
__attribute__ ((visibility("default"))) user_opcode_handler_t zend_get_user_opcode_handler(zend_uchar opcode);


typedef struct _zend_free_op {
 zval* var;

} zend_free_op;

__attribute__ ((visibility("default"))) zval *zend_get_zval_ptr(int op_type, const znode_op *node, const temp_variable *Ts, zend_free_op *should_free, int type );
__attribute__ ((visibility("default"))) zval **zend_get_zval_ptr_ptr(int op_type, const znode_op *node, const temp_variable *Ts, zend_free_op *should_free, int type );

__attribute__ ((visibility("default"))) int zend_do_fcall(zend_execute_data *execute_data );
# 451 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_execute.h"

# 31 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h" 2




typedef struct _zend_function_entry {
 const char *fname;
 void (*handler)(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
 const struct _zend_arg_info *arg_info;
 zend_uint num_args;
 zend_uint flags;
} zend_function_entry;

typedef struct _zend_fcall_info {
 size_t size;
 HashTable *function_table;
 zval *function_name;
 HashTable *symbol_table;
 zval **retval_ptr_ptr;
 zend_uint param_count;
 zval ***params;
 zval *object_ptr;
 zend_bool no_separation;
} zend_fcall_info;

typedef struct _zend_fcall_info_cache {
 zend_bool initialized;
 zend_function *function_handler;
 zend_class_entry *calling_scope;
 zend_class_entry *called_scope;
 zval *object_ptr;
} zend_fcall_info_cache;
# 228 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
int zend_next_free_module(void);


__attribute__ ((visibility("default"))) int zend_get_parameters(int ht, int param_count, ...);
__attribute__ ((visibility("default"))) int _zend_get_parameters_array(int ht, int param_count, zval **argument_array );
__attribute__ ((visibility("default"))) __attribute__((deprecated)) int zend_get_parameters_ex(int param_count, ...);
__attribute__ ((visibility("default"))) int _zend_get_parameters_array_ex(int param_count, zval ***argument_array );


__attribute__ ((visibility("default"))) int zend_copy_parameters_array(int param_count, zval *argument_array );
# 249 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) int zend_parse_parameters(int num_args , const char *type_spec, ...);
__attribute__ ((visibility("default"))) int zend_parse_parameters_ex(int flags, int num_args , const char *type_spec, ...);
__attribute__ ((visibility("default"))) char *zend_zval_type_name(const zval *arg);

__attribute__ ((visibility("default"))) int zend_parse_method_parameters(int num_args , zval *this_ptr, const char *type_spec, ...);
__attribute__ ((visibility("default"))) int zend_parse_method_parameters_ex(int flags, int num_args , zval *this_ptr, const char *type_spec, ...);



__attribute__ ((visibility("default"))) int zend_register_functions(zend_class_entry *scope, const zend_function_entry *functions, HashTable *function_table, int type );
__attribute__ ((visibility("default"))) void zend_unregister_functions(const zend_function_entry *functions, int count, HashTable *function_table );
__attribute__ ((visibility("default"))) int zend_startup_module(zend_module_entry *module_entry);
__attribute__ ((visibility("default"))) zend_module_entry* zend_register_internal_module(zend_module_entry *module_entry );
__attribute__ ((visibility("default"))) zend_module_entry* zend_register_module_ex(zend_module_entry *module );
__attribute__ ((visibility("default"))) int zend_startup_module_ex(zend_module_entry *module );
__attribute__ ((visibility("default"))) int zend_startup_modules(void);
__attribute__ ((visibility("default"))) void zend_collect_module_handlers(void);
__attribute__ ((visibility("default"))) void zend_destroy_modules(void);
__attribute__ ((visibility("default"))) void zend_check_magic_method_implementation(const zend_class_entry *ce, const zend_function *fptr, int error_type );

__attribute__ ((visibility("default"))) zend_class_entry *zend_register_internal_class(zend_class_entry *class_entry );
__attribute__ ((visibility("default"))) zend_class_entry *zend_register_internal_class_ex(zend_class_entry *class_entry, zend_class_entry *parent_ce, char *parent_name );
__attribute__ ((visibility("default"))) zend_class_entry *zend_register_internal_interface(zend_class_entry *orig_class_entry );
__attribute__ ((visibility("default"))) void zend_class_implements(zend_class_entry *class_entry , int num_interfaces, ...);

__attribute__ ((visibility("default"))) int zend_register_class_alias_ex(const char *name, int name_len, zend_class_entry *ce );






__attribute__ ((visibility("default"))) int zend_disable_function(char *function_name, uint function_name_length );
__attribute__ ((visibility("default"))) int zend_disable_class(char *class_name, uint class_name_length );

__attribute__ ((visibility("default"))) void zend_wrong_param_count(void);
# 293 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) zend_bool zend_is_callable_ex(zval *callable, zval *object_ptr, uint check_flags, char **callable_name, int *callable_name_len, zend_fcall_info_cache *fcc, char **error );
__attribute__ ((visibility("default"))) zend_bool zend_is_callable(zval *callable, uint check_flags, char **callable_name );
__attribute__ ((visibility("default"))) zend_bool zend_make_callable(zval *callable, char **callable_name );
__attribute__ ((visibility("default"))) const char *zend_get_module_version(const char *module_name);
__attribute__ ((visibility("default"))) int zend_get_module_started(const char *module_name);
__attribute__ ((visibility("default"))) int zend_declare_property(zend_class_entry *ce, const char *name, int name_length, zval *property, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_ex(zend_class_entry *ce, const char *name, int name_length, zval *property, int access_type, const char *doc_comment, int doc_comment_len );
__attribute__ ((visibility("default"))) int zend_declare_property_null(zend_class_entry *ce, const char *name, int name_length, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_bool(zend_class_entry *ce, const char *name, int name_length, long value, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_long(zend_class_entry *ce, const char *name, int name_length, long value, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_double(zend_class_entry *ce, const char *name, int name_length, double value, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_string(zend_class_entry *ce, const char *name, int name_length, const char *value, int access_type );
__attribute__ ((visibility("default"))) int zend_declare_property_stringl(zend_class_entry *ce, const char *name, int name_length, const char *value, int value_len, int access_type );

__attribute__ ((visibility("default"))) int zend_declare_class_constant(zend_class_entry *ce, const char *name, size_t name_length, zval *value );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_null(zend_class_entry *ce, const char *name, size_t name_length );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_long(zend_class_entry *ce, const char *name, size_t name_length, long value );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_bool(zend_class_entry *ce, const char *name, size_t name_length, zend_bool value );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_double(zend_class_entry *ce, const char *name, size_t name_length, double value );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_stringl(zend_class_entry *ce, const char *name, size_t name_length, const char *value, size_t value_length );
__attribute__ ((visibility("default"))) int zend_declare_class_constant_string(zend_class_entry *ce, const char *name, size_t name_length, const char *value );

__attribute__ ((visibility("default"))) void zend_update_class_constants(zend_class_entry *class_type );
__attribute__ ((visibility("default"))) void zend_update_property(zend_class_entry *scope, zval *object, const char *name, int name_length, zval *value );
__attribute__ ((visibility("default"))) void zend_update_property_null(zend_class_entry *scope, zval *object, const char *name, int name_length );
__attribute__ ((visibility("default"))) void zend_update_property_bool(zend_class_entry *scope, zval *object, const char *name, int name_length, long value );
__attribute__ ((visibility("default"))) void zend_update_property_long(zend_class_entry *scope, zval *object, const char *name, int name_length, long value );
__attribute__ ((visibility("default"))) void zend_update_property_double(zend_class_entry *scope, zval *object, const char *name, int name_length, double value );
__attribute__ ((visibility("default"))) void zend_update_property_string(zend_class_entry *scope, zval *object, const char *name, int name_length, const char *value );
__attribute__ ((visibility("default"))) void zend_update_property_stringl(zend_class_entry *scope, zval *object, const char *name, int name_length, const char *value, int value_length );

__attribute__ ((visibility("default"))) int zend_update_static_property(zend_class_entry *scope, const char *name, int name_length, zval *value );
__attribute__ ((visibility("default"))) int zend_update_static_property_null(zend_class_entry *scope, const char *name, int name_length );
__attribute__ ((visibility("default"))) int zend_update_static_property_bool(zend_class_entry *scope, const char *name, int name_length, long value );
__attribute__ ((visibility("default"))) int zend_update_static_property_long(zend_class_entry *scope, const char *name, int name_length, long value );
__attribute__ ((visibility("default"))) int zend_update_static_property_double(zend_class_entry *scope, const char *name, int name_length, double value );
__attribute__ ((visibility("default"))) int zend_update_static_property_string(zend_class_entry *scope, const char *name, int name_length, const char *value );
__attribute__ ((visibility("default"))) int zend_update_static_property_stringl(zend_class_entry *scope, const char *name, int name_length, const char *value, int value_length );

__attribute__ ((visibility("default"))) zval *zend_read_property(zend_class_entry *scope, zval *object, const char *name, int name_length, zend_bool silent );

__attribute__ ((visibility("default"))) zval *zend_read_static_property(zend_class_entry *scope, const char *name, int name_length, zend_bool silent );

__attribute__ ((visibility("default"))) zend_class_entry *zend_get_class_entry(const zval *zobject );
__attribute__ ((visibility("default"))) int zend_get_object_classname(const zval *object, const char **class_name, zend_uint *class_name_len );
__attribute__ ((visibility("default"))) char *zend_get_type_by_const(int type);
# 358 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) int _array_init(zval *arg, uint size );
__attribute__ ((visibility("default"))) int _object_init(zval *arg );
__attribute__ ((visibility("default"))) int _object_init_ex(zval *arg, zend_class_entry *ce );
__attribute__ ((visibility("default"))) int _object_and_properties_init(zval *arg, zend_class_entry *ce, HashTable *properties );
__attribute__ ((visibility("default"))) void object_properties_init(zend_object *object, zend_class_entry *class_type);

__attribute__ ((visibility("default"))) void zend_merge_properties(zval *obj, HashTable *properties, int destroy_ht );


__attribute__ ((visibility("default"))) int add_assoc_function(zval *arg, const char *key, void (*function_ptr)(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used ));

__attribute__ ((visibility("default"))) int add_assoc_long_ex(zval *arg, const char *key, uint key_len, long n);
__attribute__ ((visibility("default"))) int add_assoc_null_ex(zval *arg, const char *key, uint key_len);
__attribute__ ((visibility("default"))) int add_assoc_bool_ex(zval *arg, const char *key, uint key_len, int b);
__attribute__ ((visibility("default"))) int add_assoc_resource_ex(zval *arg, const char *key, uint key_len, int r);
__attribute__ ((visibility("default"))) int add_assoc_double_ex(zval *arg, const char *key, uint key_len, double d);
__attribute__ ((visibility("default"))) int add_assoc_string_ex(zval *arg, const char *key, uint key_len, char *str, int duplicate);
__attribute__ ((visibility("default"))) int add_assoc_stringl_ex(zval *arg, const char *key, uint key_len, char *str, uint length, int duplicate);
__attribute__ ((visibility("default"))) int add_assoc_zval_ex(zval *arg, const char *key, uint key_len, zval *value);
# 393 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) int add_index_long(zval *arg, ulong idx, long n);
__attribute__ ((visibility("default"))) int add_index_null(zval *arg, ulong idx);
__attribute__ ((visibility("default"))) int add_index_bool(zval *arg, ulong idx, int b);
__attribute__ ((visibility("default"))) int add_index_resource(zval *arg, ulong idx, int r);
__attribute__ ((visibility("default"))) int add_index_double(zval *arg, ulong idx, double d);
__attribute__ ((visibility("default"))) int add_index_string(zval *arg, ulong idx, const char *str, int duplicate);
__attribute__ ((visibility("default"))) int add_index_stringl(zval *arg, ulong idx, const char *str, uint length, int duplicate);
__attribute__ ((visibility("default"))) int add_index_zval(zval *arg, ulong index, zval *value);

__attribute__ ((visibility("default"))) int add_next_index_long(zval *arg, long n);
__attribute__ ((visibility("default"))) int add_next_index_null(zval *arg);
__attribute__ ((visibility("default"))) int add_next_index_bool(zval *arg, int b);
__attribute__ ((visibility("default"))) int add_next_index_resource(zval *arg, int r);
__attribute__ ((visibility("default"))) int add_next_index_double(zval *arg, double d);
__attribute__ ((visibility("default"))) int add_next_index_string(zval *arg, const char *str, int duplicate);
__attribute__ ((visibility("default"))) int add_next_index_stringl(zval *arg, const char *str, uint length, int duplicate);
__attribute__ ((visibility("default"))) int add_next_index_zval(zval *arg, zval *value);

__attribute__ ((visibility("default"))) int add_get_assoc_string_ex(zval *arg, const char *key, uint key_len, const char *str, void **dest, int duplicate);
__attribute__ ((visibility("default"))) int add_get_assoc_stringl_ex(zval *arg, const char *key, uint key_len, const char *str, uint length, void **dest, int duplicate);




__attribute__ ((visibility("default"))) int add_get_index_long(zval *arg, ulong idx, long l, void **dest);
__attribute__ ((visibility("default"))) int add_get_index_double(zval *arg, ulong idx, double d, void **dest);
__attribute__ ((visibility("default"))) int add_get_index_string(zval *arg, ulong idx, const char *str, void **dest, int duplicate);
__attribute__ ((visibility("default"))) int add_get_index_stringl(zval *arg, ulong idx, const char *str, uint length, void **dest, int duplicate);

__attribute__ ((visibility("default"))) int add_property_long_ex(zval *arg, const char *key, uint key_len, long l );
__attribute__ ((visibility("default"))) int add_property_null_ex(zval *arg, const char *key, uint key_len );
__attribute__ ((visibility("default"))) int add_property_bool_ex(zval *arg, const char *key, uint key_len, int b );
__attribute__ ((visibility("default"))) int add_property_resource_ex(zval *arg, const char *key, uint key_len, long r );
__attribute__ ((visibility("default"))) int add_property_double_ex(zval *arg, const char *key, uint key_len, double d );
__attribute__ ((visibility("default"))) int add_property_string_ex(zval *arg, const char *key, uint key_len, const char *str, int duplicate );
__attribute__ ((visibility("default"))) int add_property_stringl_ex(zval *arg, const char *key, uint key_len, const char *str, uint length, int duplicate );
__attribute__ ((visibility("default"))) int add_property_zval_ex(zval *arg, const char *key, uint key_len, zval *value );
# 441 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) int call_user_function(HashTable *function_table, zval **object_pp, zval *function_name, zval *retval_ptr, zend_uint param_count, zval *params[] );
__attribute__ ((visibility("default"))) int call_user_function_ex(HashTable *function_table, zval **object_pp, zval *function_name, zval **retval_ptr_ptr, zend_uint param_count, zval **params[], int no_separation, HashTable *symbol_table );

__attribute__ ((visibility("default"))) extern const zend_fcall_info empty_fcall_info;
__attribute__ ((visibility("default"))) extern const zend_fcall_info_cache empty_fcall_info_cache;
# 457 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"
__attribute__ ((visibility("default"))) int zend_fcall_info_init(zval *callable, uint check_flags, zend_fcall_info *fci, zend_fcall_info_cache *fcc, char **callable_name, char **error );




__attribute__ ((visibility("default"))) void zend_fcall_info_args_clear(zend_fcall_info *fci, int free_mem);




__attribute__ ((visibility("default"))) void zend_fcall_info_args_save(zend_fcall_info *fci, int *param_count, zval ****params);



__attribute__ ((visibility("default"))) void zend_fcall_info_args_restore(zend_fcall_info *fci, int param_count, zval ***params);




__attribute__ ((visibility("default"))) int zend_fcall_info_args(zend_fcall_info *fci, zval *args );





__attribute__ ((visibility("default"))) int zend_fcall_info_argp(zend_fcall_info *fci , int argc, zval ***argv);





__attribute__ ((visibility("default"))) int zend_fcall_info_argv(zend_fcall_info *fci , int argc, va_list *argv);





__attribute__ ((visibility("default"))) int zend_fcall_info_argn(zend_fcall_info *fci , int argc, ...);




__attribute__ ((visibility("default"))) int zend_fcall_info_call(zend_fcall_info *fci, zend_fcall_info_cache *fcc, zval **retval, zval *args );

__attribute__ ((visibility("default"))) int zend_call_function(zend_fcall_info *fci, zend_fcall_info_cache *fci_cache );

__attribute__ ((visibility("default"))) int zend_set_hash_symbol(zval *symbol, const char *name, int name_length, zend_bool is_ref, int num_symbol_tables, ...);

__attribute__ ((visibility("default"))) void zend_delete_variable(zend_execute_data *ex, HashTable *ht, const char *name, int name_len, ulong hash_value );

__attribute__ ((visibility("default"))) int zend_delete_global_variable(const char *name, int name_len );

__attribute__ ((visibility("default"))) int zend_delete_global_variable_ex(const char *name, int name_len, ulong hash_value );

__attribute__ ((visibility("default"))) void zend_reset_all_cv(HashTable *symbol_table );

__attribute__ ((visibility("default"))) void zend_rebuild_symbol_table(void);



__attribute__ ((visibility("default"))) void zif_display_disabled_function(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
__attribute__ ((visibility("default"))) void zif_display_disabled_class(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

# 731 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h"

# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 87 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/usr/include/assert.h" 1 3 4
# 88 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 103 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/build-defs.h" 1
# 104 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 128 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"

__attribute__ ((visibility("default"))) size_t php_strlcpy(char *dst, const char *src, size_t siz);







__attribute__ ((visibility("default"))) size_t php_strlcat(char *dst, const char *src, size_t siz);

# 168 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/usr/include/ctype.h" 1 3 4
# 30 "/usr/include/ctype.h" 3 4

# 48 "/usr/include/ctype.h" 3 4
enum
{
  _ISupper = ((0) < 8 ? ((1 << (0)) << 8) : ((1 << (0)) >> 8)),
  _ISlower = ((1) < 8 ? ((1 << (1)) << 8) : ((1 << (1)) >> 8)),
  _ISalpha = ((2) < 8 ? ((1 << (2)) << 8) : ((1 << (2)) >> 8)),
  _ISdigit = ((3) < 8 ? ((1 << (3)) << 8) : ((1 << (3)) >> 8)),
  _ISxdigit = ((4) < 8 ? ((1 << (4)) << 8) : ((1 << (4)) >> 8)),
  _ISspace = ((5) < 8 ? ((1 << (5)) << 8) : ((1 << (5)) >> 8)),
  _ISprint = ((6) < 8 ? ((1 << (6)) << 8) : ((1 << (6)) >> 8)),
  _ISgraph = ((7) < 8 ? ((1 << (7)) << 8) : ((1 << (7)) >> 8)),
  _ISblank = ((8) < 8 ? ((1 << (8)) << 8) : ((1 << (8)) >> 8)),
  _IScntrl = ((9) < 8 ? ((1 << (9)) << 8) : ((1 << (9)) >> 8)),
  _ISpunct = ((10) < 8 ? ((1 << (10)) << 8) : ((1 << (10)) >> 8)),
  _ISalnum = ((11) < 8 ? ((1 << (11)) << 8) : ((1 << (11)) >> 8))
};
# 81 "/usr/include/ctype.h" 3 4
extern __const unsigned short int **__ctype_b_loc (void)
     __attribute__ ((__nothrow__)) __attribute__ ((__const));
extern __const __int32_t **__ctype_tolower_loc (void)
     __attribute__ ((__nothrow__)) __attribute__ ((__const));
extern __const __int32_t **__ctype_toupper_loc (void)
     __attribute__ ((__nothrow__)) __attribute__ ((__const));
# 96 "/usr/include/ctype.h" 3 4






extern int isalnum (int) __attribute__ ((__nothrow__));
extern int isalpha (int) __attribute__ ((__nothrow__));
extern int iscntrl (int) __attribute__ ((__nothrow__));
extern int isdigit (int) __attribute__ ((__nothrow__));
extern int islower (int) __attribute__ ((__nothrow__));
extern int isgraph (int) __attribute__ ((__nothrow__));
extern int isprint (int) __attribute__ ((__nothrow__));
extern int ispunct (int) __attribute__ ((__nothrow__));
extern int isspace (int) __attribute__ ((__nothrow__));
extern int isupper (int) __attribute__ ((__nothrow__));
extern int isxdigit (int) __attribute__ ((__nothrow__));



extern int tolower (int __c) __attribute__ ((__nothrow__));


extern int toupper (int __c) __attribute__ ((__nothrow__));








extern int isblank (int) __attribute__ ((__nothrow__));






extern int isctype (int __c, int __mask) __attribute__ ((__nothrow__));






extern int isascii (int __c) __attribute__ ((__nothrow__));



extern int toascii (int __c) __attribute__ ((__nothrow__));



extern int _toupper (int) __attribute__ ((__nothrow__));
extern int _tolower (int) __attribute__ ((__nothrow__));
# 190 "/usr/include/ctype.h" 3 4
extern __inline int
__attribute__ ((__nothrow__)) tolower (int __c)
{
  return __c >= -128 && __c < 256 ? (*__ctype_tolower_loc ())[__c] : __c;
}

extern __inline int
__attribute__ ((__nothrow__)) toupper (int __c)
{
  return __c >= -128 && __c < 256 ? (*__ctype_toupper_loc ())[__c] : __c;
}
# 247 "/usr/include/ctype.h" 3 4
extern int isalnum_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isalpha_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int iscntrl_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isdigit_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int islower_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isgraph_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isprint_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int ispunct_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isspace_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isupper_l (int, __locale_t) __attribute__ ((__nothrow__));
extern int isxdigit_l (int, __locale_t) __attribute__ ((__nothrow__));

extern int isblank_l (int, __locale_t) __attribute__ ((__nothrow__));



extern int __tolower_l (int __c, __locale_t __l) __attribute__ ((__nothrow__));
extern int tolower_l (int __c, __locale_t __l) __attribute__ ((__nothrow__));


extern int __toupper_l (int __c, __locale_t __l) __attribute__ ((__nothrow__));
extern int toupper_l (int __c, __locale_t __l) __attribute__ ((__nothrow__));
# 323 "/usr/include/ctype.h" 3 4

# 169 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2

# 1 "/usr/include/unistd.h" 1 3 4
# 28 "/usr/include/unistd.h" 3 4

# 203 "/usr/include/unistd.h" 3 4
# 1 "/usr/include/bits/posix_opt.h" 1 3 4
# 204 "/usr/include/unistd.h" 2 3 4



# 1 "/usr/include/bits/environments.h" 1 3 4
# 208 "/usr/include/unistd.h" 2 3 4
# 227 "/usr/include/unistd.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 228 "/usr/include/unistd.h" 2 3 4
# 268 "/usr/include/unistd.h" 3 4
typedef __intptr_t intptr_t;






typedef __socklen_t socklen_t;
# 288 "/usr/include/unistd.h" 3 4
extern int access (__const char *__name, int __type) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));




extern int euidaccess (__const char *__name, int __type)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int eaccess (__const char *__name, int __type)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern int faccessat (int __fd, __const char *__file, int __type, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2))) ;
# 331 "/usr/include/unistd.h" 3 4
extern __off_t lseek (int __fd, __off_t __offset, int __whence) __attribute__ ((__nothrow__));
# 342 "/usr/include/unistd.h" 3 4
extern __off64_t lseek64 (int __fd, __off64_t __offset, int __whence)
     __attribute__ ((__nothrow__));






extern int close (int __fd);






extern ssize_t read (int __fd, void *__buf, size_t __nbytes) ;





extern ssize_t write (int __fd, __const void *__buf, size_t __n) ;
# 373 "/usr/include/unistd.h" 3 4
extern ssize_t pread (int __fd, void *__buf, size_t __nbytes,
        __off_t __offset) ;






extern ssize_t pwrite (int __fd, __const void *__buf, size_t __n,
         __off_t __offset) ;
# 401 "/usr/include/unistd.h" 3 4
extern ssize_t pread64 (int __fd, void *__buf, size_t __nbytes,
   __off64_t __offset) ;


extern ssize_t pwrite64 (int __fd, __const void *__buf, size_t __n,
    __off64_t __offset) ;







extern int pipe (int __pipedes[2]) __attribute__ ((__nothrow__)) ;




extern int pipe2 (int __pipedes[2], int __flags) __attribute__ ((__nothrow__)) ;
# 429 "/usr/include/unistd.h" 3 4
extern unsigned int alarm (unsigned int __seconds) __attribute__ ((__nothrow__));
# 441 "/usr/include/unistd.h" 3 4
extern unsigned int sleep (unsigned int __seconds);







extern __useconds_t ualarm (__useconds_t __value, __useconds_t __interval)
     __attribute__ ((__nothrow__));






extern int usleep (__useconds_t __useconds);
# 466 "/usr/include/unistd.h" 3 4
extern int pause (void);



extern int chown (__const char *__file, __uid_t __owner, __gid_t __group)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;



extern int fchown (int __fd, __uid_t __owner, __gid_t __group) __attribute__ ((__nothrow__)) ;




extern int lchown (__const char *__file, __uid_t __owner, __gid_t __group)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;






extern int fchownat (int __fd, __const char *__file, __uid_t __owner,
       __gid_t __group, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2))) ;



extern int chdir (__const char *__path) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;



extern int fchdir (int __fd) __attribute__ ((__nothrow__)) ;
# 508 "/usr/include/unistd.h" 3 4
extern char *getcwd (char *__buf, size_t __size) __attribute__ ((__nothrow__)) ;





extern char *get_current_dir_name (void) __attribute__ ((__nothrow__));







extern char *getwd (char *__buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) __attribute__ ((__deprecated__)) ;




extern int dup (int __fd) __attribute__ ((__nothrow__)) ;


extern int dup2 (int __fd, int __fd2) __attribute__ ((__nothrow__));




extern int dup3 (int __fd, int __fd2, int __flags) __attribute__ ((__nothrow__));



extern char **__environ;

extern char **environ;





extern int execve (__const char *__path, char *__const __argv[],
     char *__const __envp[]) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));




extern int fexecve (int __fd, char *__const __argv[], char *__const __envp[])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));




extern int execv (__const char *__path, char *__const __argv[])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern int execle (__const char *__path, __const char *__arg, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern int execl (__const char *__path, __const char *__arg, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern int execvp (__const char *__file, char *__const __argv[])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));




extern int execlp (__const char *__file, __const char *__arg, ...)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));




extern int execvpe (__const char *__file, char *__const __argv[],
      char *__const __envp[])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));





extern int nice (int __inc) __attribute__ ((__nothrow__)) ;




extern void _exit (int __status) __attribute__ ((__noreturn__));





# 1 "/usr/include/bits/confname.h" 1 3 4
# 26 "/usr/include/bits/confname.h" 3 4
enum
  {
    _PC_LINK_MAX,

    _PC_MAX_CANON,

    _PC_MAX_INPUT,

    _PC_NAME_MAX,

    _PC_PATH_MAX,

    _PC_PIPE_BUF,

    _PC_CHOWN_RESTRICTED,

    _PC_NO_TRUNC,

    _PC_VDISABLE,

    _PC_SYNC_IO,

    _PC_ASYNC_IO,

    _PC_PRIO_IO,

    _PC_SOCK_MAXBUF,

    _PC_FILESIZEBITS,

    _PC_REC_INCR_XFER_SIZE,

    _PC_REC_MAX_XFER_SIZE,

    _PC_REC_MIN_XFER_SIZE,

    _PC_REC_XFER_ALIGN,

    _PC_ALLOC_SIZE_MIN,

    _PC_SYMLINK_MAX,

    _PC_2_SYMLINKS

  };


enum
  {
    _SC_ARG_MAX,

    _SC_CHILD_MAX,

    _SC_CLK_TCK,

    _SC_NGROUPS_MAX,

    _SC_OPEN_MAX,

    _SC_STREAM_MAX,

    _SC_TZNAME_MAX,

    _SC_JOB_CONTROL,

    _SC_SAVED_IDS,

    _SC_REALTIME_SIGNALS,

    _SC_PRIORITY_SCHEDULING,

    _SC_TIMERS,

    _SC_ASYNCHRONOUS_IO,

    _SC_PRIORITIZED_IO,

    _SC_SYNCHRONIZED_IO,

    _SC_FSYNC,

    _SC_MAPPED_FILES,

    _SC_MEMLOCK,

    _SC_MEMLOCK_RANGE,

    _SC_MEMORY_PROTECTION,

    _SC_MESSAGE_PASSING,

    _SC_SEMAPHORES,

    _SC_SHARED_MEMORY_OBJECTS,

    _SC_AIO_LISTIO_MAX,

    _SC_AIO_MAX,

    _SC_AIO_PRIO_DELTA_MAX,

    _SC_DELAYTIMER_MAX,

    _SC_MQ_OPEN_MAX,

    _SC_MQ_PRIO_MAX,

    _SC_VERSION,

    _SC_PAGESIZE,


    _SC_RTSIG_MAX,

    _SC_SEM_NSEMS_MAX,

    _SC_SEM_VALUE_MAX,

    _SC_SIGQUEUE_MAX,

    _SC_TIMER_MAX,




    _SC_BC_BASE_MAX,

    _SC_BC_DIM_MAX,

    _SC_BC_SCALE_MAX,

    _SC_BC_STRING_MAX,

    _SC_COLL_WEIGHTS_MAX,

    _SC_EQUIV_CLASS_MAX,

    _SC_EXPR_NEST_MAX,

    _SC_LINE_MAX,

    _SC_RE_DUP_MAX,

    _SC_CHARCLASS_NAME_MAX,


    _SC_2_VERSION,

    _SC_2_C_BIND,

    _SC_2_C_DEV,

    _SC_2_FORT_DEV,

    _SC_2_FORT_RUN,

    _SC_2_SW_DEV,

    _SC_2_LOCALEDEF,


    _SC_PII,

    _SC_PII_XTI,

    _SC_PII_SOCKET,

    _SC_PII_INTERNET,

    _SC_PII_OSI,

    _SC_POLL,

    _SC_SELECT,

    _SC_UIO_MAXIOV,

    _SC_IOV_MAX = _SC_UIO_MAXIOV,

    _SC_PII_INTERNET_STREAM,

    _SC_PII_INTERNET_DGRAM,

    _SC_PII_OSI_COTS,

    _SC_PII_OSI_CLTS,

    _SC_PII_OSI_M,

    _SC_T_IOV_MAX,



    _SC_THREADS,

    _SC_THREAD_SAFE_FUNCTIONS,

    _SC_GETGR_R_SIZE_MAX,

    _SC_GETPW_R_SIZE_MAX,

    _SC_LOGIN_NAME_MAX,

    _SC_TTY_NAME_MAX,

    _SC_THREAD_DESTRUCTOR_ITERATIONS,

    _SC_THREAD_KEYS_MAX,

    _SC_THREAD_STACK_MIN,

    _SC_THREAD_THREADS_MAX,

    _SC_THREAD_ATTR_STACKADDR,

    _SC_THREAD_ATTR_STACKSIZE,

    _SC_THREAD_PRIORITY_SCHEDULING,

    _SC_THREAD_PRIO_INHERIT,

    _SC_THREAD_PRIO_PROTECT,

    _SC_THREAD_PROCESS_SHARED,


    _SC_NPROCESSORS_CONF,

    _SC_NPROCESSORS_ONLN,

    _SC_PHYS_PAGES,

    _SC_AVPHYS_PAGES,

    _SC_ATEXIT_MAX,

    _SC_PASS_MAX,


    _SC_XOPEN_VERSION,

    _SC_XOPEN_XCU_VERSION,

    _SC_XOPEN_UNIX,

    _SC_XOPEN_CRYPT,

    _SC_XOPEN_ENH_I18N,

    _SC_XOPEN_SHM,


    _SC_2_CHAR_TERM,

    _SC_2_C_VERSION,

    _SC_2_UPE,


    _SC_XOPEN_XPG2,

    _SC_XOPEN_XPG3,

    _SC_XOPEN_XPG4,


    _SC_CHAR_BIT,

    _SC_CHAR_MAX,

    _SC_CHAR_MIN,

    _SC_INT_MAX,

    _SC_INT_MIN,

    _SC_LONG_BIT,

    _SC_WORD_BIT,

    _SC_MB_LEN_MAX,

    _SC_NZERO,

    _SC_SSIZE_MAX,

    _SC_SCHAR_MAX,

    _SC_SCHAR_MIN,

    _SC_SHRT_MAX,

    _SC_SHRT_MIN,

    _SC_UCHAR_MAX,

    _SC_UINT_MAX,

    _SC_ULONG_MAX,

    _SC_USHRT_MAX,


    _SC_NL_ARGMAX,

    _SC_NL_LANGMAX,

    _SC_NL_MSGMAX,

    _SC_NL_NMAX,

    _SC_NL_SETMAX,

    _SC_NL_TEXTMAX,


    _SC_XBS5_ILP32_OFF32,

    _SC_XBS5_ILP32_OFFBIG,

    _SC_XBS5_LP64_OFF64,

    _SC_XBS5_LPBIG_OFFBIG,


    _SC_XOPEN_LEGACY,

    _SC_XOPEN_REALTIME,

    _SC_XOPEN_REALTIME_THREADS,


    _SC_ADVISORY_INFO,

    _SC_BARRIERS,

    _SC_BASE,

    _SC_C_LANG_SUPPORT,

    _SC_C_LANG_SUPPORT_R,

    _SC_CLOCK_SELECTION,

    _SC_CPUTIME,

    _SC_THREAD_CPUTIME,

    _SC_DEVICE_IO,

    _SC_DEVICE_SPECIFIC,

    _SC_DEVICE_SPECIFIC_R,

    _SC_FD_MGMT,

    _SC_FIFO,

    _SC_PIPE,

    _SC_FILE_ATTRIBUTES,

    _SC_FILE_LOCKING,

    _SC_FILE_SYSTEM,

    _SC_MONOTONIC_CLOCK,

    _SC_MULTI_PROCESS,

    _SC_SINGLE_PROCESS,

    _SC_NETWORKING,

    _SC_READER_WRITER_LOCKS,

    _SC_SPIN_LOCKS,

    _SC_REGEXP,

    _SC_REGEX_VERSION,

    _SC_SHELL,

    _SC_SIGNALS,

    _SC_SPAWN,

    _SC_SPORADIC_SERVER,

    _SC_THREAD_SPORADIC_SERVER,

    _SC_SYSTEM_DATABASE,

    _SC_SYSTEM_DATABASE_R,

    _SC_TIMEOUTS,

    _SC_TYPED_MEMORY_OBJECTS,

    _SC_USER_GROUPS,

    _SC_USER_GROUPS_R,

    _SC_2_PBS,

    _SC_2_PBS_ACCOUNTING,

    _SC_2_PBS_LOCATE,

    _SC_2_PBS_MESSAGE,

    _SC_2_PBS_TRACK,

    _SC_SYMLOOP_MAX,

    _SC_STREAMS,

    _SC_2_PBS_CHECKPOINT,


    _SC_V6_ILP32_OFF32,

    _SC_V6_ILP32_OFFBIG,

    _SC_V6_LP64_OFF64,

    _SC_V6_LPBIG_OFFBIG,


    _SC_HOST_NAME_MAX,

    _SC_TRACE,

    _SC_TRACE_EVENT_FILTER,

    _SC_TRACE_INHERIT,

    _SC_TRACE_LOG,


    _SC_LEVEL1_ICACHE_SIZE,

    _SC_LEVEL1_ICACHE_ASSOC,

    _SC_LEVEL1_ICACHE_LINESIZE,

    _SC_LEVEL1_DCACHE_SIZE,

    _SC_LEVEL1_DCACHE_ASSOC,

    _SC_LEVEL1_DCACHE_LINESIZE,

    _SC_LEVEL2_CACHE_SIZE,

    _SC_LEVEL2_CACHE_ASSOC,

    _SC_LEVEL2_CACHE_LINESIZE,

    _SC_LEVEL3_CACHE_SIZE,

    _SC_LEVEL3_CACHE_ASSOC,

    _SC_LEVEL3_CACHE_LINESIZE,

    _SC_LEVEL4_CACHE_SIZE,

    _SC_LEVEL4_CACHE_ASSOC,

    _SC_LEVEL4_CACHE_LINESIZE,



    _SC_IPV6 = _SC_LEVEL1_ICACHE_SIZE + 50,

    _SC_RAW_SOCKETS,


    _SC_V7_ILP32_OFF32,

    _SC_V7_ILP32_OFFBIG,

    _SC_V7_LP64_OFF64,

    _SC_V7_LPBIG_OFFBIG,


    _SC_SS_REPL_MAX,


    _SC_TRACE_EVENT_NAME_MAX,

    _SC_TRACE_NAME_MAX,

    _SC_TRACE_SYS_MAX,

    _SC_TRACE_USER_EVENT_MAX,


    _SC_XOPEN_STREAMS,


    _SC_THREAD_ROBUST_PRIO_INHERIT,

    _SC_THREAD_ROBUST_PRIO_PROTECT

  };


enum
  {
    _CS_PATH,


    _CS_V6_WIDTH_RESTRICTED_ENVS,



    _CS_GNU_LIBC_VERSION,

    _CS_GNU_LIBPTHREAD_VERSION,


    _CS_V5_WIDTH_RESTRICTED_ENVS,



    _CS_V7_WIDTH_RESTRICTED_ENVS,



    _CS_LFS_CFLAGS = 1000,

    _CS_LFS_LDFLAGS,

    _CS_LFS_LIBS,

    _CS_LFS_LINTFLAGS,

    _CS_LFS64_CFLAGS,

    _CS_LFS64_LDFLAGS,

    _CS_LFS64_LIBS,

    _CS_LFS64_LINTFLAGS,


    _CS_XBS5_ILP32_OFF32_CFLAGS = 1100,

    _CS_XBS5_ILP32_OFF32_LDFLAGS,

    _CS_XBS5_ILP32_OFF32_LIBS,

    _CS_XBS5_ILP32_OFF32_LINTFLAGS,

    _CS_XBS5_ILP32_OFFBIG_CFLAGS,

    _CS_XBS5_ILP32_OFFBIG_LDFLAGS,

    _CS_XBS5_ILP32_OFFBIG_LIBS,

    _CS_XBS5_ILP32_OFFBIG_LINTFLAGS,

    _CS_XBS5_LP64_OFF64_CFLAGS,

    _CS_XBS5_LP64_OFF64_LDFLAGS,

    _CS_XBS5_LP64_OFF64_LIBS,

    _CS_XBS5_LP64_OFF64_LINTFLAGS,

    _CS_XBS5_LPBIG_OFFBIG_CFLAGS,

    _CS_XBS5_LPBIG_OFFBIG_LDFLAGS,

    _CS_XBS5_LPBIG_OFFBIG_LIBS,

    _CS_XBS5_LPBIG_OFFBIG_LINTFLAGS,


    _CS_POSIX_V6_ILP32_OFF32_CFLAGS,

    _CS_POSIX_V6_ILP32_OFF32_LDFLAGS,

    _CS_POSIX_V6_ILP32_OFF32_LIBS,

    _CS_POSIX_V6_ILP32_OFF32_LINTFLAGS,

    _CS_POSIX_V6_ILP32_OFFBIG_CFLAGS,

    _CS_POSIX_V6_ILP32_OFFBIG_LDFLAGS,

    _CS_POSIX_V6_ILP32_OFFBIG_LIBS,

    _CS_POSIX_V6_ILP32_OFFBIG_LINTFLAGS,

    _CS_POSIX_V6_LP64_OFF64_CFLAGS,

    _CS_POSIX_V6_LP64_OFF64_LDFLAGS,

    _CS_POSIX_V6_LP64_OFF64_LIBS,

    _CS_POSIX_V6_LP64_OFF64_LINTFLAGS,

    _CS_POSIX_V6_LPBIG_OFFBIG_CFLAGS,

    _CS_POSIX_V6_LPBIG_OFFBIG_LDFLAGS,

    _CS_POSIX_V6_LPBIG_OFFBIG_LIBS,

    _CS_POSIX_V6_LPBIG_OFFBIG_LINTFLAGS,


    _CS_POSIX_V7_ILP32_OFF32_CFLAGS,

    _CS_POSIX_V7_ILP32_OFF32_LDFLAGS,

    _CS_POSIX_V7_ILP32_OFF32_LIBS,

    _CS_POSIX_V7_ILP32_OFF32_LINTFLAGS,

    _CS_POSIX_V7_ILP32_OFFBIG_CFLAGS,

    _CS_POSIX_V7_ILP32_OFFBIG_LDFLAGS,

    _CS_POSIX_V7_ILP32_OFFBIG_LIBS,

    _CS_POSIX_V7_ILP32_OFFBIG_LINTFLAGS,

    _CS_POSIX_V7_LP64_OFF64_CFLAGS,

    _CS_POSIX_V7_LP64_OFF64_LDFLAGS,

    _CS_POSIX_V7_LP64_OFF64_LIBS,

    _CS_POSIX_V7_LP64_OFF64_LINTFLAGS,

    _CS_POSIX_V7_LPBIG_OFFBIG_CFLAGS,

    _CS_POSIX_V7_LPBIG_OFFBIG_LDFLAGS,

    _CS_POSIX_V7_LPBIG_OFFBIG_LIBS,

    _CS_POSIX_V7_LPBIG_OFFBIG_LINTFLAGS,


    _CS_V6_ENV,

    _CS_V7_ENV

  };
# 607 "/usr/include/unistd.h" 2 3 4


extern long int pathconf (__const char *__path, int __name)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern long int fpathconf (int __fd, int __name) __attribute__ ((__nothrow__));


extern long int sysconf (int __name) __attribute__ ((__nothrow__));



extern size_t confstr (int __name, char *__buf, size_t __len) __attribute__ ((__nothrow__));




extern __pid_t getpid (void) __attribute__ ((__nothrow__));


extern __pid_t getppid (void) __attribute__ ((__nothrow__));




extern __pid_t getpgrp (void) __attribute__ ((__nothrow__));
# 643 "/usr/include/unistd.h" 3 4
extern __pid_t __getpgid (__pid_t __pid) __attribute__ ((__nothrow__));

extern __pid_t getpgid (__pid_t __pid) __attribute__ ((__nothrow__));






extern int setpgid (__pid_t __pid, __pid_t __pgid) __attribute__ ((__nothrow__));
# 669 "/usr/include/unistd.h" 3 4
extern int setpgrp (void) __attribute__ ((__nothrow__));
# 686 "/usr/include/unistd.h" 3 4
extern __pid_t setsid (void) __attribute__ ((__nothrow__));



extern __pid_t getsid (__pid_t __pid) __attribute__ ((__nothrow__));



extern __uid_t getuid (void) __attribute__ ((__nothrow__));


extern __uid_t geteuid (void) __attribute__ ((__nothrow__));


extern __gid_t getgid (void) __attribute__ ((__nothrow__));


extern __gid_t getegid (void) __attribute__ ((__nothrow__));




extern int getgroups (int __size, __gid_t __list[]) __attribute__ ((__nothrow__)) ;



extern int group_member (__gid_t __gid) __attribute__ ((__nothrow__));






extern int setuid (__uid_t __uid) __attribute__ ((__nothrow__));




extern int setreuid (__uid_t __ruid, __uid_t __euid) __attribute__ ((__nothrow__));




extern int seteuid (__uid_t __uid) __attribute__ ((__nothrow__));






extern int setgid (__gid_t __gid) __attribute__ ((__nothrow__));




extern int setregid (__gid_t __rgid, __gid_t __egid) __attribute__ ((__nothrow__));




extern int setegid (__gid_t __gid) __attribute__ ((__nothrow__));





extern int getresuid (__uid_t *__ruid, __uid_t *__euid, __uid_t *__suid)
     __attribute__ ((__nothrow__));



extern int getresgid (__gid_t *__rgid, __gid_t *__egid, __gid_t *__sgid)
     __attribute__ ((__nothrow__));



extern int setresuid (__uid_t __ruid, __uid_t __euid, __uid_t __suid)
     __attribute__ ((__nothrow__));



extern int setresgid (__gid_t __rgid, __gid_t __egid, __gid_t __sgid)
     __attribute__ ((__nothrow__));






extern __pid_t fork (void) __attribute__ ((__nothrow__));







extern __pid_t vfork (void) __attribute__ ((__nothrow__));





extern char *ttyname (int __fd) __attribute__ ((__nothrow__));



extern int ttyname_r (int __fd, char *__buf, size_t __buflen)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2))) ;



extern int isatty (int __fd) __attribute__ ((__nothrow__));





extern int ttyslot (void) __attribute__ ((__nothrow__));




extern int link (__const char *__from, __const char *__to)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2))) ;




extern int linkat (int __fromfd, __const char *__from, int __tofd,
     __const char *__to, int __flags)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4))) ;




extern int symlink (__const char *__from, __const char *__to)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2))) ;




extern ssize_t readlink (__const char *__restrict __path,
    char *__restrict __buf, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2))) ;




extern int symlinkat (__const char *__from, int __tofd,
        __const char *__to) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 3))) ;


extern ssize_t readlinkat (int __fd, __const char *__restrict __path,
      char *__restrict __buf, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3))) ;



extern int unlink (__const char *__name) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern int unlinkat (int __fd, __const char *__name, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));



extern int rmdir (__const char *__path) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern __pid_t tcgetpgrp (int __fd) __attribute__ ((__nothrow__));


extern int tcsetpgrp (int __fd, __pid_t __pgrp_id) __attribute__ ((__nothrow__));






extern char *getlogin (void);







extern int getlogin_r (char *__name, size_t __name_len) __attribute__ ((__nonnull__ (1)));




extern int setlogin (__const char *__name) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 890 "/usr/include/unistd.h" 3 4
# 1 "/usr/include/getopt.h" 1 3 4
# 59 "/usr/include/getopt.h" 3 4
extern char *optarg;
# 73 "/usr/include/getopt.h" 3 4
extern int optind;




extern int opterr;



extern int optopt;
# 152 "/usr/include/getopt.h" 3 4
extern int getopt (int ___argc, char *const *___argv, const char *__shortopts)
       __attribute__ ((__nothrow__));
# 891 "/usr/include/unistd.h" 2 3 4







extern int gethostname (char *__name, size_t __len) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern int sethostname (__const char *__name, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;



extern int sethostid (long int __id) __attribute__ ((__nothrow__)) ;





extern int getdomainname (char *__name, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
extern int setdomainname (__const char *__name, size_t __len)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





extern int vhangup (void) __attribute__ ((__nothrow__));


extern int revoke (__const char *__file) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;







extern int profil (unsigned short int *__sample_buffer, size_t __size,
     size_t __offset, unsigned int __scale)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int acct (__const char *__name) __attribute__ ((__nothrow__));



extern char *getusershell (void) __attribute__ ((__nothrow__));
extern void endusershell (void) __attribute__ ((__nothrow__));
extern void setusershell (void) __attribute__ ((__nothrow__));





extern int daemon (int __nochdir, int __noclose) __attribute__ ((__nothrow__)) ;






extern int chroot (__const char *__path) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;



extern char *getpass (__const char *__prompt) __attribute__ ((__nonnull__ (1)));
# 976 "/usr/include/unistd.h" 3 4
extern int fsync (int __fd);






extern long int gethostid (void);


extern void sync (void) __attribute__ ((__nothrow__));





extern int getpagesize (void) __attribute__ ((__nothrow__)) __attribute__ ((__const__));




extern int getdtablesize (void) __attribute__ ((__nothrow__));
# 1007 "/usr/include/unistd.h" 3 4
extern int truncate (__const char *__file, __off_t __length)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;
# 1019 "/usr/include/unistd.h" 3 4
extern int truncate64 (__const char *__file, __off64_t __length)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;





extern int ftruncate (int __fd, __off_t __length) __attribute__ ((__nothrow__)) ;
# 1036 "/usr/include/unistd.h" 3 4
extern int ftruncate64 (int __fd, __off64_t __length) __attribute__ ((__nothrow__)) ;
# 1047 "/usr/include/unistd.h" 3 4
extern int brk (void *__addr) __attribute__ ((__nothrow__)) ;





extern void *sbrk (intptr_t __delta) __attribute__ ((__nothrow__));
# 1068 "/usr/include/unistd.h" 3 4
extern long int syscall (long int __sysno, ...) __attribute__ ((__nothrow__));
# 1091 "/usr/include/unistd.h" 3 4
extern int lockf (int __fd, int __cmd, __off_t __len) ;
# 1101 "/usr/include/unistd.h" 3 4
extern int lockf64 (int __fd, int __cmd, __off64_t __len) ;
# 1122 "/usr/include/unistd.h" 3 4
extern int fdatasync (int __fildes);







extern char *crypt (__const char *__key, __const char *__salt)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern void encrypt (char *__block, int __edflag) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));






extern void swab (__const void *__restrict __from, void *__restrict __to,
    ssize_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));







extern char *ctermid (char *__s) __attribute__ ((__nothrow__));
# 1160 "/usr/include/unistd.h" 3 4

# 171 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 188 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h" 1
# 189 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_alloc.h" 1
# 190 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_stack.h" 1
# 191 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 211 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/usr/include/pwd.h" 1 3 4
# 28 "/usr/include/pwd.h" 3 4





# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 34 "/usr/include/pwd.h" 2 3 4
# 50 "/usr/include/pwd.h" 3 4
struct passwd
{
  char *pw_name;
  char *pw_passwd;
  __uid_t pw_uid;
  __gid_t pw_gid;
  char *pw_gecos;
  char *pw_dir;
  char *pw_shell;
};
# 73 "/usr/include/pwd.h" 3 4
extern void setpwent (void);





extern void endpwent (void);





extern struct passwd *getpwent (void);
# 95 "/usr/include/pwd.h" 3 4
extern struct passwd *fgetpwent (FILE *__stream);







extern int putpwent (__const struct passwd *__restrict __p,
       FILE *__restrict __f);






extern struct passwd *getpwuid (__uid_t __uid);





extern struct passwd *getpwnam (__const char *__name);
# 140 "/usr/include/pwd.h" 3 4
extern int getpwent_r (struct passwd *__restrict __resultbuf,
         char *__restrict __buffer, size_t __buflen,
         struct passwd **__restrict __result);


extern int getpwuid_r (__uid_t __uid,
         struct passwd *__restrict __resultbuf,
         char *__restrict __buffer, size_t __buflen,
         struct passwd **__restrict __result);

extern int getpwnam_r (__const char *__restrict __name,
         struct passwd *__restrict __resultbuf,
         char *__restrict __buffer, size_t __buflen,
         struct passwd **__restrict __result);
# 164 "/usr/include/pwd.h" 3 4
extern int fgetpwent_r (FILE *__restrict __stream,
   struct passwd *__restrict __resultbuf,
   char *__restrict __buffer, size_t __buflen,
   struct passwd **__restrict __result);
# 181 "/usr/include/pwd.h" 3 4
extern int getpw (__uid_t __uid, char *__buffer);



# 212 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/usr/include/sys/param.h" 1 3 4
# 26 "/usr/include/sys/param.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 27 "/usr/include/sys/param.h" 2 3 4

# 1 "/usr/include/linux/param.h" 1 3 4



# 1 "/usr/include/asm/param.h" 1 3 4
# 1 "/usr/include/asm-generic/param.h" 1 3 4
# 1 "/usr/include/asm/param.h" 2 3 4
# 5 "/usr/include/linux/param.h" 2 3 4
# 29 "/usr/include/sys/param.h" 2 3 4
# 213 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2




# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 218 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 240 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/snprintf.h" 1
# 73 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/snprintf.h"
typedef int bool_int;

typedef enum {
 NO = 0, YES = 1
} boolean_e;



__attribute__ ((visibility("default"))) int ap_php_slprintf(char *buf, size_t len, const char *format,...);
__attribute__ ((visibility("default"))) int ap_php_vslprintf(char *buf, size_t len, const char *format, va_list ap);
__attribute__ ((visibility("default"))) int ap_php_snprintf(char *, size_t, const char *, ...);
__attribute__ ((visibility("default"))) int ap_php_vsnprintf(char *, size_t, const char *, va_list ap);
__attribute__ ((visibility("default"))) int ap_php_vasprintf(char **buf, const char *format, va_list ap);
__attribute__ ((visibility("default"))) int ap_php_asprintf(char **buf, const char *format, ...);
__attribute__ ((visibility("default"))) int php_sprintf (char* s, const char* format, ...) __attribute__ ((format(printf, 2, 3)));
__attribute__ ((visibility("default"))) char * php_gcvt(double value, int ndigit, char dec_point, char exponent, char *buf);
__attribute__ ((visibility("default"))) char * php_conv_fp(register char format, register double num,
   boolean_e add_dp, int precision, char dec_point, bool_int * is_negative, char *buf, int *len);


# 127 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/snprintf.h"
typedef enum {
 LM_STD = 0,

 LM_INTMAX_T,


 LM_PTRDIFF_T,


 LM_LONG_LONG,

 LM_SIZE_T,
 LM_LONG,
 LM_LONG_DOUBLE
} length_modifier_e;
# 152 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/snprintf.h"
typedef long long int wide_int;
typedef unsigned long long int u_wide_int;

extern char * ap_php_conv_10(register wide_int num, register bool_int is_unsigned,
    register bool_int * is_negative, char *buf_end, register int *len);

extern char * ap_php_conv_p2(register u_wide_int num, register int nbits,
   char format, char *buf_end, register int *len);
# 242 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/spprintf.h" 1
# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/spprintf.h"

__attribute__ ((visibility("default"))) int spprintf( char **pbuf, size_t max_len, const char *format, ...) __attribute__ ((format(printf, 3, 4)));

__attribute__ ((visibility("default"))) int vspprintf(char **pbuf, size_t max_len, const char *format, va_list ap) __attribute__ ((format(printf, 3, 0)));

# 244 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 272 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
extern char **environ;
# 283 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"

void phperror(char *error);
__attribute__ ((visibility("default"))) int php_write(void *buf, uint size );
__attribute__ ((visibility("default"))) int php_printf(const char *format, ...) __attribute__ ((format(printf, 1, 2)))
    ;
__attribute__ ((visibility("default"))) int php_get_module_initialized(void);
__attribute__ ((visibility("default"))) void php_log_err(char *log_message );
int Debug(char *format, ...) __attribute__ ((format(printf, 1, 2)));
int cfgparse(void);






static inline __attribute__((deprecated)) void php_set_error_handling(zend_error_handling_t error_handling, zend_class_entry *exception_class )
{
 zend_replace_error_handling(error_handling, exception_class, ((void *)0) );
}
static inline __attribute__((deprecated)) void php_std_error_handling() {}

__attribute__ ((visibility("default"))) void php_verror(const char *docref, const char *params, int type, const char *format, va_list args ) __attribute__ ((format(printf, 4, 0)));
# 313 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
__attribute__ ((visibility("default"))) void php_error_docref0(const char *docref , int type, const char *format, ...)
 __attribute__ ((format(printf, 0 + 3, 0 + 4)));
__attribute__ ((visibility("default"))) void php_error_docref1(const char *docref , const char *param1, int type, const char *format, ...)
 __attribute__ ((format(printf, 0 + 4, 0 + 5)));
__attribute__ ((visibility("default"))) void php_error_docref2(const char *docref , const char *param1, const char *param2, int type, const char *format, ...)
 __attribute__ ((format(printf, 0 + 5, 0 + 6)));




# 336 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"

__attribute__ ((visibility("default"))) extern int (*php_register_internal_extensions_func)(void);
__attribute__ ((visibility("default"))) int php_register_internal_extensions(void);
__attribute__ ((visibility("default"))) int php_mergesort(void *base, size_t nmemb, register size_t size, int (*cmp)(const void *, const void * ) );
__attribute__ ((visibility("default"))) void php_register_pre_request_shutdown(void (*func)(void *), void *userdata);
__attribute__ ((visibility("default"))) void php_com_initialize(void);
__attribute__ ((visibility("default"))) char *php_get_current_user(void);

# 397 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h" 1
# 50 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h"
typedef enum _php_output_handler_status_t {
 PHP_OUTPUT_HANDLER_FAILURE,
 PHP_OUTPUT_HANDLER_SUCCESS,
 PHP_OUTPUT_HANDLER_NO_DATA
} php_output_handler_status_t;
# 72 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h"
typedef enum _php_output_handler_hook_t {
 PHP_OUTPUT_HANDLER_HOOK_GET_OPAQ,
 PHP_OUTPUT_HANDLER_HOOK_GET_FLAGS,
 PHP_OUTPUT_HANDLER_HOOK_GET_LEVEL,
 PHP_OUTPUT_HANDLER_HOOK_IMMUTABLE,
 PHP_OUTPUT_HANDLER_HOOK_DISABLE,

 PHP_OUTPUT_HANDLER_HOOK_LAST
} php_output_handler_hook_t;
# 90 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h"
typedef struct _php_output_buffer {
 char *data;
 size_t size;
 size_t used;
 uint free:1;
 uint _res:31;
} php_output_buffer;

typedef struct _php_output_context {
 int op;
 php_output_buffer in;
 php_output_buffer out;



} php_output_context;




typedef void (*php_output_handler_func_t)(char *output, uint output_len, char **handled_output, uint *handled_output_len, int mode );

typedef int (*php_output_handler_context_func_t)(void **handler_context, php_output_context *output_context);

typedef void (*php_output_handler_context_dtor_t)(void *opaq );

typedef int (*php_output_handler_conflict_check_t)(const char *handler_name, size_t handler_name_len );

typedef struct _php_output_handler *(*php_output_handler_alias_ctor_t)(const char *handler_name, size_t handler_name_len, size_t chunk_size, int flags );

typedef struct _php_output_handler_user_func_t {
 zend_fcall_info fci;
 zend_fcall_info_cache fcc;
 zval *zoh;
} php_output_handler_user_func_t;

typedef struct _php_output_handler {
 char *name;
 size_t name_len;
 int flags;
 int level;
 size_t size;
 php_output_buffer buffer;

 void *opaq;
 void (*dtor)(void *opaq );

 union {
  php_output_handler_user_func_t *user;
  php_output_handler_context_func_t internal;
 } func;
} php_output_handler;

typedef struct _zend_output_globals {
 int flags;
 zend_stack handlers;
 php_output_handler *active;
 php_output_handler *running;
 const char *output_start_filename;
 int output_start_lineno;
} zend_output_globals;
# 176 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h"


extern const char php_output_default_handler_name[sizeof("default output handler")];
extern const char php_output_devnull_handler_name[sizeof("null output handler")];
# 190 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_output.h"
__attribute__ ((visibility("default"))) void php_output_startup(void);

__attribute__ ((visibility("default"))) void php_output_shutdown(void);

__attribute__ ((visibility("default"))) void php_output_register_constants(void);


__attribute__ ((visibility("default"))) int php_output_activate(void);

__attribute__ ((visibility("default"))) void php_output_deactivate(void);

__attribute__ ((visibility("default"))) void php_output_set_status(int status );
__attribute__ ((visibility("default"))) int php_output_get_status(void);
__attribute__ ((visibility("default"))) void php_output_set_implicit_flush(int flush );
__attribute__ ((visibility("default"))) const char *php_output_get_start_filename(void);
__attribute__ ((visibility("default"))) int php_output_get_start_lineno(void);

__attribute__ ((visibility("default"))) int php_output_write_unbuffered(const char *str, size_t len );
__attribute__ ((visibility("default"))) int php_output_write(const char *str, size_t len );

__attribute__ ((visibility("default"))) int php_output_flush(void);
__attribute__ ((visibility("default"))) void php_output_flush_all(void);
__attribute__ ((visibility("default"))) int php_output_clean(void);
__attribute__ ((visibility("default"))) void php_output_clean_all(void);
__attribute__ ((visibility("default"))) int php_output_end(void);
__attribute__ ((visibility("default"))) void php_output_end_all(void);
__attribute__ ((visibility("default"))) int php_output_discard(void);
__attribute__ ((visibility("default"))) void php_output_discard_all(void);

__attribute__ ((visibility("default"))) int php_output_get_contents(zval *p );
__attribute__ ((visibility("default"))) int php_output_get_length(zval *p );
__attribute__ ((visibility("default"))) int php_output_get_level(void);

__attribute__ ((visibility("default"))) int php_output_start_default(void);
__attribute__ ((visibility("default"))) int php_output_start_devnull(void);

__attribute__ ((visibility("default"))) int php_output_start_user(zval *output_handler, size_t chunk_size, int flags );
__attribute__ ((visibility("default"))) int php_output_start_internal(const char *name, size_t name_len, php_output_handler_func_t output_handler, size_t chunk_size, int flags );

__attribute__ ((visibility("default"))) php_output_handler *php_output_handler_create_user(zval *handler, size_t chunk_size, int flags );
__attribute__ ((visibility("default"))) php_output_handler *php_output_handler_create_internal(const char *name, size_t name_len, php_output_handler_context_func_t handler, size_t chunk_size, int flags );

__attribute__ ((visibility("default"))) void php_output_handler_set_context(php_output_handler *handler, void *opaq, void (*dtor)(void* ) );
__attribute__ ((visibility("default"))) int php_output_handler_start(php_output_handler *handler );
__attribute__ ((visibility("default"))) int php_output_handler_started(const char *name, size_t name_len );
__attribute__ ((visibility("default"))) int php_output_handler_hook(php_output_handler_hook_t type, void *arg );
__attribute__ ((visibility("default"))) void php_output_handler_dtor(php_output_handler *handler );
__attribute__ ((visibility("default"))) void php_output_handler_free(php_output_handler **handler );

__attribute__ ((visibility("default"))) int php_output_handler_conflict(const char *handler_new, size_t handler_new_len, const char *handler_set, size_t handler_set_len );
__attribute__ ((visibility("default"))) int php_output_handler_conflict_register(const char *handler_name, size_t handler_name_len, php_output_handler_conflict_check_t check_func );
__attribute__ ((visibility("default"))) int php_output_handler_reverse_conflict_register(const char *handler_name, size_t handler_name_len, php_output_handler_conflict_check_t check_func );

__attribute__ ((visibility("default"))) php_output_handler_alias_ctor_t *php_output_handler_alias(const char *handler_name, size_t handler_name_len );
__attribute__ ((visibility("default"))) int php_output_handler_alias_register(const char *handler_name, size_t handler_name_len, php_output_handler_alias_ctor_t func );




void zif_ob_start(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_flush(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_clean(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_end_flush(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_end_clean(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_flush(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_clean(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_contents(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_length(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_level(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_get_status(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_implicit_flush(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ob_list_handlers(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_output_add_rewrite_var(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_output_reset_rewrite_vars(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
# 398 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2


# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
# 1 "/usr/include/sys/time.h" 1 3 4
# 27 "/usr/include/sys/time.h" 3 4
# 1 "/usr/include/time.h" 1 3 4
# 28 "/usr/include/sys/time.h" 2 3 4

# 1 "/usr/include/bits/time.h" 1 3 4
# 30 "/usr/include/sys/time.h" 2 3 4
# 39 "/usr/include/sys/time.h" 3 4

# 57 "/usr/include/sys/time.h" 3 4
struct timezone
  {
    int tz_minuteswest;
    int tz_dsttime;
  };

typedef struct timezone *__restrict __timezone_ptr_t;
# 73 "/usr/include/sys/time.h" 3 4
extern int gettimeofday (struct timeval *__restrict __tv,
    __timezone_ptr_t __tz) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));




extern int settimeofday (__const struct timeval *__tv,
    __const struct timezone *__tz)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int adjtime (__const struct timeval *__delta,
      struct timeval *__olddelta) __attribute__ ((__nothrow__));




enum __itimer_which
  {

    ITIMER_REAL = 0,


    ITIMER_VIRTUAL = 1,



    ITIMER_PROF = 2

  };



struct itimerval
  {

    struct timeval it_interval;

    struct timeval it_value;
  };




typedef enum __itimer_which __itimer_which_t;






extern int getitimer (__itimer_which_t __which,
        struct itimerval *__value) __attribute__ ((__nothrow__));




extern int setitimer (__itimer_which_t __which,
        __const struct itimerval *__restrict __new,
        struct itimerval *__restrict __old) __attribute__ ((__nothrow__));




extern int utimes (__const char *__file, __const struct timeval __tvp[2])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));



extern int lutimes (__const char *__file, __const struct timeval __tvp[2])
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern int futimes (int __fd, __const struct timeval __tvp[2]) __attribute__ ((__nothrow__));






extern int futimesat (int __fd, __const char *__file,
        __const struct timeval __tvp[2]) __attribute__ ((__nothrow__));
# 191 "/usr/include/sys/time.h" 3 4

# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2


# 1 "/usr/include/sys/stat.h" 1 3 4
# 39 "/usr/include/sys/stat.h" 3 4
# 1 "/usr/include/time.h" 1 3 4
# 40 "/usr/include/sys/stat.h" 2 3 4
# 105 "/usr/include/sys/stat.h" 3 4


# 1 "/usr/include/bits/stat.h" 1 3 4
# 39 "/usr/include/bits/stat.h" 3 4
struct stat
  {
    __dev_t st_dev;
    unsigned short int __pad1;

    __ino_t st_ino;



    __mode_t st_mode;
    __nlink_t st_nlink;
    __uid_t st_uid;
    __gid_t st_gid;
    __dev_t st_rdev;
    unsigned short int __pad2;

    __off_t st_size;



    __blksize_t st_blksize;


    __blkcnt_t st_blocks;
# 73 "/usr/include/bits/stat.h" 3 4
    struct timespec st_atim;
    struct timespec st_mtim;
    struct timespec st_ctim;
# 88 "/usr/include/bits/stat.h" 3 4
    unsigned long int __unused4;
    unsigned long int __unused5;



  };


struct stat64
  {
    __dev_t st_dev;
    unsigned int __pad1;

    __ino_t __st_ino;
    __mode_t st_mode;
    __nlink_t st_nlink;
    __uid_t st_uid;
    __gid_t st_gid;
    __dev_t st_rdev;
    unsigned int __pad2;
    __off64_t st_size;
    __blksize_t st_blksize;

    __blkcnt64_t st_blocks;







    struct timespec st_atim;
    struct timespec st_mtim;
    struct timespec st_ctim;
# 130 "/usr/include/bits/stat.h" 3 4
    __ino64_t st_ino;
  };
# 108 "/usr/include/sys/stat.h" 2 3 4
# 211 "/usr/include/sys/stat.h" 3 4
extern int stat (__const char *__restrict __file,
   struct stat *__restrict __buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));



extern int fstat (int __fd, struct stat *__buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));
# 230 "/usr/include/sys/stat.h" 3 4
extern int stat64 (__const char *__restrict __file,
     struct stat64 *__restrict __buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
extern int fstat64 (int __fd, struct stat64 *__buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));







extern int fstatat (int __fd, __const char *__restrict __file,
      struct stat *__restrict __buf, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
# 255 "/usr/include/sys/stat.h" 3 4
extern int fstatat64 (int __fd, __const char *__restrict __file,
        struct stat64 *__restrict __buf, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));







extern int lstat (__const char *__restrict __file,
    struct stat *__restrict __buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));
# 278 "/usr/include/sys/stat.h" 3 4
extern int lstat64 (__const char *__restrict __file,
      struct stat64 *__restrict __buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1, 2)));





extern int chmod (__const char *__file, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int lchmod (__const char *__file, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));




extern int fchmod (int __fd, __mode_t __mode) __attribute__ ((__nothrow__));





extern int fchmodat (int __fd, __const char *__file, __mode_t __mode,
       int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2))) ;






extern __mode_t umask (__mode_t __mask) __attribute__ ((__nothrow__));




extern __mode_t getumask (void) __attribute__ ((__nothrow__));



extern int mkdir (__const char *__path, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int mkdirat (int __fd, __const char *__path, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));






extern int mknod (__const char *__path, __mode_t __mode, __dev_t __dev)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int mknodat (int __fd, __const char *__path, __mode_t __mode,
      __dev_t __dev) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));





extern int mkfifo (__const char *__path, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int mkfifoat (int __fd, __const char *__path, __mode_t __mode)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));





extern int utimensat (int __fd, __const char *__path,
        __const struct timespec __times[2],
        int __flags)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2)));




extern int futimens (int __fd, __const struct timespec __times[2]) __attribute__ ((__nothrow__));
# 401 "/usr/include/sys/stat.h" 3 4
extern int __fxstat (int __ver, int __fildes, struct stat *__stat_buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3)));
extern int __xstat (int __ver, __const char *__filename,
      struct stat *__stat_buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
extern int __lxstat (int __ver, __const char *__filename,
       struct stat *__stat_buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
extern int __fxstatat (int __ver, int __fildes, __const char *__filename,
         struct stat *__stat_buf, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4)));
# 434 "/usr/include/sys/stat.h" 3 4
extern int __fxstat64 (int __ver, int __fildes, struct stat64 *__stat_buf)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3)));
extern int __xstat64 (int __ver, __const char *__filename,
        struct stat64 *__stat_buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
extern int __lxstat64 (int __ver, __const char *__filename,
         struct stat64 *__stat_buf) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 3)));
extern int __fxstatat64 (int __ver, int __fildes, __const char *__filename,
    struct stat64 *__stat_buf, int __flag)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 4)));

extern int __xmknod (int __ver, __const char *__path, __mode_t __mode,
       __dev_t *__dev) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4)));

extern int __xmknodat (int __ver, int __fd, __const char *__path,
         __mode_t __mode, __dev_t *__dev)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (3, 5)));




extern __inline int
__attribute__ ((__nothrow__)) stat (__const char *__path, struct stat *__statbuf)
{
  return __xstat (3, __path, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) lstat (__const char *__path, struct stat *__statbuf)
{
  return __lxstat (3, __path, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) fstat (int __fd, struct stat *__statbuf)
{
  return __fxstat (3, __fd, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) fstatat (int __fd, __const char *__filename, struct stat *__statbuf, int __flag)

{
  return __fxstatat (3, __fd, __filename, __statbuf, __flag);
}



extern __inline int
__attribute__ ((__nothrow__)) mknod (__const char *__path, __mode_t __mode, __dev_t __dev)
{
  return __xmknod (1, __path, __mode, &__dev);
}



extern __inline int
__attribute__ ((__nothrow__)) mknodat (int __fd, __const char *__path, __mode_t __mode, __dev_t __dev)

{
  return __xmknodat (1, __fd, __path, __mode, &__dev);
}





extern __inline int
__attribute__ ((__nothrow__)) stat64 (__const char *__path, struct stat64 *__statbuf)
{
  return __xstat64 (3, __path, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) lstat64 (__const char *__path, struct stat64 *__statbuf)
{
  return __lxstat64 (3, __path, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) fstat64 (int __fd, struct stat64 *__statbuf)
{
  return __fxstat64 (3, __fd, __statbuf);
}


extern __inline int
__attribute__ ((__nothrow__)) fstatat64 (int __fd, __const char *__filename, struct stat64 *__statbuf, int __flag)

{
  return __fxstatat64 (3, __fd, __filename, __statbuf, __flag);
}







# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2


__attribute__ ((visibility("default"))) int php_file_le_stream(void);
__attribute__ ((visibility("default"))) int php_file_le_pstream(void);
__attribute__ ((visibility("default"))) int php_file_le_stream_filter(void);

# 98 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
typedef struct _php_stream php_stream;
typedef struct _php_stream_wrapper php_stream_wrapper;
typedef struct _php_stream_context php_stream_context;
typedef struct _php_stream_filter php_stream_filter;

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_context.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_context.h"
typedef void (*php_stream_notification_func)(php_stream_context *context,
  int notifycode, int severity,
  char *xmsg, int xcode,
  size_t bytes_sofar, size_t bytes_max,
  void * ptr );
# 43 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_context.h"
typedef struct _php_stream_notifier php_stream_notifier;

struct _php_stream_notifier {
 php_stream_notification_func func;
 void (*dtor)(php_stream_notifier *notifier);
 void *ptr;
 int mask;
 size_t progress, progress_max;
};

struct _php_stream_context {
 php_stream_notifier *notifier;
 zval *options;
 zval *links;
 int rsrc_id;
};


__attribute__ ((visibility("default"))) void php_stream_context_free(php_stream_context *context);
__attribute__ ((visibility("default"))) php_stream_context *php_stream_context_alloc(void);
__attribute__ ((visibility("default"))) int php_stream_context_get_option(php_stream_context *context,
  const char *wrappername, const char *optionname, zval ***optionvalue);
__attribute__ ((visibility("default"))) int php_stream_context_set_option(php_stream_context *context,
  const char *wrappername, const char *optionname, zval *optionvalue);

__attribute__ ((visibility("default"))) int php_stream_context_get_link(php_stream_context *context,
  const char *hostent, php_stream **stream);
__attribute__ ((visibility("default"))) int php_stream_context_set_link(php_stream_context *context,
  const char *hostent, php_stream *stream);
__attribute__ ((visibility("default"))) int php_stream_context_del_link(php_stream_context *context,
  php_stream *stream);

__attribute__ ((visibility("default"))) php_stream_notifier *php_stream_notification_alloc(void);
__attribute__ ((visibility("default"))) void php_stream_notification_free(php_stream_notifier *notifier);

# 95 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_context.h"

__attribute__ ((visibility("default"))) void php_stream_notification_notify(php_stream_context *context, int notifycode, int severity,
  char *xmsg, int xcode, size_t bytes_sofar, size_t bytes_max, void * ptr );
__attribute__ ((visibility("default"))) php_stream_context *php_stream_context_set(php_stream *stream, php_stream_context *context);

# 104 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_filter_api.h" 1
# 41 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_filter_api.h"
typedef struct _php_stream_bucket php_stream_bucket;
typedef struct _php_stream_bucket_brigade php_stream_bucket_brigade;

struct _php_stream_bucket {
 php_stream_bucket *next, *prev;
 php_stream_bucket_brigade *brigade;

 char *buf;
 size_t buflen;

 int own_buf;
 int is_persistent;


 int refcount;
};

struct _php_stream_bucket_brigade {
 php_stream_bucket *head, *tail;
};

typedef enum {
 PSFS_ERR_FATAL,
 PSFS_FEED_ME,
 PSFS_PASS_ON
} php_stream_filter_status_t;



__attribute__ ((visibility("default"))) php_stream_bucket *php_stream_bucket_new(php_stream *stream, char *buf, size_t buflen, int own_buf, int buf_persistent );
__attribute__ ((visibility("default"))) int php_stream_bucket_split(php_stream_bucket *in, php_stream_bucket **left, php_stream_bucket **right, size_t length );
__attribute__ ((visibility("default"))) void php_stream_bucket_delref(php_stream_bucket *bucket );

__attribute__ ((visibility("default"))) void php_stream_bucket_prepend(php_stream_bucket_brigade *brigade, php_stream_bucket *bucket );
__attribute__ ((visibility("default"))) void php_stream_bucket_append(php_stream_bucket_brigade *brigade, php_stream_bucket *bucket );
__attribute__ ((visibility("default"))) void php_stream_bucket_unlink(php_stream_bucket *bucket );
__attribute__ ((visibility("default"))) php_stream_bucket *php_stream_bucket_make_writeable(php_stream_bucket *bucket );






typedef struct _php_stream_filter_ops {

 php_stream_filter_status_t (*filter)(
   php_stream *stream,
   php_stream_filter *thisfilter,
   php_stream_bucket_brigade *buckets_in,
   php_stream_bucket_brigade *buckets_out,
   size_t *bytes_consumed,
   int flags
   );

 void (*dtor)(php_stream_filter *thisfilter );

 const char *label;

} php_stream_filter_ops;

typedef struct _php_stream_filter_chain {
 php_stream_filter *head, *tail;


 php_stream *stream;
} php_stream_filter_chain;

struct _php_stream_filter {
 php_stream_filter_ops *fops;
 void *abstract;
 php_stream_filter *next;
 php_stream_filter *prev;
 int is_persistent;


 php_stream_filter_chain *chain;


 php_stream_bucket_brigade buffer;


 int rsrc_id;
};



__attribute__ ((visibility("default"))) void _php_stream_filter_prepend(php_stream_filter_chain *chain, php_stream_filter *filter );
__attribute__ ((visibility("default"))) int php_stream_filter_prepend_ex(php_stream_filter_chain *chain, php_stream_filter *filter );
__attribute__ ((visibility("default"))) void _php_stream_filter_append(php_stream_filter_chain *chain, php_stream_filter *filter );
__attribute__ ((visibility("default"))) int php_stream_filter_append_ex(php_stream_filter_chain *chain, php_stream_filter *filter );
__attribute__ ((visibility("default"))) int _php_stream_filter_flush(php_stream_filter *filter, int finish );
__attribute__ ((visibility("default"))) php_stream_filter *php_stream_filter_remove(php_stream_filter *filter, int call_dtor );
__attribute__ ((visibility("default"))) void php_stream_filter_free(php_stream_filter *filter );
__attribute__ ((visibility("default"))) php_stream_filter *_php_stream_filter_alloc(php_stream_filter_ops *fops, void *abstract, int persistent );

# 144 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_filter_api.h"
typedef struct _php_stream_filter_factory {
 php_stream_filter *(*create_filter)(const char *filtername, zval *filterparams, int persistent );
} php_stream_filter_factory;


__attribute__ ((visibility("default"))) int php_stream_filter_register_factory(const char *filterpattern, php_stream_filter_factory *factory );
__attribute__ ((visibility("default"))) int php_stream_filter_unregister_factory(const char *filterpattern );
__attribute__ ((visibility("default"))) int php_stream_filter_register_factory_volatile(const char *filterpattern, php_stream_filter_factory *factory );
__attribute__ ((visibility("default"))) php_stream_filter *php_stream_filter_create(const char *filtername, zval *filterparams, int persistent );

# 105 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2

typedef struct _php_stream_statbuf {
 struct stat sb;

} php_stream_statbuf;

typedef struct _php_stream_dirent {
 char d_name[4096];
} php_stream_dirent;


typedef struct _php_stream_ops {

 size_t (*write)(php_stream *stream, const char *buf, size_t count );
 size_t (*read)(php_stream *stream, char *buf, size_t count );
 int (*close)(php_stream *stream, int close_handle );
 int (*flush)(php_stream *stream );

 const char *label;


 int (*seek)(php_stream *stream, off_t offset, int whence, off_t *newoffset );
 int (*cast)(php_stream *stream, int castas, void **ret );
 int (*stat)(php_stream *stream, php_stream_statbuf *ssb );
 int (*set_option)(php_stream *stream, int option, int value, void *ptrparam );
} php_stream_ops;

typedef struct _php_stream_wrapper_ops {

 php_stream *(*stream_opener)(php_stream_wrapper *wrapper, char *filename, char *mode,
   int options, char **opened_path, php_stream_context *context );

 int (*stream_closer)(php_stream_wrapper *wrapper, php_stream *stream );

 int (*stream_stat)(php_stream_wrapper *wrapper, php_stream *stream, php_stream_statbuf *ssb );

 int (*url_stat)(php_stream_wrapper *wrapper, char *url, int flags, php_stream_statbuf *ssb, php_stream_context *context );

 php_stream *(*dir_opener)(php_stream_wrapper *wrapper, char *filename, char *mode,
   int options, char **opened_path, php_stream_context *context );

 const char *label;


 int (*unlink)(php_stream_wrapper *wrapper, char *url, int options, php_stream_context *context );


 int (*rename)(php_stream_wrapper *wrapper, char *url_from, char *url_to, int options, php_stream_context *context );


 int (*stream_mkdir)(php_stream_wrapper *wrapper, char *url, int mode, int options, php_stream_context *context );
 int (*stream_rmdir)(php_stream_wrapper *wrapper, char *url, int options, php_stream_context *context );

 int (*stream_metadata)(php_stream_wrapper *wrapper, char *url, int options, void *value, php_stream_context *context );
} php_stream_wrapper_ops;

struct _php_stream_wrapper {
 php_stream_wrapper_ops *wops;
 void *abstract;
 int is_url;


 int err_count;
 char **err_stack;
};
# 190 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
struct _php_stream {
 php_stream_ops *ops;
 void *abstract;

 php_stream_filter_chain readfilters, writefilters;

 php_stream_wrapper *wrapper;
 void *wrapperthis;
 zval *wrapperdata;

 int fgetss_state;
 int is_persistent;
 char mode[16];
 int rsrc_id;
 int in_free;


 int fclose_stdiocast;
 FILE *stdiocast;



 char *orig_path;

 php_stream_context *context;
 int flags;


 off_t position;
 unsigned char *readbuf;
 size_t readbuflen;
 off_t readpos;
 off_t writepos;


 size_t chunk_size;

 int eof;






 struct _php_stream *enclosing_stream;
};








__attribute__ ((visibility("default"))) php_stream *_php_stream_alloc(php_stream_ops *ops, void *abstract,
  const char *persistent_id, const char *mode );

# 265 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"

__attribute__ ((visibility("default"))) php_stream *php_stream_encloses(php_stream *enclosing, php_stream *enclosed);

__attribute__ ((visibility("default"))) int _php_stream_free_enclosed(php_stream *stream_enclosed, int close_options );

__attribute__ ((visibility("default"))) int php_stream_from_persistent_id(const char *persistent_id, php_stream **stream );
# 285 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
__attribute__ ((visibility("default"))) int _php_stream_free(php_stream *stream, int close_options );




__attribute__ ((visibility("default"))) int _php_stream_seek(php_stream *stream, off_t offset, int whence );



__attribute__ ((visibility("default"))) off_t _php_stream_tell(php_stream *stream );


__attribute__ ((visibility("default"))) size_t _php_stream_read(php_stream *stream, char *buf, size_t count );


__attribute__ ((visibility("default"))) size_t _php_stream_write(php_stream *stream, const char *buf, size_t count );






__attribute__ ((visibility("default"))) size_t _php_stream_printf(php_stream *stream , const char *fmt, ...) __attribute__ ((format(printf, 2, 3)));





__attribute__ ((visibility("default"))) int _php_stream_eof(php_stream *stream );


__attribute__ ((visibility("default"))) int _php_stream_getc(php_stream *stream );


__attribute__ ((visibility("default"))) int _php_stream_putc(php_stream *stream, int c );


__attribute__ ((visibility("default"))) int _php_stream_flush(php_stream *stream, int closing );


__attribute__ ((visibility("default"))) char *_php_stream_get_line(php_stream *stream, char *buf, size_t maxlen, size_t *returned_len );



__attribute__ ((visibility("default"))) char *php_stream_get_record(php_stream *stream, size_t maxlen, size_t *returned_len, char *delim, size_t delim_len );


__attribute__ ((visibility("default"))) int _php_stream_puts(php_stream *stream, char *buf );


__attribute__ ((visibility("default"))) int _php_stream_stat(php_stream *stream, php_stream_statbuf *ssb );


__attribute__ ((visibility("default"))) int _php_stream_stat_path(char *path, int flags, php_stream_statbuf *ssb, php_stream_context *context );



__attribute__ ((visibility("default"))) int _php_stream_mkdir(char *path, int mode, int options, php_stream_context *context );


__attribute__ ((visibility("default"))) int _php_stream_rmdir(char *path, int options, php_stream_context *context );


__attribute__ ((visibility("default"))) php_stream *_php_stream_opendir(char *path, int options, php_stream_context *context );

__attribute__ ((visibility("default"))) php_stream_dirent *_php_stream_readdir(php_stream *dirstream, php_stream_dirent *ent );




__attribute__ ((visibility("default"))) int php_stream_dirent_alphasort(const char **a, const char **b);
__attribute__ ((visibility("default"))) int php_stream_dirent_alphasortr(const char **a, const char **b);

__attribute__ ((visibility("default"))) int _php_stream_scandir(char *dirname, char **namelist[], int flags, php_stream_context *context,
   int (*compare) (const char **a, const char **b) );


__attribute__ ((visibility("default"))) int _php_stream_set_option(php_stream *stream, int option, int value, void *ptrparam );





# 417 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"

__attribute__ ((visibility("default"))) int _php_stream_truncate_set_size(php_stream *stream, size_t newsize );


# 437 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"

__attribute__((deprecated))
__attribute__ ((visibility("default"))) size_t _php_stream_copy_to_stream(php_stream *src, php_stream *dest, size_t maxlen );

__attribute__ ((visibility("default"))) int _php_stream_copy_to_stream_ex(php_stream *src, php_stream *dest, size_t maxlen, size_t *len );





__attribute__ ((visibility("default"))) size_t _php_stream_copy_to_mem(php_stream *src, char **buf, size_t maxlen,
  int persistent );



__attribute__ ((visibility("default"))) size_t _php_stream_passthru(php_stream * src );



# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_transport.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_transport.h"
# 1 "/usr/include/sys/socket.h" 1 3 4
# 26 "/usr/include/sys/socket.h" 3 4


# 1 "/usr/include/sys/uio.h" 1 3 4
# 26 "/usr/include/sys/uio.h" 3 4



# 1 "/usr/include/bits/uio.h" 1 3 4
# 44 "/usr/include/bits/uio.h" 3 4
struct iovec
  {
    void *iov_base;
    size_t iov_len;
  };
# 30 "/usr/include/sys/uio.h" 2 3 4
# 40 "/usr/include/sys/uio.h" 3 4
extern ssize_t readv (int __fd, __const struct iovec *__iovec, int __count)
  ;
# 51 "/usr/include/sys/uio.h" 3 4
extern ssize_t writev (int __fd, __const struct iovec *__iovec, int __count)
  ;
# 66 "/usr/include/sys/uio.h" 3 4
extern ssize_t preadv (int __fd, __const struct iovec *__iovec, int __count,
         __off_t __offset) ;
# 78 "/usr/include/sys/uio.h" 3 4
extern ssize_t pwritev (int __fd, __const struct iovec *__iovec, int __count,
   __off_t __offset) ;
# 104 "/usr/include/sys/uio.h" 3 4
extern ssize_t preadv64 (int __fd, __const struct iovec *__iovec, int __count,
    __off64_t __offset) ;
# 116 "/usr/include/sys/uio.h" 3 4
extern ssize_t pwritev64 (int __fd, __const struct iovec *__iovec, int __count,
     __off64_t __offset) ;




# 29 "/usr/include/sys/socket.h" 2 3 4

# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 31 "/usr/include/sys/socket.h" 2 3 4


# 1 "/usr/include/bits/sigset.h" 1 3 4
# 34 "/usr/include/sys/socket.h" 2 3 4






# 1 "/usr/include/bits/socket.h" 1 3 4
# 29 "/usr/include/bits/socket.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 30 "/usr/include/bits/socket.h" 2 3 4
# 40 "/usr/include/bits/socket.h" 3 4
enum __socket_type
{
  SOCK_STREAM = 1,


  SOCK_DGRAM = 2,


  SOCK_RAW = 3,

  SOCK_RDM = 4,

  SOCK_SEQPACKET = 5,


  SOCK_DCCP = 6,

  SOCK_PACKET = 10,







  SOCK_CLOEXEC = 02000000,


  SOCK_NONBLOCK = 04000


};
# 171 "/usr/include/bits/socket.h" 3 4
# 1 "/usr/include/bits/sockaddr.h" 1 3 4
# 29 "/usr/include/bits/sockaddr.h" 3 4
typedef unsigned short int sa_family_t;
# 172 "/usr/include/bits/socket.h" 2 3 4


struct sockaddr
  {
    sa_family_t sa_family;
    char sa_data[14];
  };
# 187 "/usr/include/bits/socket.h" 3 4
struct sockaddr_storage
  {
    sa_family_t ss_family;
    unsigned long int __ss_align;
    char __ss_padding[(128 - (2 * sizeof (unsigned long int)))];
  };



enum
  {
    MSG_OOB = 0x01,

    MSG_PEEK = 0x02,

    MSG_DONTROUTE = 0x04,



    MSG_TRYHARD = MSG_DONTROUTE,


    MSG_CTRUNC = 0x08,

    MSG_PROXY = 0x10,

    MSG_TRUNC = 0x20,

    MSG_DONTWAIT = 0x40,

    MSG_EOR = 0x80,

    MSG_WAITALL = 0x100,

    MSG_FIN = 0x200,

    MSG_SYN = 0x400,

    MSG_CONFIRM = 0x800,

    MSG_RST = 0x1000,

    MSG_ERRQUEUE = 0x2000,

    MSG_NOSIGNAL = 0x4000,

    MSG_MORE = 0x8000,

    MSG_WAITFORONE = 0x10000,


    MSG_CMSG_CLOEXEC = 0x40000000



  };




struct msghdr
  {
    void *msg_name;
    socklen_t msg_namelen;

    struct iovec *msg_iov;
    size_t msg_iovlen;

    void *msg_control;
    size_t msg_controllen;




    int msg_flags;
  };



struct mmsghdr
  {
    struct msghdr msg_hdr;
    unsigned int msg_len;
  };



struct cmsghdr
  {
    size_t cmsg_len;




    int cmsg_level;
    int cmsg_type;

    __extension__ unsigned char __cmsg_data [];

  };
# 304 "/usr/include/bits/socket.h" 3 4
extern struct cmsghdr *__cmsg_nxthdr (struct msghdr *__mhdr,
          struct cmsghdr *__cmsg) __attribute__ ((__nothrow__));




extern __inline struct cmsghdr *
__attribute__ ((__nothrow__)) __cmsg_nxthdr (struct msghdr *__mhdr, struct cmsghdr *__cmsg)
{
  if ((size_t) __cmsg->cmsg_len < sizeof (struct cmsghdr))

    return 0;

  __cmsg = (struct cmsghdr *) ((unsigned char *) __cmsg
          + (((__cmsg->cmsg_len) + sizeof (size_t) - 1) & (size_t) ~(sizeof (size_t) - 1)));
  if ((unsigned char *) (__cmsg + 1) > ((unsigned char *) __mhdr->msg_control
     + __mhdr->msg_controllen)
      || ((unsigned char *) __cmsg + (((__cmsg->cmsg_len) + sizeof (size_t) - 1) & (size_t) ~(sizeof (size_t) - 1))
   > ((unsigned char *) __mhdr->msg_control + __mhdr->msg_controllen)))

    return 0;
  return __cmsg;
}




enum
  {
    SCM_RIGHTS = 0x01


    , SCM_CREDENTIALS = 0x02


  };



struct ucred
{
  pid_t pid;
  uid_t uid;
  gid_t gid;
};
# 377 "/usr/include/bits/socket.h" 3 4
# 1 "/usr/include/asm/socket.h" 1 3 4
# 1 "/usr/include/asm-generic/socket.h" 1 3 4



# 1 "/usr/include/asm/sockios.h" 1 3 4
# 1 "/usr/include/asm-generic/sockios.h" 1 3 4
# 1 "/usr/include/asm/sockios.h" 2 3 4
# 5 "/usr/include/asm-generic/socket.h" 2 3 4
# 1 "/usr/include/asm/socket.h" 2 3 4
# 378 "/usr/include/bits/socket.h" 2 3 4
# 411 "/usr/include/bits/socket.h" 3 4
struct linger
  {
    int l_onoff;
    int l_linger;
  };









extern int recvmmsg (int __fd, struct mmsghdr *__vmessages,
       unsigned int __vlen, int __flags,
       __const struct timespec *__tmo);


# 41 "/usr/include/sys/socket.h" 2 3 4




struct osockaddr
  {
    unsigned short int sa_family;
    unsigned char sa_data[14];
  };




enum
{
  SHUT_RD = 0,

  SHUT_WR,

  SHUT_RDWR

};
# 92 "/usr/include/sys/socket.h" 3 4
typedef union { struct sockaddr *__restrict __sockaddr__; struct sockaddr_at *__restrict __sockaddr_at__; struct sockaddr_ax25 *__restrict __sockaddr_ax25__; struct sockaddr_dl *__restrict __sockaddr_dl__; struct sockaddr_eon *__restrict __sockaddr_eon__; struct sockaddr_in *__restrict __sockaddr_in__; struct sockaddr_in6 *__restrict __sockaddr_in6__; struct sockaddr_inarp *__restrict __sockaddr_inarp__; struct sockaddr_ipx *__restrict __sockaddr_ipx__; struct sockaddr_iso *__restrict __sockaddr_iso__; struct sockaddr_ns *__restrict __sockaddr_ns__; struct sockaddr_un *__restrict __sockaddr_un__; struct sockaddr_x25 *__restrict __sockaddr_x25__;
       } __SOCKADDR_ARG __attribute__ ((__transparent_union__));


typedef union { __const struct sockaddr *__restrict __sockaddr__; __const struct sockaddr_at *__restrict __sockaddr_at__; __const struct sockaddr_ax25 *__restrict __sockaddr_ax25__; __const struct sockaddr_dl *__restrict __sockaddr_dl__; __const struct sockaddr_eon *__restrict __sockaddr_eon__; __const struct sockaddr_in *__restrict __sockaddr_in__; __const struct sockaddr_in6 *__restrict __sockaddr_in6__; __const struct sockaddr_inarp *__restrict __sockaddr_inarp__; __const struct sockaddr_ipx *__restrict __sockaddr_ipx__; __const struct sockaddr_iso *__restrict __sockaddr_iso__; __const struct sockaddr_ns *__restrict __sockaddr_ns__; __const struct sockaddr_un *__restrict __sockaddr_un__; __const struct sockaddr_x25 *__restrict __sockaddr_x25__;
       } __CONST_SOCKADDR_ARG __attribute__ ((__transparent_union__));







extern int socket (int __domain, int __type, int __protocol) __attribute__ ((__nothrow__));





extern int socketpair (int __domain, int __type, int __protocol,
         int __fds[2]) __attribute__ ((__nothrow__));


extern int bind (int __fd, __CONST_SOCKADDR_ARG __addr, socklen_t __len)
     __attribute__ ((__nothrow__));


extern int getsockname (int __fd, __SOCKADDR_ARG __addr,
   socklen_t *__restrict __len) __attribute__ ((__nothrow__));
# 129 "/usr/include/sys/socket.h" 3 4
extern int connect (int __fd, __CONST_SOCKADDR_ARG __addr, socklen_t __len);



extern int getpeername (int __fd, __SOCKADDR_ARG __addr,
   socklen_t *__restrict __len) __attribute__ ((__nothrow__));






extern ssize_t send (int __fd, __const void *__buf, size_t __n, int __flags);






extern ssize_t recv (int __fd, void *__buf, size_t __n, int __flags);






extern ssize_t sendto (int __fd, __const void *__buf, size_t __n,
         int __flags, __CONST_SOCKADDR_ARG __addr,
         socklen_t __addr_len);
# 166 "/usr/include/sys/socket.h" 3 4
extern ssize_t recvfrom (int __fd, void *__restrict __buf, size_t __n,
    int __flags, __SOCKADDR_ARG __addr,
    socklen_t *__restrict __addr_len);







extern ssize_t sendmsg (int __fd, __const struct msghdr *__message,
   int __flags);






extern ssize_t recvmsg (int __fd, struct msghdr *__message, int __flags);





extern int getsockopt (int __fd, int __level, int __optname,
         void *__restrict __optval,
         socklen_t *__restrict __optlen) __attribute__ ((__nothrow__));




extern int setsockopt (int __fd, int __level, int __optname,
         __const void *__optval, socklen_t __optlen) __attribute__ ((__nothrow__));





extern int listen (int __fd, int __n) __attribute__ ((__nothrow__));
# 214 "/usr/include/sys/socket.h" 3 4
extern int accept (int __fd, __SOCKADDR_ARG __addr,
     socklen_t *__restrict __addr_len);






extern int accept4 (int __fd, __SOCKADDR_ARG __addr,
      socklen_t *__restrict __addr_len, int __flags);
# 232 "/usr/include/sys/socket.h" 3 4
extern int shutdown (int __fd, int __how) __attribute__ ((__nothrow__));




extern int sockatmark (int __fd) __attribute__ ((__nothrow__));







extern int isfdtype (int __fd, int __fdtype) __attribute__ ((__nothrow__));
# 254 "/usr/include/sys/socket.h" 3 4

# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_transport.h" 2


typedef php_stream *(php_stream_transport_factory_func)(const char *proto, long protolen,
  char *resourcename, long resourcenamelen,
  const char *persistent_id, int options, int flags,
  struct timeval *timeout,
  php_stream_context *context );
typedef php_stream_transport_factory_func *php_stream_transport_factory;


__attribute__ ((visibility("default"))) int php_stream_xport_register(char *protocol, php_stream_transport_factory factory );
__attribute__ ((visibility("default"))) int php_stream_xport_unregister(char *protocol );
# 49 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_transport.h"
__attribute__ ((visibility("default"))) php_stream *_php_stream_xport_create(const char *name, long namelen, int options,
  int flags, const char *persistent_id,
  struct timeval *timeout,
  php_stream_context *context,
  char **error_string,
  int *error_code
  );





__attribute__ ((visibility("default"))) int php_stream_xport_bind(php_stream *stream,
  const char *name, long namelen,
  char **error_text
  );


__attribute__ ((visibility("default"))) int php_stream_xport_connect(php_stream *stream,
  const char *name, long namelen,
  int asynchronous,
  struct timeval *timeout,
  char **error_text,
  int *error_code
  );


__attribute__ ((visibility("default"))) int php_stream_xport_listen(php_stream *stream,
  int backlog,
  char **error_text
  );



__attribute__ ((visibility("default"))) int php_stream_xport_accept(php_stream *stream, php_stream **client,
  char **textaddr, int *textaddrlen,
  void **addr, socklen_t *addrlen,
  struct timeval *timeout,
  char **error_text
  );


__attribute__ ((visibility("default"))) int php_stream_xport_get_name(php_stream *stream, int want_peer,
  char **textaddr, int *textaddrlen,
  void **addr, socklen_t *addrlen
  );

enum php_stream_xport_send_recv_flags {
 STREAM_OOB = 1,
 STREAM_PEEK = 2
};



__attribute__ ((visibility("default"))) int php_stream_xport_recvfrom(php_stream *stream, char *buf, size_t buflen,
  long flags, void **addr, socklen_t *addrlen,
  char **textaddr, int *textaddrlen );



__attribute__ ((visibility("default"))) int php_stream_xport_sendto(php_stream *stream, const char *buf, size_t buflen,
  long flags, void *addr, socklen_t addrlen );

typedef enum {
 STREAM_SHUT_RD,
 STREAM_SHUT_WR,
 STREAM_SHUT_RDWR
} stream_shutdown_t;



__attribute__ ((visibility("default"))) int php_stream_xport_shutdown(php_stream *stream, stream_shutdown_t how );





typedef struct _php_stream_xport_param {
 enum {
  STREAM_XPORT_OP_BIND, STREAM_XPORT_OP_CONNECT,
  STREAM_XPORT_OP_LISTEN, STREAM_XPORT_OP_ACCEPT,
  STREAM_XPORT_OP_CONNECT_ASYNC,
  STREAM_XPORT_OP_GET_NAME,
  STREAM_XPORT_OP_GET_PEER_NAME,
  STREAM_XPORT_OP_RECV,
  STREAM_XPORT_OP_SEND,
  STREAM_XPORT_OP_SHUTDOWN
 } op;
 unsigned int want_addr:1;
 unsigned int want_textaddr:1;
 unsigned int want_errortext:1;
 unsigned int how:2;

 struct {
  char *name;
  long namelen;
  int backlog;
  struct timeval *timeout;
  struct sockaddr *addr;
  socklen_t addrlen;
  char *buf;
  size_t buflen;
  long flags;
 } inputs;
 struct {
  php_stream *client;
  int returncode;
  struct sockaddr *addr;
  socklen_t addrlen;
  char *textaddr;
  long textaddrlen;

  char *error_text;
  int error_code;
 } outputs;
} php_stream_xport_param;



typedef enum {
 STREAM_CRYPTO_METHOD_SSLv2_CLIENT,
 STREAM_CRYPTO_METHOD_SSLv3_CLIENT,
 STREAM_CRYPTO_METHOD_SSLv23_CLIENT,
 STREAM_CRYPTO_METHOD_TLS_CLIENT,
 STREAM_CRYPTO_METHOD_SSLv2_SERVER,
 STREAM_CRYPTO_METHOD_SSLv3_SERVER,
 STREAM_CRYPTO_METHOD_SSLv23_SERVER,
 STREAM_CRYPTO_METHOD_TLS_SERVER
} php_stream_xport_crypt_method_t;


__attribute__ ((visibility("default"))) int php_stream_xport_crypto_setup(php_stream *stream, php_stream_xport_crypt_method_t crypto_method, php_stream *session_stream );
__attribute__ ((visibility("default"))) int php_stream_xport_crypto_enable(php_stream *stream, int activate );


typedef struct _php_stream_xport_crypto_param {
 enum {
  STREAM_XPORT_CRYPTO_OP_SETUP,
  STREAM_XPORT_CRYPTO_OP_ENABLE
 } op;
 struct {
  int activate;
  php_stream_xport_crypt_method_t method;
  php_stream *session;
 } inputs;
 struct {
  int returncode;
 } outputs;
} php_stream_xport_crypto_param;


__attribute__ ((visibility("default"))) HashTable *php_stream_xport_get_hash(void);
__attribute__ ((visibility("default"))) php_stream_transport_factory_func php_stream_generic_socket_factory;

# 457 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_plain_wrapper.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_plain_wrapper.h"
__attribute__ ((visibility("default"))) extern php_stream_ops php_stream_stdio_ops;
__attribute__ ((visibility("default"))) extern php_stream_wrapper php_plain_files_wrapper;




__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen(const char *filename, const char *mode, char **opened_path, int options );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_with_path(char *filename, char *mode, char *path, char **opened_path, int options );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_from_file(FILE *file, const char *mode );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_from_fd(int fd, const char *mode, const char *persistent_id );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_from_pipe(FILE *file, const char *mode );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_tmpfile(int dummy );


__attribute__ ((visibility("default"))) php_stream *_php_stream_fopen_temporary_file(const char *dir, const char *pfx, char **opened_path );





__attribute__ ((visibility("default"))) FILE * _php_stream_open_wrapper_as_file(char * path, char * mode, int options, char **opened_path );



# 458 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_glob_wrapper.h" 1
# 21 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_glob_wrapper.h"
__attribute__ ((visibility("default"))) extern php_stream_wrapper php_glob_stream_wrapper;
__attribute__ ((visibility("default"))) extern php_stream_ops php_glob_stream_ops;



__attribute__ ((visibility("default"))) char* _php_glob_stream_get_path(php_stream *stream, int copy, int *plen );


__attribute__ ((visibility("default"))) char* _php_glob_stream_get_pattern(php_stream *stream, int copy, int *plen );


__attribute__ ((visibility("default"))) int _php_glob_stream_get_count(php_stream *stream, int *pflags );



# 459 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_userspace.h" 1
# 23 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_userspace.h"
__attribute__ ((visibility("default"))) extern php_stream_ops php_stream_userspace_ops;
__attribute__ ((visibility("default"))) extern php_stream_ops php_stream_userspace_dir_ops;
# 460 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_mmap.h" 1
# 30 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_mmap.h"
typedef enum {

 PHP_STREAM_MMAP_SUPPORTED,



 PHP_STREAM_MMAP_MAP_RANGE,

 PHP_STREAM_MMAP_UNMAP
} php_stream_mmap_operation_t;

typedef enum {
 PHP_STREAM_MAP_MODE_READONLY,
 PHP_STREAM_MAP_MODE_READWRITE,
 PHP_STREAM_MAP_MODE_SHARED_READONLY,
 PHP_STREAM_MAP_MODE_SHARED_READWRITE
} php_stream_mmap_access_t;

typedef struct {


 size_t offset;
 size_t length;

 php_stream_mmap_access_t mode;


 char *mapped;

} php_stream_mmap_range;
# 69 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/streams/php_stream_mmap.h"

__attribute__ ((visibility("default"))) char *_php_stream_mmap_range(php_stream *stream, size_t offset, size_t length, php_stream_mmap_operation_t mode, size_t *mapped_len );



__attribute__ ((visibility("default"))) int _php_stream_mmap_unmap(php_stream *stream );


__attribute__ ((visibility("default"))) int _php_stream_mmap_unmap_ex(php_stream *stream, off_t readden );


# 461 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 2
# 477 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"

__attribute__ ((visibility("default"))) int _php_stream_cast(php_stream *stream, int castas, void **ret, int show_err );

# 541 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
int php_init_stream_wrappers(int module_number );
int php_shutdown_stream_wrappers(int module_number );
void php_shutdown_stream_hashes(void);
int zm_deactivate_streams(int type, int module_number );


__attribute__ ((visibility("default"))) int php_register_url_stream_wrapper(char *protocol, php_stream_wrapper *wrapper );
__attribute__ ((visibility("default"))) int php_unregister_url_stream_wrapper(char *protocol );
__attribute__ ((visibility("default"))) int php_register_url_stream_wrapper_volatile(char *protocol, php_stream_wrapper *wrapper );
__attribute__ ((visibility("default"))) int php_unregister_url_stream_wrapper_volatile(char *protocol );
__attribute__ ((visibility("default"))) php_stream *_php_stream_open_wrapper_ex(char *path, char *mode, int options, char **opened_path, php_stream_context *context );
__attribute__ ((visibility("default"))) php_stream_wrapper *php_stream_locate_url_wrapper(const char *path, char **path_for_open, int options );
__attribute__ ((visibility("default"))) char *php_stream_locate_eol(php_stream *stream, char *buf, size_t buf_len );
# 568 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
__attribute__ ((visibility("default"))) void php_stream_wrapper_log_error(php_stream_wrapper *wrapper, int options , const char *fmt, ...) __attribute__ ((format(printf, 3, 4)));
# 579 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h"
__attribute__ ((visibility("default"))) int _php_stream_make_seekable(php_stream *origstream, php_stream **newstream, int flags );



__attribute__ ((visibility("default"))) HashTable *_php_stream_get_url_stream_wrappers_hash(void);

__attribute__ ((visibility("default"))) HashTable *php_stream_get_url_stream_wrappers_hash_global(void);
__attribute__ ((visibility("default"))) HashTable *_php_get_stream_filters_hash(void);

__attribute__ ((visibility("default"))) HashTable *php_get_stream_filters_hash_global(void);
extern php_stream_wrapper_ops *php_stream_user_wrapper_ops;

# 401 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_memory_streams.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_memory_streams.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_memory_streams.h" 2
# 42 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_memory_streams.h"

__attribute__ ((visibility("default"))) php_stream *_php_stream_memory_create(int mode );
__attribute__ ((visibility("default"))) php_stream *_php_stream_memory_open(int mode, char *buf, size_t length );
__attribute__ ((visibility("default"))) char *_php_stream_memory_get_buffer(php_stream *stream, size_t *length );

__attribute__ ((visibility("default"))) php_stream *_php_stream_temp_create(int mode, size_t max_memory_usage );
__attribute__ ((visibility("default"))) php_stream *_php_stream_temp_open(int mode, size_t max_memory_usage, char *buf, size_t length );


extern __attribute__ ((visibility("default"))) php_stream_ops php_stream_memory_ops;
extern __attribute__ ((visibility("default"))) php_stream_ops php_stream_temp_ops;
extern __attribute__ ((visibility("default"))) php_stream_ops php_stream_rfc2397_ops;
extern __attribute__ ((visibility("default"))) php_stream_wrapper php_stream_rfc2397_wrapper;
# 402 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/fopen_wrappers.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/fopen_wrappers.h"

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_globals.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_globals.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_globals.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_globals.h" 2

typedef struct _php_core_globals php_core_globals;






extern __attribute__ ((visibility("default"))) struct _php_core_globals core_globals;
# 49 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_globals.h"
struct _php_tick_function_entry;

typedef struct _arg_separators {
 char *output;
 char *input;
} arg_separators;

struct _php_core_globals {
 zend_bool implicit_flush;

 long output_buffering;

 zend_bool sql_safe_mode;
 zend_bool enable_dl;

 char *output_handler;

 char *unserialize_callback_func;
 long serialize_precision;

 long memory_limit;
 long max_input_time;

 zend_bool track_errors;
 zend_bool display_errors;
 zend_bool display_startup_errors;
 zend_bool log_errors;
 long log_errors_max_len;
 zend_bool ignore_repeated_errors;
 zend_bool ignore_repeated_source;
 zend_bool report_memleaks;
 char *error_log;

 char *doc_root;
 char *user_dir;
 char *include_path;
 char *open_basedir;
 char *extension_dir;

 char *upload_tmp_dir;
 long upload_max_filesize;

 char *error_append_string;
 char *error_prepend_string;

 char *auto_prepend_file;
 char *auto_append_file;

 arg_separators arg_separator;

 char *variables_order;

 HashTable rfc1867_protected_variables;

 short connection_status;
 short ignore_user_abort;

 unsigned char header_is_being_sent;

 zend_llist tick_functions;

 zval *http_globals[6];

 zend_bool expose_php;

 zend_bool register_argc_argv;
 zend_bool auto_globals_jit;

 char *docref_root;
 char *docref_ext;

 zend_bool html_errors;
 zend_bool xmlrpc_errors;

 long xmlrpc_error_number;

 zend_bool activated_auto_globals[8];

 zend_bool modules_activated;
 zend_bool file_uploads;
 zend_bool during_request_startup;
 zend_bool allow_url_fopen;
 zend_bool enable_post_data_reading;
 zend_bool always_populate_raw_post_data;
 zend_bool report_zend_debug;

 int last_error_type;
 char *last_error_message;
 char *last_error_file;
 int last_error_lineno;

 char *disable_functions;
 char *disable_classes;
 zend_bool allow_url_include;
 zend_bool exit_on_timeout;



 long max_input_nesting_level;
 zend_bool in_user_include;

 char *user_ini_filename;
 long user_ini_cache_ttl;

 char *request_order;

 zend_bool mail_x_header;
 char *mail_log;

 zend_bool in_error_log;




};
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/fopen_wrappers.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_ini.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_ini.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ini.h" 1
# 63 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ini.h"
struct _zend_ini_entry {
 int module_number;
 int modifiable;
 char *name;
 uint name_length;
 int (*on_modify)(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
 void *mh_arg1;
 void *mh_arg2;
 void *mh_arg3;

 char *value;
 uint value_length;

 char *orig_value;
 uint orig_value_length;
 int orig_modifiable;
 int modified;

 void (*displayer)(zend_ini_entry *ini_entry, int type);
};


__attribute__ ((visibility("default"))) int zend_ini_startup(void);
__attribute__ ((visibility("default"))) int zend_ini_shutdown(void);
__attribute__ ((visibility("default"))) int zend_ini_global_shutdown(void);
__attribute__ ((visibility("default"))) int zend_ini_deactivate(void);

__attribute__ ((visibility("default"))) int zend_copy_ini_directives(void);

__attribute__ ((visibility("default"))) void zend_ini_sort_entries(void);

__attribute__ ((visibility("default"))) int zend_register_ini_entries(const zend_ini_entry *ini_entry, int module_number );
__attribute__ ((visibility("default"))) void zend_unregister_ini_entries(int module_number );
__attribute__ ((visibility("default"))) void zend_ini_refresh_caches(int stage );
__attribute__ ((visibility("default"))) int zend_alter_ini_entry(char *name, uint name_length, char *new_value, uint new_value_length, int modify_type, int stage);
__attribute__ ((visibility("default"))) int zend_alter_ini_entry_ex(char *name, uint name_length, char *new_value, uint new_value_length, int modify_type, int stage, int force_change );
__attribute__ ((visibility("default"))) int zend_restore_ini_entry(char *name, uint name_length, int stage);
__attribute__ ((visibility("default"))) void display_ini_entries(zend_module_entry *module);

__attribute__ ((visibility("default"))) long zend_ini_long(char *name, uint name_length, int orig);
__attribute__ ((visibility("default"))) double zend_ini_double(char *name, uint name_length, int orig);
__attribute__ ((visibility("default"))) char *zend_ini_string(char *name, uint name_length, int orig);
__attribute__ ((visibility("default"))) char *zend_ini_string_ex(char *name, uint name_length, int orig, zend_bool *exists);

__attribute__ ((visibility("default"))) int zend_ini_register_displayer(char *name, uint name_length, void (*displayer)(zend_ini_entry *ini_entry, int type));

__attribute__ ((visibility("default"))) void zend_ini_boolean_displayer_cb(zend_ini_entry *ini_entry, int type);
__attribute__ ((visibility("default"))) void zend_ini_color_displayer_cb(zend_ini_entry *ini_entry, int type);
__attribute__ ((visibility("default"))) void display_link_numbers(zend_ini_entry *ini_entry, int type);

# 175 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ini.h"

__attribute__ ((visibility("default"))) int OnUpdateBool(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
__attribute__ ((visibility("default"))) int OnUpdateLong(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
__attribute__ ((visibility("default"))) int OnUpdateLongGEZero(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
__attribute__ ((visibility("default"))) int OnUpdateReal(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
__attribute__ ((visibility("default"))) int OnUpdateString(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );
__attribute__ ((visibility("default"))) int OnUpdateStringUnempty(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );

# 195 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_ini.h"
typedef void (*zend_ini_parser_cb_t)(zval *arg1, zval *arg2, zval *arg3, int callback_type, void *arg );

__attribute__ ((visibility("default"))) int zend_parse_ini_file(zend_file_handle *fh, zend_bool unbuffered_errors, int scanner_mode, zend_ini_parser_cb_t ini_parser_cb, void *arg );
__attribute__ ((visibility("default"))) int zend_parse_ini_string(char *str, zend_bool unbuffered_errors, int scanner_mode, zend_ini_parser_cb_t ini_parser_cb, void *arg );







typedef struct _zend_ini_parser_param {
 zend_ini_parser_cb_t ini_parser_cb;
 void *arg;
} zend_ini_parser_param;
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_ini.h" 2


__attribute__ ((visibility("default"))) void config_zval_dtor(zval *zvalue);
int php_init_config(void);
int php_shutdown_config(void);
void php_ini_register_extensions(void);
__attribute__ ((visibility("default"))) zval *cfg_get_entry(const char *name, uint name_length);
__attribute__ ((visibility("default"))) int cfg_get_long(const char *varname, long *result);
__attribute__ ((visibility("default"))) int cfg_get_double(const char *varname, double *result);
__attribute__ ((visibility("default"))) int cfg_get_string(const char *varname, char **result);
__attribute__ ((visibility("default"))) int php_parse_user_ini_file(const char *dirname, char *ini_filename, HashTable *target_hash );
__attribute__ ((visibility("default"))) void php_ini_activate_config(HashTable *source_hash, int modify_type, int stage );
__attribute__ ((visibility("default"))) int php_ini_has_per_dir_config(void);
__attribute__ ((visibility("default"))) int php_ini_has_per_host_config(void);
__attribute__ ((visibility("default"))) void php_ini_activate_per_dir_config(char *path, uint path_len );
__attribute__ ((visibility("default"))) void php_ini_activate_per_host_config(const char *host, uint host_len );
__attribute__ ((visibility("default"))) HashTable* php_ini_get_configuration_hash(void);

# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/fopen_wrappers.h" 2

__attribute__ ((visibility("default"))) int php_fopen_primary_script(zend_file_handle *file_handle );
__attribute__ ((visibility("default"))) char *expand_filepath(const char *filepath, char *real_path );
__attribute__ ((visibility("default"))) char *expand_filepath_ex(const char *filepath, char *real_path, const char *relative_to, size_t relative_to_len );
__attribute__ ((visibility("default"))) char *expand_filepath_with_mode(const char *filepath, char *real_path, const char *relative_to, size_t relative_to_len, int realpath_mode );

__attribute__ ((visibility("default"))) int php_check_open_basedir(const char *path );
__attribute__ ((visibility("default"))) int php_check_open_basedir_ex(const char *path, int warn );
__attribute__ ((visibility("default"))) int php_check_specific_open_basedir(const char *basedir, const char *path );
# 47 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/fopen_wrappers.h"
__attribute__ ((visibility("default"))) int php_check_safe_mode_include_dir(const char *path );

__attribute__ ((visibility("default"))) char *php_resolve_path(const char *filename, int filename_len, const char *path );

__attribute__ ((visibility("default"))) FILE *php_fopen_with_path(const char *filename, const char *mode, const char *path, char **opened_path );

__attribute__ ((visibility("default"))) char *php_strip_url_passwd(char *path);

__attribute__ ((visibility("default"))) int OnUpdateBaseDir(zend_ini_entry *entry, char *new_value, uint new_value_length, void *mh_arg1, void *mh_arg2, void *mh_arg3, int stage );

# 403 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2



# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/TSRM.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config_common.h" 1
# 13 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config_common.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config.h" 1
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/include/../main/php_config.h" 1
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config.h" 2
# 14 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config_common.h" 2
# 41 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config_common.h"
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 42 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_config_common.h" 2
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h" 2






# 1 "/usr/include/utime.h" 1 3 4
# 28 "/usr/include/utime.h" 3 4






# 1 "/usr/include/time.h" 1 3 4
# 35 "/usr/include/utime.h" 2 3 4



struct utimbuf
  {
    __time_t actime;
    __time_t modtime;
  };



extern int utime (__const char *__file,
    __const struct utimbuf *__file_times)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


# 35 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h" 2
# 90 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h"
# 1 "/usr/include/dirent.h" 1 3 4
# 28 "/usr/include/dirent.h" 3 4

# 62 "/usr/include/dirent.h" 3 4
# 1 "/usr/include/bits/dirent.h" 1 3 4
# 23 "/usr/include/bits/dirent.h" 3 4
struct dirent
  {

    __ino_t d_ino;
    __off_t d_off;




    unsigned short int d_reclen;
    unsigned char d_type;
    char d_name[256];
  };


struct dirent64
  {
    __ino64_t d_ino;
    __off64_t d_off;
    unsigned short int d_reclen;
    unsigned char d_type;
    char d_name[256];
  };
# 63 "/usr/include/dirent.h" 2 3 4
# 98 "/usr/include/dirent.h" 3 4
enum
  {
    DT_UNKNOWN = 0,

    DT_FIFO = 1,

    DT_CHR = 2,

    DT_DIR = 4,

    DT_BLK = 6,

    DT_REG = 8,

    DT_LNK = 10,

    DT_SOCK = 12,

    DT_WHT = 14

  };
# 128 "/usr/include/dirent.h" 3 4
typedef struct __dirstream DIR;






extern DIR *opendir (__const char *__name) __attribute__ ((__nonnull__ (1)));






extern DIR *fdopendir (int __fd);







extern int closedir (DIR *__dirp) __attribute__ ((__nonnull__ (1)));
# 163 "/usr/include/dirent.h" 3 4
extern struct dirent *readdir (DIR *__dirp) __attribute__ ((__nonnull__ (1)));
# 174 "/usr/include/dirent.h" 3 4
extern struct dirent64 *readdir64 (DIR *__dirp) __attribute__ ((__nonnull__ (1)));
# 184 "/usr/include/dirent.h" 3 4
extern int readdir_r (DIR *__restrict __dirp,
        struct dirent *__restrict __entry,
        struct dirent **__restrict __result)
     __attribute__ ((__nonnull__ (1, 2, 3)));
# 201 "/usr/include/dirent.h" 3 4
extern int readdir64_r (DIR *__restrict __dirp,
   struct dirent64 *__restrict __entry,
   struct dirent64 **__restrict __result)
     __attribute__ ((__nonnull__ (1, 2, 3)));




extern void rewinddir (DIR *__dirp) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern void seekdir (DIR *__dirp, long int __pos) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));


extern long int telldir (DIR *__dirp) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));





extern int dirfd (DIR *__dirp) __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1)));
# 245 "/usr/include/dirent.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 246 "/usr/include/dirent.h" 2 3 4






extern int scandir (__const char *__restrict __dir,
      struct dirent ***__restrict __namelist,
      int (*__selector) (__const struct dirent *),
      int (*__cmp) (__const struct dirent **,
      __const struct dirent **))
     __attribute__ ((__nonnull__ (1, 2)));
# 275 "/usr/include/dirent.h" 3 4
extern int scandir64 (__const char *__restrict __dir,
        struct dirent64 ***__restrict __namelist,
        int (*__selector) (__const struct dirent64 *),
        int (*__cmp) (__const struct dirent64 **,
        __const struct dirent64 **))
     __attribute__ ((__nonnull__ (1, 2)));




extern int alphasort (__const struct dirent **__e1,
        __const struct dirent **__e2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 300 "/usr/include/dirent.h" 3 4
extern int alphasort64 (__const struct dirent64 **__e1,
   __const struct dirent64 **__e2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 313 "/usr/include/dirent.h" 3 4
extern __ssize_t getdirentries (int __fd, char *__restrict __buf,
    size_t __nbytes,
    __off_t *__restrict __basep)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4)));
# 330 "/usr/include/dirent.h" 3 4
extern __ssize_t getdirentries64 (int __fd, char *__restrict __buf,
      size_t __nbytes,
      __off64_t *__restrict __basep)
     __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (2, 4)));






extern int versionsort (__const struct dirent **__e1,
   __const struct dirent **__e2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));
# 356 "/usr/include/dirent.h" 3 4
extern int versionsort64 (__const struct dirent64 **__e1,
     __const struct dirent64 **__e2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__)) __attribute__ ((__nonnull__ (1, 2)));




# 91 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h" 2
# 145 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h"
typedef struct _cwd_state {
 char *cwd;
 int cwd_length;
} cwd_state;

typedef int (*verify_path_func)(const cwd_state *);

__attribute__ ((visibility("default"))) void virtual_cwd_startup(void);
__attribute__ ((visibility("default"))) void virtual_cwd_shutdown(void);
__attribute__ ((visibility("default"))) char *virtual_getcwd_ex(size_t *length );
__attribute__ ((visibility("default"))) char *virtual_getcwd(char *buf, size_t size );
__attribute__ ((visibility("default"))) int virtual_chdir(const char *path );
__attribute__ ((visibility("default"))) int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path ) );
__attribute__ ((visibility("default"))) int virtual_filepath(const char *path, char **filepath );
__attribute__ ((visibility("default"))) int virtual_filepath_ex(const char *path, char **filepath, verify_path_func verify_path );
__attribute__ ((visibility("default"))) char *virtual_realpath(const char *path, char *real_path );
__attribute__ ((visibility("default"))) FILE *virtual_fopen(const char *path, const char *mode );
__attribute__ ((visibility("default"))) int virtual_open(const char *path , int flags, ...);
__attribute__ ((visibility("default"))) int virtual_creat(const char *path, mode_t mode );
__attribute__ ((visibility("default"))) int virtual_rename(char *oldname, char *newname );
__attribute__ ((visibility("default"))) int virtual_stat(const char *path, struct stat *buf );
__attribute__ ((visibility("default"))) int virtual_lstat(const char *path, struct stat *buf );
__attribute__ ((visibility("default"))) int virtual_unlink(const char *path );
__attribute__ ((visibility("default"))) int virtual_mkdir(const char *pathname, mode_t mode );
__attribute__ ((visibility("default"))) int virtual_rmdir(const char *pathname );
__attribute__ ((visibility("default"))) DIR *virtual_opendir(const char *pathname );
__attribute__ ((visibility("default"))) FILE *virtual_popen(const char *command, const char *type );
__attribute__ ((visibility("default"))) int virtual_access(const char *pathname, int mode );
# 190 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h"
__attribute__ ((visibility("default"))) int virtual_utime(const char *filename, struct utimbuf *buf );

__attribute__ ((visibility("default"))) int virtual_chmod(const char *filename, mode_t mode );

__attribute__ ((visibility("default"))) int virtual_chown(const char *filename, uid_t owner, gid_t group, int link );
# 204 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/TSRM/tsrm_virtual_cwd.h"
__attribute__ ((visibility("default"))) int virtual_file_ex(cwd_state *state, const char *path, verify_path_func verify_path, int use_realpath );

__attribute__ ((visibility("default"))) char *tsrm_realpath(const char *path, char *real_path );




typedef struct _realpath_cache_bucket {
 unsigned long key;
 char *path;
 int path_len;
 char *realpath;
 int realpath_len;
 int is_dir;
 time_t expires;






 struct _realpath_cache_bucket *next;
} realpath_cache_bucket;

typedef struct _virtual_cwd_globals {
 cwd_state cwd;
 long realpath_cache_size;
 long realpath_cache_size_limit;
 long realpath_cache_ttl;
 realpath_cache_bucket *realpath_cache[1024];
} virtual_cwd_globals;





extern virtual_cwd_globals cwd_globals;



__attribute__ ((visibility("default"))) void realpath_cache_clean(void);
__attribute__ ((visibility("default"))) void realpath_cache_del(const char *path, int path_len );
__attribute__ ((visibility("default"))) realpath_cache_bucket* realpath_cache_lookup(const char *path, int path_len, time_t t );
__attribute__ ((visibility("default"))) int realpath_cache_size(void);
__attribute__ ((visibility("default"))) int realpath_cache_max_buckets(void);
__attribute__ ((visibility("default"))) realpath_cache_bucket** realpath_cache_get_buckets(void);
# 407 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_constants.h" 1
# 33 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_constants.h"
typedef struct _zend_constant {
 zval value;
 int flags;
 char *name;
 uint name_len;
 int module_number;
} zend_constant;
# 56 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_constants.h"

void clean_module_constants(int module_number );
void free_zend_constant(zend_constant *c);
int zend_startup_constants(void);
int zend_shutdown_constants(void);
void zend_register_standard_constants(void);
void clean_non_persistent_constants(void);
__attribute__ ((visibility("default"))) int zend_get_constant(const char *name, uint name_len, zval *result );
__attribute__ ((visibility("default"))) int zend_get_constant_ex(const char *name, uint name_len, zval *result, zend_class_entry *scope, ulong flags );
__attribute__ ((visibility("default"))) void zend_register_long_constant(const char *name, uint name_len, long lval, int flags, int module_number );
__attribute__ ((visibility("default"))) void zend_register_double_constant(const char *name, uint name_len, double dval, int flags, int module_number );
__attribute__ ((visibility("default"))) void zend_register_string_constant(const char *name, uint name_len, char *strval, int flags, int module_number );
__attribute__ ((visibility("default"))) void zend_register_stringl_constant(const char *name, uint name_len, char *strval, uint strlen, int flags, int module_number );
__attribute__ ((visibility("default"))) int zend_register_constant(zend_constant *c );
void zend_copy_constants(HashTable *target, HashTable *sourc);
void copy_zend_constant(zend_constant *c);
zend_constant *zend_quick_get_constant(const zend_literal *key, ulong flags );

# 409 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2






# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h" 2





# 1 "/usr/include/time.h" 1 3 4
# 30 "/usr/include/time.h" 3 4








# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 39 "/usr/include/time.h" 2 3 4



# 1 "/usr/include/bits/time.h" 1 3 4
# 43 "/usr/include/time.h" 2 3 4
# 131 "/usr/include/time.h" 3 4


struct tm
{
  int tm_sec;
  int tm_min;
  int tm_hour;
  int tm_mday;
  int tm_mon;
  int tm_year;
  int tm_wday;
  int tm_yday;
  int tm_isdst;


  long int tm_gmtoff;
  __const char *tm_zone;




};








struct itimerspec
  {
    struct timespec it_interval;
    struct timespec it_value;
  };


struct sigevent;
# 180 "/usr/include/time.h" 3 4



extern clock_t clock (void) __attribute__ ((__nothrow__));


extern time_t time (time_t *__timer) __attribute__ ((__nothrow__));


extern double difftime (time_t __time1, time_t __time0)
     __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern time_t mktime (struct tm *__tp) __attribute__ ((__nothrow__));





extern size_t strftime (char *__restrict __s, size_t __maxsize,
   __const char *__restrict __format,
   __const struct tm *__restrict __tp) __attribute__ ((__nothrow__));





extern char *strptime (__const char *__restrict __s,
         __const char *__restrict __fmt, struct tm *__tp)
     __attribute__ ((__nothrow__));







extern size_t strftime_l (char *__restrict __s, size_t __maxsize,
     __const char *__restrict __format,
     __const struct tm *__restrict __tp,
     __locale_t __loc) __attribute__ ((__nothrow__));



extern char *strptime_l (__const char *__restrict __s,
    __const char *__restrict __fmt, struct tm *__tp,
    __locale_t __loc) __attribute__ ((__nothrow__));






extern struct tm *gmtime (__const time_t *__timer) __attribute__ ((__nothrow__));



extern struct tm *localtime (__const time_t *__timer) __attribute__ ((__nothrow__));





extern struct tm *gmtime_r (__const time_t *__restrict __timer,
       struct tm *__restrict __tp) __attribute__ ((__nothrow__));



extern struct tm *localtime_r (__const time_t *__restrict __timer,
          struct tm *__restrict __tp) __attribute__ ((__nothrow__));





extern char *asctime (__const struct tm *__tp) __attribute__ ((__nothrow__));


extern char *ctime (__const time_t *__timer) __attribute__ ((__nothrow__));







extern char *asctime_r (__const struct tm *__restrict __tp,
   char *__restrict __buf) __attribute__ ((__nothrow__));


extern char *ctime_r (__const time_t *__restrict __timer,
        char *__restrict __buf) __attribute__ ((__nothrow__));




extern char *__tzname[2];
extern int __daylight;
extern long int __timezone;




extern char *tzname[2];



extern void tzset (void) __attribute__ ((__nothrow__));



extern int daylight;
extern long int timezone;





extern int stime (__const time_t *__when) __attribute__ ((__nothrow__));
# 313 "/usr/include/time.h" 3 4
extern time_t timegm (struct tm *__tp) __attribute__ ((__nothrow__));


extern time_t timelocal (struct tm *__tp) __attribute__ ((__nothrow__));


extern int dysize (int __year) __attribute__ ((__nothrow__)) __attribute__ ((__const__));
# 328 "/usr/include/time.h" 3 4
extern int nanosleep (__const struct timespec *__requested_time,
        struct timespec *__remaining);



extern int clock_getres (clockid_t __clock_id, struct timespec *__res) __attribute__ ((__nothrow__));


extern int clock_gettime (clockid_t __clock_id, struct timespec *__tp) __attribute__ ((__nothrow__));


extern int clock_settime (clockid_t __clock_id, __const struct timespec *__tp)
     __attribute__ ((__nothrow__));






extern int clock_nanosleep (clockid_t __clock_id, int __flags,
       __const struct timespec *__req,
       struct timespec *__rem);


extern int clock_getcpuclockid (pid_t __pid, clockid_t *__clock_id) __attribute__ ((__nothrow__));




extern int timer_create (clockid_t __clock_id,
    struct sigevent *__restrict __evp,
    timer_t *__restrict __timerid) __attribute__ ((__nothrow__));


extern int timer_delete (timer_t __timerid) __attribute__ ((__nothrow__));


extern int timer_settime (timer_t __timerid, int __flags,
     __const struct itimerspec *__restrict __value,
     struct itimerspec *__restrict __ovalue) __attribute__ ((__nothrow__));


extern int timer_gettime (timer_t __timerid, struct itimerspec *__value)
     __attribute__ ((__nothrow__));


extern int timer_getoverrun (timer_t __timerid) __attribute__ ((__nothrow__));
# 390 "/usr/include/time.h" 3 4
extern int getdate_err;
# 399 "/usr/include/time.h" 3 4
extern struct tm *getdate (__const char *__string);
# 413 "/usr/include/time.h" 3 4
extern int getdate_r (__const char *__restrict __string,
        struct tm *__restrict __resbufp);



# 31 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h" 2
# 52 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h"

# 119 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_reentrancy.h"

# 416 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php.h" 2
# 22 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_streams.h" 1
# 23 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_main.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_main.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_llist.h" 1
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_operators.h" 1
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h" 2
# 51 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h"
typedef struct {
 char *header;
 uint header_len;
} sapi_header_struct;


typedef struct {
 zend_llist headers;
 int http_response_code;
 unsigned char send_default_content_type;
 char *mimetype;
 char *http_status_line;
} sapi_headers_struct;


typedef struct _sapi_post_entry sapi_post_entry;
typedef struct _sapi_module_struct sapi_module_struct;


extern __attribute__ ((visibility("default"))) sapi_module_struct sapi_module;








typedef struct {
 const char *request_method;
 char *query_string;
 char *post_data, *raw_post_data;
 char *cookie_data;
 long content_length;
 uint post_data_length, raw_post_data_length;

 char *path_translated;
 char *request_uri;

 const char *content_type;

 zend_bool headers_only;
 zend_bool no_headers;
 zend_bool headers_read;

 sapi_post_entry *post_entry;

 char *content_type_dup;


 char *auth_user;
 char *auth_password;
 char *auth_digest;


 char *argv0;


 char *current_user;
 int current_user_length;


 int argc;
 char **argv;
 int proto_num;
} sapi_request_info;


typedef struct _sapi_globals_struct {
 void *server_context;
 sapi_request_info request_info;
 sapi_headers_struct sapi_headers;
 int read_post_bytes;
 unsigned char headers_sent;
 struct stat global_stat;
 char *default_mimetype;
 char *default_charset;
 HashTable *rfc1867_uploaded_files;
 long post_max_size;
 int options;
 zend_bool sapi_started;
 double global_request_time;
 HashTable known_post_content_types;
 zval *callback_func;
 zend_fcall_info_cache fci_cache;
 zend_bool callback_run;
} sapi_globals_struct;








extern __attribute__ ((visibility("default"))) sapi_globals_struct sapi_globals;


__attribute__ ((visibility("default"))) void sapi_startup(sapi_module_struct *sf);
__attribute__ ((visibility("default"))) void sapi_shutdown(void);
__attribute__ ((visibility("default"))) void sapi_activate(void);
__attribute__ ((visibility("default"))) void sapi_deactivate(void);
__attribute__ ((visibility("default"))) void sapi_initialize_empty_request(void);

# 167 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h"
typedef struct {
 char *line;
 uint line_len;
 long response_code;
} sapi_header_line;

typedef enum {
 SAPI_HEADER_REPLACE,
 SAPI_HEADER_ADD,
 SAPI_HEADER_DELETE,
 SAPI_HEADER_DELETE_ALL,
 SAPI_HEADER_SET_STATUS
} sapi_header_op_enum;


__attribute__ ((visibility("default"))) int sapi_header_op(sapi_header_op_enum op, void *arg );


__attribute__ ((visibility("default"))) int sapi_add_header_ex(char *header_line, uint header_line_len, zend_bool duplicate, zend_bool replace );



__attribute__ ((visibility("default"))) int sapi_send_headers(void);
__attribute__ ((visibility("default"))) void sapi_free_header(sapi_header_struct *sapi_header);
__attribute__ ((visibility("default"))) void sapi_handle_post(void *arg );

__attribute__ ((visibility("default"))) int sapi_register_post_entries(sapi_post_entry *post_entry );
__attribute__ ((visibility("default"))) int sapi_register_post_entry(sapi_post_entry *post_entry );
__attribute__ ((visibility("default"))) void sapi_unregister_post_entry(sapi_post_entry *post_entry );
__attribute__ ((visibility("default"))) int sapi_register_default_post_reader(void (*default_post_reader)(void) );
__attribute__ ((visibility("default"))) int sapi_register_treat_data(void (*treat_data)(int arg, char *str, zval *destArray ) );
__attribute__ ((visibility("default"))) int sapi_register_input_filter(unsigned int (*input_filter)(int arg, char *var, char **val, unsigned int val_len, unsigned int *new_val_len ), unsigned int (*input_filter_init)(void) );

__attribute__ ((visibility("default"))) int sapi_flush(void);
__attribute__ ((visibility("default"))) struct stat *sapi_get_stat(void);
__attribute__ ((visibility("default"))) char *sapi_getenv(char *name, size_t name_len );

__attribute__ ((visibility("default"))) char *sapi_get_default_content_type(void);
__attribute__ ((visibility("default"))) void sapi_get_default_content_type_header(sapi_header_struct *default_header );
__attribute__ ((visibility("default"))) size_t sapi_apply_default_charset(char **mimetype, size_t len );
__attribute__ ((visibility("default"))) void sapi_activate_headers_only(void);

__attribute__ ((visibility("default"))) int sapi_get_fd(int *fd );
__attribute__ ((visibility("default"))) int sapi_force_http_10(void);

__attribute__ ((visibility("default"))) int sapi_get_target_uid(uid_t * );
__attribute__ ((visibility("default"))) int sapi_get_target_gid(gid_t * );
__attribute__ ((visibility("default"))) double sapi_get_request_time(void);
__attribute__ ((visibility("default"))) void sapi_terminate_process(void);


struct _sapi_module_struct {
 char *name;
 char *pretty_name;

 int (*startup)(struct _sapi_module_struct *sapi_module);
 int (*shutdown)(struct _sapi_module_struct *sapi_module);

 int (*activate)(void);
 int (*deactivate)(void);

 int (*ub_write)(const char *str, unsigned int str_length );
 void (*flush)(void *server_context);
 struct stat *(*get_stat)(void);
 char *(*getenv)(char *name, size_t name_len );

 void (*sapi_error)(int type, const char *error_msg, ...);

 int (*header_handler)(sapi_header_struct *sapi_header, sapi_header_op_enum op, sapi_headers_struct *sapi_headers );
 int (*send_headers)(sapi_headers_struct *sapi_headers );
 void (*send_header)(sapi_header_struct *sapi_header, void *server_context );

 int (*read_post)(char *buffer, uint count_bytes );
 char *(*read_cookies)(void);

 void (*register_server_variables)(zval *track_vars_array );
 void (*log_message)(char *message );
 double (*get_request_time)(void);
 void (*terminate_process)(void);

 char *php_ini_path_override;

 void (*block_interruptions)(void);
 void (*unblock_interruptions)(void);

 void (*default_post_reader)(void);
 void (*treat_data)(int arg, char *str, zval *destArray );
 char *executable_location;

 int php_ini_ignore;
 int php_ini_ignore_cwd;

 int (*get_fd)(int *fd );

 int (*force_http_10)(void);

 int (*get_target_uid)(uid_t * );
 int (*get_target_gid)(gid_t * );

 unsigned int (*input_filter)(int arg, char *var, char **val, unsigned int val_len, unsigned int *new_val_len );

 void (*ini_defaults)(HashTable *configuration_hash);
 int phpinfo_as_text;

 char *ini_entries;
 const zend_function_entry *additional_functions;
 unsigned int (*input_filter_init)(void);
};


struct _sapi_post_entry {
 char *content_type;
 uint content_type_len;
 void (*post_reader)(void);
 void (*post_handler)(char *content_type_dup, void *arg );
};
# 302 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/SAPI.h"

__attribute__ ((visibility("default"))) void sapi_read_standard_form_data(void);
__attribute__ ((visibility("default"))) void php_default_post_reader(void);
__attribute__ ((visibility("default"))) void php_default_treat_data(int arg, char *str, zval* destArray );
__attribute__ ((visibility("default"))) unsigned int php_default_input_filter(int arg, char *var, char **val, unsigned int val_len, unsigned int *new_val_len );

# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_main.h" 2


__attribute__ ((visibility("default"))) int php_request_startup(void);
__attribute__ ((visibility("default"))) void php_request_shutdown(void *dummy);
__attribute__ ((visibility("default"))) void php_request_shutdown_for_exec(void *dummy);
__attribute__ ((visibility("default"))) int php_module_startup(sapi_module_struct *sf, zend_module_entry *additional_modules, uint num_additional_modules);
__attribute__ ((visibility("default"))) void php_module_shutdown(void);
__attribute__ ((visibility("default"))) void php_module_shutdown_for_exec(void);
__attribute__ ((visibility("default"))) int php_module_shutdown_wrapper(sapi_module_struct *sapi_globals);
__attribute__ ((visibility("default"))) int php_request_startup_for_hook(void);
__attribute__ ((visibility("default"))) void php_request_shutdown_for_hook(void *dummy);

__attribute__ ((visibility("default"))) int php_register_extensions(zend_module_entry **ptr, int count );

__attribute__ ((visibility("default"))) int php_execute_script(zend_file_handle *primary_file );
__attribute__ ((visibility("default"))) int php_execute_simple_script(zend_file_handle *primary_file, zval **ret );
__attribute__ ((visibility("default"))) int php_handle_special_queries(void);
__attribute__ ((visibility("default"))) int php_lint_script(zend_file_handle *file );

__attribute__ ((visibility("default"))) void php_handle_aborted_connection(void);
__attribute__ ((visibility("default"))) int php_handle_auth_data(const char *auth );

__attribute__ ((visibility("default"))) void php_html_puts(const char *str, uint siz );
__attribute__ ((visibility("default"))) int php_stream_open_for_zend_ex(const char *filename, zend_file_handle *handle, int mode );

extern void php_call_shutdown_functions(void);
extern void php_free_shutdown_functions(void);


extern int php_init_environ(void);
extern int php_shutdown_environ(void);

# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_globals.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_ini.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/info.h" 1
# 58 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/info.h"

void zif_phpversion(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_phpinfo(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_phpcredits(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_logo_guid(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_real_logo_guid(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_zend_logo_guid(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_egg_logo_guid(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_sapi_name(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_uname(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_ini_scanned_files(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_ini_loaded_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
__attribute__ ((visibility("default"))) char *php_info_html_esc(char *string );
__attribute__ ((visibility("default"))) void php_info_html_esc_write(char *string, int str_len );
__attribute__ ((visibility("default"))) void php_print_info_htmlhead(void);
__attribute__ ((visibility("default"))) void php_print_info(int flag );
__attribute__ ((visibility("default"))) void php_print_style(void);
__attribute__ ((visibility("default"))) void php_info_print_style(void);
__attribute__ ((visibility("default"))) void php_info_print_table_colspan_header(int num_cols, char *header);
__attribute__ ((visibility("default"))) void php_info_print_table_header(int num_cols, ...);
__attribute__ ((visibility("default"))) void php_info_print_table_row(int num_cols, ...);
__attribute__ ((visibility("default"))) void php_info_print_table_row_ex(int num_cols, const char *, ...);
__attribute__ ((visibility("default"))) void php_info_print_table_start(void);
__attribute__ ((visibility("default"))) void php_info_print_table_end(void);
__attribute__ ((visibility("default"))) void php_info_print_box_start(int bg);
__attribute__ ((visibility("default"))) void php_info_print_box_end(void);
__attribute__ ((visibility("default"))) void php_info_print_hr(void);
__attribute__ ((visibility("default"))) void php_info_print_module(zend_module_entry *module );
__attribute__ ((visibility("default"))) char *php_logo_guid(void);
__attribute__ ((visibility("default"))) char *php_get_uname(char mode);

void register_phpinfo_constants(int type, int module_number );

# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_versioning.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_versioning.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h" 1
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h"
# 1 "/usr/include/wchar.h" 1 3 4
# 42 "/usr/include/wchar.h" 3 4
# 1 "/usr/include/bits/wchar.h" 1 3 4
# 43 "/usr/include/wchar.h" 2 3 4
# 52 "/usr/include/wchar.h" 3 4
# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 1 3 4
# 352 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/stddef.h" 3 4
typedef unsigned int wint_t;
# 53 "/usr/include/wchar.h" 2 3 4
# 104 "/usr/include/wchar.h" 3 4


typedef __mbstate_t mbstate_t;



# 129 "/usr/include/wchar.h" 3 4





struct tm;









extern wchar_t *wcscpy (wchar_t *__restrict __dest,
   __const wchar_t *__restrict __src) __attribute__ ((__nothrow__));

extern wchar_t *wcsncpy (wchar_t *__restrict __dest,
    __const wchar_t *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__));


extern wchar_t *wcscat (wchar_t *__restrict __dest,
   __const wchar_t *__restrict __src) __attribute__ ((__nothrow__));

extern wchar_t *wcsncat (wchar_t *__restrict __dest,
    __const wchar_t *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__));


extern int wcscmp (__const wchar_t *__s1, __const wchar_t *__s2)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));

extern int wcsncmp (__const wchar_t *__s1, __const wchar_t *__s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));




extern int wcscasecmp (__const wchar_t *__s1, __const wchar_t *__s2) __attribute__ ((__nothrow__));


extern int wcsncasecmp (__const wchar_t *__s1, __const wchar_t *__s2,
   size_t __n) __attribute__ ((__nothrow__));





extern int wcscasecmp_l (__const wchar_t *__s1, __const wchar_t *__s2,
    __locale_t __loc) __attribute__ ((__nothrow__));

extern int wcsncasecmp_l (__const wchar_t *__s1, __const wchar_t *__s2,
     size_t __n, __locale_t __loc) __attribute__ ((__nothrow__));





extern int wcscoll (__const wchar_t *__s1, __const wchar_t *__s2) __attribute__ ((__nothrow__));



extern size_t wcsxfrm (wchar_t *__restrict __s1,
         __const wchar_t *__restrict __s2, size_t __n) __attribute__ ((__nothrow__));








extern int wcscoll_l (__const wchar_t *__s1, __const wchar_t *__s2,
        __locale_t __loc) __attribute__ ((__nothrow__));




extern size_t wcsxfrm_l (wchar_t *__s1, __const wchar_t *__s2,
    size_t __n, __locale_t __loc) __attribute__ ((__nothrow__));


extern wchar_t *wcsdup (__const wchar_t *__s) __attribute__ ((__nothrow__)) __attribute__ ((__malloc__));










extern wchar_t *wcschr (__const wchar_t *__wcs, wchar_t __wc)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));
# 234 "/usr/include/wchar.h" 3 4
extern wchar_t *wcsrchr (__const wchar_t *__wcs, wchar_t __wc)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));






extern wchar_t *wcschrnul (__const wchar_t *__s, wchar_t __wc)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));





extern size_t wcscspn (__const wchar_t *__wcs, __const wchar_t *__reject)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));


extern size_t wcsspn (__const wchar_t *__wcs, __const wchar_t *__accept)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));
# 263 "/usr/include/wchar.h" 3 4
extern wchar_t *wcspbrk (__const wchar_t *__wcs, __const wchar_t *__accept)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));
# 274 "/usr/include/wchar.h" 3 4
extern wchar_t *wcsstr (__const wchar_t *__haystack, __const wchar_t *__needle)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));



extern wchar_t *wcstok (wchar_t *__restrict __s,
   __const wchar_t *__restrict __delim,
   wchar_t **__restrict __ptr) __attribute__ ((__nothrow__));


extern size_t wcslen (__const wchar_t *__s) __attribute__ ((__nothrow__)) __attribute__ ((__pure__));

# 296 "/usr/include/wchar.h" 3 4
extern wchar_t *wcswcs (__const wchar_t *__haystack, __const wchar_t *__needle)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));





extern size_t wcsnlen (__const wchar_t *__s, size_t __maxlen)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));




# 317 "/usr/include/wchar.h" 3 4
extern wchar_t *wmemchr (__const wchar_t *__s, wchar_t __c, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));



extern int wmemcmp (__const wchar_t *__restrict __s1,
      __const wchar_t *__restrict __s2, size_t __n)
     __attribute__ ((__nothrow__)) __attribute__ ((__pure__));


extern wchar_t *wmemcpy (wchar_t *__restrict __s1,
    __const wchar_t *__restrict __s2, size_t __n) __attribute__ ((__nothrow__));



extern wchar_t *wmemmove (wchar_t *__s1, __const wchar_t *__s2, size_t __n)
     __attribute__ ((__nothrow__));


extern wchar_t *wmemset (wchar_t *__s, wchar_t __c, size_t __n) __attribute__ ((__nothrow__));





extern wchar_t *wmempcpy (wchar_t *__restrict __s1,
     __const wchar_t *__restrict __s2, size_t __n)
     __attribute__ ((__nothrow__));






extern wint_t btowc (int __c) __attribute__ ((__nothrow__));



extern int wctob (wint_t __c) __attribute__ ((__nothrow__));



extern int mbsinit (__const mbstate_t *__ps) __attribute__ ((__nothrow__)) __attribute__ ((__pure__));



extern size_t mbrtowc (wchar_t *__restrict __pwc,
         __const char *__restrict __s, size_t __n,
         mbstate_t *__p) __attribute__ ((__nothrow__));


extern size_t wcrtomb (char *__restrict __s, wchar_t __wc,
         mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));


extern size_t __mbrlen (__const char *__restrict __s, size_t __n,
   mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));
extern size_t mbrlen (__const char *__restrict __s, size_t __n,
        mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));








extern wint_t __btowc_alias (int __c) __asm ("btowc");
extern __inline wint_t
__attribute__ ((__nothrow__)) btowc (int __c)
{ return (__builtin_constant_p (__c) && __c >= '\0' && __c <= '\x7f'
   ? (wint_t) __c : __btowc_alias (__c)); }

extern int __wctob_alias (wint_t __c) __asm ("wctob");
extern __inline int
__attribute__ ((__nothrow__)) wctob (wint_t __wc)
{ return (__builtin_constant_p (__wc) && __wc >= L'\0' && __wc <= L'\x7f'
   ? (int) __wc : __wctob_alias (__wc)); }

extern __inline size_t
__attribute__ ((__nothrow__)) mbrlen (__const char *__restrict __s, size_t __n, mbstate_t *__restrict __ps)

{ return (__ps != ((void *)0)
   ? mbrtowc (((void *)0), __s, __n, __ps) : __mbrlen (__s, __n, ((void *)0))); }





extern size_t mbsrtowcs (wchar_t *__restrict __dst,
    __const char **__restrict __src, size_t __len,
    mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));



extern size_t wcsrtombs (char *__restrict __dst,
    __const wchar_t **__restrict __src, size_t __len,
    mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));






extern size_t mbsnrtowcs (wchar_t *__restrict __dst,
     __const char **__restrict __src, size_t __nmc,
     size_t __len, mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));



extern size_t wcsnrtombs (char *__restrict __dst,
     __const wchar_t **__restrict __src,
     size_t __nwc, size_t __len,
     mbstate_t *__restrict __ps) __attribute__ ((__nothrow__));






extern int wcwidth (wchar_t __c) __attribute__ ((__nothrow__));



extern int wcswidth (__const wchar_t *__s, size_t __n) __attribute__ ((__nothrow__));






extern double wcstod (__const wchar_t *__restrict __nptr,
        wchar_t **__restrict __endptr) __attribute__ ((__nothrow__));





extern float wcstof (__const wchar_t *__restrict __nptr,
       wchar_t **__restrict __endptr) __attribute__ ((__nothrow__));
extern long double wcstold (__const wchar_t *__restrict __nptr,
       wchar_t **__restrict __endptr) __attribute__ ((__nothrow__));







extern long int wcstol (__const wchar_t *__restrict __nptr,
   wchar_t **__restrict __endptr, int __base) __attribute__ ((__nothrow__));



extern unsigned long int wcstoul (__const wchar_t *__restrict __nptr,
      wchar_t **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__));






__extension__
extern long long int wcstoll (__const wchar_t *__restrict __nptr,
         wchar_t **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__));



__extension__
extern unsigned long long int wcstoull (__const wchar_t *__restrict __nptr,
     wchar_t **__restrict __endptr,
     int __base) __attribute__ ((__nothrow__));






__extension__
extern long long int wcstoq (__const wchar_t *__restrict __nptr,
        wchar_t **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__));



__extension__
extern unsigned long long int wcstouq (__const wchar_t *__restrict __nptr,
           wchar_t **__restrict __endptr,
           int __base) __attribute__ ((__nothrow__));
# 528 "/usr/include/wchar.h" 3 4
extern long int wcstol_l (__const wchar_t *__restrict __nptr,
     wchar_t **__restrict __endptr, int __base,
     __locale_t __loc) __attribute__ ((__nothrow__));

extern unsigned long int wcstoul_l (__const wchar_t *__restrict __nptr,
        wchar_t **__restrict __endptr,
        int __base, __locale_t __loc) __attribute__ ((__nothrow__));

__extension__
extern long long int wcstoll_l (__const wchar_t *__restrict __nptr,
    wchar_t **__restrict __endptr,
    int __base, __locale_t __loc) __attribute__ ((__nothrow__));

__extension__
extern unsigned long long int wcstoull_l (__const wchar_t *__restrict __nptr,
       wchar_t **__restrict __endptr,
       int __base, __locale_t __loc)
     __attribute__ ((__nothrow__));

extern double wcstod_l (__const wchar_t *__restrict __nptr,
   wchar_t **__restrict __endptr, __locale_t __loc)
     __attribute__ ((__nothrow__));

extern float wcstof_l (__const wchar_t *__restrict __nptr,
         wchar_t **__restrict __endptr, __locale_t __loc)
     __attribute__ ((__nothrow__));

extern long double wcstold_l (__const wchar_t *__restrict __nptr,
         wchar_t **__restrict __endptr,
         __locale_t __loc) __attribute__ ((__nothrow__));




extern wchar_t *wcpcpy (wchar_t *__restrict __dest,
   __const wchar_t *__restrict __src) __attribute__ ((__nothrow__));



extern wchar_t *wcpncpy (wchar_t *__restrict __dest,
    __const wchar_t *__restrict __src, size_t __n)
     __attribute__ ((__nothrow__));
# 578 "/usr/include/wchar.h" 3 4
extern __FILE *open_wmemstream (wchar_t **__bufloc, size_t *__sizeloc) __attribute__ ((__nothrow__));






extern int fwide (__FILE *__fp, int __mode) __attribute__ ((__nothrow__));






extern int fwprintf (__FILE *__restrict __stream,
       __const wchar_t *__restrict __format, ...)
                                                           ;




extern int wprintf (__const wchar_t *__restrict __format, ...)
                                                           ;

extern int swprintf (wchar_t *__restrict __s, size_t __n,
       __const wchar_t *__restrict __format, ...)
     __attribute__ ((__nothrow__)) ;





extern int vfwprintf (__FILE *__restrict __s,
        __const wchar_t *__restrict __format,
        __gnuc_va_list __arg)
                                                           ;




extern int vwprintf (__const wchar_t *__restrict __format,
       __gnuc_va_list __arg)
                                                           ;


extern int vswprintf (wchar_t *__restrict __s, size_t __n,
        __const wchar_t *__restrict __format,
        __gnuc_va_list __arg)
     __attribute__ ((__nothrow__)) ;






extern int fwscanf (__FILE *__restrict __stream,
      __const wchar_t *__restrict __format, ...)
                                                          ;




extern int wscanf (__const wchar_t *__restrict __format, ...)
                                                          ;

extern int swscanf (__const wchar_t *__restrict __s,
      __const wchar_t *__restrict __format, ...)
     __attribute__ ((__nothrow__)) ;
# 678 "/usr/include/wchar.h" 3 4









extern int vfwscanf (__FILE *__restrict __s,
       __const wchar_t *__restrict __format,
       __gnuc_va_list __arg)
                                                          ;




extern int vwscanf (__const wchar_t *__restrict __format,
      __gnuc_va_list __arg)
                                                          ;

extern int vswscanf (__const wchar_t *__restrict __s,
       __const wchar_t *__restrict __format,
       __gnuc_va_list __arg)
     __attribute__ ((__nothrow__)) ;
# 734 "/usr/include/wchar.h" 3 4









extern wint_t fgetwc (__FILE *__stream);
extern wint_t getwc (__FILE *__stream);





extern wint_t getwchar (void);






extern wint_t fputwc (wchar_t __wc, __FILE *__stream);
extern wint_t putwc (wchar_t __wc, __FILE *__stream);





extern wint_t putwchar (wchar_t __wc);







extern wchar_t *fgetws (wchar_t *__restrict __ws, int __n,
   __FILE *__restrict __stream);





extern int fputws (__const wchar_t *__restrict __ws,
     __FILE *__restrict __stream);






extern wint_t ungetwc (wint_t __wc, __FILE *__stream);

# 799 "/usr/include/wchar.h" 3 4
extern wint_t getwc_unlocked (__FILE *__stream);
extern wint_t getwchar_unlocked (void);







extern wint_t fgetwc_unlocked (__FILE *__stream);







extern wint_t fputwc_unlocked (wchar_t __wc, __FILE *__stream);
# 825 "/usr/include/wchar.h" 3 4
extern wint_t putwc_unlocked (wchar_t __wc, __FILE *__stream);
extern wint_t putwchar_unlocked (wchar_t __wc);
# 835 "/usr/include/wchar.h" 3 4
extern wchar_t *fgetws_unlocked (wchar_t *__restrict __ws, int __n,
     __FILE *__restrict __stream);







extern int fputws_unlocked (__const wchar_t *__restrict __ws,
       __FILE *__restrict __stream);







extern size_t wcsftime (wchar_t *__restrict __s, size_t __maxsize,
   __const wchar_t *__restrict __format,
   __const struct tm *__restrict __tp) __attribute__ ((__nothrow__));







extern size_t wcsftime_l (wchar_t *__restrict __s, size_t __maxsize,
     __const wchar_t *__restrict __format,
     __const struct tm *__restrict __tp,
     __locale_t __loc) __attribute__ ((__nothrow__));
# 889 "/usr/include/wchar.h" 3 4

# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h" 2


# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_filestat.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_filestat.h"
int zm_activate_filestat(int type, int module_number );
int zm_deactivate_filestat(int type, int module_number );

void zif_realpath_cache_size(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_realpath_cache_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_clearstatcache(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fileatime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_filectime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_filegroup(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fileinode(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_filemtime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fileowner(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fileperms(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_filesize(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_filetype(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_writable(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_readable(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_executable(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_dir(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_link(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_file_exists(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void php_if_stat(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void php_if_lstat(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_disk_total_space(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_disk_free_space(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_chown(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_chgrp(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_lchown(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_lchgrp(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_chmod(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_touch(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_clearstatcache(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
# 89 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_filestat.h"
typedef int php_stat_len;


__attribute__ ((visibility("default"))) void php_clear_stat_cache(zend_bool clear_realpath_cache, const char *filename, int filename_len );
__attribute__ ((visibility("default"))) void php_stat(const char *filename, php_stat_len filename_length, int type, zval *return_value );
# 32 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_highlight.h" 1
# 32 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_highlight.h"
typedef struct _zend_syntax_highlighter_ini {
 char *highlight_html;
 char *highlight_comment;
 char *highlight_default;
 char *highlight_string;
 char *highlight_keyword;
} zend_syntax_highlighter_ini;



__attribute__ ((visibility("default"))) void zend_highlight(zend_syntax_highlighter_ini *syntax_highlighter_ini );
__attribute__ ((visibility("default"))) void zend_strip(void);
__attribute__ ((visibility("default"))) int highlight_file(char *filename, zend_syntax_highlighter_ini *syntax_highlighter_ini );
__attribute__ ((visibility("default"))) int highlight_string(zval *str, zend_syntax_highlighter_ini *syntax_highlighter_ini, char *str_name );
__attribute__ ((visibility("default"))) void zend_html_putc(char c);
__attribute__ ((visibility("default"))) void zend_html_puts(const char *s, uint len );


extern zend_syntax_highlighter_ini syntax_highlighter_ini;
# 34 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h" 2

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/url_scanner_ex.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/url_scanner_ex.h"
int zm_startup_url_scanner_ex(int type, int module_number );
int zm_shutdown_url_scanner_ex(int type, int module_number );

int zm_activate_url_scanner_ex(int type, int module_number );
int zm_deactivate_url_scanner_ex(int type, int module_number );

__attribute__ ((visibility("default"))) char *php_url_scanner_adapt_single_url(const char *url, size_t urllen, const char *name, const char *value, size_t *newlen );
__attribute__ ((visibility("default"))) int php_url_scanner_add_var(char *name, int name_len, char *value, int value_len, int urlencode );
__attribute__ ((visibility("default"))) int php_url_scanner_reset_vars(void);

# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_smart_str_public.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_smart_str_public.h"
typedef struct {
 char *c;
 size_t len;
 size_t a;
} smart_str;
# 35 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/url_scanner_ex.h" 2

typedef struct {

 smart_str tag;
 smart_str arg;
 smart_str val;
 smart_str buf;


 smart_str result;


 smart_str form_app, url_app;

 int active;

 char *lookup_data;
 int state;


 HashTable *tags;
} url_adapt_state_ex_t;
# 36 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h" 2

extern zend_module_entry basic_functions_module;


int zm_startup_basic(int type, int module_number );
int zm_shutdown_basic(int type, int module_number );
int zm_activate_basic(int type, int module_number );
int zm_deactivate_basic(int type, int module_number );
void zm_info_basic(zend_module_entry *zend_module );

void zif_constant(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_sleep(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_usleep(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_time_nanosleep(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_time_sleep_until(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_flush(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void php_inet_ntop(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void php_inet_pton(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_ip2long(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_long2ip(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_getenv(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_putenv(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_getopt(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_get_current_user(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_set_time_limit(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_header_register_callback(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_get_cfg_var(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_set_magic_quotes_runtime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_get_magic_quotes_runtime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_get_magic_quotes_gpc(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_error_log(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_error_get_last(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_call_user_func(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_call_user_func_array(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_call_user_method(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_call_user_method_array(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_forward_static_call(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_forward_static_call_array(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_register_shutdown_function(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_highlight_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_highlight_string(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_php_strip_whitespace(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
__attribute__ ((visibility("default"))) void php_get_highlight_struct(zend_syntax_highlighter_ini *syntax_highlighter_ini);

void zif_ini_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ini_get_all(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ini_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ini_restore(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_get_include_path(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_set_include_path(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_restore_include_path(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_print_r(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fprintf(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_vfprintf(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_connection_aborted(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_connection_status(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ignore_user_abort(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_getservbyname(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_getservbyport(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_getprotobyname(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_getprotobynumber(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void php_if_crc32(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_register_tick_function(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_unregister_tick_function(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_sys_getloadavg(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_is_uploaded_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_move_uploaded_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_parse_ini_file(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_parse_ini_string(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );




void zif_str_rot13(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_get_filters(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_filter_register(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_bucket_make_writeable(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_bucket_prepend(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_bucket_append(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_stream_bucket_new(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
int zm_startup_user_filters(int type, int module_number );
int zm_deactivate_user_filters(int type, int module_number );
int zm_deactivate_browscap(int type, int module_number );


__attribute__ ((visibility("default"))) int _php_error_log(int opt_err, char *message, char *opt, char *headers );
__attribute__ ((visibility("default"))) int _php_error_log_ex(int opt_err, char *message, int message_len, char *opt, char *headers );
__attribute__ ((visibility("default"))) int php_prefix_varname(zval *result, zval *prefix, char *var_name, int var_name_len, zend_bool add_underscore );



typedef unsigned int php_uint32;
typedef signed int php_int32;
# 164 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/basic_functions.h"
typedef struct _php_basic_globals {
 HashTable *user_shutdown_function_names;
 HashTable putenv_ht;
 zval *strtok_zval;
 char *strtok_string;
 char *locale_string;
 char *strtok_last;
 char strtok_table[256];
 ulong strtok_len;
 char str_ebuf[40];
 zend_fcall_info array_walk_fci;
 zend_fcall_info_cache array_walk_fci_cache;
 zend_fcall_info user_compare_fci;
 zend_fcall_info_cache user_compare_fci_cache;
 zend_llist *user_tick_functions;

 zval *active_ini_file_section;


 long page_uid;
 long page_gid;
 long page_inode;
 time_t page_mtime;


 char *CurrentStatFile, *CurrentLStatFile;
 php_stream_statbuf ssb, lssb;


 php_uint32 state[(624)+1];
 php_uint32 *next;
 int left;

 unsigned int rand_seed;

 zend_bool rand_is_seeded;
 zend_bool mt_rand_is_seeded;


 char *syslog_device;


 zend_class_entry *incomplete_class;
 unsigned serialize_lock;
 struct {
  void *var_hash;
  unsigned level;
 } serialize;
 struct {
  void *var_hash;
  unsigned level;
 } unserialize;


 url_adapt_state_ex_t url_adapt_state_ex;


 void *mmap_file;
 size_t mmap_len;


 HashTable *user_filter_map;






 int umask;
} php_basic_globals;






__attribute__ ((visibility("default"))) extern php_basic_globals basic_globals;



typedef struct {
 char *putenv_string;
 char *previous_value;
 char *key;
 int key_len;
} putenv_entry;


__attribute__ ((visibility("default"))) double php_get_nan(void);
__attribute__ ((visibility("default"))) double php_get_inf(void);

typedef struct _php_shutdown_function_entry {
 zval **arguments;
 int arg_count;
} php_shutdown_function_entry;

__attribute__ ((visibility("default"))) extern zend_bool register_user_shutdown_function(char *function_name, size_t function_len, php_shutdown_function_entry *shutdown_function_entry );
__attribute__ ((visibility("default"))) extern zend_bool remove_user_shutdown_function(char *function_name, size_t function_len );
__attribute__ ((visibility("default"))) extern zend_bool append_user_shutdown_function(php_shutdown_function_entry shutdown_function_entry );
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_versioning.h" 2

__attribute__ ((visibility("default"))) char *php_canonicalize_version(const char *);
__attribute__ ((visibility("default"))) int php_version_compare(const char *, const char *);
void zif_version_compare(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
# 28 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_math.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_math.h"
__attribute__ ((visibility("default"))) char *_php_math_number_format(double, int, char, char);
__attribute__ ((visibility("default"))) char *_php_math_number_format_ex(double, int, char *, size_t, char *, size_t);
__attribute__ ((visibility("default"))) char * _php_math_longtobase(zval *arg, int base);
__attribute__ ((visibility("default"))) long _php_math_basetolong(zval *arg, int base);
__attribute__ ((visibility("default"))) int _php_math_basetozval(zval *arg, int base, zval *ret);
__attribute__ ((visibility("default"))) char * _php_math_zvaltobase(zval *arg, int base );

void zif_sin(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_cos(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_tan(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_asin(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_acos(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_atan(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_atan2(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_pi(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_exp(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_log(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_log10(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_finite(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_infinite(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_is_nan(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_pow(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_sqrt(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_srand(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_rand(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_getrandmax(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_mt_srand(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_mt_rand(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_mt_getrandmax(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_abs(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_ceil(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_floor(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_round(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_decbin(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_dechex(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_decoct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_bindec(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_hexdec(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_octdec(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_base_convert(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_number_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_fmod(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_deg2rad(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_rad2deg(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );





void zif_hypot(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_expm1(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_log1p(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_sinh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_cosh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_tanh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_asinh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_acosh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_atanh(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
# 29 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_structs.h" 1
# 24 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_structs.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_config.h" 1



# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/main/php_config.h" 1
# 5 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_config.h" 2
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_structs.h" 2






# 1 "/usr/include/inttypes.h" 1 3 4
# 28 "/usr/include/inttypes.h" 3 4
# 1 "/usr/include/stdint.h" 1 3 4
# 28 "/usr/include/stdint.h" 3 4
# 1 "/usr/include/bits/wordsize.h" 1 3 4
# 29 "/usr/include/stdint.h" 2 3 4
# 49 "/usr/include/stdint.h" 3 4
typedef unsigned char uint8_t;
typedef unsigned short int uint16_t;

typedef unsigned int uint32_t;





__extension__
typedef unsigned long long int uint64_t;






typedef signed char int_least8_t;
typedef short int int_least16_t;
typedef int int_least32_t;



__extension__
typedef long long int int_least64_t;



typedef unsigned char uint_least8_t;
typedef unsigned short int uint_least16_t;
typedef unsigned int uint_least32_t;



__extension__
typedef unsigned long long int uint_least64_t;






typedef signed char int_fast8_t;





typedef int int_fast16_t;
typedef int int_fast32_t;
__extension__
typedef long long int int_fast64_t;



typedef unsigned char uint_fast8_t;





typedef unsigned int uint_fast16_t;
typedef unsigned int uint_fast32_t;
__extension__
typedef unsigned long long int uint_fast64_t;
# 129 "/usr/include/stdint.h" 3 4
typedef unsigned int uintptr_t;
# 138 "/usr/include/stdint.h" 3 4
__extension__
typedef long long int intmax_t;
__extension__
typedef unsigned long long int uintmax_t;
# 29 "/usr/include/inttypes.h" 2 3 4






typedef long int __gwchar_t;
# 274 "/usr/include/inttypes.h" 3 4

# 288 "/usr/include/inttypes.h" 3 4
typedef struct
  {
    long long int quot;
    long long int rem;
  } imaxdiv_t;





extern intmax_t imaxabs (intmax_t __n) __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern imaxdiv_t imaxdiv (intmax_t __numer, intmax_t __denom)
      __attribute__ ((__nothrow__)) __attribute__ ((__const__));


extern intmax_t strtoimax (__const char *__restrict __nptr,
      char **__restrict __endptr, int __base) __attribute__ ((__nothrow__));


extern uintmax_t strtoumax (__const char *__restrict __nptr,
       char ** __restrict __endptr, int __base) __attribute__ ((__nothrow__));


extern intmax_t wcstoimax (__const __gwchar_t *__restrict __nptr,
      __gwchar_t **__restrict __endptr, int __base)
     __attribute__ ((__nothrow__));


extern uintmax_t wcstoumax (__const __gwchar_t *__restrict __nptr,
       __gwchar_t ** __restrict __endptr, int __base)
     __attribute__ ((__nothrow__));
# 379 "/usr/include/inttypes.h" 3 4
__extension__
extern long long int __strtoll_internal (__const char *__restrict __nptr,
      char **__restrict __endptr,
      int __base, int __group)
  __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern __inline intmax_t
__attribute__ ((__nothrow__)) strtoimax (__const char *__restrict nptr, char **__restrict endptr, int base)

{
  return __strtoll_internal (nptr, endptr, base, 0);
}

__extension__
extern unsigned long long int __strtoull_internal (__const char *
         __restrict __nptr,
         char **
         __restrict __endptr,
         int __base,
         int __group)
  __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern __inline uintmax_t
__attribute__ ((__nothrow__)) strtoumax (__const char *__restrict nptr, char **__restrict endptr, int base)

{
  return __strtoull_internal (nptr, endptr, base, 0);
}

__extension__
extern long long int __wcstoll_internal (__const __gwchar_t *
      __restrict __nptr,
      __gwchar_t **__restrict __endptr,
      int __base, int __group)
  __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern __inline intmax_t
__attribute__ ((__nothrow__)) wcstoimax (__const __gwchar_t *__restrict nptr, __gwchar_t **__restrict endptr, int base)

{
  return __wcstoll_internal (nptr, endptr, base, 0);
}


__extension__
extern unsigned long long int __wcstoull_internal (__const __gwchar_t *
         __restrict __nptr,
         __gwchar_t **
         __restrict __endptr,
         int __base,
         int __group)
  __attribute__ ((__nothrow__)) __attribute__ ((__nonnull__ (1))) ;

extern __inline uintmax_t
__attribute__ ((__nothrow__)) wcstoumax (__const __gwchar_t *__restrict nptr, __gwchar_t **__restrict endptr, int base)

{
  return __wcstoull_internal (nptr, endptr, base, 0);
}





# 32 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_structs.h" 2
# 75 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib_structs.h"
typedef unsigned long long timelib_ull;
typedef signed long long timelib_sll;



typedef struct ttinfo
{
 int32_t offset;
 int isdst;
 unsigned int abbr_idx;

 unsigned int isstdcnt;
 unsigned int isgmtcnt;
} ttinfo;

typedef struct tlinfo
{
 int32_t trans;
 int32_t offset;
} tlinfo;

typedef struct tlocinfo
{
 char country_code[3];
 double latitude;
 double longitude;
 char *comments;
} tlocinfo;

typedef struct timelib_tzinfo
{
 char *name;
 uint32_t ttisgmtcnt;
 uint32_t ttisstdcnt;
 uint32_t leapcnt;
 uint32_t timecnt;
 uint32_t typecnt;
 uint32_t charcnt;

 int32_t *trans;
 unsigned char *trans_idx;

 ttinfo *type;
 char *timezone_abbr;

 tlinfo *leap_times;
 unsigned char bc;
 tlocinfo location;
} timelib_tzinfo;

typedef struct timelib_special {
 unsigned int type;
 timelib_sll amount;
} timelib_special;

typedef struct timelib_rel_time {
 timelib_sll y, m, d;
 timelib_sll h, i, s;

 int weekday;
 int weekday_behavior;

 int first_last_day_of;
 int invert;
 timelib_sll days;

 timelib_special special;
 unsigned int have_weekday_relative, have_special_relative;
} timelib_rel_time;

typedef struct timelib_time_offset {
 int32_t offset;
 unsigned int leap_secs;
 unsigned int is_dst;
 char *abbr;
 timelib_sll transistion_time;
} timelib_time_offset;

typedef struct timelib_time {
 timelib_sll y, m, d;
 timelib_sll h, i, s;
 double f;
 int z;
 char *tz_abbr;
 timelib_tzinfo *tz_info;
 signed int dst;
 timelib_rel_time relative;

 timelib_sll sse;

 unsigned int have_time, have_date, have_zone, have_relative, have_weeknr_day;

 unsigned int sse_uptodate;
 unsigned int tim_uptodate;
 unsigned int is_localtime;
 unsigned int zone_type;


} timelib_time;

typedef struct timelib_error_message {
 int position;
 char character;
 char *message;
} timelib_error_message;

typedef struct timelib_error_container {
 int warning_count;
 struct timelib_error_message *warning_messages;
 int error_count;
 struct timelib_error_message *error_messages;
} timelib_error_container;

typedef struct _timelib_tz_lookup_table {
 char *name;
 int type;
 float gmtoffset;
 char *full_tz_name;
} timelib_tz_lookup_table;

typedef struct _timelib_tzdb_index_entry {
 char *id;
 unsigned int pos;
} timelib_tzdb_index_entry;

typedef struct _timelib_tzdb {
 char *version;
 int index_size;
 const timelib_tzdb_index_entry *index;
 const unsigned char *data;
} timelib_tzdb;
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib.h" 2

# 1 "/usr/lib/gcc/i686-redhat-linux/4.4.5/include/limits.h" 1 3 4
# 27 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib.h" 2
# 58 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/lib/timelib.h"
typedef timelib_tzinfo* (*timelib_tz_get_wrapper)(char *tzname, const timelib_tzdb *tzdb);


timelib_sll timelib_day_of_week(timelib_sll y, timelib_sll m, timelib_sll d);
timelib_sll timelib_iso_day_of_week(timelib_sll y, timelib_sll m, timelib_sll d);
timelib_sll timelib_day_of_year(timelib_sll y, timelib_sll m, timelib_sll d);
timelib_sll timelib_daynr_from_weeknr(timelib_sll y, timelib_sll w, timelib_sll d);
timelib_sll timelib_days_in_month(timelib_sll y, timelib_sll m);
void timelib_isoweek_from_date(timelib_sll y, timelib_sll m, timelib_sll d, timelib_sll *iw, timelib_sll *iy);
int timelib_valid_time(timelib_sll h, timelib_sll i, timelib_sll s);
int timelib_valid_date(timelib_sll y, timelib_sll m, timelib_sll d);


timelib_time *timelib_strtotime(char *s, int len, timelib_error_container **errors, const timelib_tzdb *tzdb, timelib_tz_get_wrapper tz_get_wrapper);
timelib_time *timelib_parse_from_format(char *format, char *s, int len, timelib_error_container **errors, const timelib_tzdb *tzdb, timelib_tz_get_wrapper tz_get_wrapper);
void timelib_fill_holes(timelib_time *parsed, timelib_time *now, int options);
char *timelib_timezone_id_from_abbr(const char *abbr, long gmtoffset, int isdst);
const timelib_tz_lookup_table *timelib_timezone_abbreviations_list(void);


void timelib_strtointerval(char *s, int len,
                           timelib_time **begin, timelib_time **end,
         timelib_rel_time **period, int *recurrences,
         struct timelib_error_container **errors);



void timelib_update_ts(timelib_time* time, timelib_tzinfo* tzi);
void timelib_do_normalize(timelib_time *base);
void timelib_do_rel_normalize(timelib_time *base, timelib_rel_time *rt);


int timelib_apply_localtime(timelib_time *t, unsigned int localtime);
void timelib_unixtime2gmt(timelib_time* tm, timelib_sll ts);
void timelib_unixtime2local(timelib_time *tm, timelib_sll ts);
void timelib_update_from_sse(timelib_time *tm);
void timelib_set_timezone(timelib_time *t, timelib_tzinfo *tz);


int timelib_timezone_id_is_valid(char *timezone, const timelib_tzdb *tzdb);
timelib_tzinfo *timelib_parse_tzfile(char *timezone, const timelib_tzdb *tzdb);
int timelib_timestamp_is_in_dst(timelib_sll ts, timelib_tzinfo *tz);
timelib_time_offset *timelib_get_time_zone_info(timelib_sll ts, timelib_tzinfo *tz);
timelib_sll timelib_get_current_offset(timelib_time *t);
void timelib_dump_tzinfo(timelib_tzinfo *tz);
const timelib_tzdb *timelib_builtin_db(void);
const timelib_tzdb_index_entry *timelib_timezone_builtin_identifiers_list(int *count);


timelib_tzinfo* timelib_tzinfo_ctor(char *name);
void timelib_time_tz_abbr_update(timelib_time* tm, char* tz_abbr);
void timelib_time_tz_name_update(timelib_time* tm, char* tz_name);
void timelib_tzinfo_dtor(timelib_tzinfo *tz);
timelib_tzinfo* timelib_tzinfo_clone(timelib_tzinfo *tz);

timelib_rel_time* timelib_rel_time_ctor(void);
void timelib_rel_time_dtor(timelib_rel_time* t);
timelib_rel_time* timelib_rel_time_clone(timelib_rel_time *tz);

timelib_time* timelib_time_ctor(void);
void timelib_time_set_option(timelib_time* tm, int option, void* option_value);
void timelib_time_dtor(timelib_time* t);
timelib_time* timelib_time_clone(timelib_time* orig);

timelib_time_offset* timelib_time_offset_ctor(void);
void timelib_time_offset_dtor(timelib_time_offset* t);

void timelib_error_container_dtor(timelib_error_container *errors);

signed long timelib_date_to_int(timelib_time *d, int *error);
void timelib_dump_date(timelib_time *d, int options);
void timelib_dump_rel_time(timelib_rel_time *d);

void timelib_decimal_hour_to_hms(double h, int *hour, int *min, int *sec);


double timelib_ts_to_juliandate(timelib_sll ts);
int timelib_astro_rise_set_altitude(timelib_time *time, double lon, double lat, double altit, int upper_limb, double *h_rise, double *h_set, timelib_sll *ts_rise, timelib_sll *ts_set, timelib_sll *ts_transit);


timelib_rel_time *timelib_diff(timelib_time *one, timelib_time *two);
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.h" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_hash.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.h" 2

extern zend_module_entry date_module_entry;


void zif_date(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_idate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_gmdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_strtotime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_mktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_gmmktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_checkdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_strftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_gmstrftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_time(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_localtime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_getdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zim_DateTime___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zim_DateTime___wakeup(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zim_DateTime___set_state(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_create(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_create_from_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_parse(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_parse_from_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_get_last_errors(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_modify(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_add(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_sub(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_timezone_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_timezone_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_offset_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_diff(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zif_date_time_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_date_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_isodate_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_timestamp_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_timestamp_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zim_DateTimeZone___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_open(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_name_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_name_from_abbr(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_offset_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_transitions_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_location_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_identifiers_list(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_abbreviations_list(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_timezone_version_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zim_DateInterval___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zim_DateInterval___wakeup(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zim_DateInterval___set_state(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_interval_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_interval_create_from_date_string(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

void zim_DatePeriod___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_date_default_timezone_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_default_timezone_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );


void zif_date_sunrise(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_sunset(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );
void zif_date_sun_info(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used );

int zm_activate_date(int type, int module_number );
int zm_deactivate_date(int type, int module_number );
int zm_startup_date(int type, int module_number );
int zm_shutdown_date(int type, int module_number );
void zm_info_date(zend_module_entry *zend_module );

typedef struct _php_date_obj php_date_obj;
typedef struct _php_timezone_obj php_timezone_obj;
typedef struct _php_interval_obj php_interval_obj;
typedef struct _php_period_obj php_period_obj;

struct _php_date_obj {
 zend_object std;
 timelib_time *time;
 HashTable *props;
};

struct _php_timezone_obj {
 zend_object std;
 int initialized;
 int type;
 union {
  timelib_tzinfo *tz;
  timelib_sll utc_offset;
  struct
  {
   timelib_sll utc_offset;
   char *abbr;
   int dst;
  } z;
 } tzi;
};

struct _php_interval_obj {
 zend_object std;
 timelib_rel_time *diff;
 HashTable *props;
 int initialized;
};

struct _php_period_obj {
 zend_object std;
 timelib_time *start;
 timelib_time *current;
 timelib_time *end;
 timelib_rel_time *interval;
 int recurrences;
 int initialized;
 int include_start_date;
};

typedef struct _zend_date_globals {
 char *default_timezone;
 char *timezone;
 HashTable *tzcache;
 timelib_error_container *last_errors;
} zend_date_globals;
# 166 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.h"
__attribute__ ((visibility("default"))) signed long php_parse_date(char *string, signed long *now);
__attribute__ ((visibility("default"))) void php_mktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int gmt);
__attribute__ ((visibility("default"))) int php_idate(char format, time_t ts, int localtime );


__attribute__ ((visibility("default"))) void php_strftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int gm);

__attribute__ ((visibility("default"))) char *php_format_date(char *format, int format_len, time_t ts, int localtime );


__attribute__ ((visibility("default"))) void php_date_set_tzdb(timelib_tzdb *tzdb);
__attribute__ ((visibility("default"))) timelib_tzinfo *get_timezone_info(void);


__attribute__ ((visibility("default"))) zend_class_entry *php_date_get_date_ce(void);
__attribute__ ((visibility("default"))) zend_class_entry *php_date_get_timezone_ce(void);


__attribute__ ((visibility("default"))) zval *php_date_instantiate(zend_class_entry *pce, zval *object );
__attribute__ ((visibility("default"))) int php_date_initialize(php_date_obj *dateobj, char *time_str, int time_str_len, char *format, zval *timezone_object, int ctor );
# 30 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_interfaces.h" 1
# 25 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_interfaces.h"
# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_API.h" 1
# 26 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_interfaces.h" 2



extern __attribute__ ((visibility("default"))) zend_class_entry *zend_ce_traversable;
extern __attribute__ ((visibility("default"))) zend_class_entry *zend_ce_aggregate;
extern __attribute__ ((visibility("default"))) zend_class_entry *zend_ce_iterator;
extern __attribute__ ((visibility("default"))) zend_class_entry *zend_ce_arrayaccess;
extern __attribute__ ((visibility("default"))) zend_class_entry *zend_ce_serializable;

typedef struct _zend_user_iterator {
 zend_object_iterator it;
 zend_class_entry *ce;
 zval *value;
} zend_user_iterator;

__attribute__ ((visibility("default"))) zval* zend_call_method(zval **object_pp, zend_class_entry *obj_ce, zend_function **fn_proxy, const char *function_name, int function_name_len, zval **retval_ptr_ptr, int param_count, zval* arg1, zval* arg2 );
# 52 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/Zend/zend_interfaces.h"
__attribute__ ((visibility("default"))) void zend_user_it_rewind(zend_object_iterator *_iter );
__attribute__ ((visibility("default"))) int zend_user_it_valid(zend_object_iterator *_iter );
__attribute__ ((visibility("default"))) int zend_user_it_get_current_key(zend_object_iterator *_iter, char **str_key, uint *str_key_len, ulong *int_key );
__attribute__ ((visibility("default"))) void zend_user_it_get_current_data(zend_object_iterator *_iter, zval ***data );
__attribute__ ((visibility("default"))) void zend_user_it_move_forward(zend_object_iterator *_iter );
__attribute__ ((visibility("default"))) void zend_user_it_invalidate_current(zend_object_iterator *_iter );

__attribute__ ((visibility("default"))) zval *zend_user_it_new_iterator(zend_class_entry *ce, zval *object );
__attribute__ ((visibility("default"))) zend_object_iterator *zend_user_it_get_new_iterator(zend_class_entry *ce, zval *object, int by_ref );

__attribute__ ((visibility("default"))) void zend_register_interfaces(void);

__attribute__ ((visibility("default"))) int zend_user_serialize(zval *object, unsigned char **buffer, zend_uint *buf_len, zend_serialize_data *data );
__attribute__ ((visibility("default"))) int zend_user_unserialize(zval **object, zend_class_entry *ce, const unsigned char *buf, zend_uint buf_len, zend_unserialize_data *data );

__attribute__ ((visibility("default"))) int zend_class_serialize_deny(zval *object, unsigned char **buffer, zend_uint *buf_len, zend_serialize_data *data );
__attribute__ ((visibility("default"))) int zend_class_unserialize_deny(zval **object, zend_class_entry *ce, const unsigned char *buf, zend_uint buf_len, zend_unserialize_data *data );


# 31 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2
# 39 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
static inline long long php_date_llabs( long long i ) { return i >= 0 ? i : -i; }



static const zend_arg_info arginfo_date[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_gmdate[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_idate[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_strtotime[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "now", sizeof("now")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_mktime[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "hour", sizeof("hour")-1, ((void *)0), 0, 0, 0, 0},
 { "min", sizeof("min")-1, ((void *)0), 0, 0, 0, 0},
 { "sec", sizeof("sec")-1, ((void *)0), 0, 0, 0, 0},
 { "mon", sizeof("mon")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_gmmktime[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "hour", sizeof("hour")-1, ((void *)0), 0, 0, 0, 0},
 { "min", sizeof("min")-1, ((void *)0), 0, 0, 0, 0},
 { "sec", sizeof("sec")-1, ((void *)0), 0, 0, 0, 0},
 { "mon", sizeof("mon")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_checkdate[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
 { "month", sizeof("month")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_strftime[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_gmstrftime[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_time[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_localtime[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
 { "associative_array", sizeof("associative_array")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_getdate[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "timestamp", sizeof("timestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_default_timezone_set[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
 { "timezone_identifier", sizeof("timezone_identifier")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_default_timezone_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_date_sunrise[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "latitude", sizeof("latitude")-1, ((void *)0), 0, 0, 0, 0},
 { "longitude", sizeof("longitude")-1, ((void *)0), 0, 0, 0, 0},
 { "zenith", sizeof("zenith")-1, ((void *)0), 0, 0, 0, 0},
 { "gmt_offset", sizeof("gmt_offset")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_sunset[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "latitude", sizeof("latitude")-1, ((void *)0), 0, 0, 0, 0},
 { "longitude", sizeof("longitude")-1, ((void *)0), 0, 0, 0, 0},
 { "zenith", sizeof("zenith")-1, ((void *)0), 0, 0, 0, 0},
 { "gmt_offset", sizeof("gmt_offset")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_sun_info[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "latitude", sizeof("latitude")-1, ((void *)0), 0, 0, 0, 0},
 { "longitude", sizeof("longitude")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_create[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_create_from_format[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_parse[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "date", sizeof("date")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_parse_from_format[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
 { "date", sizeof("date")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_get_last_errors[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_date_format[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_format[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_modify[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "modify", sizeof("modify")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_modify[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "modify", sizeof("modify")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_add[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "interval", sizeof("interval")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_add[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "interval", sizeof("interval")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_sub[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "interval", sizeof("interval")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_sub[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "interval", sizeof("interval")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_timezone_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_timezone_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_date_timezone_set[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "timezone", sizeof("timezone")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_timezone_set[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "timezone", sizeof("timezone")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_offset_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_offset_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_date_diff[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "object2", sizeof("object2")-1, ((void *)0), 0, 0, 0, 0},
 { "absolute", sizeof("absolute")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_diff[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "absolute", sizeof("absolute")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_time_set[] = { { ((void *)0), 0, ((void *)0), 3, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "hour", sizeof("hour")-1, ((void *)0), 0, 0, 0, 0},
 { "minute", sizeof("minute")-1, ((void *)0), 0, 0, 0, 0},
 { "second", sizeof("second")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_time_set[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "hour", sizeof("hour")-1, ((void *)0), 0, 0, 0, 0},
 { "minute", sizeof("minute")-1, ((void *)0), 0, 0, 0, 0},
 { "second", sizeof("second")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_date_set[] = { { ((void *)0), 0, ((void *)0), 4, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
 { "month", sizeof("month")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_date_set[] = { { ((void *)0), 0, ((void *)0), 3, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
 { "month", sizeof("month")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_isodate_set[] = { { ((void *)0), 0, ((void *)0), 3, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
 { "week", sizeof("week")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_isodate_set[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "year", sizeof("year")-1, ((void *)0), 0, 0, 0, 0},
 { "week", sizeof("week")-1, ((void *)0), 0, 0, 0, 0},
 { "day", sizeof("day")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_timestamp_set[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "unixtimestamp", sizeof("unixtimestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_timestamp_set[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "unixtimestamp", sizeof("unixtimestamp")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_timestamp_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_timestamp_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_open[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "timezone", sizeof("timezone")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_name_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_method_name_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_name_from_abbr[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "abbr", sizeof("abbr")-1, ((void *)0), 0, 0, 0, 0},
 { "gmtoffset", sizeof("gmtoffset")-1, ((void *)0), 0, 0, 0, 0},
 { "isdst", sizeof("isdst")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_offset_get[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "datetime", sizeof("datetime")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_method_offset_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "datetime", sizeof("datetime")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_transitions_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp_begin", sizeof("timestamp_begin")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp_end", sizeof("timestamp_end")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_method_transitions_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
 { "timestamp_begin", sizeof("timestamp_begin")-1, ((void *)0), 0, 0, 0, 0},
 { "timestamp_end", sizeof("timestamp_end")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_location_get[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_method_location_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_identifiers_list[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "what", sizeof("what")-1, ((void *)0), 0, 0, 0, 0},
 { "country", sizeof("country")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_abbreviations_list[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_timezone_version_get[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
};

static const zend_arg_info arginfo_date_interval_create_from_date_string[] = { { ((void *)0), 0, ((void *)0), 1, 0, 0, 0},
 { "time", sizeof("time")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_interval_format[] = { { ((void *)0), 0, ((void *)0), 2, 0, 0, 0},
 { "object", sizeof("object")-1, ((void *)0), 0, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_method_interval_format[] = { { ((void *)0), 0, ((void *)0), -1, 0, 0, 0},
 { "format", sizeof("format")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_period_construct[] = { { ((void *)0), 0, ((void *)0), 3, 0, 0, 0},
 { "start", sizeof("start")-1, ((void *)0), 0, 0, 0, 0},
 { "interval", sizeof("interval")-1, ((void *)0), 0, 0, 0, 0},
 { "end", sizeof("end")-1, ((void *)0), 0, 0, 0, 0},
};

static const zend_arg_info arginfo_date_interval_construct[] = { { ((void *)0), 0, ((void *)0), 0, 0, 0, 0},
 { "interval_spec", sizeof("interval_spec")-1, ((void *)0), 0, 0, 0, 0},
};



const zend_function_entry date_functions[] = {
 { "strtotime", zif_strtotime, arginfo_strtotime, (zend_uint) (sizeof(arginfo_strtotime)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date", zif_date, arginfo_date, (zend_uint) (sizeof(arginfo_date)/sizeof(struct _zend_arg_info)-1), 0 },
 { "idate", zif_idate, arginfo_idate, (zend_uint) (sizeof(arginfo_idate)/sizeof(struct _zend_arg_info)-1), 0 },
 { "gmdate", zif_gmdate, arginfo_gmdate, (zend_uint) (sizeof(arginfo_gmdate)/sizeof(struct _zend_arg_info)-1), 0 },
 { "mktime", zif_mktime, arginfo_mktime, (zend_uint) (sizeof(arginfo_mktime)/sizeof(struct _zend_arg_info)-1), 0 },
 { "gmmktime", zif_gmmktime, arginfo_gmmktime, (zend_uint) (sizeof(arginfo_gmmktime)/sizeof(struct _zend_arg_info)-1), 0 },
 { "checkdate", zif_checkdate, arginfo_checkdate, (zend_uint) (sizeof(arginfo_checkdate)/sizeof(struct _zend_arg_info)-1), 0 },


 { "strftime", zif_strftime, arginfo_strftime, (zend_uint) (sizeof(arginfo_strftime)/sizeof(struct _zend_arg_info)-1), 0 },
 { "gmstrftime", zif_gmstrftime, arginfo_gmstrftime, (zend_uint) (sizeof(arginfo_gmstrftime)/sizeof(struct _zend_arg_info)-1), 0 },


 { "time", zif_time, arginfo_time, (zend_uint) (sizeof(arginfo_time)/sizeof(struct _zend_arg_info)-1), 0 },
 { "localtime", zif_localtime, arginfo_localtime, (zend_uint) (sizeof(arginfo_localtime)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getdate", zif_getdate, arginfo_getdate, (zend_uint) (sizeof(arginfo_getdate)/sizeof(struct _zend_arg_info)-1), 0 },


 { "date_create", zif_date_create, arginfo_date_create, (zend_uint) (sizeof(arginfo_date_create)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_create_from_format", zif_date_create_from_format, arginfo_date_create_from_format, (zend_uint) (sizeof(arginfo_date_create_from_format)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_parse", zif_date_parse, arginfo_date_parse, (zend_uint) (sizeof(arginfo_date_parse)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_parse_from_format", zif_date_parse_from_format, arginfo_date_parse_from_format, (zend_uint) (sizeof(arginfo_date_parse_from_format)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_get_last_errors", zif_date_get_last_errors, arginfo_date_get_last_errors, (zend_uint) (sizeof(arginfo_date_get_last_errors)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_format", zif_date_format, arginfo_date_format, (zend_uint) (sizeof(arginfo_date_format)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_modify", zif_date_modify, arginfo_date_modify, (zend_uint) (sizeof(arginfo_date_modify)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_add", zif_date_add, arginfo_date_add, (zend_uint) (sizeof(arginfo_date_add)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_sub", zif_date_sub, arginfo_date_sub, (zend_uint) (sizeof(arginfo_date_sub)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_timezone_get", zif_date_timezone_get, arginfo_date_timezone_get, (zend_uint) (sizeof(arginfo_date_timezone_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_timezone_set", zif_date_timezone_set, arginfo_date_timezone_set, (zend_uint) (sizeof(arginfo_date_timezone_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_offset_get", zif_date_offset_get, arginfo_date_offset_get, (zend_uint) (sizeof(arginfo_date_offset_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_diff", zif_date_diff, arginfo_date_diff, (zend_uint) (sizeof(arginfo_date_diff)/sizeof(struct _zend_arg_info)-1), 0 },

 { "date_time_set", zif_date_time_set, arginfo_date_time_set, (zend_uint) (sizeof(arginfo_date_time_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_date_set", zif_date_date_set, arginfo_date_date_set, (zend_uint) (sizeof(arginfo_date_date_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_isodate_set", zif_date_isodate_set, arginfo_date_isodate_set, (zend_uint) (sizeof(arginfo_date_isodate_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_timestamp_set", zif_date_timestamp_set, arginfo_date_timestamp_set, (zend_uint) (sizeof(arginfo_date_timestamp_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_timestamp_get", zif_date_timestamp_get, arginfo_date_timestamp_get, (zend_uint) (sizeof(arginfo_date_timestamp_get)/sizeof(struct _zend_arg_info)-1), 0 },

 { "timezone_open", zif_timezone_open, arginfo_timezone_open, (zend_uint) (sizeof(arginfo_timezone_open)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_name_get", zif_timezone_name_get, arginfo_timezone_name_get, (zend_uint) (sizeof(arginfo_timezone_name_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_name_from_abbr", zif_timezone_name_from_abbr, arginfo_timezone_name_from_abbr, (zend_uint) (sizeof(arginfo_timezone_name_from_abbr)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_offset_get", zif_timezone_offset_get, arginfo_timezone_offset_get, (zend_uint) (sizeof(arginfo_timezone_offset_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_transitions_get", zif_timezone_transitions_get, arginfo_timezone_transitions_get, (zend_uint) (sizeof(arginfo_timezone_transitions_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_location_get", zif_timezone_location_get, arginfo_timezone_location_get, (zend_uint) (sizeof(arginfo_timezone_location_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_identifiers_list", zif_timezone_identifiers_list, arginfo_timezone_identifiers_list, (zend_uint) (sizeof(arginfo_timezone_identifiers_list)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_abbreviations_list", zif_timezone_abbreviations_list, arginfo_timezone_abbreviations_list, (zend_uint) (sizeof(arginfo_timezone_abbreviations_list)/sizeof(struct _zend_arg_info)-1), 0 },
 { "timezone_version_get", zif_timezone_version_get, arginfo_timezone_version_get, (zend_uint) (sizeof(arginfo_timezone_version_get)/sizeof(struct _zend_arg_info)-1), 0 },

 { "date_interval_create_from_date_string", zif_date_interval_create_from_date_string, arginfo_date_interval_create_from_date_string, (zend_uint) (sizeof(arginfo_date_interval_create_from_date_string)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_interval_format", zif_date_interval_format, arginfo_date_interval_format, (zend_uint) (sizeof(arginfo_date_interval_format)/sizeof(struct _zend_arg_info)-1), 0 },


 { "date_default_timezone_set", zif_date_default_timezone_set, arginfo_date_default_timezone_set, (zend_uint) (sizeof(arginfo_date_default_timezone_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_default_timezone_get", zif_date_default_timezone_get, arginfo_date_default_timezone_get, (zend_uint) (sizeof(arginfo_date_default_timezone_get)/sizeof(struct _zend_arg_info)-1), 0 },


 { "date_sunrise", zif_date_sunrise, arginfo_date_sunrise, (zend_uint) (sizeof(arginfo_date_sunrise)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_sunset", zif_date_sunset, arginfo_date_sunset, (zend_uint) (sizeof(arginfo_date_sunset)/sizeof(struct _zend_arg_info)-1), 0 },
 { "date_sun_info", zif_date_sun_info, arginfo_date_sun_info, (zend_uint) (sizeof(arginfo_date_sun_info)/sizeof(struct _zend_arg_info)-1), 0 },
 { ((void *)0), ((void *)0), ((void *)0), 0, 0 }
};

const zend_function_entry date_funcs_date[] = {
 { "__construct", zim_DateTime___construct, arginfo_date_create, (zend_uint) (sizeof(arginfo_date_create)/sizeof(struct _zend_arg_info)-1), 0x2000|0x100 },
 { "__wakeup", zim_DateTime___wakeup, ((void *)0), (zend_uint) (sizeof(((void *)0))/sizeof(struct _zend_arg_info)-1), 0x100 },
 { "__set_state", zim_DateTime___set_state, ((void *)0), (zend_uint) (sizeof(((void *)0))/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { "createFromFormat", zif_date_create_from_format, arginfo_date_create_from_format, (zend_uint) (sizeof(arginfo_date_create_from_format)/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { "getLastErrors", zif_date_get_last_errors, arginfo_date_get_last_errors, (zend_uint) (sizeof(arginfo_date_get_last_errors)/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { "format", zif_date_format, arginfo_date_method_format, (zend_uint) (sizeof(arginfo_date_method_format)/sizeof(struct _zend_arg_info)-1), 0 },
 { "modify", zif_date_modify, arginfo_date_method_modify, (zend_uint) (sizeof(arginfo_date_method_modify)/sizeof(struct _zend_arg_info)-1), 0 },
 { "add", zif_date_add, arginfo_date_method_add, (zend_uint) (sizeof(arginfo_date_method_add)/sizeof(struct _zend_arg_info)-1), 0 },
 { "sub", zif_date_sub, arginfo_date_method_sub, (zend_uint) (sizeof(arginfo_date_method_sub)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getTimezone", zif_date_timezone_get, arginfo_date_method_timezone_get, (zend_uint) (sizeof(arginfo_date_method_timezone_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "setTimezone", zif_date_timezone_set, arginfo_date_method_timezone_set, (zend_uint) (sizeof(arginfo_date_method_timezone_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getOffset", zif_date_offset_get, arginfo_date_method_offset_get, (zend_uint) (sizeof(arginfo_date_method_offset_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "setTime", zif_date_time_set, arginfo_date_method_time_set, (zend_uint) (sizeof(arginfo_date_method_time_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "setDate", zif_date_date_set, arginfo_date_method_date_set, (zend_uint) (sizeof(arginfo_date_method_date_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "setISODate", zif_date_isodate_set, arginfo_date_method_isodate_set, (zend_uint) (sizeof(arginfo_date_method_isodate_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "setTimestamp", zif_date_timestamp_set, arginfo_date_method_timestamp_set, (zend_uint) (sizeof(arginfo_date_method_timestamp_set)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getTimestamp", zif_date_timestamp_get, arginfo_date_method_timestamp_get, (zend_uint) (sizeof(arginfo_date_method_timestamp_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "diff", zif_date_diff, arginfo_date_method_diff, (zend_uint) (sizeof(arginfo_date_method_diff)/sizeof(struct _zend_arg_info)-1), 0 },
 { ((void *)0), ((void *)0), ((void *)0), 0, 0 }
};

const zend_function_entry date_funcs_timezone[] = {
 { "__construct", zim_DateTimeZone___construct, arginfo_timezone_open, (zend_uint) (sizeof(arginfo_timezone_open)/sizeof(struct _zend_arg_info)-1), 0x2000|0x100 },
 { "getName", zif_timezone_name_get, arginfo_timezone_method_name_get, (zend_uint) (sizeof(arginfo_timezone_method_name_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getOffset", zif_timezone_offset_get, arginfo_timezone_method_offset_get, (zend_uint) (sizeof(arginfo_timezone_method_offset_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getTransitions", zif_timezone_transitions_get, arginfo_timezone_method_transitions_get, (zend_uint) (sizeof(arginfo_timezone_method_transitions_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "getLocation", zif_timezone_location_get, arginfo_timezone_method_location_get, (zend_uint) (sizeof(arginfo_timezone_method_location_get)/sizeof(struct _zend_arg_info)-1), 0 },
 { "listAbbreviations", zif_timezone_abbreviations_list, arginfo_timezone_abbreviations_list, (zend_uint) (sizeof(arginfo_timezone_abbreviations_list)/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { "listIdentifiers", zif_timezone_identifiers_list, arginfo_timezone_identifiers_list, (zend_uint) (sizeof(arginfo_timezone_identifiers_list)/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { ((void *)0), ((void *)0), ((void *)0), 0, 0 }
};

const zend_function_entry date_funcs_interval[] = {
 { "__construct", zim_DateInterval___construct, arginfo_date_interval_construct, (zend_uint) (sizeof(arginfo_date_interval_construct)/sizeof(struct _zend_arg_info)-1), 0x2000|0x100 },
 { "__wakeup", zim_DateInterval___wakeup, ((void *)0), (zend_uint) (sizeof(((void *)0))/sizeof(struct _zend_arg_info)-1), 0x100 },
 { "__set_state", zim_DateInterval___set_state, ((void *)0), (zend_uint) (sizeof(((void *)0))/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { "format", zif_date_interval_format, arginfo_date_method_interval_format, (zend_uint) (sizeof(arginfo_date_method_interval_format)/sizeof(struct _zend_arg_info)-1), 0 },
 { "createFromDateString", zif_date_interval_create_from_date_string, arginfo_date_interval_create_from_date_string, (zend_uint) (sizeof(arginfo_date_interval_create_from_date_string)/sizeof(struct _zend_arg_info)-1), 0x100|0x01 },
 { ((void *)0), ((void *)0), ((void *)0), 0, 0 }
};

const zend_function_entry date_funcs_period[] = {
 { "__construct", zim_DatePeriod___construct, arginfo_date_period_construct, (zend_uint) (sizeof(arginfo_date_period_construct)/sizeof(struct _zend_arg_info)-1), 0x2000|0x100 },
 { ((void *)0), ((void *)0), ((void *)0), 0, 0 }
};

static char* guess_timezone(const timelib_tzdb *tzdb );
static void date_register_classes(void);


zend_date_globals date_globals;
static void zm_globals_ctor_date(zend_date_globals *date_globals );


timelib_tzdb *php_date_global_timezone_db;
int php_date_global_timezone_db_enabled;
# 499 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
static const zend_ini_entry ini_entries[] = {
 { 0, ((1<<0)|(1<<1)|(1<<2)), "date.timezone", sizeof("date.timezone"), OnUpdateString, (void *) ((long) (((char *) (&(((zend_date_globals*)((void *)0))->default_timezone))) - ((char *) ((void *)0)))), (void *) &date_globals, ((void *)0), "", sizeof("")-1, ((void *)0), 0, 0, 0, ((void *)0) },
 { 0, ((1<<0)|(1<<1)|(1<<2)), "date.default_latitude", sizeof("date.default_latitude"), ((void *)0), ((void *)0), ((void *)0), ((void *)0), "31.7667", sizeof("31.7667")-1, ((void *)0), 0, 0, 0, ((void *)0) },
 { 0, ((1<<0)|(1<<1)|(1<<2)), "date.default_longitude", sizeof("date.default_longitude"), ((void *)0), ((void *)0), ((void *)0), ((void *)0), "35.2333", sizeof("35.2333")-1, ((void *)0), 0, 0, 0, ((void *)0) },
 { 0, ((1<<0)|(1<<1)|(1<<2)), "date.sunset_zenith", sizeof("date.sunset_zenith"), ((void *)0), ((void *)0), ((void *)0), ((void *)0), "90.583333", sizeof("90.583333")-1, ((void *)0), 0, 0, 0, ((void *)0) },
 { 0, ((1<<0)|(1<<1)|(1<<2)), "date.sunrise_zenith", sizeof("date.sunrise_zenith"), ((void *)0), ((void *)0), ((void *)0), ((void *)0), "90.583333", sizeof("90.583333")-1, ((void *)0), 0, 0, 0, ((void *)0) },
{ 0, 0, ((void *)0), 0, ((void *)0), ((void *)0), ((void *)0), ((void *)0), ((void *)0), 0, ((void *)0), 0, 0, 0, ((void *)0) } };


zend_class_entry *date_ce_date, *date_ce_timezone, *date_ce_interval, *date_ce_period;


__attribute__ ((visibility("default"))) zend_class_entry *php_date_get_date_ce(void)
{
 return date_ce_date;
}

__attribute__ ((visibility("default"))) zend_class_entry *php_date_get_timezone_ce(void)
{
 return date_ce_timezone;
}

static zend_object_handlers date_object_handlers_date;
static zend_object_handlers date_object_handlers_timezone;
static zend_object_handlers date_object_handlers_interval;
static zend_object_handlers date_object_handlers_period;
# 550 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
static void date_object_free_storage_date(void *object );
static void date_object_free_storage_timezone(void *object );
static void date_object_free_storage_interval(void *object );
static void date_object_free_storage_period(void *object );

static zend_object_value date_object_new_date(zend_class_entry *class_type );
static zend_object_value date_object_new_timezone(zend_class_entry *class_type );
static zend_object_value date_object_new_interval(zend_class_entry *class_type );
static zend_object_value date_object_new_period(zend_class_entry *class_type );

static zend_object_value date_object_clone_date(zval *this_ptr );
static zend_object_value date_object_clone_timezone(zval *this_ptr );
static zend_object_value date_object_clone_interval(zval *this_ptr );
static zend_object_value date_object_clone_period(zval *this_ptr );

static int date_object_compare_date(zval *d1, zval *d2 );
static HashTable *date_object_get_properties(zval *object );
static HashTable *date_object_get_properties_interval(zval *object );

zval *date_interval_read_property(zval *object, zval *member, int type, const zend_literal *key );
void date_interval_write_property(zval *object, zval *member, zval *value, const zend_literal *key );


zend_module_entry date_module_entry = {
 sizeof(zend_module_entry), 20100525, 0, 0,
 ((void *)0),
 ((void *)0),
 "date",
 date_functions,
 zm_startup_date,
 zm_shutdown_date,
 zm_activate_date,
 zm_deactivate_date,
 zm_info_date,
 "5.4.0RC3-dev",
 sizeof(zend_date_globals), &date_globals,
 ((void (*)(void* ))(zm_globals_ctor_date)),
 ((void *)0),
 ((void *)0),
 0, 0, ((void *)0), 0, "API" "20100525" ",NTS"
};




static void zm_globals_ctor_date(zend_date_globals *date_globals )
{
 date_globals->default_timezone = ((void *)0);
 date_globals->timezone = ((void *)0);
 date_globals->tzcache = ((void *)0);
}



static void _php_date_tzinfo_dtor(void *tzinfo)
{
 timelib_tzinfo **tzi = (timelib_tzinfo **)tzinfo;

 timelib_tzinfo_dtor(*tzi);
}


int zm_activate_date(int type, int module_number )
{
 if ((date_globals.timezone)) {
  _efree(((date_globals.timezone)) );
 }
 (date_globals.timezone) = ((void *)0);
 (date_globals.tzcache) = ((void *)0);
 (date_globals.last_errors) = ((void *)0);

 return 0;
}



int zm_deactivate_date(int type, int module_number )
{
 if ((date_globals.timezone)) {
  _efree(((date_globals.timezone)) );
 }
 (date_globals.timezone) = ((void *)0);
 if((date_globals.tzcache)) {
  zend_hash_destroy((date_globals.tzcache));
  _efree(((date_globals.tzcache)) );
  (date_globals.tzcache) = ((void *)0);
 }
 if ((date_globals.last_errors)) {
  timelib_error_container_dtor((date_globals.last_errors));
  (date_globals.last_errors) = ((void *)0);
 }

 return 0;
}
# 743 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
int zm_startup_date(int type, int module_number )
{
 zend_register_ini_entries(ini_entries, module_number );
 date_register_classes();







 zend_register_string_constant(("DATE_ATOM"), sizeof("DATE_ATOM"), ("Y-m-d\\TH:i:sP"), ((1<<0) | (1<<1)), module_number );






 zend_register_string_constant(("DATE_COOKIE"), sizeof("DATE_COOKIE"), ("l, d-M-y H:i:s T"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_ISO8601"), sizeof("DATE_ISO8601"), ("Y-m-d\\TH:i:sO"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_RFC822"), sizeof("DATE_RFC822"), ("D, d M y H:i:s O"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_RFC850"), sizeof("DATE_RFC850"), ("l, d-M-y H:i:s T"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_RFC1036"), sizeof("DATE_RFC1036"), ("D, d M y H:i:s O"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_RFC1123"), sizeof("DATE_RFC1123"), ("D, d M Y H:i:s O"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_RFC2822"), sizeof("DATE_RFC2822"), ("D, d M Y H:i:s O"), ((1<<0) | (1<<1)), module_number );
  zend_register_string_constant(("DATE_RFC3339"), sizeof("DATE_RFC3339"), ("Y-m-d\\TH:i:sP"), ((1<<0) | (1<<1)), module_number );





 zend_register_string_constant(("DATE_RSS"), sizeof("DATE_RSS"), ("D, d M Y H:i:s O"), ((1<<0) | (1<<1)), module_number );
 zend_register_string_constant(("DATE_W3C"), sizeof("DATE_W3C"), ("Y-m-d\\TH:i:sP"), ((1<<0) | (1<<1)), module_number );

 zend_register_long_constant(("SUNFUNCS_RET_TIMESTAMP"), sizeof("SUNFUNCS_RET_TIMESTAMP"), (0), ((1<<0) | (1<<1)), module_number );
 zend_register_long_constant(("SUNFUNCS_RET_STRING"), sizeof("SUNFUNCS_RET_STRING"), (1), ((1<<0) | (1<<1)), module_number );
 zend_register_long_constant(("SUNFUNCS_RET_DOUBLE"), sizeof("SUNFUNCS_RET_DOUBLE"), (2), ((1<<0) | (1<<1)), module_number );

 php_date_global_timezone_db = ((void *)0);
 php_date_global_timezone_db_enabled = 0;
 (date_globals.last_errors) = ((void *)0);
 return 0;
}



int zm_shutdown_date(int type, int module_number )
{
 zend_unregister_ini_entries(module_number );

 if ((date_globals.last_errors)) {
  timelib_error_container_dtor((date_globals.last_errors));
 }

 return 0;
}



void zm_info_date(zend_module_entry *zend_module )
{
 const timelib_tzdb *tzdb = php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db();

 php_info_print_table_start();
 php_info_print_table_row(2, "date/time support", "enabled");
 php_info_print_table_row(2, "\"Olson\" Timezone Database Version", tzdb->version);
 php_info_print_table_row(2, "Timezone Database", php_date_global_timezone_db_enabled ? "external" : "internal");
 php_info_print_table_row(2, "Default timezone", guess_timezone(tzdb ));
 php_info_print_table_end();

 display_ini_entries(zend_module);
}



static timelib_tzinfo *php_date_parse_tzfile(char *formal_tzname, const timelib_tzdb *tzdb )
{
 timelib_tzinfo *tzi, **ptzi;

 if(!(date_globals.tzcache)) {
  ((date_globals.tzcache)) = (HashTable *) _emalloc((sizeof(HashTable)) );
  _zend_hash_init(((date_globals.tzcache)), (4), (((void *)0)), (_php_date_tzinfo_dtor), (0) );
 }

 if (zend_hash_find((date_globals.tzcache), formal_tzname, strlen(formal_tzname) + 1, (void **) &ptzi) == 0) {
  return *ptzi;
 }

 tzi = timelib_parse_tzfile(formal_tzname, tzdb);
 if (tzi) {
  _zend_hash_add_or_update((date_globals.tzcache), formal_tzname, strlen(formal_tzname) + 1, (void *) &tzi, sizeof(timelib_tzinfo*), ((void *)0), (1<<1) );
 }
 return tzi;
}

timelib_tzinfo *php_date_parse_tzfile_wrapper(char *formal_tzname, const timelib_tzdb *tzdb)
{
 ;
 return php_date_parse_tzfile(formal_tzname, tzdb );
}



static char* guess_timezone(const timelib_tzdb *tzdb )
{

 if ((date_globals.timezone) && (strlen((date_globals.timezone)) > 0)) {
  return (date_globals.timezone);
 }

 if (!(date_globals.default_timezone)) {

  zval ztz;

  if (0 == zend_get_configuration_directive("date.timezone", sizeof("date.timezone"), &ztz) &&
      (ztz).type == 6 &&
      (ztz).value.str.len > 0 &&
      timelib_timezone_id_is_valid((ztz).value.str.val, tzdb)) {
   return (ztz).value.str.val;
  }
 } else if (*(date_globals.default_timezone) && timelib_timezone_id_is_valid((date_globals.default_timezone), tzdb)) {
  return (date_globals.default_timezone);
 }

 php_error_docref0(((void *)0) , (1<<1L), "It is not safe to rely on the system's timezone settings. You are " "*required* to use the date.timezone setting or the " "date_default_timezone_set() function. In case you used any of those " "methods and you are still getting this warning, you most likely " "misspelled the timezone identifier. " "We selected the timezone 'UTC' for now, but please set date.timezone to select your timezone.");
 return "UTC";
}

__attribute__ ((visibility("default"))) timelib_tzinfo *get_timezone_info(void)
{
 char *tz;
 timelib_tzinfo *tzi;

 tz = guess_timezone(php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db() );
 tzi = php_date_parse_tzfile(tz, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db() );
 if (! tzi) {
  php_error_docref0(((void *)0) , (1<<0L), "Timezone database is corrupt - this should *never* happen!");
 }
 return tzi;
}




# 1 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_smart_str.h" 1
# 153 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/standard/php_smart_str.h"
static inline char *smart_str_print_long(char *buf, long num) {
 char *r;
 do { if (num < 0) { do { char *__p = ((buf)); unsigned long __num = (-(num)); *__p = '\0'; do { *--__p = (char) (__num % 10) + '0'; __num /= 10; } while (__num > 0); (r) = __p; } while (0); *--(r) = '-'; } else { do { char *__p = ((buf)); unsigned long __num = ((num)); *__p = '\0'; do { *--__p = (char) (__num % 10) + '0'; __num /= 10; } while (__num > 0); (r) = __p; } while (0); } } while (0);
 return r;
}

static inline char *smart_str_print_unsigned(char *buf, long num) {
 char *r;
 do { char *__p = (buf); unsigned long __num = (num); *__p = '\0'; do { *--__p = (char) (__num % 10) + '0'; __num /= 10; } while (__num > 0); r = __p; } while (0);
 return r;
}
# 888 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c" 2

static char *mon_full_names[] = {
 "January", "February", "March", "April",
 "May", "June", "July", "August",
 "September", "October", "November", "December"
};

static char *mon_short_names[] = {
 "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static char *day_full_names[] = {
 "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
};

static char *day_short_names[] = {
 "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

static char *english_suffix(timelib_sll number)
{
 if (number >= 10 && number <= 19) {
  return "th";
 } else {
  switch (number % 10) {
   case 1: return "st";
   case 2: return "nd";
   case 3: return "rd";
  }
 }
 return "th";
}



char *php_date_full_day_name(timelib_sll y, timelib_sll m, timelib_sll d)
{
 timelib_sll day_of_week = timelib_day_of_week(y, m, d);
 if (day_of_week < 0) {
  return "Unknown";
 }
 return day_full_names[day_of_week];
}

char *php_date_short_day_name(timelib_sll y, timelib_sll m, timelib_sll d)
{
 timelib_sll day_of_week = timelib_day_of_week(y, m, d);
 if (day_of_week < 0) {
  return "Unknown";
 }
 return day_short_names[day_of_week];
}



static char *date_format(char *format, int format_len, timelib_time *t, int localtime)
{
 smart_str string = {0};
 int i, length;
 char buffer[97];
 timelib_time_offset *offset = ((void *)0);
 timelib_sll isoweek, isoyear;
 int rfc_colon;

 if (!format_len) {
  return _estrdup(("") );
 }

 if (localtime) {
  if (t->zone_type == 2) {
   offset = timelib_time_offset_ctor();
   offset->offset = (t->z - (t->dst * 60)) * -60;
   offset->leap_secs = 0;
   offset->is_dst = t->dst;
   offset->abbr = (__extension__ (__builtin_constant_p (t->tz_abbr) && ((size_t)(const void *)((t->tz_abbr) + 1) - (size_t)(const void *)(t->tz_abbr) == 1) ? (((__const char *) (t->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (t->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, t->tz_abbr, __len); __retval; })) : __strdup (t->tz_abbr)));
  } else if (t->zone_type == 1) {
   offset = timelib_time_offset_ctor();
   offset->offset = (t->z) * -60;
   offset->leap_secs = 0;
   offset->is_dst = 0;
   offset->abbr = malloc(9);
   ap_php_snprintf(offset->abbr, 9, "GMT%c%02d%02d",
                             localtime ? ((offset->offset < 0) ? '-' : '+') : '+',
                             localtime ? abs(offset->offset / 3600) : 0,
                             localtime ? abs((offset->offset % 3600) / 60) : 0 );
  } else {
   offset = timelib_get_time_zone_info(t->sse, t->tz_info);
  }
 }
 timelib_isoweek_from_date(t->y, t->m, t->d, &isoweek, &isoyear);

 for (i = 0; i < format_len; i++) {
  rfc_colon = 0;
  switch (format[i]) {

   case 'd': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->d); break;
   case 'D': length = ap_php_slprintf(buffer, 32, "%s", php_date_short_day_name(t->y, t->m, t->d)); break;
   case 'j': length = ap_php_slprintf(buffer, 32, "%d", (int) t->d); break;
   case 'l': length = ap_php_slprintf(buffer, 32, "%s", php_date_full_day_name(t->y, t->m, t->d)); break;
   case 'S': length = ap_php_slprintf(buffer, 32, "%s", english_suffix(t->d)); break;
   case 'w': length = ap_php_slprintf(buffer, 32, "%d", (int) timelib_day_of_week(t->y, t->m, t->d)); break;
   case 'N': length = ap_php_slprintf(buffer, 32, "%d", (int) timelib_iso_day_of_week(t->y, t->m, t->d)); break;
   case 'z': length = ap_php_slprintf(buffer, 32, "%d", (int) timelib_day_of_year(t->y, t->m, t->d)); break;


   case 'W': length = ap_php_slprintf(buffer, 32, "%02d", (int) isoweek); break;
   case 'o': length = ap_php_slprintf(buffer, 32, "%d", (int) isoyear); break;


   case 'F': length = ap_php_slprintf(buffer, 32, "%s", mon_full_names[t->m - 1]); break;
   case 'm': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->m); break;
   case 'M': length = ap_php_slprintf(buffer, 32, "%s", mon_short_names[t->m - 1]); break;
   case 'n': length = ap_php_slprintf(buffer, 32, "%d", (int) t->m); break;
   case 't': length = ap_php_slprintf(buffer, 32, "%d", (int) timelib_days_in_month(t->y, t->m)); break;


   case 'L': length = ap_php_slprintf(buffer, 32, "%d", (((int) t->y) % 4 == 0 && (((int) t->y) % 100 != 0 || ((int) t->y) % 400 == 0))); break;
   case 'y': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->y % 100); break;
   case 'Y': length = ap_php_slprintf(buffer, 32, "%s%04lld", t->y < 0 ? "-" : "", php_date_llabs((timelib_sll) t->y)); break;


   case 'a': length = ap_php_slprintf(buffer, 32, "%s", t->h >= 12 ? "pm" : "am"); break;
   case 'A': length = ap_php_slprintf(buffer, 32, "%s", t->h >= 12 ? "PM" : "AM"); break;
   case 'B': {
    int retval = (((((long)t->sse)-(((long)t->sse) - ((((long)t->sse) % 86400) + 3600))) * 10) / 864);
    while (retval < 0) {
     retval += 1000;
    }
    retval = retval % 1000;
    length = ap_php_slprintf(buffer, 32, "%03d", retval);
    break;
   }
   case 'g': length = ap_php_slprintf(buffer, 32, "%d", (t->h % 12) ? (int) t->h % 12 : 12); break;
   case 'G': length = ap_php_slprintf(buffer, 32, "%d", (int) t->h); break;
   case 'h': length = ap_php_slprintf(buffer, 32, "%02d", (t->h % 12) ? (int) t->h % 12 : 12); break;
   case 'H': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->h); break;
   case 'i': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->i); break;
   case 's': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->s); break;
   case 'u': length = ap_php_slprintf(buffer, 32, "%06d", (int) floor(t->f * 1000000)); break;


   case 'I': length = ap_php_slprintf(buffer, 32, "%d", localtime ? offset->is_dst : 0); break;
   case 'P': rfc_colon = 1;
   case 'O': length = ap_php_slprintf(buffer, 32, "%c%02d%s%02d",
           localtime ? ((offset->offset < 0) ? '-' : '+') : '+',
           localtime ? abs(offset->offset / 3600) : 0,
           rfc_colon ? ":" : "",
           localtime ? abs((offset->offset % 3600) / 60) : 0
         );
       break;
   case 'T': length = ap_php_slprintf(buffer, 32, "%s", localtime ? offset->abbr : "GMT"); break;
   case 'e': if (!localtime) {
           length = ap_php_slprintf(buffer, 32, "%s", "UTC");
       } else {
        switch (t->zone_type) {
         case 3:
          length = ap_php_slprintf(buffer, 32, "%s", t->tz_info->name);
          break;
         case 2:
          length = ap_php_slprintf(buffer, 32, "%s", offset->abbr);
          break;
         case 1:
          length = ap_php_slprintf(buffer, 32, "%c%02d:%02d",
            ((offset->offset < 0) ? '-' : '+'),
            abs(offset->offset / 3600),
            abs((offset->offset % 3600) / 60)
             );
          break;
        }
       }
       break;
   case 'Z': length = ap_php_slprintf(buffer, 32, "%d", localtime ? offset->offset : 0); break;


   case 'c': length = ap_php_slprintf(buffer, 96, "%04d-%02d-%02dT%02d:%02d:%02d%c%02d:%02d",
                       (int) t->y, (int) t->m, (int) t->d,
           (int) t->h, (int) t->i, (int) t->s,
           localtime ? ((offset->offset < 0) ? '-' : '+') : '+',
           localtime ? abs(offset->offset / 3600) : 0,
           localtime ? abs((offset->offset % 3600) / 60) : 0
         );
       break;
   case 'r': length = ap_php_slprintf(buffer, 96, "%3s, %02d %3s %04d %02d:%02d:%02d %c%02d%02d",
                       php_date_short_day_name(t->y, t->m, t->d),
           (int) t->d, mon_short_names[t->m - 1],
           (int) t->y, (int) t->h, (int) t->i, (int) t->s,
           localtime ? ((offset->offset < 0) ? '-' : '+') : '+',
           localtime ? abs(offset->offset / 3600) : 0,
           localtime ? abs((offset->offset % 3600) / 60) : 0
         );
       break;
   case 'U': length = ap_php_slprintf(buffer, 32, "%lld", (timelib_sll) t->sse); break;

   case '\\': if (i < format_len) i++;

   default: buffer[0] = format[i]; buffer[1] = '\0'; length = 1; break;
  }
  do { register size_t __nl; smart_str *__dest = (smart_str *) ((&string)); do { if (!(__dest)->c) { (__dest)->len = 0; __nl = (((length))); (__dest)->a = __nl < 78 ? 78 : __nl + 128; (__dest)->c = (((((0))))?__zend_realloc((((__dest)->c)), (((__dest)->a + 1))):_erealloc(((((__dest)->c))), ((((__dest)->a + 1))), 0 )); } else { __nl = (__dest)->len + (((length))); if (__nl >= (__dest)->a) { (__dest)->a = __nl + 128; (__dest)->c = (((((0))))?__zend_realloc((((__dest)->c)), (((__dest)->a + 1))):_erealloc(((((__dest)->c))), ((((__dest)->a + 1))), 0 )); } } } while (0); memcpy(__dest->c + __dest->len, ((buffer)), ((length))); __dest->len = __nl; } while (0);
 }

 do { if ((&string)->c) { (&string)->c[(&string)->len] = '\0'; } } while (0);

 if (localtime) {
  timelib_time_offset_dtor(offset);
 }

 return string.c;
}

static void php_date(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int localtime)
{
 char *format;
 int format_len;
 long ts;
 char *string;

 if (zend_parse_parameters((ht) , "s|l", &format, &format_len, &ts) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 if ((ht) == 1) {
  ts = time(((void *)0));
 }

 string = php_format_date(format, format_len, ts, localtime );

 do { const char *__s=(string); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
}


__attribute__ ((visibility("default"))) char *php_format_date(char *format, int format_len, time_t ts, int localtime )
{
 timelib_time *t;
 timelib_tzinfo *tzi;
 char *string;

 t = timelib_time_ctor();

 if (localtime) {
  tzi = get_timezone_info();
  t->tz_info = tzi;
  t->zone_type = 3;
  timelib_unixtime2local(t, ts);
 } else {
  tzi = ((void *)0);
  timelib_unixtime2gmt(t, ts);
 }

 string = date_format(format, format_len, t, localtime);

 timelib_time_dtor(t);
 return string;
}




__attribute__ ((visibility("default"))) int php_idate(char format, time_t ts, int localtime )
{
 timelib_time *t;
 timelib_tzinfo *tzi;
 int retval = -1;
 timelib_time_offset *offset = ((void *)0);
 timelib_sll isoweek, isoyear;

 t = timelib_time_ctor();

 if (!localtime) {
  tzi = get_timezone_info();
  t->tz_info = tzi;
  t->zone_type = 3;
  timelib_unixtime2local(t, ts);
 } else {
  tzi = ((void *)0);
  timelib_unixtime2gmt(t, ts);
 }

 if (!localtime) {
  if (t->zone_type == 2) {
   offset = timelib_time_offset_ctor();
   offset->offset = (t->z - (t->dst * 60)) * -60;
   offset->leap_secs = 0;
   offset->is_dst = t->dst;
   offset->abbr = (__extension__ (__builtin_constant_p (t->tz_abbr) && ((size_t)(const void *)((t->tz_abbr) + 1) - (size_t)(const void *)(t->tz_abbr) == 1) ? (((__const char *) (t->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (t->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, t->tz_abbr, __len); __retval; })) : __strdup (t->tz_abbr)));
  } else if (t->zone_type == 1) {
   offset = timelib_time_offset_ctor();
   offset->offset = (t->z - (t->dst * 60)) * -60;
   offset->leap_secs = 0;
   offset->is_dst = t->dst;
   offset->abbr = malloc(9);
   ap_php_snprintf(offset->abbr, 9, "GMT%c%02d%02d",
                             !localtime ? ((offset->offset < 0) ? '-' : '+') : '+',
                             !localtime ? abs(offset->offset / 3600) : 0,
                             !localtime ? abs((offset->offset % 3600) / 60) : 0 );
  } else {
   offset = timelib_get_time_zone_info(t->sse, t->tz_info);
  }
 }

 timelib_isoweek_from_date(t->y, t->m, t->d, &isoweek, &isoyear);

 switch (format) {

  case 'd': case 'j': retval = (int) t->d; break;

  case 'w': retval = (int) timelib_day_of_week(t->y, t->m, t->d); break;
  case 'z': retval = (int) timelib_day_of_year(t->y, t->m, t->d); break;


  case 'W': retval = (int) isoweek; break;


  case 'm': case 'n': retval = (int) t->m; break;
  case 't': retval = (int) timelib_days_in_month(t->y, t->m); break;


  case 'L': retval = (int) (((int) t->y) % 4 == 0 && (((int) t->y) % 100 != 0 || ((int) t->y) % 400 == 0)); break;
  case 'y': retval = (int) (t->y % 100); break;
  case 'Y': retval = (int) t->y; break;


  case 'B':
   retval = (((((long)t->sse)-(((long)t->sse) - ((((long)t->sse) % 86400) + 3600))) * 10) / 864);
   while (retval < 0) {
    retval += 1000;
   }
   retval = retval % 1000;
   break;


  case 'g': case 'h': retval = (int) ((t->h % 12) ? (int) t->h % 12 : 12); break;
  case 'H': case 'G': retval = (int) t->h; break;
  case 'i': retval = (int) t->i; break;
  case 's': retval = (int) t->s; break;


  case 'I': retval = (int) (!localtime ? offset->is_dst : 0); break;
  case 'Z': retval = (int) (!localtime ? offset->offset : 0); break;

  case 'U': retval = (int) t->sse; break;
 }

 if (!localtime) {
  timelib_time_offset_dtor(offset);
 }
 timelib_time_dtor(t);

 return retval;
}




void zif_date(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_date(ht, return_value, return_value_ptr, this_ptr, return_value_used , 1);
}




void zif_gmdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_date(ht, return_value, return_value_ptr, this_ptr, return_value_used , 0);
}




void zif_idate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *format;
 int format_len;
 long ts = 0;
 int ret;

 if (zend_parse_parameters((ht) , "s|l", &format, &format_len, &ts) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if (format_len != 1) {
  php_error_docref0(((void *)0) , (1<<1L), "idate format is one char");
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if ((ht) == 1) {
  ts = time(((void *)0));
 }

 ret = php_idate(format[0], ts, 0 );
 if (ret == -1) {
  php_error_docref0(((void *)0) , (1<<1L), "Unrecognized date format token.");
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 { { zval *__z = (return_value); (*__z).value.lval = ret; (*__z).type = 1; }; return; };
}



__attribute__ ((visibility("default"))) void php_date_set_tzdb(timelib_tzdb *tzdb)
{
 const timelib_tzdb *builtin = timelib_builtin_db();

 if (php_version_compare(tzdb->version, builtin->version) > 0) {
  php_date_global_timezone_db = tzdb;
  php_date_global_timezone_db_enabled = 1;
 }
}



__attribute__ ((visibility("default"))) signed long php_parse_date(char *string, signed long *now)
{
 timelib_time *parsed_time;
 timelib_error_container *error = ((void *)0);
 int error2;
 signed long retval;

 parsed_time = timelib_strtotime(string, strlen(string), &error, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 if (error->error_count) {
  timelib_error_container_dtor(error);
  return -1;
 }
 timelib_error_container_dtor(error);
 timelib_update_ts(parsed_time, ((void *)0));
 retval = timelib_date_to_int(parsed_time, &error2);
 timelib_time_dtor(parsed_time);
 if (error2) {
  return -1;
 }
 return retval;
}





void zif_strtotime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *times, *initial_ts;
 int time_len, error1, error2;
 struct timelib_error_container *error;
 long preset_ts = 0, ts;

 timelib_time *t, *now;
 timelib_tzinfo *tzi;

 tzi = get_timezone_info();

 if (zend_parse_parameters_ex(1<<1, (ht) , "sl", &times, &time_len, &preset_ts) != -1) {

  now = timelib_time_ctor();

  initial_ts = _emalloc((25) );
  ap_php_snprintf(initial_ts, 24, "@%ld UTC", preset_ts);
  t = timelib_strtotime(initial_ts, strlen(initial_ts), ((void *)0), php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
  timelib_update_ts(t, tzi);
  now->tz_info = tzi;
  now->zone_type = 3;
  timelib_unixtime2local(now, t->sse);
  timelib_time_dtor(t);
  _efree((initial_ts) );
 } else if (zend_parse_parameters((ht) , "s|l", &times, &time_len, &preset_ts) != -1) {

  now = timelib_time_ctor();
  now->tz_info = tzi;
  now->zone_type = 3;
  timelib_unixtime2local(now, (timelib_sll) time(((void *)0)));
 } else {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if (!time_len) {
  timelib_time_dtor(now);
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 t = timelib_strtotime(times, time_len, &error, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 error1 = error->error_count;
 timelib_error_container_dtor(error);
 timelib_fill_holes(t, now, 0x02);
 timelib_update_ts(t, tzi);
 ts = timelib_date_to_int(t, &error2);

 timelib_time_dtor(now);
 timelib_time_dtor(t);

 if (error1 || error2) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 } else {
  { { zval *__z = (return_value); (*__z).value.lval = ts; (*__z).type = 1; }; return; };
 }
}




__attribute__ ((visibility("default"))) void php_mktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int gmt)
{
 long hou = 0, min = 0, sec = 0, mon = 0, day = 0, yea = 0, dst = -1;
 timelib_time *now;
 timelib_tzinfo *tzi = ((void *)0);
 long ts, adjust_seconds = 0;
 int error;

 if (zend_parse_parameters((ht) , "|lllllll", &hou, &min, &sec, &mon, &day, &yea, &dst) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 now = timelib_time_ctor();
 if (gmt) {
  timelib_unixtime2gmt(now, (timelib_sll) time(((void *)0)));
 } else {
  tzi = get_timezone_info();
  now->tz_info = tzi;
  now->zone_type = 3;
  timelib_unixtime2local(now, (timelib_sll) time(((void *)0)));
 }

 switch ((ht)) {
  case 7:

  case 6:
   if (yea >= 0 && yea < 70) {
    yea += 2000;
   } else if (yea >= 70 && yea <= 100) {
    yea += 1900;
   }
   now->y = yea;

  case 5:
   now->d = day;

  case 4:
   now->m = mon;

  case 3:
   now->s = sec;

  case 2:
   now->i = min;

  case 1:
   now->h = hou;
   break;
  default:
   php_error_docref0(((void *)0) , (1<<11L), "You should be using the time() function instead");
 }

 if (gmt) {
  timelib_update_ts(now, ((void *)0));
 } else {
  timelib_update_ts(now, tzi);
 }

 if (dst != -1) {
  php_error_docref0(((void *)0) , (1<<13L), "The is_dst parameter is deprecated");
  if (gmt) {

   if (dst == 1) {
    adjust_seconds = -3600;
   }
  } else {

   timelib_time_offset *tmp_offset;
   tmp_offset = timelib_get_time_zone_info(now->sse, tzi);
   if (dst == 1 && tmp_offset->is_dst == 0) {
    adjust_seconds = -3600;
   }
   if (dst == 0 && tmp_offset->is_dst == 1) {
    adjust_seconds = +3600;
   }
   timelib_time_offset_dtor(tmp_offset);
  }
 }

 ts = timelib_date_to_int(now, &error);
 ts += adjust_seconds;
 timelib_time_dtor(now);

 if (error) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 } else {
  { { zval *__z = (return_value); (*__z).value.lval = ts; (*__z).type = 1; }; return; };
 }
}




void zif_mktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_mktime(ht, return_value, return_value_ptr, this_ptr, return_value_used , 0);
}




void zif_gmmktime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_mktime(ht, return_value, return_value_ptr, this_ptr, return_value_used , 1);
}





void zif_checkdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 long m, d, y;

 if (zend_parse_parameters((ht) , "lll", &m, &d, &y) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if (y < 1 || y > 32767 || !timelib_valid_date(y, m, d)) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 { do { zval *__z = (return_value); (*__z).value.lval = ((1) != 0); (*__z).type = 3; } while (0); return; };
}




__attribute__ ((visibility("default"))) void php_strftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int gmt)
{
 char *format, *buf;
 int format_len;
 long timestamp = 0;
 struct tm ta;
 int max_reallocs = 5;
 size_t buf_len = 64, real_len;
 timelib_time *ts;
 timelib_tzinfo *tzi;
 timelib_time_offset *offset = ((void *)0);

 timestamp = (long) time(((void *)0));

 if (zend_parse_parameters((ht) , "s|l", &format, &format_len, &timestamp) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if (format_len == 0) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 ts = timelib_time_ctor();
 if (gmt) {
  tzi = ((void *)0);
  timelib_unixtime2gmt(ts, (timelib_sll) timestamp);
 } else {
  tzi = get_timezone_info();
  ts->tz_info = tzi;
  ts->zone_type = 3;
  timelib_unixtime2local(ts, (timelib_sll) timestamp);
 }
 ta.tm_sec = ts->s;
 ta.tm_min = ts->i;
 ta.tm_hour = ts->h;
 ta.tm_mday = ts->d;
 ta.tm_mon = ts->m - 1;
 ta.tm_year = ts->y - 1900;
 ta.tm_wday = timelib_day_of_week(ts->y, ts->m, ts->d);
 ta.tm_yday = timelib_day_of_year(ts->y, ts->m, ts->d);
 if (gmt) {
  ta.tm_isdst = 0;

  ta.tm_gmtoff = 0;


  ta.tm_zone = "GMT";

 } else {
  offset = timelib_get_time_zone_info(timestamp, tzi);

  ta.tm_isdst = offset->is_dst;

  ta.tm_gmtoff = offset->offset;


  ta.tm_zone = offset->abbr;

 }

 buf = (char *) _emalloc((buf_len) );
 while ((real_len=strftime(buf, buf_len, format, &ta))==buf_len || real_len==0) {
  buf_len *= 2;
  buf = (char *) _erealloc((buf), (buf_len), 0 );
  if (!--max_reallocs) {
   break;
  }
 }

 timelib_time_dtor(ts);
 if (!gmt) {
  timelib_time_offset_dtor(offset);
 }

 if (real_len && real_len != buf_len) {
  buf = (char *) _erealloc((buf), (real_len + 1), 0 );
  { do { const char *__s=(buf); int __l=real_len; zval *__z = (return_value); (*__z).value.str.len = __l; (*__z).value.str.val = (0?_estrndup((__s), (__l) ):(char*)__s); (*__z).type = 6; } while (0); return; };
 }
 _efree((buf) );
 { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
}




void zif_strftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_strftime(ht, return_value, return_value_ptr, this_ptr, return_value_used , 0);
}




void zif_gmstrftime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_strftime(ht, return_value, return_value_ptr, this_ptr, return_value_used , 1);
}





void zif_time(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 { { zval *__z = (return_value); (*__z).value.lval = (long)time(((void *)0)); (*__z).type = 1; }; return; };
}




void zif_localtime(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 long timestamp = (long)time(((void *)0));
 zend_bool associative = 0;
 timelib_tzinfo *tzi;
 timelib_time *ts;

 if (zend_parse_parameters((ht) , "|lb", &timestamp, &associative) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 tzi = get_timezone_info();
 ts = timelib_time_ctor();
 ts->tz_info = tzi;
 ts->zone_type = 3;
 timelib_unixtime2local(ts, (timelib_sll) timestamp);

 _array_init((return_value), 0 );

 if (associative) {
  add_assoc_long_ex(return_value, "tm_sec", strlen("tm_sec")+1, ts->s);
  add_assoc_long_ex(return_value, "tm_min", strlen("tm_min")+1, ts->i);
  add_assoc_long_ex(return_value, "tm_hour", strlen("tm_hour")+1, ts->h);
  add_assoc_long_ex(return_value, "tm_mday", strlen("tm_mday")+1, ts->d);
  add_assoc_long_ex(return_value, "tm_mon", strlen("tm_mon")+1, ts->m - 1);
  add_assoc_long_ex(return_value, "tm_year", strlen("tm_year")+1, ts->y - 1900);
  add_assoc_long_ex(return_value, "tm_wday", strlen("tm_wday")+1, timelib_day_of_week(ts->y, ts->m, ts->d));
  add_assoc_long_ex(return_value, "tm_yday", strlen("tm_yday")+1, timelib_day_of_year(ts->y, ts->m, ts->d));
  add_assoc_long_ex(return_value, "tm_isdst", strlen("tm_isdst")+1, ts->dst);
 } else {
  add_next_index_long(return_value, ts->s);
  add_next_index_long(return_value, ts->i);
  add_next_index_long(return_value, ts->h);
  add_next_index_long(return_value, ts->d);
  add_next_index_long(return_value, ts->m - 1);
  add_next_index_long(return_value, ts->y- 1900);
  add_next_index_long(return_value, timelib_day_of_week(ts->y, ts->m, ts->d));
  add_next_index_long(return_value, timelib_day_of_year(ts->y, ts->m, ts->d));
  add_next_index_long(return_value, ts->dst);
 }

 timelib_time_dtor(ts);
}




void zif_getdate(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 long timestamp = (long)time(((void *)0));
 timelib_tzinfo *tzi;
 timelib_time *ts;

 if (zend_parse_parameters((ht) , "|l", &timestamp) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 tzi = get_timezone_info();
 ts = timelib_time_ctor();
 ts->tz_info = tzi;
 ts->zone_type = 3;
 timelib_unixtime2local(ts, (timelib_sll) timestamp);

 _array_init((return_value), 0 );

 add_assoc_long_ex(return_value, "seconds", strlen("seconds")+1, ts->s);
 add_assoc_long_ex(return_value, "minutes", strlen("minutes")+1, ts->i);
 add_assoc_long_ex(return_value, "hours", strlen("hours")+1, ts->h);
 add_assoc_long_ex(return_value, "mday", strlen("mday")+1, ts->d);
 add_assoc_long_ex(return_value, "wday", strlen("wday")+1, timelib_day_of_week(ts->y, ts->m, ts->d));
 add_assoc_long_ex(return_value, "mon", strlen("mon")+1, ts->m);
 add_assoc_long_ex(return_value, "year", strlen("year")+1, ts->y);
 add_assoc_long_ex(return_value, "yday", strlen("yday")+1, timelib_day_of_year(ts->y, ts->m, ts->d));
 add_assoc_string_ex(return_value, "weekday", strlen("weekday")+1, php_date_full_day_name(ts->y, ts->m, ts->d), 1);
 add_assoc_string_ex(return_value, "month", strlen("month")+1, mon_full_names[ts->m - 1], 1);
 add_index_long(return_value, 0, timestamp);

 timelib_time_dtor(ts);
}
# 1721 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
typedef struct {
 zend_object_iterator intern;
 zval *date_period_zval;
 zval *current;
 php_period_obj *object;
 int current_index;
} date_period_it;


static void date_period_it_invalidate_current(zend_object_iterator *iter )
{
 date_period_it *iterator = (date_period_it *)iter;

 if (iterator->current) {
  _zval_ptr_dtor((&iterator->current) );
  iterator->current = ((void *)0);
 }
}




static void date_period_it_dtor(zend_object_iterator *iter )
{
 date_period_it *iterator = (date_period_it *)iter;

 date_period_it_invalidate_current(iter );

 _zval_ptr_dtor((&iterator->date_period_zval) );

 _efree((iterator) );
}




static int date_period_it_has_more(zend_object_iterator *iter )
{
 date_period_it *iterator = (date_period_it *)iter;
 php_period_obj *object = iterator->object;
 timelib_time *it_time = object->current;


 if (!object->include_start_date || iterator->current_index > 0) {
  it_time->have_relative = 1;
  it_time->relative = *object->interval;
  it_time->sse_uptodate = 0;
  timelib_update_ts(it_time, ((void *)0));
  timelib_update_from_sse(it_time);
 }

 if (object->end) {
  return object->current->sse < object->end->sse ? 0 : -1;
 } else {
  return (iterator->current_index < object->recurrences) ? 0 : -1;
 }
}




static void date_period_it_current_data(zend_object_iterator *iter, zval ***data )
{
 date_period_it *iterator = (date_period_it *)iter;
 php_period_obj *object = iterator->object;
 timelib_time *it_time = object->current;
 php_date_obj *newdateobj;


 do { (iterator->current) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(iterator->current))->u.buffered = ((void *)0); } while (0); (iterator->current)->refcount__gc = 1; (iterator->current)->is_ref__gc = 0;;;
 php_date_instantiate(date_ce_date, iterator->current );
 newdateobj = (php_date_obj *) zend_object_store_get_object(iterator->current );
 newdateobj->time = timelib_time_ctor();
 *newdateobj->time = *it_time;
 if (it_time->tz_abbr) {
  newdateobj->time->tz_abbr = (__extension__ (__builtin_constant_p (it_time->tz_abbr) && ((size_t)(const void *)((it_time->tz_abbr) + 1) - (size_t)(const void *)(it_time->tz_abbr) == 1) ? (((__const char *) (it_time->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (it_time->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, it_time->tz_abbr, __len); __retval; })) : __strdup (it_time->tz_abbr)));
 }
 if (it_time->tz_info) {
  newdateobj->time->tz_info = it_time->tz_info;
 }

 *data = &iterator->current;
}




static int date_period_it_current_key(zend_object_iterator *iter, char **str_key, uint *str_key_len, ulong *int_key )
{
 date_period_it *iterator = (date_period_it *)iter;
 *int_key = iterator->current_index;
 return 2;
}




static void date_period_it_move_forward(zend_object_iterator *iter )
{
 date_period_it *iterator = (date_period_it *)iter;

 iterator->current_index++;
 date_period_it_invalidate_current(iter );
}




static void date_period_it_rewind(zend_object_iterator *iter )
{
 date_period_it *iterator = (date_period_it *)iter;

 iterator->current_index = 0;
 if (iterator->object->current) {
  timelib_time_dtor(iterator->object->current);
 }
 iterator->object->current = timelib_time_clone(iterator->object->start);
 date_period_it_invalidate_current(iter );
}




zend_object_iterator_funcs date_period_it_funcs = {
 date_period_it_dtor,
 date_period_it_has_more,
 date_period_it_current_data,
 date_period_it_current_key,
 date_period_it_move_forward,
 date_period_it_rewind,
 date_period_it_invalidate_current
};



zend_object_iterator *date_object_period_get_iterator(zend_class_entry *ce, zval *object, int by_ref )
{
 date_period_it *iterator = _emalloc((sizeof(date_period_it)) );
 php_period_obj *dpobj = (php_period_obj *)zend_object_store_get_object(object );

 if (by_ref) {
  zend_error((1<<0L), "An iterator cannot be used with foreach by reference");
 }

 zval_addref_p(object);
 iterator->intern.data = (void*) dpobj;
 iterator->intern.funcs = &date_period_it_funcs;
 iterator->date_period_zval = object;
 iterator->object = dpobj;
 iterator->current = ((void *)0);

 return (zend_object_iterator*)iterator;
}

static void date_register_classes(void)
{
 zend_class_entry ce_date, ce_timezone, ce_interval, ce_period;

 { const char *cl_name = "DateTime"; int _len = sizeof("DateTime")-1; ce_date.name = zend_new_interned_string(cl_name, _len+1, 0 ); if (ce_date.name == cl_name) { ce_date.name = zend_strndup(cl_name, _len); } ce_date.name_length = _len; ce_date.constructor = ((void *)0); ce_date.destructor = ((void *)0); ce_date.clone = ((void *)0); ce_date.serialize = ((void *)0); ce_date.unserialize = ((void *)0); ce_date.create_object = ((void *)0); ce_date.interface_gets_implemented = ((void *)0); ce_date.get_static_method = ((void *)0); ce_date.__call = ((void *)0); ce_date.__callstatic = ((void *)0); ce_date.__tostring = ((void *)0); ce_date.__get = ((void *)0); ce_date.__set = ((void *)0); ce_date.__unset = ((void *)0); ce_date.__isset = ((void *)0); ce_date.serialize_func = ((void *)0); ce_date.unserialize_func = ((void *)0); ce_date.serialize = ((void *)0); ce_date.unserialize = ((void *)0); ce_date.parent = ((void *)0); ce_date.num_interfaces = 0; ce_date.traits = ((void *)0); ce_date.num_traits = 0; ce_date.trait_aliases = ((void *)0); ce_date.trait_precedences = ((void *)0); ce_date.interfaces = ((void *)0); ce_date.get_iterator = ((void *)0); ce_date.iterator_funcs.funcs = ((void *)0); ce_date.info.internal.module = ((void *)0); ce_date.info.internal.builtin_functions = date_funcs_date; };
 ce_date.create_object = date_object_new_date;
 date_ce_date = zend_register_internal_class_ex(&ce_date, ((void *)0), ((void *)0) );
 memcpy(&date_object_handlers_date, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 date_object_handlers_date.clone_obj = date_object_clone_date;
 date_object_handlers_date.compare_objects = date_object_compare_date;
 date_object_handlers_date.get_properties = date_object_get_properties;




 zend_declare_class_constant_stringl(date_ce_date, "ATOM", sizeof("ATOM")-1, "Y-m-d\\TH:i:sP", sizeof("Y-m-d\\TH:i:sP")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "COOKIE", sizeof("COOKIE")-1, "l, d-M-y H:i:s T", sizeof("l, d-M-y H:i:s T")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "ISO8601", sizeof("ISO8601")-1, "Y-m-d\\TH:i:sO", sizeof("Y-m-d\\TH:i:sO")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC822", sizeof("RFC822")-1, "D, d M y H:i:s O", sizeof("D, d M y H:i:s O")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC850", sizeof("RFC850")-1, "l, d-M-y H:i:s T", sizeof("l, d-M-y H:i:s T")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC1036", sizeof("RFC1036")-1, "D, d M y H:i:s O", sizeof("D, d M y H:i:s O")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC1123", sizeof("RFC1123")-1, "D, d M Y H:i:s O", sizeof("D, d M Y H:i:s O")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC2822", sizeof("RFC2822")-1, "D, d M Y H:i:s O", sizeof("D, d M Y H:i:s O")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RFC3339", sizeof("RFC3339")-1, "Y-m-d\\TH:i:sP", sizeof("Y-m-d\\TH:i:sP")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "RSS", sizeof("RSS")-1, "D, d M Y H:i:s O", sizeof("D, d M Y H:i:s O")-1 );;
 zend_declare_class_constant_stringl(date_ce_date, "W3C", sizeof("W3C")-1, "Y-m-d\\TH:i:sP", sizeof("Y-m-d\\TH:i:sP")-1 );;


 { const char *cl_name = "DateTimeZone"; int _len = sizeof("DateTimeZone")-1; ce_timezone.name = zend_new_interned_string(cl_name, _len+1, 0 ); if (ce_timezone.name == cl_name) { ce_timezone.name = zend_strndup(cl_name, _len); } ce_timezone.name_length = _len; ce_timezone.constructor = ((void *)0); ce_timezone.destructor = ((void *)0); ce_timezone.clone = ((void *)0); ce_timezone.serialize = ((void *)0); ce_timezone.unserialize = ((void *)0); ce_timezone.create_object = ((void *)0); ce_timezone.interface_gets_implemented = ((void *)0); ce_timezone.get_static_method = ((void *)0); ce_timezone.__call = ((void *)0); ce_timezone.__callstatic = ((void *)0); ce_timezone.__tostring = ((void *)0); ce_timezone.__get = ((void *)0); ce_timezone.__set = ((void *)0); ce_timezone.__unset = ((void *)0); ce_timezone.__isset = ((void *)0); ce_timezone.serialize_func = ((void *)0); ce_timezone.unserialize_func = ((void *)0); ce_timezone.serialize = ((void *)0); ce_timezone.unserialize = ((void *)0); ce_timezone.parent = ((void *)0); ce_timezone.num_interfaces = 0; ce_timezone.traits = ((void *)0); ce_timezone.num_traits = 0; ce_timezone.trait_aliases = ((void *)0); ce_timezone.trait_precedences = ((void *)0); ce_timezone.interfaces = ((void *)0); ce_timezone.get_iterator = ((void *)0); ce_timezone.iterator_funcs.funcs = ((void *)0); ce_timezone.info.internal.module = ((void *)0); ce_timezone.info.internal.builtin_functions = date_funcs_timezone; };
 ce_timezone.create_object = date_object_new_timezone;
 date_ce_timezone = zend_register_internal_class_ex(&ce_timezone, ((void *)0), ((void *)0) );
 memcpy(&date_object_handlers_timezone, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 date_object_handlers_timezone.clone_obj = date_object_clone_timezone;




 zend_declare_class_constant_long(date_ce_timezone, "AFRICA", sizeof("AFRICA")-1, 0x0001 );;
 zend_declare_class_constant_long(date_ce_timezone, "AMERICA", sizeof("AMERICA")-1, 0x0002 );;
 zend_declare_class_constant_long(date_ce_timezone, "ANTARCTICA", sizeof("ANTARCTICA")-1, 0x0004 );;
 zend_declare_class_constant_long(date_ce_timezone, "ARCTIC", sizeof("ARCTIC")-1, 0x0008 );;
 zend_declare_class_constant_long(date_ce_timezone, "ASIA", sizeof("ASIA")-1, 0x0010 );;
 zend_declare_class_constant_long(date_ce_timezone, "ATLANTIC", sizeof("ATLANTIC")-1, 0x0020 );;
 zend_declare_class_constant_long(date_ce_timezone, "AUSTRALIA", sizeof("AUSTRALIA")-1, 0x0040 );;
 zend_declare_class_constant_long(date_ce_timezone, "EUROPE", sizeof("EUROPE")-1, 0x0080 );;
 zend_declare_class_constant_long(date_ce_timezone, "INDIAN", sizeof("INDIAN")-1, 0x0100 );;
 zend_declare_class_constant_long(date_ce_timezone, "PACIFIC", sizeof("PACIFIC")-1, 0x0200 );;
 zend_declare_class_constant_long(date_ce_timezone, "UTC", sizeof("UTC")-1, 0x0400 );;
 zend_declare_class_constant_long(date_ce_timezone, "ALL", sizeof("ALL")-1, 0x07FF );;
 zend_declare_class_constant_long(date_ce_timezone, "ALL_WITH_BC", sizeof("ALL_WITH_BC")-1, 0x0FFF );;
 zend_declare_class_constant_long(date_ce_timezone, "PER_COUNTRY", sizeof("PER_COUNTRY")-1, 0x1000 );;

 { const char *cl_name = "DateInterval"; int _len = sizeof("DateInterval")-1; ce_interval.name = zend_new_interned_string(cl_name, _len+1, 0 ); if (ce_interval.name == cl_name) { ce_interval.name = zend_strndup(cl_name, _len); } ce_interval.name_length = _len; ce_interval.constructor = ((void *)0); ce_interval.destructor = ((void *)0); ce_interval.clone = ((void *)0); ce_interval.serialize = ((void *)0); ce_interval.unserialize = ((void *)0); ce_interval.create_object = ((void *)0); ce_interval.interface_gets_implemented = ((void *)0); ce_interval.get_static_method = ((void *)0); ce_interval.__call = ((void *)0); ce_interval.__callstatic = ((void *)0); ce_interval.__tostring = ((void *)0); ce_interval.__get = ((void *)0); ce_interval.__set = ((void *)0); ce_interval.__unset = ((void *)0); ce_interval.__isset = ((void *)0); ce_interval.serialize_func = ((void *)0); ce_interval.unserialize_func = ((void *)0); ce_interval.serialize = ((void *)0); ce_interval.unserialize = ((void *)0); ce_interval.parent = ((void *)0); ce_interval.num_interfaces = 0; ce_interval.traits = ((void *)0); ce_interval.num_traits = 0; ce_interval.trait_aliases = ((void *)0); ce_interval.trait_precedences = ((void *)0); ce_interval.interfaces = ((void *)0); ce_interval.get_iterator = ((void *)0); ce_interval.iterator_funcs.funcs = ((void *)0); ce_interval.info.internal.module = ((void *)0); ce_interval.info.internal.builtin_functions = date_funcs_interval; };
 ce_interval.create_object = date_object_new_interval;
 date_ce_interval = zend_register_internal_class_ex(&ce_interval, ((void *)0), ((void *)0) );
 memcpy(&date_object_handlers_interval, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 date_object_handlers_interval.clone_obj = date_object_clone_interval;
 date_object_handlers_interval.read_property = date_interval_read_property;
 date_object_handlers_interval.write_property = date_interval_write_property;
 date_object_handlers_interval.get_properties = date_object_get_properties_interval;
 date_object_handlers_interval.get_property_ptr_ptr = ((void *)0);

 { const char *cl_name = "DatePeriod"; int _len = sizeof("DatePeriod")-1; ce_period.name = zend_new_interned_string(cl_name, _len+1, 0 ); if (ce_period.name == cl_name) { ce_period.name = zend_strndup(cl_name, _len); } ce_period.name_length = _len; ce_period.constructor = ((void *)0); ce_period.destructor = ((void *)0); ce_period.clone = ((void *)0); ce_period.serialize = ((void *)0); ce_period.unserialize = ((void *)0); ce_period.create_object = ((void *)0); ce_period.interface_gets_implemented = ((void *)0); ce_period.get_static_method = ((void *)0); ce_period.__call = ((void *)0); ce_period.__callstatic = ((void *)0); ce_period.__tostring = ((void *)0); ce_period.__get = ((void *)0); ce_period.__set = ((void *)0); ce_period.__unset = ((void *)0); ce_period.__isset = ((void *)0); ce_period.serialize_func = ((void *)0); ce_period.unserialize_func = ((void *)0); ce_period.serialize = ((void *)0); ce_period.unserialize = ((void *)0); ce_period.parent = ((void *)0); ce_period.num_interfaces = 0; ce_period.traits = ((void *)0); ce_period.num_traits = 0; ce_period.trait_aliases = ((void *)0); ce_period.trait_precedences = ((void *)0); ce_period.interfaces = ((void *)0); ce_period.get_iterator = ((void *)0); ce_period.iterator_funcs.funcs = ((void *)0); ce_period.info.internal.module = ((void *)0); ce_period.info.internal.builtin_functions = date_funcs_period; };
 ce_period.create_object = date_object_new_period;
 date_ce_period = zend_register_internal_class_ex(&ce_period, ((void *)0), ((void *)0) );
 date_ce_period->get_iterator = date_object_period_get_iterator;
 date_ce_period->iterator_funcs.funcs = &date_period_it_funcs;
 zend_class_implements(date_ce_period , 1, zend_ce_traversable);
 memcpy(&date_object_handlers_period, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 date_object_handlers_period.clone_obj = date_object_clone_period;




 zend_declare_class_constant_long(date_ce_period, "EXCLUDE_START_DATE", sizeof("EXCLUDE_START_DATE")-1, 0x0001 );;
}

static inline zend_object_value date_object_new_date_ex(zend_class_entry *class_type, php_date_obj **ptr )
{
 php_date_obj *intern;
 zend_object_value retval;

 intern = _emalloc((sizeof(php_date_obj)) );
 memset(intern, 0, sizeof(php_date_obj));
 if (ptr) {
  *ptr = intern;
 }

 zend_object_std_init(&intern->std, class_type );
 object_properties_init(&intern->std, class_type);

 retval.handle = zend_objects_store_put(intern, (zend_objects_store_dtor_t)zend_objects_destroy_object, (zend_objects_free_object_storage_t) date_object_free_storage_date, ((void *)0) );
 retval.handlers = &date_object_handlers_date;

 return retval;
}

static zend_object_value date_object_new_date(zend_class_entry *class_type )
{
 return date_object_new_date_ex(class_type, ((void *)0) );
}

static zend_object_value date_object_clone_date(zval *this_ptr )
{
 php_date_obj *new_obj = ((void *)0);
 php_date_obj *old_obj = (php_date_obj *) zend_object_store_get_object(this_ptr );
 zend_object_value new_ov = date_object_new_date_ex(old_obj->std.ce, &new_obj );

 zend_objects_clone_members(&new_obj->std, new_ov, &old_obj->std, (*this_ptr).value.obj.handle );
 if (!old_obj->time) {
  return new_ov;
 }


 new_obj->time = timelib_time_ctor();
 *new_obj->time = *old_obj->time;
 if (old_obj->time->tz_abbr) {
  new_obj->time->tz_abbr = (__extension__ (__builtin_constant_p (old_obj->time->tz_abbr) && ((size_t)(const void *)((old_obj->time->tz_abbr) + 1) - (size_t)(const void *)(old_obj->time->tz_abbr) == 1) ? (((__const char *) (old_obj->time->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (old_obj->time->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, old_obj->time->tz_abbr, __len); __retval; })) : __strdup (old_obj->time->tz_abbr)));
 }
 if (old_obj->time->tz_info) {
  new_obj->time->tz_info = old_obj->time->tz_info;
 }

 return new_ov;
}

static int date_object_compare_date(zval *d1, zval *d2 )
{
 if ((*d1).type == 5 && (*d2).type == 5 &&
  instanceof_function(zend_get_class_entry(&(*d1) ), date_ce_date ) &&
  instanceof_function(zend_get_class_entry(&(*d2) ), date_ce_date )) {
  php_date_obj *o1 = zend_object_store_get_object(d1 );
  php_date_obj *o2 = zend_object_store_get_object(d2 );

  if (!o1->time->sse_uptodate) {
   timelib_update_ts(o1->time, o1->time->tz_info);
  }
  if (!o2->time->sse_uptodate) {
   timelib_update_ts(o2->time, o2->time->tz_info);
  }

  return (o1->time->sse == o2->time->sse) ? 0 : ((o1->time->sse < o2->time->sse) ? -1 : 1);
 }

 return 1;
}

static HashTable *date_object_get_properties(zval *object )
{
 HashTable *props;
 zval *zv;
 php_date_obj *dateobj;


 dateobj = (php_date_obj *) zend_object_store_get_object(object );

 props = zend_std_get_properties(object );

 if (!dateobj->time || (gc_globals.gc_active)) {
  return props;
 }


 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;;
 do { const char *__s=(date_format("Y-m-d H:i:s", 12, dateobj->time, 1)); zval *__z = (zv); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
 _zend_hash_add_or_update(props, "date", 5, &zv, sizeof(zval), ((void *)0), (1<<0) );


 if (dateobj->time->is_localtime) {
  do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;;
  { zval *__z = (zv); (*__z).value.lval = dateobj->time->zone_type; (*__z).type = 1; };
  _zend_hash_add_or_update(props, "timezone_type", 14, &zv, sizeof(zval), ((void *)0), (1<<0) );

  do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;;
  switch (dateobj->time->zone_type) {
   case 3:
    do { const char *__s=(dateobj->time->tz_info->name); zval *__z = (zv); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
    break;
   case 1: {
    char *tmpstr = _emalloc((sizeof("UTC+05:00")) );
    timelib_sll utc_offset = dateobj->time->z;

    ap_php_snprintf(tmpstr, sizeof("+05:00"), "%c%02d:%02d",
     utc_offset > 0 ? '-' : '+',
     abs(utc_offset / 60),
     abs((utc_offset % 60)));

    do { const char *__s=(tmpstr); zval *__z = (zv); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
    }
    break;
   case 2:
    do { const char *__s=(dateobj->time->tz_abbr); zval *__z = (zv); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
    break;
  }
  _zend_hash_add_or_update(props, "timezone", 9, &zv, sizeof(zval), ((void *)0), (1<<0) );
 }

 return props;
}

static inline zend_object_value date_object_new_timezone_ex(zend_class_entry *class_type, php_timezone_obj **ptr )
{
 php_timezone_obj *intern;
 zend_object_value retval;

 intern = _emalloc((sizeof(php_timezone_obj)) );
 memset(intern, 0, sizeof(php_timezone_obj));
 if (ptr) {
  *ptr = intern;
 }

 zend_object_std_init(&intern->std, class_type );
 object_properties_init(&intern->std, class_type);

 retval.handle = zend_objects_store_put(intern, (zend_objects_store_dtor_t)zend_objects_destroy_object, (zend_objects_free_object_storage_t) date_object_free_storage_timezone, ((void *)0) );
 retval.handlers = &date_object_handlers_timezone;

 return retval;
}

static zend_object_value date_object_new_timezone(zend_class_entry *class_type )
{
 return date_object_new_timezone_ex(class_type, ((void *)0) );
}

static zend_object_value date_object_clone_timezone(zval *this_ptr )
{
 php_timezone_obj *new_obj = ((void *)0);
 php_timezone_obj *old_obj = (php_timezone_obj *) zend_object_store_get_object(this_ptr );
 zend_object_value new_ov = date_object_new_timezone_ex(old_obj->std.ce, &new_obj );

 zend_objects_clone_members(&new_obj->std, new_ov, &old_obj->std, (*this_ptr).value.obj.handle );
 if (!old_obj->initialized) {
  return new_ov;
 }

 new_obj->type = old_obj->type;
 new_obj->initialized = 1;
 switch (new_obj->type) {
  case 3:
   new_obj->tzi.tz = old_obj->tzi.tz;
   break;
  case 1:
   new_obj->tzi.utc_offset = old_obj->tzi.utc_offset;
   break;
  case 2:
   new_obj->tzi.z.utc_offset = old_obj->tzi.z.utc_offset;
   new_obj->tzi.z.dst = old_obj->tzi.z.dst;
   new_obj->tzi.z.abbr = old_obj->tzi.z.abbr;
   break;
 }

 return new_ov;
}

static inline zend_object_value date_object_new_interval_ex(zend_class_entry *class_type, php_interval_obj **ptr )
{
 php_interval_obj *intern;
 zend_object_value retval;

 intern = _emalloc((sizeof(php_interval_obj)) );
 memset(intern, 0, sizeof(php_interval_obj));
 if (ptr) {
  *ptr = intern;
 }

 zend_object_std_init(&intern->std, class_type );
 object_properties_init(&intern->std, class_type);

 retval.handle = zend_objects_store_put(intern, (zend_objects_store_dtor_t)zend_objects_destroy_object, (zend_objects_free_object_storage_t) date_object_free_storage_interval, ((void *)0) );
 retval.handlers = &date_object_handlers_interval;

 return retval;
}

static zend_object_value date_object_new_interval(zend_class_entry *class_type )
{
 return date_object_new_interval_ex(class_type, ((void *)0) );
}

static zend_object_value date_object_clone_interval(zval *this_ptr )
{
 php_interval_obj *new_obj = ((void *)0);
 php_interval_obj *old_obj = (php_interval_obj *) zend_object_store_get_object(this_ptr );
 zend_object_value new_ov = date_object_new_interval_ex(old_obj->std.ce, &new_obj );

 zend_objects_clone_members(&new_obj->std, new_ov, &old_obj->std, (*this_ptr).value.obj.handle );


 return new_ov;
}

static HashTable *date_object_get_properties_interval(zval *object )
{
 HashTable *props;
 zval *zv;
 php_interval_obj *intervalobj;


 intervalobj = (php_interval_obj *) zend_object_store_get_object(object );

 props = zend_std_get_properties(object );

 if (!intervalobj->initialized || (gc_globals.gc_active)) {
  return props;
 }






 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->y; (*__z).type = 1; }; _zend_hash_add_or_update(props, "y", strlen("y") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->m; (*__z).type = 1; }; _zend_hash_add_or_update(props, "m", strlen("m") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->d; (*__z).type = 1; }; _zend_hash_add_or_update(props, "d", strlen("d") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->h; (*__z).type = 1; }; _zend_hash_add_or_update(props, "h", strlen("h") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->i; (*__z).type = 1; }; _zend_hash_add_or_update(props, "i", strlen("i") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->s; (*__z).type = 1; }; _zend_hash_add_or_update(props, "s", strlen("s") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->invert; (*__z).type = 1; }; _zend_hash_add_or_update(props, "invert", strlen("invert") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 if (intervalobj->diff->days != -99999) {
  do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;; { zval *__z = (zv); (*__z).value.lval = intervalobj->diff->days; (*__z).type = 1; }; _zend_hash_add_or_update(props, "days", strlen("days") + 1, &zv, sizeof(zval), ((void *)0), (1<<0) );;
 } else {
  do { (zv) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(zv))->u.buffered = ((void *)0); } while (0); (zv)->refcount__gc = 1; (zv)->is_ref__gc = 0;;;
  do { zval *__z = (zv); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0);
  _zend_hash_add_or_update(props, "days", 5, &zv, sizeof(zval), ((void *)0), (1<<0) );
 }

 return props;
}

static inline zend_object_value date_object_new_period_ex(zend_class_entry *class_type, php_period_obj **ptr )
{
 php_period_obj *intern;
 zend_object_value retval;

 intern = _emalloc((sizeof(php_period_obj)) );
 memset(intern, 0, sizeof(php_period_obj));
 if (ptr) {
  *ptr = intern;
 }

 zend_object_std_init(&intern->std, class_type );
 object_properties_init(&intern->std, class_type);

 retval.handle = zend_objects_store_put(intern, (zend_objects_store_dtor_t)zend_objects_destroy_object, (zend_objects_free_object_storage_t) date_object_free_storage_period, ((void *)0) );
 retval.handlers = &date_object_handlers_period;

 return retval;
}

static zend_object_value date_object_new_period(zend_class_entry *class_type )
{
 return date_object_new_period_ex(class_type, ((void *)0) );
}

static zend_object_value date_object_clone_period(zval *this_ptr )
{
 php_period_obj *new_obj = ((void *)0);
 php_period_obj *old_obj = (php_period_obj *) zend_object_store_get_object(this_ptr );
 zend_object_value new_ov = date_object_new_period_ex(old_obj->std.ce, &new_obj );

 zend_objects_clone_members(&new_obj->std, new_ov, &old_obj->std, (*this_ptr).value.obj.handle );


 return new_ov;
}

static void date_object_free_storage_date(void *object )
{
 php_date_obj *intern = (php_date_obj *)object;

 if (intern->time) {
  timelib_time_dtor(intern->time);
 }

 zend_object_std_dtor(&intern->std );
 _efree((object) );
}

static void date_object_free_storage_timezone(void *object )
{
 php_timezone_obj *intern = (php_timezone_obj *)object;

 if (intern->type == 2) {
  free(intern->tzi.z.abbr);
 }
 zend_object_std_dtor(&intern->std );
 _efree((object) );
}

static void date_object_free_storage_interval(void *object )
{
 php_interval_obj *intern = (php_interval_obj *)object;

 timelib_rel_time_dtor(intern->diff);
 zend_object_std_dtor(&intern->std );
 _efree((object) );
}

static void date_object_free_storage_period(void *object )
{
 php_period_obj *intern = (php_period_obj *)object;

 if (intern->start) {
  timelib_time_dtor(intern->start);
 }

 if (intern->current) {
  timelib_time_dtor(intern->current);
 }

 if (intern->end) {
  timelib_time_dtor(intern->end);
 }

 timelib_rel_time_dtor(intern->interval);
 zend_object_std_dtor(&intern->std );
 _efree((object) );
}


__attribute__ ((visibility("default"))) zval *php_date_instantiate(zend_class_entry *pce, zval *object )
{
 (*object).type = 5;
 _object_init_ex((object), (pce) );
 zval_set_refcount_p(object, 1);
 zval_unset_isref_p(object);
 return object;
}



static void update_errors_warnings(timelib_error_container *last_errors )
{
 if ((date_globals.last_errors)) {
  timelib_error_container_dtor((date_globals.last_errors));
  (date_globals.last_errors) = ((void *)0);
 }
 (date_globals.last_errors) = last_errors;
}

__attribute__ ((visibility("default"))) int php_date_initialize(php_date_obj *dateobj, char *time_str, int time_str_len, char *format, zval *timezone_object, int ctor )
{
 timelib_time *now;
 timelib_tzinfo *tzi = ((void *)0);
 timelib_error_container *err = ((void *)0);
 int type = 3, new_dst;
 char *new_abbr;
 timelib_sll new_offset;

 if (dateobj->time) {
  timelib_time_dtor(dateobj->time);
 }
 if (format) {
  dateobj->time = timelib_parse_from_format(format, time_str_len ? time_str : "", time_str_len ? time_str_len : 0, &err, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 } else {
  dateobj->time = timelib_strtotime(time_str_len ? time_str : "now", time_str_len ? time_str_len : sizeof("now") -1, &err, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 }


 update_errors_warnings(err );


 if (ctor && err && err->error_count) {

  php_error_docref0(((void *)0) , (1<<1L), "Failed to parse time string (%s) at position %d (%c): %s", time_str,
   err->error_messages[0].position, err->error_messages[0].character, err->error_messages[0].message);
 }
 if (err && err->error_count) {
  return 0;
 }

 if (timezone_object) {
  php_timezone_obj *tzobj;

  tzobj = (php_timezone_obj *) zend_object_store_get_object(timezone_object );
  switch (tzobj->type) {
   case 3:
    tzi = tzobj->tzi.tz;
    break;
   case 1:
    new_offset = tzobj->tzi.utc_offset;
    break;
   case 2:
    new_offset = tzobj->tzi.z.utc_offset;
    new_dst = tzobj->tzi.z.dst;
    new_abbr = (__extension__ (__builtin_constant_p (tzobj->tzi.z.abbr) && ((size_t)(const void *)((tzobj->tzi.z.abbr) + 1) - (size_t)(const void *)(tzobj->tzi.z.abbr) == 1) ? (((__const char *) (tzobj->tzi.z.abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (tzobj->tzi.z.abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, tzobj->tzi.z.abbr, __len); __retval; })) : __strdup (tzobj->tzi.z.abbr)));
    break;
  }
  type = tzobj->type;
 } else if (dateobj->time->tz_info) {
  tzi = dateobj->time->tz_info;
 } else {
  tzi = get_timezone_info();
 }

 now = timelib_time_ctor();
 now->zone_type = type;
 switch (type) {
  case 3:
   now->tz_info = tzi;
   break;
  case 1:
   now->z = new_offset;
   break;
  case 2:
   now->z = new_offset;
   now->dst = new_dst;
   now->tz_abbr = new_abbr;
   break;
 }
 timelib_unixtime2local(now, (timelib_sll) time(((void *)0)));

 timelib_fill_holes(dateobj->time, now, 0x02);
 timelib_update_ts(dateobj->time, tzi);

 dateobj->time->have_relative = 0;

 timelib_time_dtor(now);

 return 1;
}




void zif_date_create(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *timezone_object = ((void *)0);
 char *time_str = ((void *)0);
 int time_str_len = 0;

 if (zend_parse_parameters((ht) , "|sO!", &time_str, &time_str_len, &timezone_object, date_ce_timezone) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 php_date_instantiate(date_ce_date, return_value );
 if (!php_date_initialize(zend_object_store_get_object(return_value ), time_str, time_str_len, ((void *)0), timezone_object, 0 )) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
}





void zif_date_create_from_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *timezone_object = ((void *)0);
 char *time_str = ((void *)0), *format_str = ((void *)0);
 int time_str_len = 0, format_str_len = 0;

 if (zend_parse_parameters((ht) , "ss|O", &format_str, &format_str_len, &time_str, &time_str_len, &timezone_object, date_ce_timezone) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 php_date_instantiate(date_ce_date, return_value );
 if (!php_date_initialize(zend_object_store_get_object(return_value ), time_str, time_str_len, format_str, timezone_object, 0 )) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
}





void zim_DateTime___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *timezone_object = ((void *)0);
 char *time_str = ((void *)0);
 int time_str_len = 0;
 zend_error_handling error_handling;

 zend_replace_error_handling(EH_THROW, ((void *)0), &error_handling );
 if (0 == zend_parse_parameters((ht) , "|sO!", &time_str, &time_str_len, &timezone_object, date_ce_timezone)) {
  php_date_initialize(zend_object_store_get_object((this_ptr) ), time_str, time_str_len, ((void *)0), timezone_object, 1 );
 }
 zend_restore_error_handling(&error_handling );
}


static int php_date_initialize_from_hash(zval **return_value, php_date_obj **dateobj, HashTable *myht )
{
 zval **z_date = ((void *)0);
 zval **z_timezone = ((void *)0);
 zval **z_timezone_type = ((void *)0);
 zval *tmp_obj = ((void *)0);
 timelib_tzinfo *tzi;
 php_timezone_obj *tzobj;

 if (zend_hash_find(myht, "date", 5, (void**) &z_date) == 0) {
  if ((*z_date)->type != 6) { _convert_to_string((*z_date) ); };
  if (zend_hash_find(myht, "timezone_type", 14, (void**) &z_timezone_type) == 0) {
   convert_to_long(*z_timezone_type);
   if (zend_hash_find(myht, "timezone", 9, (void**) &z_timezone) == 0) {
    if ((*z_timezone)->type != 6) { _convert_to_string((*z_timezone) ); };

    switch ((**z_timezone_type).value.lval) {
     case 1:
     case 2: {
      char *tmp = _emalloc(((**z_date).value.str.len + (**z_timezone).value.str.len + 2) );
      ap_php_snprintf(tmp, (**z_date).value.str.len + (**z_timezone).value.str.len + 2, "%s %s", (**z_date).value.str.val, (**z_timezone).value.str.val);
      php_date_initialize(*dateobj, tmp, (**z_date).value.str.len + (**z_timezone).value.str.len + 1, ((void *)0), ((void *)0), 0 );
      _efree((tmp) );
      return 1;
     }

     case 3:
      if ((*z_timezone)->type != 6) { _convert_to_string((*z_timezone) ); };

      tzi = php_date_parse_tzfile((**z_timezone).value.str.val, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db() );

      do { (tmp_obj) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(tmp_obj))->u.buffered = ((void *)0); } while (0); *tmp_obj = zval_used_for_init;;;
      tzobj = zend_object_store_get_object(php_date_instantiate(date_ce_timezone, tmp_obj ) );
      tzobj->type = 3;
      tzobj->tzi.tz = tzi;
      tzobj->initialized = 1;

      php_date_initialize(*dateobj, (**z_date).value.str.val, (**z_date).value.str.len, ((void *)0), tmp_obj, 0 );
      _zval_ptr_dtor((&tmp_obj) );
      return 1;
    }
   }
  }
 }
 return 0;
}



void zim_DateTime___set_state(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_date_obj *dateobj;
 zval *array;
 HashTable *myht;

 if (zend_parse_parameters((ht) , "a", &array) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 myht = ((*array).type==4 ? (*array).value.ht : (((*array).type==5 ? (*array).value.obj.handlers->get_properties((array) ) : ((void *)0))));

 php_date_instantiate(date_ce_date, return_value );
 dateobj = (php_date_obj *) zend_object_store_get_object(return_value );
 php_date_initialize_from_hash(&return_value, &dateobj, myht );
}




void zim_DateTime___wakeup(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object = (this_ptr);
 php_date_obj *dateobj;
 HashTable *myht;

 dateobj = (php_date_obj *) zend_object_store_get_object(object );

 myht = ((*object)).value.obj.handlers->get_properties(&(*object) );

 php_date_initialize_from_hash(&return_value, &dateobj, myht );
}



static void zval_from_error_container(zval *z, timelib_error_container *error)
{
 int i;
 zval *element;

 add_assoc_long_ex(z, "warning_count", strlen("warning_count")+1, error->warning_count);
 do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;;
 _array_init((element), 0 );
 for (i = 0; i < error->warning_count; i++) {
  add_index_string(element, error->warning_messages[i].position, error->warning_messages[i].message, 1);
 }
 add_assoc_zval_ex(z, "warnings", strlen("warnings")+1, element);

 add_assoc_long_ex(z, "error_count", strlen("error_count")+1, error->error_count);
 do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;;
 _array_init((element), 0 );
 for (i = 0; i < error->error_count; i++) {
  add_index_string(element, error->error_messages[i].position, error->error_messages[i].message, 1);
 }
 add_assoc_zval_ex(z, "errors", strlen("errors")+1, element);
}




void zif_date_get_last_errors(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 if ((date_globals.last_errors)) {
  _array_init((return_value), 0 );
  zval_from_error_container(return_value, (date_globals.last_errors));
 } else {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
}


void php_date_do_return_parsed_time(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , timelib_time *parsed_time, struct timelib_error_container *error)
{
 zval *element;

 _array_init((return_value), 0 );






 if (parsed_time->y == -99999) { add_assoc_bool_ex(return_value, "year", strlen("year")+1, 0); } else { add_assoc_long_ex(return_value, "year", strlen("year")+1, parsed_time->y); };
 if (parsed_time->m == -99999) { add_assoc_bool_ex(return_value, "month", strlen("month")+1, 0); } else { add_assoc_long_ex(return_value, "month", strlen("month")+1, parsed_time->m); };
 if (parsed_time->d == -99999) { add_assoc_bool_ex(return_value, "day", strlen("day")+1, 0); } else { add_assoc_long_ex(return_value, "day", strlen("day")+1, parsed_time->d); };
 if (parsed_time->h == -99999) { add_assoc_bool_ex(return_value, "hour", strlen("hour")+1, 0); } else { add_assoc_long_ex(return_value, "hour", strlen("hour")+1, parsed_time->h); };
 if (parsed_time->i == -99999) { add_assoc_bool_ex(return_value, "minute", strlen("minute")+1, 0); } else { add_assoc_long_ex(return_value, "minute", strlen("minute")+1, parsed_time->i); };
 if (parsed_time->s == -99999) { add_assoc_bool_ex(return_value, "second", strlen("second")+1, 0); } else { add_assoc_long_ex(return_value, "second", strlen("second")+1, parsed_time->s); };

 if (parsed_time->f == -99999) {
  add_assoc_bool_ex(return_value, "fraction", strlen("fraction")+1, 0);
 } else {
  add_assoc_double_ex(return_value, "fraction", strlen("fraction")+1, parsed_time->f);
 }

 zval_from_error_container(return_value, error);

 timelib_error_container_dtor(error);

 add_assoc_bool_ex(return_value, "is_localtime", strlen("is_localtime")+1, parsed_time->is_localtime);

 if (parsed_time->is_localtime) {
  if (parsed_time->zone_type == -99999) { add_assoc_bool_ex(return_value, "zone_type", strlen("zone_type")+1, 0); } else { add_assoc_long_ex(return_value, "zone_type", strlen("zone_type")+1, parsed_time->zone_type); };
  switch (parsed_time->zone_type) {
   case 1:
    if (parsed_time->z == -99999) { add_assoc_bool_ex(return_value, "zone", strlen("zone")+1, 0); } else { add_assoc_long_ex(return_value, "zone", strlen("zone")+1, parsed_time->z); };
    add_assoc_bool_ex(return_value, "is_dst", strlen("is_dst")+1, parsed_time->dst);
    break;
   case 3:
    if (parsed_time->tz_abbr) {
     add_assoc_string_ex(return_value, "tz_abbr", strlen("tz_abbr")+1, parsed_time->tz_abbr, 1);
    }
    if (parsed_time->tz_info) {
     add_assoc_string_ex(return_value, "tz_id", strlen("tz_id")+1, parsed_time->tz_info->name, 1);
    }
    break;
   case 2:
    if (parsed_time->z == -99999) { add_assoc_bool_ex(return_value, "zone", strlen("zone")+1, 0); } else { add_assoc_long_ex(return_value, "zone", strlen("zone")+1, parsed_time->z); };
    add_assoc_bool_ex(return_value, "is_dst", strlen("is_dst")+1, parsed_time->dst);
    add_assoc_string_ex(return_value, "tz_abbr", strlen("tz_abbr")+1, parsed_time->tz_abbr, 1);
    break;
  }
 }
 if (parsed_time->have_relative) {
  do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;;
  _array_init((element), 0 );
  add_assoc_long_ex(element, "year", strlen("year")+1, parsed_time->relative.y);
  add_assoc_long_ex(element, "month", strlen("month")+1, parsed_time->relative.m);
  add_assoc_long_ex(element, "day", strlen("day")+1, parsed_time->relative.d);
  add_assoc_long_ex(element, "hour", strlen("hour")+1, parsed_time->relative.h);
  add_assoc_long_ex(element, "minute", strlen("minute")+1, parsed_time->relative.i);
  add_assoc_long_ex(element, "second", strlen("second")+1, parsed_time->relative.s);
  if (parsed_time->relative.have_weekday_relative) {
   add_assoc_long_ex(element, "weekday", strlen("weekday")+1, parsed_time->relative.weekday);
  }
  if (parsed_time->relative.have_special_relative && (parsed_time->relative.special.type == 0x01)) {
   add_assoc_long_ex(element, "weekdays", strlen("weekdays")+1, parsed_time->relative.special.amount);
  }
  if (parsed_time->relative.first_last_day_of) {
   add_assoc_bool_ex(element, parsed_time->relative.first_last_day_of == 1 ? "first_day_of_month" : "last_day_of_month", strlen(parsed_time->relative.first_last_day_of == 1 ? "first_day_of_month" : "last_day_of_month")+1, 1);
  }
  add_assoc_zval_ex(return_value, "relative", strlen("relative")+1, element);
 }
 timelib_time_dtor(parsed_time);
}




void zif_date_parse(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *date;
 int date_len;
 struct timelib_error_container *error;
 timelib_time *parsed_time;

 if (zend_parse_parameters((ht) , "s", &date, &date_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 parsed_time = timelib_strtotime(date, date_len, &error, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 php_date_do_return_parsed_time(ht, return_value, return_value_ptr, this_ptr, return_value_used , parsed_time, error);
}





void zif_date_parse_from_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *date, *format;
 int date_len, format_len;
 struct timelib_error_container *error;
 timelib_time *parsed_time;

 if (zend_parse_parameters((ht) , "ss", &format, &format_len, &date, &date_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 parsed_time = timelib_parse_from_format(format, date, date_len, &error, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 php_date_do_return_parsed_time(ht, return_value, return_value_ptr, this_ptr, return_value_used , parsed_time, error);
}





void zif_date_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 char *format;
 int format_len;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Os", &object, date_ce_date, &format, &format_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 { do { const char *__s=(date_format(format, format_len, dateobj->time, dateobj->time->is_localtime)); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
}





void zif_date_modify(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 char *modify;
 int modify_len;
 timelib_time *tmp_time;
 timelib_error_container *err = ((void *)0);

 if (zend_parse_method_parameters((ht) , (this_ptr), "Os", &object, date_ce_date, &modify, &modify_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };

 tmp_time = timelib_strtotime(modify, modify_len, &err, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);


 update_errors_warnings(err );
 if (err && err->error_count) {

  php_error_docref0(((void *)0) , (1<<1L), "Failed to parse time string (%s) at position %d (%c): %s", modify,
   err->error_messages[0].position, err->error_messages[0].character, err->error_messages[0].message);
  timelib_time_dtor(tmp_time);
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 memcpy(&dateobj->time->relative, &tmp_time->relative, sizeof(struct timelib_rel_time));
 dateobj->time->have_relative = tmp_time->have_relative;
 dateobj->time->sse_uptodate = 0;

 if (tmp_time->y != -99999) {
  dateobj->time->y = tmp_time->y;
 }
 if (tmp_time->m != -99999) {
  dateobj->time->m = tmp_time->m;
 }
 if (tmp_time->d != -99999) {
  dateobj->time->d = tmp_time->d;
 }

 if (tmp_time->h != -99999) {
  dateobj->time->h = tmp_time->h;
  if (tmp_time->i != -99999) {
   dateobj->time->i = tmp_time->i;
   if (tmp_time->s != -99999) {
    dateobj->time->s = tmp_time->s;
   } else {
    dateobj->time->s = 0;
   }
  } else {
   dateobj->time->i = 0;
   dateobj->time->s = 0;
  }
 }
 timelib_time_dtor(tmp_time);

 timelib_update_ts(dateobj->time, ((void *)0));
 timelib_update_from_sse(dateobj->time);
 dateobj->time->have_relative = 0;

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_add(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object, *interval;
 php_date_obj *dateobj;
 php_interval_obj *intobj;
 int bias = 1;

 if (zend_parse_method_parameters((ht) , (this_ptr), "OO", &object, date_ce_date, &interval, date_ce_interval) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 intobj = (php_interval_obj *) zend_object_store_get_object(interval );
 if (!(intobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateInterval" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };


 if (intobj->diff->have_weekday_relative || intobj->diff->have_special_relative) {
  memcpy(&dateobj->time->relative, intobj->diff, sizeof(struct timelib_rel_time));
 } else {
  if (intobj->diff->invert) {
   bias = -1;
  }
  memset(&dateobj->time->relative, 0, sizeof(struct timelib_rel_time));
  dateobj->time->relative.y = intobj->diff->y * bias;
  dateobj->time->relative.m = intobj->diff->m * bias;
  dateobj->time->relative.d = intobj->diff->d * bias;
  dateobj->time->relative.h = intobj->diff->h * bias;
  dateobj->time->relative.i = intobj->diff->i * bias;
  dateobj->time->relative.s = intobj->diff->s * bias;
 }
 dateobj->time->have_relative = 1;
 dateobj->time->sse_uptodate = 0;

 timelib_update_ts(dateobj->time, ((void *)0));
 timelib_update_from_sse(dateobj->time);
 dateobj->time->have_relative = 0;

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_sub(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object, *interval;
 php_date_obj *dateobj;
 php_interval_obj *intobj;
 int bias = 1;

 if (zend_parse_method_parameters((ht) , (this_ptr), "OO", &object, date_ce_date, &interval, date_ce_interval) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 intobj = (php_interval_obj *) zend_object_store_get_object(interval );
 if (!(intobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateInterval" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };

 if (intobj->diff->have_special_relative) {
  php_error_docref0(((void *)0) , (1<<1L), "Only non-special relative time specifications are supported for subtraction");
  return;
 }

 if (intobj->diff->invert) {
  bias = -1;
 }

 memset(&dateobj->time->relative, 0, sizeof(struct timelib_rel_time));
 dateobj->time->relative.y = 0 - (intobj->diff->y * bias);
 dateobj->time->relative.m = 0 - (intobj->diff->m * bias);
 dateobj->time->relative.d = 0 - (intobj->diff->d * bias);
 dateobj->time->relative.h = 0 - (intobj->diff->h * bias);
 dateobj->time->relative.i = 0 - (intobj->diff->i * bias);
 dateobj->time->relative.s = 0 - (intobj->diff->s * bias);
 dateobj->time->have_relative = 1;
 dateobj->time->sse_uptodate = 0;

 timelib_update_ts(dateobj->time, ((void *)0));
 timelib_update_from_sse(dateobj->time);

 dateobj->time->have_relative = 0;

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_timezone_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 php_timezone_obj *tzobj;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O", &object, date_ce_date) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 if (dateobj->time->is_localtime ) {
  php_date_instantiate(date_ce_timezone, return_value );
  tzobj = (php_timezone_obj *) zend_object_store_get_object(return_value );
  tzobj->initialized = 1;
  tzobj->type = dateobj->time->zone_type;
  switch (dateobj->time->zone_type) {
   case 3:
    tzobj->tzi.tz = dateobj->time->tz_info;
    break;
   case 1:
    tzobj->tzi.utc_offset = dateobj->time->z;
    break;
   case 2:
    tzobj->tzi.z.utc_offset = dateobj->time->z;
    tzobj->tzi.z.dst = dateobj->time->dst;
    tzobj->tzi.z.abbr = (__extension__ (__builtin_constant_p (dateobj->time->tz_abbr) && ((size_t)(const void *)((dateobj->time->tz_abbr) + 1) - (size_t)(const void *)(dateobj->time->tz_abbr) == 1) ? (((__const char *) (dateobj->time->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (dateobj->time->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, dateobj->time->tz_abbr, __len); __retval; })) : __strdup (dateobj->time->tz_abbr)));
    break;
  }
 } else {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
}





void zif_date_timezone_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 zval *timezone_object;
 php_date_obj *dateobj;
 php_timezone_obj *tzobj;

 if (zend_parse_method_parameters((ht) , (this_ptr), "OO", &object, date_ce_date, &timezone_object, date_ce_timezone) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 tzobj = (php_timezone_obj *) zend_object_store_get_object(timezone_object );
 if (tzobj->type != 3) {
  php_error_docref0(((void *)0) , (1<<1L), "Can only do this for zones with ID for now");
  return;
 }
 timelib_set_timezone(dateobj->time, tzobj->tzi.tz);
 timelib_unixtime2local(dateobj->time, dateobj->time->sse);

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_offset_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 timelib_time_offset *offset;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O", &object, date_ce_date) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 if (dateobj->time->is_localtime ) {
  switch (dateobj->time->zone_type) {
   case 3:
    offset = timelib_get_time_zone_info(dateobj->time->sse, dateobj->time->tz_info);
    { zval *__z = (return_value); (*__z).value.lval = offset->offset; (*__z).type = 1; };
    timelib_time_offset_dtor(offset);
    break;
   case 1:
    { zval *__z = (return_value); (*__z).value.lval = dateobj->time->z * -60; (*__z).type = 1; };
    break;
   case 2:
    { zval *__z = (return_value); (*__z).value.lval = (dateobj->time->z - (60 * dateobj->time->dst)) * -60; (*__z).type = 1; };
    break;
  }
  return;
 } else {
  { { zval *__z = (return_value); (*__z).value.lval = 0; (*__z).type = 1; }; return; };
 }
}





void zif_date_time_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 long h, i, s = 0;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Oll|l", &object, date_ce_date, &h, &i, &s) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 dateobj->time->h = h;
 dateobj->time->i = i;
 dateobj->time->s = s;
 timelib_update_ts(dateobj->time, ((void *)0));

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_date_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 long y, m, d;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Olll", &object, date_ce_date, &y, &m, &d) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 dateobj->time->y = y;
 dateobj->time->m = m;
 dateobj->time->d = d;
 timelib_update_ts(dateobj->time, ((void *)0));

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_isodate_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 long y, w, d = 1;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Oll|l", &object, date_ce_date, &y, &w, &d) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 dateobj->time->y = y;
 dateobj->time->m = 1;
 dateobj->time->d = 1;
 memset(&dateobj->time->relative, 0, sizeof(dateobj->time->relative));
 dateobj->time->relative.d = timelib_daynr_from_weeknr(y, w, d);
 dateobj->time->have_relative = 1;

 timelib_update_ts(dateobj->time, ((void *)0));

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_timestamp_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 long timestamp;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Ol", &object, date_ce_date, &timestamp) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 timelib_unixtime2local(dateobj->time, (timelib_sll)timestamp);
 timelib_update_ts(dateobj->time, ((void *)0));

 { { zend_uchar is_ref = zval_isref_p(return_value); zend_uint refcount = zval_refcount_p(return_value); do { (return_value)->value = (object)->value; (*return_value).type = (*object).type; } while (0); if (1) { _zval_copy_ctor((return_value) ); } if (0) { if (!1) { { (*object).type = 0; }; } _zval_ptr_dtor((&object) ); } zval_set_isref_to_p(return_value, is_ref); zval_set_refcount_p(return_value, refcount); }; return; };
}





void zif_date_timestamp_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_date_obj *dateobj;
 long timestamp;
 int error;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O", &object, date_ce_date) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj = (php_date_obj *) zend_object_store_get_object(object );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 timelib_update_ts(dateobj->time, ((void *)0));

 timestamp = timelib_date_to_int(dateobj->time, &error);
 if (error) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 } else {
  { zval *__z = (return_value); (*__z).value.lval = timestamp; (*__z).type = 1; };
 }
}





void zif_date_diff(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object1, *object2;
 php_date_obj *dateobj1, *dateobj2;
 php_interval_obj *interval;
 long absolute = 0;

 if (zend_parse_method_parameters((ht) , (this_ptr), "OO|l", &object1, date_ce_date, &object2, date_ce_date, &absolute) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 dateobj1 = (php_date_obj *) zend_object_store_get_object(object1 );
 dateobj2 = (php_date_obj *) zend_object_store_get_object(object2 );
 if (!(dateobj1->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 if (!(dateobj2->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 timelib_update_ts(dateobj1->time, ((void *)0));
 timelib_update_ts(dateobj2->time, ((void *)0));

 php_date_instantiate(date_ce_interval, return_value );
 interval = zend_object_store_get_object(return_value );
 interval->diff = timelib_diff(dateobj1->time, dateobj2->time);
 if (absolute) {
  interval->diff->invert = 0;
 }
 interval->initialized = 1;
}


static int timezone_initialize(timelib_tzinfo **tzi, char *tz )
{
 char *tzid;

 *tzi = ((void *)0);

 if ((tzid = timelib_timezone_id_from_abbr(tz, -1, 0))) {
  *tzi = php_date_parse_tzfile(tzid, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db() );
 } else {
  *tzi = php_date_parse_tzfile(tz, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db() );
 }

 if (*tzi) {
  return 0;
 } else {
  php_error_docref0(((void *)0) , (1<<1L), "Unknown or bad timezone (%s)", tz);
  return -1;
 }
}




void zif_timezone_open(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *tz;
 int tz_len;
 timelib_tzinfo *tzi = ((void *)0);
 php_timezone_obj *tzobj;

 if (zend_parse_parameters((ht) , "s", &tz, &tz_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 if (0 != timezone_initialize(&tzi, tz )) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzobj = zend_object_store_get_object(php_date_instantiate(date_ce_timezone, return_value ) );
 tzobj->type = 3;
 tzobj->tzi.tz = tzi;
 tzobj->initialized = 1;
}





void zim_DateTimeZone___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *tz;
 int tz_len;
 timelib_tzinfo *tzi = ((void *)0);
 php_timezone_obj *tzobj;
 zend_error_handling error_handling;

 zend_replace_error_handling(EH_THROW, ((void *)0), &error_handling );
 if (0 == zend_parse_parameters((ht) , "s", &tz, &tz_len)) {
  if (0 == timezone_initialize(&tzi, tz )) {
   tzobj = zend_object_store_get_object((this_ptr) );
   tzobj->type = 3;
   tzobj->tzi.tz = tzi;
   tzobj->initialized = 1;
  } else {
   { (*(this_ptr)).type = 0; };
  }
 }
 zend_restore_error_handling(&error_handling );
}





void zif_timezone_name_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_timezone_obj *tzobj;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O", &object, date_ce_timezone) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzobj = (php_timezone_obj *) zend_object_store_get_object(object );
 if (!(tzobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTimeZone" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };

 switch (tzobj->type) {
  case 3:
   { do { const char *__s=(tzobj->tzi.tz->name); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
   break;
  case 1: {
   char *tmpstr = _emalloc((sizeof("UTC+05:00")) );
   timelib_sll utc_offset = tzobj->tzi.utc_offset;

   ap_php_snprintf(tmpstr, sizeof("+05:00"), "%c%02d:%02d",
    utc_offset > 0 ? '-' : '+',
    abs(utc_offset / 60),
    abs((utc_offset % 60)));

   { do { const char *__s=(tmpstr); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
   }
   break;
  case 2:
   { do { const char *__s=(tzobj->tzi.z.abbr); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
   break;
 }
}





void zif_timezone_name_from_abbr(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *abbr;
 char *tzid;
 int abbr_len;
 long gmtoffset = -1;
 long isdst = -1;

 if (zend_parse_parameters((ht) , "s|ll", &abbr, &abbr_len, &gmtoffset, &isdst) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzid = timelib_timezone_id_from_abbr(abbr, gmtoffset, isdst);

 if (tzid) {
  { do { const char *__s=(tzid); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
 } else {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
}





void zif_timezone_offset_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object, *dateobject;
 php_timezone_obj *tzobj;
 php_date_obj *dateobj;
 timelib_time_offset *offset;

 if (zend_parse_method_parameters((ht) , (this_ptr), "OO", &object, date_ce_timezone, &dateobject, date_ce_date) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzobj = (php_timezone_obj *) zend_object_store_get_object(object );
 if (!(tzobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTimeZone" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 dateobj = (php_date_obj *) zend_object_store_get_object(dateobject );
 if (!(dateobj->time)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTime" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };

 switch (tzobj->type) {
  case 3:
   offset = timelib_get_time_zone_info(dateobj->time->sse, tzobj->tzi.tz);
   { zval *__z = (return_value); (*__z).value.lval = offset->offset; (*__z).type = 1; };
   timelib_time_offset_dtor(offset);
   break;
  case 1:
   { { zval *__z = (return_value); (*__z).value.lval = tzobj->tzi.utc_offset * -60; (*__z).type = 1; }; return; };
   break;
  case 2:
   { { zval *__z = (return_value); (*__z).value.lval = (tzobj->tzi.z.utc_offset - (tzobj->tzi.z.dst*60)) * -60; (*__z).type = 1; }; return; };
   break;
 }
}





void zif_timezone_transitions_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object, *element;
 php_timezone_obj *tzobj;
 unsigned int i, begin = 0, found;
 long timestamp_begin = (-2147483647L - 1L), timestamp_end = 2147483647L;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O|ll", &object, date_ce_timezone, &timestamp_begin, &timestamp_end) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzobj = (php_timezone_obj *) zend_object_store_get_object(object );
 if (!(tzobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTimeZone" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 if (tzobj->type != 3) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
# 3321 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
 _array_init((return_value), 0 );

 if (timestamp_begin == (-2147483647L - 1L)) {
  do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, timestamp_begin); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, timestamp_begin, 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[0].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[0].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[0].abbr_idx], 1); add_next_index_zval(return_value, element);;
  begin = 0;
  found = 1;
 } else {
  begin = 0;
  found = 0;
  if (tzobj->tzi.tz->timecnt > 0) {
   do {
    if (tzobj->tzi.tz->trans[begin] > timestamp_begin) {
     if (begin > 0) {
      do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, timestamp_begin); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, timestamp_begin, 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[begin - 1]].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[begin - 1]].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[begin - 1]].abbr_idx], 1); add_next_index_zval(return_value, element);;
     } else {
      do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, timestamp_begin); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, timestamp_begin, 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[0].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[0].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[0].abbr_idx], 1); add_next_index_zval(return_value, element);;
     }
     found = 1;
     break;
    }
    begin++;
   } while (begin < tzobj->tzi.tz->timecnt);
  }
 }

 if (!found) {
  if (tzobj->tzi.tz->timecnt > 0) {
   do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, timestamp_begin); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, timestamp_begin, 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[tzobj->tzi.tz->timecnt - 1]].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[tzobj->tzi.tz->timecnt - 1]].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[tzobj->tzi.tz->timecnt - 1]].abbr_idx], 1); add_next_index_zval(return_value, element);;
  } else {
   do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, timestamp_begin); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, timestamp_begin, 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[0].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[0].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[0].abbr_idx], 1); add_next_index_zval(return_value, element);;
  }
 } else {
  for (i = begin; i < tzobj->tzi.tz->timecnt; ++i) {
   if (tzobj->tzi.tz->trans[i] < timestamp_end) {
    do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;; _array_init((element), 0 ); add_assoc_long_ex(element, "ts", strlen("ts")+1, tzobj->tzi.tz->trans[i]); add_assoc_string_ex(element, "time", strlen("time")+1, php_format_date("Y-m-d\\TH:i:sO", 13, tzobj->tzi.tz->trans[i], 0 ), 0); add_assoc_long_ex(element, "offset", strlen("offset")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[i]].offset); add_assoc_bool_ex(element, "isdst", strlen("isdst")+1, tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[i]].isdst); add_assoc_string_ex(element, "abbr", strlen("abbr")+1, &tzobj->tzi.tz->timezone_abbr[tzobj->tzi.tz->type[tzobj->tzi.tz->trans_idx[i]].abbr_idx], 1); add_next_index_zval(return_value, element);;
   }
  }
 }
}





void zif_timezone_location_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_timezone_obj *tzobj;

 if (zend_parse_method_parameters((ht) , (this_ptr), "O", &object, date_ce_timezone) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 tzobj = (php_timezone_obj *) zend_object_store_get_object(object );
 if (!(tzobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateTimeZone" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };
 if (tzobj->type != 3) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 _array_init((return_value), 0 );
 add_assoc_string_ex(return_value, "country_code", strlen("country_code")+1, tzobj->tzi.tz->location.country_code, 1);
 add_assoc_double_ex(return_value, "latitude", strlen("latitude")+1, tzobj->tzi.tz->location.latitude);
 add_assoc_double_ex(return_value, "longitude", strlen("longitude")+1, tzobj->tzi.tz->location.longitude);
 add_assoc_string_ex(return_value, "comments", strlen("comments")+1, tzobj->tzi.tz->location.comments, 1);
}


static int date_interval_initialize(timelib_rel_time **rt, char *format, int format_length )
{
 timelib_time *b = ((void *)0), *e = ((void *)0);
 timelib_rel_time *p = ((void *)0);
 int r = 0;
 int retval = 0;
 struct timelib_error_container *errors;

 timelib_strtointerval(format, format_length, &b, &e, &p, &r, &errors);

 if (errors->error_count > 0) {
  php_error_docref0(((void *)0) , (1<<1L), "Unknown or bad format (%s)", format);
  retval = -1;
 } else {
  if(p) {
   *rt = p;
   retval = 0;
  } else {
   if(b && e) {
    timelib_update_ts(b, ((void *)0));
    timelib_update_ts(e, ((void *)0));
    *rt = timelib_diff(b, e);
    retval = 0;
   } else {
    php_error_docref0(((void *)0) , (1<<1L), "Failed to parse interval (%s)", format);
    retval = -1;
   }
  }
 }
 timelib_error_container_dtor(errors);
 return retval;
}


zval *date_interval_read_property(zval *object, zval *member, int type, const zend_literal *key )
{
 php_interval_obj *obj;
 zval *retval;
 zval tmp_member;
 timelib_sll value = -1;

  if (member->type != 6) {
  tmp_member = *member;
  _zval_copy_ctor((&tmp_member) );
  if ((&tmp_member)->type != 6) { _convert_to_string((&tmp_member) ); };
  member = &tmp_member;
 }

 obj = (php_interval_obj *)zend_objects_get_address(object );






 do {
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("y") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("y"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "y") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("y") && ((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) ? __builtin_strcmp ((*member).value.str.val, "y") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("y"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("y") && ((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) && (__s2_len = strlen ("y"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "y") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("y"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("y"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("y"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("y"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "y")))); }) == 0) { value = obj->diff->y; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("m") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("m"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "m") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("m") && ((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) ? __builtin_strcmp ((*member).value.str.val, "m") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("m"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("m") && ((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) && (__s2_len = strlen ("m"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "m") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("m"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("m"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("m"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("m"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "m")))); }) == 0) { value = obj->diff->m; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("d") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("d"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "d") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("d") && ((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) ? __builtin_strcmp ((*member).value.str.val, "d") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("d"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("d") && ((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) && (__s2_len = strlen ("d"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "d") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("d"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("d"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("d"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("d"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "d")))); }) == 0) { value = obj->diff->d; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("h") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("h"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "h") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("h") && ((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) ? __builtin_strcmp ((*member).value.str.val, "h") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("h"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("h") && ((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) && (__s2_len = strlen ("h"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "h") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("h"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("h"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("h"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("h"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "h")))); }) == 0) { value = obj->diff->h; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("i") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("i"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "i") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("i") && ((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) ? __builtin_strcmp ((*member).value.str.val, "i") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("i"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("i") && ((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) && (__s2_len = strlen ("i"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "i") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("i"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("i"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("i"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("i"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "i")))); }) == 0) { value = obj->diff->i; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("s") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("s"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "s") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("s") && ((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) ? __builtin_strcmp ((*member).value.str.val, "s") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("s"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("s") && ((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) && (__s2_len = strlen ("s"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "s") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("s"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("s"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("s"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("s"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "s")))); }) == 0) { value = obj->diff->s; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("invert") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("invert"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("invert") && ((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("invert"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("invert") && ((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) && (__s2_len = strlen ("invert"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("invert"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("invert"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("invert"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("invert"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "invert")))); }) == 0) { value = obj->diff->invert; break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("days") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("days"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("days") + 1) - (size_t)(const void *)("days") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "days") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("days") && ((size_t)(const void *)(("days") + 1) - (size_t)(const void *)("days") == 1) ? __builtin_strcmp ((*member).value.str.val, "days") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("days"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("days") && ((size_t)(const void *)(("days") + 1) - (size_t)(const void *)("days") == 1) && (__s2_len = strlen ("days"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "days") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("days"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("days"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("days"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("days"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "days")))); }) == 0) { value = obj->diff->days; break; };

  retval = (zend_get_std_object_handlers())->read_property(object, member, type, key );

  if (member == &tmp_member) {
   _zval_dtor((member) );
  }

  return retval;
 } while(0);

 do { (retval) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(retval))->u.buffered = ((void *)0); } while (0); *retval = zval_used_for_init;;;
 zval_set_refcount_p(retval, 0);

 { zval *__z = (retval); (*__z).value.lval = value; (*__z).type = 1; };

 if (member == &tmp_member) {
  _zval_dtor((member) );
 }

 return retval;
}



void date_interval_write_property(zval *object, zval *member, zval *value, const zend_literal *key )
{
 php_interval_obj *obj;
 zval tmp_member, tmp_value;

  if (member->type != 6) {
  tmp_member = *member;
  _zval_copy_ctor((&tmp_member) );
  if ((&tmp_member)->type != 6) { _convert_to_string((&tmp_member) ); };
  member = &tmp_member;
 }
 obj = (php_interval_obj *)zend_objects_get_address(object );
# 3503 "/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/php/ext/date/php_date.c"
 do {
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("y") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("y"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "y") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("y") && ((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) ? __builtin_strcmp ((*member).value.str.val, "y") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("y"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("y") && ((size_t)(const void *)(("y") + 1) - (size_t)(const void *)("y") == 1) && (__s2_len = strlen ("y"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "y") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("y"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("y"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("y"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("y"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "y")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->y = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("m") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("m"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "m") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("m") && ((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) ? __builtin_strcmp ((*member).value.str.val, "m") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("m"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("m") && ((size_t)(const void *)(("m") + 1) - (size_t)(const void *)("m") == 1) && (__s2_len = strlen ("m"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "m") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("m"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("m"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("m"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("m"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "m")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->m = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("d") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("d"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "d") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("d") && ((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) ? __builtin_strcmp ((*member).value.str.val, "d") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("d"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("d") && ((size_t)(const void *)(("d") + 1) - (size_t)(const void *)("d") == 1) && (__s2_len = strlen ("d"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "d") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("d"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("d"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("d"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("d"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "d")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->d = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("h") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("h"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "h") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("h") && ((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) ? __builtin_strcmp ((*member).value.str.val, "h") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("h"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("h") && ((size_t)(const void *)(("h") + 1) - (size_t)(const void *)("h") == 1) && (__s2_len = strlen ("h"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "h") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("h"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("h"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("h"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("h"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "h")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->h = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("i") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("i"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "i") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("i") && ((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) ? __builtin_strcmp ((*member).value.str.val, "i") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("i"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("i") && ((size_t)(const void *)(("i") + 1) - (size_t)(const void *)("i") == 1) && (__s2_len = strlen ("i"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "i") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("i"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("i"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("i"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("i"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "i")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->i = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("s") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("s"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "s") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("s") && ((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) ? __builtin_strcmp ((*member).value.str.val, "s") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("s"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("s") && ((size_t)(const void *)(("s") + 1) - (size_t)(const void *)("s") == 1) && (__s2_len = strlen ("s"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "s") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("s"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("s"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("s"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("s"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "s")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->s = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };
  if (__extension__ ({ size_t __s1_len, __s2_len; (__builtin_constant_p ((*member).value.str.val) && __builtin_constant_p ("invert") && (__s1_len = strlen ((*member).value.str.val), __s2_len = strlen ("invert"), (!((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) || __s1_len >= 4) && (!((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) || __s2_len >= 4)) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) && (__s1_len = strlen ((*member).value.str.val), __s1_len < 4) ? (__builtin_constant_p ("invert") && ((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__extension__ ({ __const unsigned char *__s2 = (__const unsigned char *) (__const char *) ("invert"); register int __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[0] - __s2[0]); if (__s1_len > 0 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[1] - __s2[1]); if (__s1_len > 1 && __result == 0) { __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[2] - __s2[2]); if (__s1_len > 2 && __result == 0) __result = (((__const unsigned char *) (__const char *) ((*member).value.str.val))[3] - __s2[3]); } } __result; }))) : (__builtin_constant_p ("invert") && ((size_t)(const void *)(("invert") + 1) - (size_t)(const void *)("invert") == 1) && (__s2_len = strlen ("invert"), __s2_len < 4) ? (__builtin_constant_p ((*member).value.str.val) && ((size_t)(const void *)(((*member).value.str.val) + 1) - (size_t)(const void *)((*member).value.str.val) == 1) ? __builtin_strcmp ((*member).value.str.val, "invert") : (__extension__ ({ __const unsigned char *__s1 = (__const unsigned char *) (__const char *) ((*member).value.str.val); register int __result = __s1[0] - ((__const unsigned char *) (__const char *) ("invert"))[0]; if (__s2_len > 0 && __result == 0) { __result = (__s1[1] - ((__const unsigned char *) (__const char *) ("invert"))[1]); if (__s2_len > 1 && __result == 0) { __result = (__s1[2] - ((__const unsigned char *) (__const char *) ("invert"))[2]); if (__s2_len > 2 && __result == 0) __result = (__s1[3] - ((__const unsigned char *) (__const char *) ("invert"))[3]); } } __result; }))) : __builtin_strcmp ((*member).value.str.val, "invert")))); }) == 0) { if (value->type != 1) { tmp_value = *value; _zval_copy_ctor((&tmp_value) ); convert_to_long(&tmp_value); value = &tmp_value; } obj->diff->invert = (*value).value.lval; if (value == &tmp_value) { _zval_dtor((value) ); } break; };

  (zend_get_std_object_handlers())->write_property(object, member, value, key );
 } while(0);

 if (member == &tmp_member) {
  _zval_dtor((member) );
 }
}






void zim_DateInterval___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *interval_string = ((void *)0);
 int interval_string_length;
 php_interval_obj *diobj;
 timelib_rel_time *reltime;
 zend_error_handling error_handling;

 zend_replace_error_handling(EH_THROW, ((void *)0), &error_handling );
 if (zend_parse_parameters((ht) , "s", &interval_string, &interval_string_length) == 0) {
  if (date_interval_initialize(&reltime, interval_string, interval_string_length ) == 0) {
   diobj = zend_object_store_get_object((this_ptr) );
   diobj->diff = reltime;
   diobj->initialized = 1;
  } else {
   { (*(this_ptr)).type = 0; };
  }
 }
 zend_restore_error_handling(&error_handling );
}


static long php_date_long_from_hash_element(HashTable *myht, char *element, size_t size)
{
 zval **z_arg = ((void *)0);

 if (zend_hash_find(myht, element, size + 1, (void**) &z_arg) == 0) {
  convert_to_long(*z_arg);
  return (**z_arg).value.lval;
 } else {
  return -1;
 }
}

static int php_date_interval_initialize_from_hash(zval **return_value, php_interval_obj **intobj, HashTable *myht )
{
 (*intobj)->diff = timelib_rel_time_ctor();

 (*intobj)->diff->y = php_date_long_from_hash_element(myht, "y", 1);
 (*intobj)->diff->m = php_date_long_from_hash_element(myht, "m", 1);
 (*intobj)->diff->d = php_date_long_from_hash_element(myht, "d", 1);
 (*intobj)->diff->h = php_date_long_from_hash_element(myht, "h", 1);
 (*intobj)->diff->i = php_date_long_from_hash_element(myht, "i", 1);
 (*intobj)->diff->s = php_date_long_from_hash_element(myht, "s", 1);
 (*intobj)->diff->invert = php_date_long_from_hash_element(myht, "invert", 6);
 (*intobj)->diff->days = php_date_long_from_hash_element(myht, "days", 4);
 (*intobj)->initialized = 1;

 return 0;
}



void zim_DateInterval___set_state(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_interval_obj *intobj;
 zval *array;
 HashTable *myht;

 if (zend_parse_parameters((ht) , "a", &array) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 myht = ((*array).type==4 ? (*array).value.ht : (((*array).type==5 ? (*array).value.obj.handlers->get_properties((array) ) : ((void *)0))));

 php_date_instantiate(date_ce_interval, return_value );
 intobj = (php_interval_obj *) zend_object_store_get_object(return_value );
 php_date_interval_initialize_from_hash(&return_value, &intobj, myht );
}




void zim_DateInterval___wakeup(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object = (this_ptr);
 php_interval_obj *intobj;
 HashTable *myht;

 intobj = (php_interval_obj *) zend_object_store_get_object(object );

 myht = ((*object)).value.obj.handlers->get_properties(&(*object) );

 php_date_interval_initialize_from_hash(&return_value, &intobj, myht );
}




void zif_date_interval_create_from_date_string(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *time_str = ((void *)0);
 int time_str_len = 0;
 timelib_time *time;
 timelib_error_container *err = ((void *)0);
 php_interval_obj *diobj;

 if (zend_parse_parameters((ht) , "s", &time_str, &time_str_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 php_date_instantiate(date_ce_interval, return_value );

 time = timelib_strtotime(time_str, time_str_len, &err, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db(), php_date_parse_tzfile_wrapper);
 diobj = (php_interval_obj *) zend_object_store_get_object(return_value );
 diobj->diff = timelib_rel_time_clone(&time->relative);
 diobj->initialized = 1;
 timelib_time_dtor(time);
 timelib_error_container_dtor(err);
}



static char *date_interval_format(char *format, int format_len, timelib_rel_time *t)
{
 smart_str string = {0};
 int i, length, have_format_spec = 0;
 char buffer[33];

 if (!format_len) {
  return _estrdup(("") );
 }

 for (i = 0; i < format_len; i++) {
  if (have_format_spec) {
   switch (format[i]) {
    case 'Y': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->y); break;
    case 'y': length = ap_php_slprintf(buffer, 32, "%d", (int) t->y); break;

    case 'M': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->m); break;
    case 'm': length = ap_php_slprintf(buffer, 32, "%d", (int) t->m); break;

    case 'D': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->d); break;
    case 'd': length = ap_php_slprintf(buffer, 32, "%d", (int) t->d); break;

    case 'H': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->h); break;
    case 'h': length = ap_php_slprintf(buffer, 32, "%d", (int) t->h); break;

    case 'I': length = ap_php_slprintf(buffer, 32, "%02d", (int) t->i); break;
    case 'i': length = ap_php_slprintf(buffer, 32, "%d", (int) t->i); break;

    case 'S': length = ap_php_slprintf(buffer, 32, "%02ld", (long) t->s); break;
    case 's': length = ap_php_slprintf(buffer, 32, "%ld", (long) t->s); break;

    case 'a': {
     if ((int) t->days != -99999) {
      length = ap_php_slprintf(buffer, 32, "%d", (int) t->days);
     } else {
      length = ap_php_slprintf(buffer, 32, "(unknown)");
     }
    } break;
    case 'r': length = ap_php_slprintf(buffer, 32, "%s", t->invert ? "-" : ""); break;
    case 'R': length = ap_php_slprintf(buffer, 32, "%c", t->invert ? '-' : '+'); break;

    case '%': length = ap_php_slprintf(buffer, 32, "%%"); break;
    default: buffer[0] = '%'; buffer[1] = format[i]; buffer[2] = '\0'; length = 2; break;
   }
   do { register size_t __nl; smart_str *__dest = (smart_str *) ((&string)); do { if (!(__dest)->c) { (__dest)->len = 0; __nl = (((length))); (__dest)->a = __nl < 78 ? 78 : __nl + 128; (__dest)->c = (((((0))))?__zend_realloc((((__dest)->c)), (((__dest)->a + 1))):_erealloc(((((__dest)->c))), ((((__dest)->a + 1))), 0 )); } else { __nl = (__dest)->len + (((length))); if (__nl >= (__dest)->a) { (__dest)->a = __nl + 128; (__dest)->c = (((((0))))?__zend_realloc((((__dest)->c)), (((__dest)->a + 1))):_erealloc(((((__dest)->c))), ((((__dest)->a + 1))), 0 )); } } } while (0); memcpy(__dest->c + __dest->len, ((buffer)), ((length))); __dest->len = __nl; } while (0);
   have_format_spec = 0;
  } else {
   if (format[i] == '%') {
    have_format_spec = 1;
   } else {
    do { register size_t __nl; do { if (!(((&string)))->c) { (((&string)))->len = 0; __nl = (1); (((&string)))->a = __nl < 78 ? 78 : __nl + 128; (((&string)))->c = (((((0))))?__zend_realloc((((((&string)))->c)), (((((&string)))->a + 1))):_erealloc(((((((&string)))->c))), ((((((&string)))->a + 1))), 0 )); } else { __nl = (((&string)))->len + (1); if (__nl >= (((&string)))->a) { (((&string)))->a = __nl + 128; (((&string)))->c = (((((0))))?__zend_realloc((((((&string)))->c)), (((((&string)))->a + 1))):_erealloc(((((((&string)))->c))), ((((((&string)))->a + 1))), 0 )); } } } while (0); ((&string))->len = __nl; ((unsigned char *) ((&string))->c)[((&string))->len - 1] = ((format[i])); } while (0);
   }
  }
 }

 do { if ((&string)->c) { (&string)->c[(&string)->len] = '\0'; } } while (0);

 return string.c;
}





void zif_date_interval_format(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 zval *object;
 php_interval_obj *diobj;
 char *format;
 int format_len;

 if (zend_parse_method_parameters((ht) , (this_ptr), "Os", &object, date_ce_interval, &format, &format_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 diobj = (php_interval_obj *) zend_object_store_get_object(object );
 if (!(diobj->initialized)) { php_error_docref0(((void *)0) , (1<<1L), "The " "DateInterval" " object has not been correctly initialized by its constructor"); { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; }; };

 { do { const char *__s=(date_interval_format(format, format_len, diobj->diff)); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (0?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
}


static int date_period_initialize(timelib_time **st, timelib_time **et, timelib_rel_time **d, long *recurrences, char *format, int format_length )
{
 timelib_time *b = ((void *)0), *e = ((void *)0);
 timelib_rel_time *p = ((void *)0);
 int r = 0;
 int retval = 0;
 struct timelib_error_container *errors;

 timelib_strtointerval(format, format_length, &b, &e, &p, &r, &errors);

 if (errors->error_count > 0) {
  php_error_docref0(((void *)0) , (1<<1L), "Unknown or bad format (%s)", format);
  retval = -1;
 } else {
  *st = b;
  *et = e;
  *d = p;
  *recurrences = r;
  retval = 0;
 }
 timelib_error_container_dtor(errors);
 return retval;
}




void zim_DatePeriod___construct(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_period_obj *dpobj;
 php_date_obj *dateobj;
 php_interval_obj *intobj;
 zval *start, *end = ((void *)0), *interval;
 long recurrences = 0, options = 0;
 char *isostr = ((void *)0);
 int isostr_len = 0;
 timelib_time *clone;
 zend_error_handling error_handling;

 zend_replace_error_handling(EH_THROW, ((void *)0), &error_handling );
 if (zend_parse_parameters_ex(1<<1, (ht) , "OOl|l", &start, date_ce_date, &interval, date_ce_interval, &recurrences, &options) == -1) {
  if (zend_parse_parameters_ex(1<<1, (ht) , "OOO|l", &start, date_ce_date, &interval, date_ce_interval, &end, date_ce_date, &options) == -1) {
   if (zend_parse_parameters_ex(1<<1, (ht) , "s|l", &isostr, &isostr_len, &options) == -1) {
    php_error_docref0(((void *)0) , (1<<1L), "This constructor accepts either (DateTime, DateInterval, int) OR (DateTime, DateInterval, DateTime) OR (string) as arguments.");
    zend_restore_error_handling(&error_handling );
    return;
   }
  }
 }

 dpobj = zend_object_store_get_object((this_ptr) );
 dpobj->current = ((void *)0);

 if (isostr) {
  date_period_initialize(&(dpobj->start), &(dpobj->end), &(dpobj->interval), &recurrences, isostr, isostr_len );
  if (dpobj->start == ((void *)0)) {
   php_error_docref0(((void *)0) , (1<<1L), "The ISO interval '%s' did not contain a start date.", isostr);
  }
  if (dpobj->interval == ((void *)0)) {
   php_error_docref0(((void *)0) , (1<<1L), "The ISO interval '%s' did not contain an interval.", isostr);
  }
  if (dpobj->end == ((void *)0) && recurrences == 0) {
   php_error_docref0(((void *)0) , (1<<1L), "The ISO interval '%s' did not contain an end date or a recurrence count.", isostr);
  }

  if (dpobj->start) {
   timelib_update_ts(dpobj->start, ((void *)0));
  }
  if (dpobj->end) {
   timelib_update_ts(dpobj->end, ((void *)0));
  }
 } else {

  intobj = (php_interval_obj *) zend_object_store_get_object(interval );


  dateobj = (php_date_obj *) zend_object_store_get_object(start );
  clone = timelib_time_ctor();
  memcpy(clone, dateobj->time, sizeof(timelib_time));
  if (dateobj->time->tz_abbr) {
   clone->tz_abbr = (__extension__ (__builtin_constant_p (dateobj->time->tz_abbr) && ((size_t)(const void *)((dateobj->time->tz_abbr) + 1) - (size_t)(const void *)(dateobj->time->tz_abbr) == 1) ? (((__const char *) (dateobj->time->tz_abbr))[0] == '\0' ? (char *) calloc ((size_t) 1, (size_t) 1) : ({ size_t __len = strlen (dateobj->time->tz_abbr) + 1; char *__retval = (char *) malloc (__len); if (__retval != ((void *)0)) __retval = (char *) memcpy (__retval, dateobj->time->tz_abbr, __len); __retval; })) : __strdup (dateobj->time->tz_abbr)));
  }
  if (dateobj->time->tz_info) {
   clone->tz_info = dateobj->time->tz_info;
  }
  dpobj->start = clone;


  dpobj->interval = timelib_rel_time_clone(intobj->diff);


  if (end) {
   dateobj = (php_date_obj *) zend_object_store_get_object(end );
   clone = timelib_time_clone(dateobj->time);
   dpobj->end = clone;
  }
 }


 dpobj->include_start_date = !(options & 0x0001);


 dpobj->recurrences = recurrences + dpobj->include_start_date;

 dpobj->initialized = 1;

 zend_restore_error_handling(&error_handling );
}


static int check_id_allowed(char *id, long what)
{
 if (what & 0x0001 && strncasecmp(id, "Africa/", 7) == 0) return 1;
 if (what & 0x0002 && strncasecmp(id, "America/", 8) == 0) return 1;
 if (what & 0x0004 && strncasecmp(id, "Antarctica/", 11) == 0) return 1;
 if (what & 0x0008 && strncasecmp(id, "Arctic/", 7) == 0) return 1;
 if (what & 0x0010 && strncasecmp(id, "Asia/", 5) == 0) return 1;
 if (what & 0x0020 && strncasecmp(id, "Atlantic/", 9) == 0) return 1;
 if (what & 0x0040 && strncasecmp(id, "Australia/", 10) == 0) return 1;
 if (what & 0x0080 && strncasecmp(id, "Europe/", 7) == 0) return 1;
 if (what & 0x0100 && strncasecmp(id, "Indian/", 7) == 0) return 1;
 if (what & 0x0200 && strncasecmp(id, "Pacific/", 8) == 0) return 1;
 if (what & 0x0400 && strncasecmp(id, "UTC", 3) == 0) return 1;
 return 0;
}




void zif_timezone_identifiers_list(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 const timelib_tzdb *tzdb;
 const timelib_tzdb_index_entry *table;
 int i, item_count;
 long what = 0x07FF;
 char *option = ((void *)0);
 int option_len = 0;

 if (zend_parse_parameters((ht) , "|ls", &what, &option, &option_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }


 if (what == 0x1000 && option_len != 2) {
  php_error_docref0(((void *)0) , (1<<3L), "A two-letter ISO 3166-1 compatible country code is expected");
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 tzdb = php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db();
 item_count = tzdb->index_size;
 table = tzdb->index;

 _array_init((return_value), 0 );

 for (i = 0; i < item_count; ++i) {
  if (what == 0x1000) {
   if (tzdb->data[table[i].pos + 5] == option[0] && tzdb->data[table[i].pos + 6] == option[1]) {
    add_next_index_string(return_value, table[i].id, 1);
   }
  } else if (what == 0x0FFF || (check_id_allowed(table[i].id, what) && (tzdb->data[table[i].pos + 4] == '\1'))) {
   add_next_index_string(return_value, table[i].id, 1);
  }
 };
}





void zif_timezone_version_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 const timelib_tzdb *tzdb;

 tzdb = php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db();
 { do { const char *__s=(tzdb->version); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0); return; };
}





void zif_timezone_abbreviations_list(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 const timelib_tz_lookup_table *table, *entry;
 zval *element, **abbr_array_pp, *abbr_array;

 table = timelib_timezone_abbreviations_list();
 _array_init((return_value), 0 );
 entry = table;

 do {
  do { (element) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(element))->u.buffered = ((void *)0); } while (0); (element)->refcount__gc = 1; (element)->is_ref__gc = 0;;;
  _array_init((element), 0 );
  add_assoc_bool_ex(element, "dst", strlen("dst")+1, entry->type);
  add_assoc_long_ex(element, "offset", strlen("offset")+1, entry->gmtoffset);
  if (entry->full_tz_name) {
   add_assoc_string_ex(element, "timezone_id", strlen("timezone_id")+1, entry->full_tz_name, 1);
  } else {
   add_assoc_null_ex(element, "timezone_id", strlen("timezone_id") + 1);
  }

  if (zend_hash_find(((*return_value).type==4 ? (*return_value).value.ht : (((*return_value).type==5 ? (*return_value).value.obj.handlers->get_properties((return_value) ) : ((void *)0)))), entry->name, strlen(entry->name) + 1, (void **) &abbr_array_pp) == -1) {
   do { (abbr_array) = (zval*)_emalloc((sizeof(zval_gc_info)) ); ((zval_gc_info*)(abbr_array))->u.buffered = ((void *)0); } while (0); (abbr_array)->refcount__gc = 1; (abbr_array)->is_ref__gc = 0;;;
   _array_init((abbr_array), 0 );
   add_assoc_zval_ex(return_value, entry->name, strlen(entry->name)+1, abbr_array);
  } else {
   abbr_array = *abbr_array_pp;
  }
  add_next_index_zval(abbr_array, element);
  entry++;
 } while (entry->name);
}




void zif_date_default_timezone_set(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 char *zone;
 int zone_len;

 if (zend_parse_parameters((ht) , "s", &zone, &zone_len) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 if (!timelib_timezone_id_is_valid(zone, php_date_global_timezone_db ? php_date_global_timezone_db : timelib_builtin_db())) {
  php_error_docref0(((void *)0) , (1<<3L), "Timezone ID '%s' is invalid", zone);
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 if ((date_globals.timezone)) {
  _efree(((date_globals.timezone)) );
  (date_globals.timezone) = ((void *)0);
 }
 (date_globals.timezone) = _estrndup((zone), (zone_len) );
 { do { zval *__z = (return_value); (*__z).value.lval = ((1) != 0); (*__z).type = 3; } while (0); return; };
}




void zif_date_default_timezone_get(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 timelib_tzinfo *default_tz;

 default_tz = get_timezone_info();
 do { const char *__s=(default_tz->name); zval *__z = (return_value); (*__z).value.str.len = strlen(__s); (*__z).value.str.val = (1?_estrndup((__s), ((*__z).value.str.len) ):(char*)__s); (*__z).type = 6; } while (0);
}





static void php_do_date_sunrise_sunset(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used , int calc_sunset)
{
 double latitude = 0.0, longitude = 0.0, zenith = 0.0, gmt_offset = 0, altitude;
 double h_rise, h_set, N;
 timelib_sll rise, set, transit;
 long time, retformat = 0;
 int rs;
 timelib_time *t;
 timelib_tzinfo *tzi;
 char *retstr;

 if (zend_parse_parameters((ht) , "l|ldddd", &time, &retformat, &latitude, &longitude, &zenith, &gmt_offset) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 switch ((ht)) {
  case 1:
   retformat = 1;
  case 2:
   latitude = zend_ini_double(("date.default_latitude"), sizeof("date.default_latitude"), 0);
  case 3:
   longitude = zend_ini_double(("date.default_longitude"), sizeof("date.default_longitude"), 0);
  case 4:
   if (calc_sunset) {
    zenith = zend_ini_double(("date.sunset_zenith"), sizeof("date.sunset_zenith"), 0);
   } else {
    zenith = zend_ini_double(("date.sunrise_zenith"), sizeof("date.sunrise_zenith"), 0);
   }
  case 5:
  case 6:
   break;
  default:
   php_error_docref0(((void *)0) , (1<<1L), "invalid format");
   { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
   break;
 }
 if (retformat != 0 &&
  retformat != 1 &&
  retformat != 2)
 {
  php_error_docref0(((void *)0) , (1<<1L), "Wrong return format given, pick one of SUNFUNCS_RET_TIMESTAMP, SUNFUNCS_RET_STRING or SUNFUNCS_RET_DOUBLE");
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }
 altitude = 90 - zenith;


 t = timelib_time_ctor();
 tzi = get_timezone_info();
 t->tz_info = tzi;
 t->zone_type = 3;

 if ((ht) <= 5) {
  gmt_offset = timelib_get_current_offset(t) / 3600;
 }

 timelib_unixtime2local(t, time);
 rs = timelib_astro_rise_set_altitude(t, longitude, latitude, altitude, 1, &h_rise, &h_set, &rise, &set, &transit);
 timelib_time_dtor(t);

 if (rs != 0) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 if (retformat == 0) {
  { { zval *__z = (return_value); (*__z).value.lval = calc_sunset ? set : rise; (*__z).type = 1; }; return; };
 }
 N = (calc_sunset ? h_set : h_rise) + gmt_offset;

 if (N > 24 || N < 0) {
  N -= floor(N / 24) * 24;
 }

 switch (retformat) {
  case 1:
   spprintf(&retstr, 0, "%02d:%02d", (int) N, (int) (60 * (N - (int) N)));
   { do { const char *__s=(retstr); int __l=5; zval *__z = (return_value); (*__z).value.str.len = __l; (*__z).value.str.val = (0?_estrndup((__s), (__l) ):(char*)__s); (*__z).type = 6; } while (0); return; };
   break;
  case 2:
   { { zval *__z = (return_value); (*__z).value.dval = N; (*__z).type = 2; }; return; };
   break;
 }
}




void zif_date_sunrise(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_do_date_sunrise_sunset(ht, return_value, return_value_ptr, this_ptr, return_value_used , 0);
}




void zif_date_sunset(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 php_do_date_sunrise_sunset(ht, return_value, return_value_ptr, this_ptr, return_value_used , 1);
}




void zif_date_sun_info(int ht, zval *return_value, zval **return_value_ptr, zval *this_ptr, int return_value_used )
{
 long time;
 double latitude, longitude;
 timelib_time *t, *t2;
 timelib_tzinfo *tzi;
 int rs;
 timelib_sll rise, set, transit;
 int dummy;
 double ddummy;

 if (zend_parse_parameters((ht) , "ldd", &time, &latitude, &longitude) == -1) {
  { do { zval *__z = (return_value); (*__z).value.lval = ((0) != 0); (*__z).type = 3; } while (0); return; };
 }

 t = timelib_time_ctor();
 tzi = get_timezone_info();
 t->tz_info = tzi;
 t->zone_type = 3;
 timelib_unixtime2local(t, time);


 t2 = timelib_time_ctor();
 _array_init((return_value), 0 );


 rs = timelib_astro_rise_set_altitude(t, longitude, latitude, -35.0/60, 1, &ddummy, &ddummy, &rise, &set, &transit);
 switch (rs) {
  case -1:
   add_assoc_bool_ex(return_value, "sunrise", strlen("sunrise")+1, 0);
   add_assoc_bool_ex(return_value, "sunset", strlen("sunset")+1, 0);
   break;
  case 1:
   add_assoc_bool_ex(return_value, "sunrise", strlen("sunrise")+1, 1);
   add_assoc_bool_ex(return_value, "sunset", strlen("sunset")+1, 1);
   break;
  default:
   t2->sse = rise;
   add_assoc_long_ex(return_value, "sunrise", strlen("sunrise")+1, timelib_date_to_int(t2, &dummy));
   t2->sse = set;
   add_assoc_long_ex(return_value, "sunset", strlen("sunset")+1, timelib_date_to_int(t2, &dummy));
 }
 t2->sse = transit;
 add_assoc_long_ex(return_value, "transit", strlen("transit")+1, timelib_date_to_int(t2, &dummy));


 rs = timelib_astro_rise_set_altitude(t, longitude, latitude, -6.0, 0, &ddummy, &ddummy, &rise, &set, &transit);
 switch (rs) {
  case -1:
   add_assoc_bool_ex(return_value, "civil_twilight_begin", strlen("civil_twilight_begin")+1, 0);
   add_assoc_bool_ex(return_value, "civil_twilight_end", strlen("civil_twilight_end")+1, 0);
   break;
  case 1:
   add_assoc_bool_ex(return_value, "civil_twilight_begin", strlen("civil_twilight_begin")+1, 1);
   add_assoc_bool_ex(return_value, "civil_twilight_end", strlen("civil_twilight_end")+1, 1);
   break;
  default:
   t2->sse = rise;
   add_assoc_long_ex(return_value, "civil_twilight_begin", strlen("civil_twilight_begin")+1, timelib_date_to_int(t2, &dummy));
   t2->sse = set;
   add_assoc_long_ex(return_value, "civil_twilight_end", strlen("civil_twilight_end")+1, timelib_date_to_int(t2, &dummy));
 }


 rs = timelib_astro_rise_set_altitude(t, longitude, latitude, -12.0, 0, &ddummy, &ddummy, &rise, &set, &transit);
 switch (rs) {
  case -1:
   add_assoc_bool_ex(return_value, "nautical_twilight_begin", strlen("nautical_twilight_begin")+1, 0);
   add_assoc_bool_ex(return_value, "nautical_twilight_end", strlen("nautical_twilight_end")+1, 0);
   break;
  case 1:
   add_assoc_bool_ex(return_value, "nautical_twilight_begin", strlen("nautical_twilight_begin")+1, 1);
   add_assoc_bool_ex(return_value, "nautical_twilight_end", strlen("nautical_twilight_end")+1, 1);
   break;
  default:
   t2->sse = rise;
   add_assoc_long_ex(return_value, "nautical_twilight_begin", strlen("nautical_twilight_begin")+1, timelib_date_to_int(t2, &dummy));
   t2->sse = set;
   add_assoc_long_ex(return_value, "nautical_twilight_end", strlen("nautical_twilight_end")+1, timelib_date_to_int(t2, &dummy));
 }


 rs = timelib_astro_rise_set_altitude(t, longitude, latitude, -18.0, 0, &ddummy, &ddummy, &rise, &set, &transit);
 switch (rs) {
  case -1:
   add_assoc_bool_ex(return_value, "astronomical_twilight_begin", strlen("astronomical_twilight_begin")+1, 0);
   add_assoc_bool_ex(return_value, "astronomical_twilight_end", strlen("astronomical_twilight_end")+1, 0);
   break;
  case 1:
   add_assoc_bool_ex(return_value, "astronomical_twilight_begin", strlen("astronomical_twilight_begin")+1, 1);
   add_assoc_bool_ex(return_value, "astronomical_twilight_end", strlen("astronomical_twilight_end")+1, 1);
   break;
  default:
   t2->sse = rise;
   add_assoc_long_ex(return_value, "astronomical_twilight_begin", strlen("astronomical_twilight_begin")+1, timelib_date_to_int(t2, &dummy));
   t2->sse = set;
   add_assoc_long_ex(return_value, "astronomical_twilight_end", strlen("astronomical_twilight_end")+1, timelib_date_to_int(t2, &dummy));
 }
 timelib_time_dtor(t);
 timelib_time_dtor(t2);
}
