struct _IO_FILE;
struct _IO_FILE *_coverage_fout  ;
typedef unsigned int size_t;
typedef long wchar_t;
struct __anonstruct___wait_terminated_1 {
   unsigned int __w_termsig : 7 ;
   unsigned int __w_coredump : 1 ;
   unsigned int __w_retcode : 8 ;
   unsigned int  : 16 ;
};
struct __anonstruct___wait_stopped_2 {
   unsigned int __w_stopval : 8 ;
   unsigned int __w_stopsig : 8 ;
   unsigned int  : 16 ;
};
union wait {
   int w_status ;
   struct __anonstruct___wait_terminated_1 __wait_terminated ;
   struct __anonstruct___wait_stopped_2 __wait_stopped ;
};
union __anonunion___WAIT_STATUS_3 {
   union wait *__uptr ;
   int *__iptr ;
};
typedef union __anonunion___WAIT_STATUS_3  __attribute__((__transparent_union__)) __WAIT_STATUS;
struct __anonstruct_div_t_4 {
   int quot ;
   int rem ;
};
typedef struct __anonstruct_div_t_4 div_t;
struct __anonstruct_ldiv_t_5 {
   long quot ;
   long rem ;
};
typedef struct __anonstruct_ldiv_t_5 ldiv_t;
struct __anonstruct_lldiv_t_6 {
   long long quot ;
   long long rem ;
};
typedef struct __anonstruct_lldiv_t_6 lldiv_t;
struct __locale_data;
struct __locale_struct {
   struct __locale_data *__locales[13] ;
   unsigned short const   *__ctype_b ;
   int const   *__ctype_tolower ;
   int const   *__ctype_toupper ;
   char const   *__names[13] ;
};
typedef struct __locale_struct *__locale_t;
typedef __locale_t locale_t;
typedef unsigned char __u_char;
typedef unsigned short __u_short;
typedef unsigned int __u_int;
typedef unsigned long __u_long;
typedef signed char __int8_t;
typedef unsigned char __uint8_t;
typedef short __int16_t;
typedef unsigned short __uint16_t;
typedef int __int32_t;
typedef unsigned int __uint32_t;
typedef long long __int64_t;
typedef unsigned long long __uint64_t;
typedef long long __quad_t;
typedef unsigned long long __u_quad_t;
typedef __u_quad_t __dev_t;
typedef unsigned int __uid_t;
typedef unsigned int __gid_t;
typedef unsigned long __ino_t;
typedef __u_quad_t __ino64_t;
typedef unsigned int __mode_t;
typedef unsigned int __nlink_t;
typedef long __off_t;
typedef __quad_t __off64_t;
typedef int __pid_t;
struct __anonstruct___fsid_t_7 {
   int __val[2] ;
};
typedef struct __anonstruct___fsid_t_7 __fsid_t;
typedef long __clock_t;
typedef unsigned long __rlim_t;
typedef __u_quad_t __rlim64_t;
typedef unsigned int __id_t;
typedef long __time_t;
typedef unsigned int __useconds_t;
typedef long __suseconds_t;
typedef int __daddr_t;
typedef long __swblk_t;
typedef int __key_t;
typedef int __clockid_t;
typedef void *__timer_t;
typedef long __blksize_t;
typedef long __blkcnt_t;
typedef __quad_t __blkcnt64_t;
typedef unsigned long __fsblkcnt_t;
typedef __u_quad_t __fsblkcnt64_t;
typedef unsigned long __fsfilcnt_t;
typedef __u_quad_t __fsfilcnt64_t;
typedef int __ssize_t;
typedef __off64_t __loff_t;
typedef __quad_t *__qaddr_t;
typedef char *__caddr_t;
typedef int __intptr_t;
typedef unsigned int __socklen_t;
typedef __u_char u_char;
typedef __u_short u_short;
typedef __u_int u_int;
typedef __u_long u_long;
typedef __quad_t quad_t;
typedef __u_quad_t u_quad_t;
typedef __fsid_t fsid_t;
typedef __loff_t loff_t;
typedef __ino_t ino_t;
typedef __ino64_t ino64_t;
typedef __dev_t dev_t;
typedef __gid_t gid_t;
typedef __mode_t mode_t;
typedef __nlink_t nlink_t;
typedef __uid_t uid_t;
typedef __off_t off_t;
typedef __off64_t off64_t;
typedef __pid_t pid_t;
typedef __id_t id_t;
typedef __ssize_t ssize_t;
typedef __daddr_t daddr_t;
typedef __caddr_t caddr_t;
typedef __key_t key_t;
typedef __clock_t clock_t;
typedef __time_t time_t;
typedef __clockid_t clockid_t;
typedef __timer_t timer_t;
typedef __useconds_t useconds_t;
typedef __suseconds_t suseconds_t;
typedef unsigned long ulong;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef signed char int8_t;
typedef short int16_t;
typedef int int32_t;
typedef long long int64_t;
typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;
typedef unsigned int u_int32_t;
typedef unsigned long long u_int64_t;
typedef int register_t;
typedef int __sig_atomic_t;
struct __anonstruct___sigset_t_8 {
   unsigned long __val[1024U / (8U * sizeof(unsigned long ))] ;
};
typedef struct __anonstruct___sigset_t_8 __sigset_t;
typedef __sigset_t sigset_t;
struct timespec {
   __time_t tv_sec ;
   long tv_nsec ;
};
struct timeval {
   __time_t tv_sec ;
   __suseconds_t tv_usec ;
};
typedef long __fd_mask;
struct __anonstruct_fd_set_9 {
   __fd_mask fds_bits[1024 / (8 * (int )sizeof(__fd_mask ))] ;
};
typedef struct __anonstruct_fd_set_9 fd_set;
typedef __fd_mask fd_mask;
typedef __blksize_t blksize_t;
typedef __blkcnt_t blkcnt_t;
typedef __fsblkcnt_t fsblkcnt_t;
typedef __fsfilcnt_t fsfilcnt_t;
typedef __blkcnt64_t blkcnt64_t;
typedef __fsblkcnt64_t fsblkcnt64_t;
typedef __fsfilcnt64_t fsfilcnt64_t;
typedef unsigned long pthread_t;
union __anonunion_pthread_attr_t_10 {
   char __size[36] ;
   long __align ;
};
typedef union __anonunion_pthread_attr_t_10 pthread_attr_t;
struct __pthread_internal_slist {
   struct __pthread_internal_slist *__next ;
};
typedef struct __pthread_internal_slist __pthread_slist_t;
union __anonunion____missing_field_name_12 {
   int __spins ;
   __pthread_slist_t __list ;
};
struct __pthread_mutex_s {
   int __lock ;
   unsigned int __count ;
   int __owner ;
   int __kind ;
   unsigned int __nusers ;
   union __anonunion____missing_field_name_12 __annonCompField1 ;
};
union __anonunion_pthread_mutex_t_11 {
   struct __pthread_mutex_s __data ;
   char __size[24] ;
   long __align ;
};
typedef union __anonunion_pthread_mutex_t_11 pthread_mutex_t;
union __anonunion_pthread_mutexattr_t_13 {
   char __size[4] ;
   long __align ;
};
typedef union __anonunion_pthread_mutexattr_t_13 pthread_mutexattr_t;
struct __anonstruct___data_15 {
   int __lock ;
   unsigned int __futex ;
   unsigned long long __total_seq ;
   unsigned long long __wakeup_seq ;
   unsigned long long __woken_seq ;
   void *__mutex ;
   unsigned int __nwaiters ;
   unsigned int __broadcast_seq ;
};
union __anonunion_pthread_cond_t_14 {
   struct __anonstruct___data_15 __data ;
   char __size[48] ;
   long long __align ;
};
typedef union __anonunion_pthread_cond_t_14 pthread_cond_t;
union __anonunion_pthread_condattr_t_16 {
   char __size[4] ;
   long __align ;
};
typedef union __anonunion_pthread_condattr_t_16 pthread_condattr_t;
typedef unsigned int pthread_key_t;
typedef int pthread_once_t;
struct __anonstruct___data_18 {
   int __lock ;
   unsigned int __nr_readers ;
   unsigned int __readers_wakeup ;
   unsigned int __writer_wakeup ;
   unsigned int __nr_readers_queued ;
   unsigned int __nr_writers_queued ;
   unsigned char __flags ;
   unsigned char __shared ;
   unsigned char __pad1 ;
   unsigned char __pad2 ;
   int __writer ;
};
union __anonunion_pthread_rwlock_t_17 {
   struct __anonstruct___data_18 __data ;
   char __size[32] ;
   long __align ;
};
typedef union __anonunion_pthread_rwlock_t_17 pthread_rwlock_t;
union __anonunion_pthread_rwlockattr_t_19 {
   char __size[8] ;
   long __align ;
};
typedef union __anonunion_pthread_rwlockattr_t_19 pthread_rwlockattr_t;
typedef int volatile   pthread_spinlock_t;
union __anonunion_pthread_barrier_t_20 {
   char __size[20] ;
   long __align ;
};
typedef union __anonunion_pthread_barrier_t_20 pthread_barrier_t;
union __anonunion_pthread_barrierattr_t_21 {
   char __size[4] ;
   int __align ;
};
typedef union __anonunion_pthread_barrierattr_t_21 pthread_barrierattr_t;
struct random_data {
   int32_t *fptr ;
   int32_t *rptr ;
   int32_t *state ;
   int rand_type ;
   int rand_deg ;
   int rand_sep ;
   int32_t *end_ptr ;
};
struct drand48_data {
   unsigned short __x[3] ;
   unsigned short __old_x[3] ;
   unsigned short __c ;
   unsigned short __init ;
   unsigned long long __a ;
};
typedef int (*__compar_fn_t)(void const   * , void const   * );
typedef int (*comparison_fn_t)(void const   * , void const   * );
typedef int (*__compar_d_fn_t)(void const   * , void const   * , void * );
typedef long double float_t;
typedef long double double_t;
enum __anonenum_22 {
    FP_NAN = 0,
    FP_INFINITE = 1,
    FP_ZERO = 2,
    FP_SUBNORMAL = 3,
    FP_NORMAL = 4
} ;
enum __anonenum__LIB_VERSION_TYPE_23 {
    _IEEE_ = -1,
    _SVID_ = 0,
    _XOPEN_ = 1,
    _POSIX_ = 2,
    _ISOC_ = 3
} ;
typedef enum __anonenum__LIB_VERSION_TYPE_23 _LIB_VERSION_TYPE;
struct exception {
   int type ;
   char *name ;
   double arg1 ;
   double arg2 ;
   double retval ;
};
union __anonunion___u_24 {
   float __f ;
   int __i ;
};
union __anonunion___u_25 {
   double __d ;
   int __i[2] ;
};
union __anonunion___u_26 {
   long double __l ;
   int __i[3] ;
};
union __anonunion___n_27 {
   long double __xld ;
   unsigned int __xi[3] ;
};
union __anonunion_28 {
   double __d ;
   int __i[2] ;
};
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned long long uint64_t;
typedef signed char int_least8_t;
typedef short int_least16_t;
typedef int int_least32_t;
typedef long long int_least64_t;
typedef unsigned char uint_least8_t;
typedef unsigned short uint_least16_t;
typedef unsigned int uint_least32_t;
typedef unsigned long long uint_least64_t;
typedef signed char int_fast8_t;
typedef int int_fast16_t;
typedef int int_fast32_t;
typedef long long int_fast64_t;
typedef unsigned char uint_fast8_t;
typedef unsigned int uint_fast16_t;
typedef unsigned int uint_fast32_t;
typedef unsigned long long uint_fast64_t;
typedef int intptr_t;
typedef unsigned int uintptr_t;
typedef long long intmax_t;
typedef unsigned long long uintmax_t;
typedef long __gwchar_t;
struct __anonstruct_imaxdiv_t_29 {
   long long quot ;
   long long rem ;
};
typedef struct __anonstruct_imaxdiv_t_29 imaxdiv_t;
typedef struct _IO_FILE FILE;
typedef struct _IO_FILE __FILE;
union __anonunion___value_31 {
   unsigned int __wch ;
   char __wchb[4] ;
};
struct __anonstruct___mbstate_t_30 {
   int __count ;
   union __anonunion___value_31 __value ;
};
typedef struct __anonstruct___mbstate_t_30 __mbstate_t;
struct __anonstruct__G_fpos_t_32 {
   __off_t __pos ;
   __mbstate_t __state ;
};
typedef struct __anonstruct__G_fpos_t_32 _G_fpos_t;
struct __anonstruct__G_fpos64_t_33 {
   __off64_t __pos ;
   __mbstate_t __state ;
};
typedef struct __anonstruct__G_fpos64_t_33 _G_fpos64_t;
typedef short _G_int16_t;
typedef int _G_int32_t;
typedef unsigned short _G_uint16_t;
typedef unsigned int _G_uint32_t;
typedef __builtin_va_list __gnuc_va_list;
struct _IO_jump_t;
typedef void _IO_lock_t;
struct _IO_marker {
   struct _IO_marker *_next ;
   struct _IO_FILE *_sbuf ;
   int _pos ;
};
enum __codecvt_result {
    __codecvt_ok = 0,
    __codecvt_partial = 1,
    __codecvt_error = 2,
    __codecvt_noconv = 3
} ;
struct _IO_FILE {
   int _flags ;
   char *_IO_read_ptr ;
   char *_IO_read_end ;
   char *_IO_read_base ;
   char *_IO_write_base ;
   char *_IO_write_ptr ;
   char *_IO_write_end ;
   char *_IO_buf_base ;
   char *_IO_buf_end ;
   char *_IO_save_base ;
   char *_IO_backup_base ;
   char *_IO_save_end ;
   struct _IO_marker *_markers ;
   struct _IO_FILE *_chain ;
   int _fileno ;
   int _flags2 ;
   __off_t _old_offset ;
   unsigned short _cur_column ;
   signed char _vtable_offset ;
   char _shortbuf[1] ;
   _IO_lock_t *_lock ;
   __off64_t _offset ;
   void *__pad1 ;
   void *__pad2 ;
   void *__pad3 ;
   void *__pad4 ;
   size_t __pad5 ;
   int _mode ;
   char _unused2[(15U * sizeof(int ) - 4U * sizeof(void *)) - sizeof(size_t )] ;
};
typedef struct _IO_FILE _IO_FILE;
struct _IO_FILE_plus;
typedef __ssize_t __io_read_fn(void *__cookie , char *__buf , size_t __nbytes );
typedef __ssize_t __io_write_fn(void *__cookie , char const   *__buf ,
                                size_t __n );
typedef int __io_seek_fn(void *__cookie , __off64_t *__pos , int __w );
typedef int __io_close_fn(void *__cookie );
typedef __io_read_fn cookie_read_function_t;
typedef __io_write_fn cookie_write_function_t;
typedef __io_seek_fn cookie_seek_function_t;
typedef __io_close_fn cookie_close_function_t;
struct __anonstruct__IO_cookie_io_functions_t_34 {
   __io_read_fn *read ;
   __io_write_fn *write ;
   __io_seek_fn *seek ;
   __io_close_fn *close ;
};
typedef struct __anonstruct__IO_cookie_io_functions_t_34 _IO_cookie_io_functions_t;
typedef _IO_cookie_io_functions_t cookie_io_functions_t;
struct _IO_cookie_file;
typedef __gnuc_va_list va_list;
typedef _G_fpos_t fpos_t;
typedef _G_fpos64_t fpos64_t;
struct obstack;
typedef unsigned long long timelib_ull;
typedef long long timelib_sll;
struct ttinfo {
   int32_t offset ;
   int isdst ;
   unsigned int abbr_idx ;
   unsigned int isstdcnt ;
   unsigned int isgmtcnt ;
};
typedef struct ttinfo ttinfo;
struct tlinfo {
   int32_t trans ;
   int32_t offset ;
};
typedef struct tlinfo tlinfo;
struct tlocinfo {
   char country_code[3] ;
   double latitude ;
   double longitude ;
   char *comments ;
};
typedef struct tlocinfo tlocinfo;
struct timelib_tzinfo {
   char *name ;
   uint32_t ttisgmtcnt ;
   uint32_t ttisstdcnt ;
   uint32_t leapcnt ;
   uint32_t timecnt ;
   uint32_t typecnt ;
   uint32_t charcnt ;
   int32_t *trans ;
   unsigned char *trans_idx ;
   ttinfo *type ;
   char *timezone_abbr ;
   tlinfo *leap_times ;
   unsigned char bc ;
   tlocinfo location ;
};
typedef struct timelib_tzinfo timelib_tzinfo;
struct timelib_special {
   unsigned int type ;
   timelib_sll amount ;
};
typedef struct timelib_special timelib_special;
struct timelib_rel_time {
   timelib_sll y ;
   timelib_sll m ;
   timelib_sll d ;
   timelib_sll h ;
   timelib_sll i ;
   timelib_sll s ;
   int weekday ;
   int weekday_behavior ;
   int first_last_day_of ;
   int invert ;
   timelib_sll days ;
   timelib_special special ;
   unsigned int have_weekday_relative ;
   unsigned int have_special_relative ;
};
typedef struct timelib_rel_time timelib_rel_time;
struct timelib_time_offset {
   int32_t offset ;
   unsigned int leap_secs ;
   unsigned int is_dst ;
   char *abbr ;
   timelib_sll transistion_time ;
};
typedef struct timelib_time_offset timelib_time_offset;
struct timelib_time {
   timelib_sll y ;
   timelib_sll m ;
   timelib_sll d ;
   timelib_sll h ;
   timelib_sll i ;
   timelib_sll s ;
   double f ;
   int z ;
   char *tz_abbr ;
   timelib_tzinfo *tz_info ;
   int dst ;
   timelib_rel_time relative ;
   timelib_sll sse ;
   unsigned int have_time ;
   unsigned int have_date ;
   unsigned int have_zone ;
   unsigned int have_relative ;
   unsigned int have_weeknr_day ;
   unsigned int sse_uptodate ;
   unsigned int tim_uptodate ;
   unsigned int is_localtime ;
   unsigned int zone_type ;
};
typedef struct timelib_time timelib_time;
struct timelib_error_message {
   int position ;
   char character ;
   char *message ;
};
typedef struct timelib_error_message timelib_error_message;
struct timelib_error_container {
   int warning_count ;
   struct timelib_error_message *warning_messages ;
   int error_count ;
   struct timelib_error_message *error_messages ;
};
typedef struct timelib_error_container timelib_error_container;
struct _timelib_tz_lookup_table {
   char *name ;
   int type ;
   float gmtoffset ;
   char *full_tz_name ;
};
typedef struct _timelib_tz_lookup_table timelib_tz_lookup_table;
struct _timelib_tzdb_index_entry {
   char *id ;
   unsigned int pos ;
};
typedef struct _timelib_tzdb_index_entry timelib_tzdb_index_entry;
struct _timelib_tzdb {
   char *version ;
   int index_size ;
   timelib_tzdb_index_entry const   *index ;
   unsigned char const   *data ;
};
typedef struct _timelib_tzdb timelib_tzdb;
extern  __attribute__((__nothrow__)) size_t __ctype_get_mb_cur_max(void) ;
__inline extern  __attribute__((__nothrow__)) double atof(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) int atoi(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) long atol(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) long long atoll(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) double strtod(char const   * __restrict  __nptr ,
                                                   char ** __restrict  __endptr )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) float strtof(char const   * __restrict  __nptr ,
                                                  char ** __restrict  __endptr )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long double strtold(char const   * __restrict  __nptr ,
                                                         char ** __restrict  __endptr )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long strtol(char const   * __restrict  __nptr ,
                                                 char ** __restrict  __endptr ,
                                                 int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) unsigned long strtoul(char const   * __restrict  __nptr ,
                                                           char ** __restrict  __endptr ,
                                                           int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long long strtoq(char const   * __restrict  __nptr ,
                                                      char ** __restrict  __endptr ,
                                                      int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) unsigned long long strtouq(char const   * __restrict  __nptr ,
                                                                char ** __restrict  __endptr ,
                                                                int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long long strtoll(char const   * __restrict  __nptr ,
                                                       char ** __restrict  __endptr ,
                                                       int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) unsigned long long strtoull(char const   * __restrict  __nptr ,
                                                                 char ** __restrict  __endptr ,
                                                                 int __base )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long strtol_l(char const   * __restrict  __nptr ,
                                                   char ** __restrict  __endptr ,
                                                   int __base ,
                                                   __locale_t __loc )  __attribute__((__nonnull__(1,4))) ;
extern  __attribute__((__nothrow__)) unsigned long strtoul_l(char const   * __restrict  __nptr ,
                                                             char ** __restrict  __endptr ,
                                                             int __base ,
                                                             __locale_t __loc )  __attribute__((__nonnull__(1,4))) ;
extern  __attribute__((__nothrow__)) long long strtoll_l(char const   * __restrict  __nptr ,
                                                         char ** __restrict  __endptr ,
                                                         int __base ,
                                                         __locale_t __loc )  __attribute__((__nonnull__(1,4))) ;
extern  __attribute__((__nothrow__)) unsigned long long strtoull_l(char const   * __restrict  __nptr ,
                                                                   char ** __restrict  __endptr ,
                                                                   int __base ,
                                                                   __locale_t __loc )  __attribute__((__nonnull__(1,4))) ;
extern  __attribute__((__nothrow__)) double strtod_l(char const   * __restrict  __nptr ,
                                                     char ** __restrict  __endptr ,
                                                     __locale_t __loc )  __attribute__((__nonnull__(1,3))) ;
extern  __attribute__((__nothrow__)) float strtof_l(char const   * __restrict  __nptr ,
                                                    char ** __restrict  __endptr ,
                                                    __locale_t __loc )  __attribute__((__nonnull__(1,3))) ;
extern  __attribute__((__nothrow__)) long double strtold_l(char const   * __restrict  __nptr ,
                                                           char ** __restrict  __endptr ,
                                                           __locale_t __loc )  __attribute__((__nonnull__(1,3))) ;
extern FILE *fopen(char const   * __restrict  __filename ,
                   char const   * __restrict  __modes ) ;
extern int fprintf(FILE * __restrict  __stream ,
                   char const   * __restrict  __format  , ...) ;
extern int fflush(FILE *__stream ) ;
extern int fclose(FILE *__stream ) ;
__inline extern  __attribute__((__nothrow__)) double atof(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern double atof(char const   *__nptr ) 
{ double tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22785\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = strtod((char const   */* __restrict  */)__nptr,
               (char **/* __restrict  */)((char **)((void *)0)));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22786\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) int atoi(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern int atoi(char const   *__nptr ) 
{ long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22787\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = strtol((char const   */* __restrict  */)__nptr,
               (char **/* __restrict  */)((char **)((void *)0)), 10);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22788\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((int )tmp);
}
}
__inline extern  __attribute__((__nothrow__)) long atol(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern long atol(char const   *__nptr ) 
{ long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22789\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = strtol((char const   */* __restrict  */)__nptr,
               (char **/* __restrict  */)((char **)((void *)0)), 10);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22790\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) long long atoll(char const   *__nptr )  __attribute__((__pure__,
__nonnull__(1))) ;
__inline extern long long atoll(char const   *__nptr ) 
{ long long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22791\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = strtoll((char const   */* __restrict  */)__nptr,
                (char **/* __restrict  */)((char **)((void *)0)), 10);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22792\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
extern  __attribute__((__nothrow__)) char *l64a(long __n ) ;
extern  __attribute__((__nothrow__)) long a64l(char const   *__s )  __attribute__((__pure__,
__nonnull__(1))) ;
extern int select(int __nfds , fd_set * __restrict  __readfds ,
                  fd_set * __restrict  __writefds ,
                  fd_set * __restrict  __exceptfds ,
                  struct timeval * __restrict  __timeout ) ;
extern int pselect(int __nfds , fd_set * __restrict  __readfds ,
                   fd_set * __restrict  __writefds ,
                   fd_set * __restrict  __exceptfds ,
                   struct timespec  const  * __restrict  __timeout ,
                   __sigset_t const   * __restrict  __sigmask ) ;
__inline extern  __attribute__((__nothrow__)) unsigned int gnu_dev_major(unsigned long long __dev ) ;
__inline extern  __attribute__((__nothrow__)) unsigned int gnu_dev_minor(unsigned long long __dev ) ;
__inline extern  __attribute__((__nothrow__)) unsigned long long gnu_dev_makedev(unsigned int __major ,
                                                                                 unsigned int __minor ) ;
__inline extern  __attribute__((__nothrow__)) unsigned int gnu_dev_major(unsigned long long __dev ) ;
__inline extern unsigned int gnu_dev_major(unsigned long long __dev ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22793\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((unsigned int )(((__dev >> 8) & 4095ULL) | (unsigned long long )((unsigned int )(__dev >> 32) & 4294963200U)));
}
}
__inline extern  __attribute__((__nothrow__)) unsigned int gnu_dev_minor(unsigned long long __dev ) ;
__inline extern unsigned int gnu_dev_minor(unsigned long long __dev ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22794\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((unsigned int )((__dev & 255ULL) | (unsigned long long )((unsigned int )(__dev >> 12) & 4294967040U)));
}
}
__inline extern  __attribute__((__nothrow__)) unsigned long long gnu_dev_makedev(unsigned int __major ,
                                                                                 unsigned int __minor ) ;
__inline extern unsigned long long gnu_dev_makedev(unsigned int __major ,
                                                   unsigned int __minor ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22795\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (((unsigned long long )((__minor & 255U) | ((__major & 4095U) << 8)) | ((unsigned long long )(__minor & 4294967040U) << 12)) | ((unsigned long long )(__major & 4294963200U) << 32));
}
}
extern  __attribute__((__nothrow__)) long random(void) ;
extern  __attribute__((__nothrow__)) void srandom(unsigned int __seed ) ;
extern  __attribute__((__nothrow__)) char *initstate(unsigned int __seed ,
                                                     char *__statebuf ,
                                                     size_t __statelen )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) char *setstate(char *__statebuf )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int random_r(struct random_data * __restrict  __buf ,
                                                  int32_t * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int srandom_r(unsigned int __seed ,
                                                   struct random_data *__buf )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) int initstate_r(unsigned int __seed ,
                                                     char * __restrict  __statebuf ,
                                                     size_t __statelen ,
                                                     struct random_data * __restrict  __buf )  __attribute__((__nonnull__(2,4))) ;
extern  __attribute__((__nothrow__)) int setstate_r(char * __restrict  __statebuf ,
                                                    struct random_data * __restrict  __buf )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int rand(void) ;
extern  __attribute__((__nothrow__)) void srand(unsigned int __seed ) ;
extern  __attribute__((__nothrow__)) int rand_r(unsigned int *__seed ) ;
extern  __attribute__((__nothrow__)) double drand48(void) ;
extern  __attribute__((__nothrow__)) double erand48(unsigned short *__xsubi )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long lrand48(void) ;
extern  __attribute__((__nothrow__)) long nrand48(unsigned short *__xsubi )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) long mrand48(void) ;
extern  __attribute__((__nothrow__)) long jrand48(unsigned short *__xsubi )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void srand48(long __seedval ) ;
extern  __attribute__((__nothrow__)) unsigned short *seed48(unsigned short *__seed16v )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void lcong48(unsigned short *__param )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int drand48_r(struct drand48_data * __restrict  __buffer ,
                                                   double * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int erand48_r(unsigned short *__xsubi ,
                                                   struct drand48_data * __restrict  __buffer ,
                                                   double * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int lrand48_r(struct drand48_data * __restrict  __buffer ,
                                                   long * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int nrand48_r(unsigned short *__xsubi ,
                                                   struct drand48_data * __restrict  __buffer ,
                                                   long * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int mrand48_r(struct drand48_data * __restrict  __buffer ,
                                                   long * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int jrand48_r(unsigned short *__xsubi ,
                                                   struct drand48_data * __restrict  __buffer ,
                                                   long * __restrict  __result )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int srand48_r(long __seedval ,
                                                   struct drand48_data *__buffer )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) int seed48_r(unsigned short *__seed16v ,
                                                  struct drand48_data *__buffer )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int lcong48_r(unsigned short *__param ,
                                                   struct drand48_data *__buffer )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *malloc(size_t __size )  __attribute__((__malloc__)) ;
extern  __attribute__((__nothrow__)) void *calloc(size_t __nmemb ,
                                                  size_t __size )  __attribute__((__malloc__)) ;
extern  __attribute__((__nothrow__)) void *realloc(void *__ptr , size_t __size )  __attribute__((__warn_unused_result__)) ;
extern  __attribute__((__nothrow__)) void free(void *__ptr ) ;
extern  __attribute__((__nothrow__)) void cfree(void *__ptr ) ;
extern  __attribute__((__nothrow__)) void *alloca(size_t __size ) ;
extern  __attribute__((__nothrow__)) void *valloc(size_t __size )  __attribute__((__malloc__)) ;
extern  __attribute__((__nothrow__)) int posix_memalign(void **__memptr ,
                                                        size_t __alignment ,
                                                        size_t __size )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__, __noreturn__)) void abort(void) ;
extern  __attribute__((__nothrow__)) int atexit(void (*__func)(void) )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int at_quick_exit(void (*__func)(void) )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int on_exit(void (*__func)(int __status ,
                                                                void *__arg ) ,
                                                 void *__arg )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__, __noreturn__)) void exit(int __status ) ;
extern  __attribute__((__nothrow__,
__noreturn__)) void quick_exit(int __status ) ;
extern  __attribute__((__nothrow__, __noreturn__)) void _Exit(int __status ) ;
extern  __attribute__((__nothrow__)) char *getenv(char const   *__name )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *__secure_getenv(char const   *__name )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int putenv(char *__string )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int setenv(char const   *__name ,
                                                char const   *__value ,
                                                int __replace )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) int unsetenv(char const   *__name )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int clearenv(void) ;
extern  __attribute__((__nothrow__)) char *mktemp(char *__template )  __attribute__((__nonnull__(1))) ;
extern int mkstemp(char *__template )  __attribute__((__nonnull__(1))) ;
extern int mkstemp64(char *__template )  __attribute__((__nonnull__(1))) ;
extern int mkstemps(char *__template , int __suffixlen )  __attribute__((__nonnull__(1))) ;
extern int mkstemps64(char *__template , int __suffixlen )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *mkdtemp(char *__template )  __attribute__((__nonnull__(1))) ;
extern int mkostemp(char *__template , int __flags )  __attribute__((__nonnull__(1))) ;
extern int mkostemp64(char *__template , int __flags )  __attribute__((__nonnull__(1))) ;
extern int mkostemps(char *__template , int __suffixlen , int __flags )  __attribute__((__nonnull__(1))) ;
extern int mkostemps64(char *__template , int __suffixlen , int __flags )  __attribute__((__nonnull__(1))) ;
extern int system(char const   *__command ) ;
extern  __attribute__((__nothrow__)) char *canonicalize_file_name(char const   *__name )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *realpath(char const   * __restrict  __name ,
                                                    char * __restrict  __resolved ) ;
extern void *bsearch(void const   *__key , void const   *__base ,
                     size_t __nmemb , size_t __size ,
                     int (*__compar)(void const   * , void const   * ) )  __attribute__((__nonnull__(1,2,5))) ;
extern void qsort(void *__base , size_t __nmemb , size_t __size ,
                  int (*__compar)(void const   * , void const   * ) )  __attribute__((__nonnull__(1,4))) ;
extern void qsort_r(void *__base , size_t __nmemb , size_t __size ,
                    int (*__compar)(void const   * , void const   * , void * ) ,
                    void *__arg )  __attribute__((__nonnull__(1,4))) ;
extern  __attribute__((__nothrow__)) int abs(int __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long labs(long __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long long llabs(long long __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) div_t div(int __numer , int __denom )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) ldiv_t ldiv(long __numer , long __denom )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) lldiv_t lldiv(long long __numer ,
                                                   long long __denom )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) char *ecvt(double __value , int __ndigit ,
                                                int * __restrict  __decpt ,
                                                int * __restrict  __sign )  __attribute__((__nonnull__(3,4))) ;
extern  __attribute__((__nothrow__)) char *fcvt(double __value , int __ndigit ,
                                                int * __restrict  __decpt ,
                                                int * __restrict  __sign )  __attribute__((__nonnull__(3,4))) ;
extern  __attribute__((__nothrow__)) char *gcvt(double __value , int __ndigit ,
                                                char *__buf )  __attribute__((__nonnull__(3))) ;
extern  __attribute__((__nothrow__)) char *qecvt(long double __value ,
                                                 int __ndigit ,
                                                 int * __restrict  __decpt ,
                                                 int * __restrict  __sign )  __attribute__((__nonnull__(3,4))) ;
extern  __attribute__((__nothrow__)) char *qfcvt(long double __value ,
                                                 int __ndigit ,
                                                 int * __restrict  __decpt ,
                                                 int * __restrict  __sign )  __attribute__((__nonnull__(3,4))) ;
extern  __attribute__((__nothrow__)) char *qgcvt(long double __value ,
                                                 int __ndigit , char *__buf )  __attribute__((__nonnull__(3))) ;
extern  __attribute__((__nothrow__)) int ecvt_r(double __value , int __ndigit ,
                                                int * __restrict  __decpt ,
                                                int * __restrict  __sign ,
                                                char * __restrict  __buf ,
                                                size_t __len )  __attribute__((__nonnull__(3,4,5))) ;
extern  __attribute__((__nothrow__)) int fcvt_r(double __value , int __ndigit ,
                                                int * __restrict  __decpt ,
                                                int * __restrict  __sign ,
                                                char * __restrict  __buf ,
                                                size_t __len )  __attribute__((__nonnull__(3,4,5))) ;
extern  __attribute__((__nothrow__)) int qecvt_r(long double __value ,
                                                 int __ndigit ,
                                                 int * __restrict  __decpt ,
                                                 int * __restrict  __sign ,
                                                 char * __restrict  __buf ,
                                                 size_t __len )  __attribute__((__nonnull__(3,4,5))) ;
extern  __attribute__((__nothrow__)) int qfcvt_r(long double __value ,
                                                 int __ndigit ,
                                                 int * __restrict  __decpt ,
                                                 int * __restrict  __sign ,
                                                 char * __restrict  __buf ,
                                                 size_t __len )  __attribute__((__nonnull__(3,4,5))) ;
extern  __attribute__((__nothrow__)) int mblen(char const   *__s , size_t __n ) ;
extern  __attribute__((__nothrow__)) int mbtowc(wchar_t * __restrict  __pwc ,
                                                char const   * __restrict  __s ,
                                                size_t __n ) ;
extern  __attribute__((__nothrow__)) int wctomb(char *__s , wchar_t __wchar ) ;
extern  __attribute__((__nothrow__)) size_t mbstowcs(wchar_t * __restrict  __pwcs ,
                                                     char const   * __restrict  __s ,
                                                     size_t __n ) ;
extern  __attribute__((__nothrow__)) size_t wcstombs(char * __restrict  __s ,
                                                     wchar_t const   * __restrict  __pwcs ,
                                                     size_t __n ) ;
extern  __attribute__((__nothrow__)) int rpmatch(char const   *__response )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int getsubopt(char ** __restrict  __optionp ,
                                                   char * const  * __restrict  __tokens ,
                                                   char ** __restrict  __valuep )  __attribute__((__nonnull__(1,2,3))) ;
extern  __attribute__((__nothrow__)) void setkey(char const   *__key )  __attribute__((__nonnull__(1))) ;
extern int posix_openpt(int __oflag ) ;
extern  __attribute__((__nothrow__)) int grantpt(int __fd ) ;
extern  __attribute__((__nothrow__)) int unlockpt(int __fd ) ;
extern  __attribute__((__nothrow__)) char *ptsname(int __fd ) ;
extern  __attribute__((__nothrow__)) int ptsname_r(int __fd , char *__buf ,
                                                   size_t __buflen )  __attribute__((__nonnull__(2))) ;
extern int getpt(void) ;
extern  __attribute__((__nothrow__)) int getloadavg(double *__loadavg ,
                                                    int __nelem )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void *memcpy(void * __restrict  __dest ,
                                                  void const   * __restrict  __src ,
                                                  size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *memmove(void *__dest ,
                                                   void const   *__src ,
                                                   size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *memccpy(void * __restrict  __dest ,
                                                   void const   * __restrict  __src ,
                                                   int __c , size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *memset(void *__s , int __c ,
                                                  size_t __n )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int memcmp(void const   *__s1 ,
                                                void const   *__s2 , size_t __n )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *memchr(void const   *__s , int __c ,
                                                  size_t __n )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void *rawmemchr(void const   *__s ,
                                                     int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void *memrchr(void const   *__s , int __c ,
                                                   size_t __n )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *strcpy(char * __restrict  __dest ,
                                                  char const   * __restrict  __src )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strncpy(char * __restrict  __dest ,
                                                   char const   * __restrict  __src ,
                                                   size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strcat(char * __restrict  __dest ,
                                                  char const   * __restrict  __src )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strncat(char * __restrict  __dest ,
                                                   char const   * __restrict  __src ,
                                                   size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strcmp(char const   *__s1 ,
                                                char const   *__s2 )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strncmp(char const   *__s1 ,
                                                 char const   *__s2 ,
                                                 size_t __n )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strcoll(char const   *__s1 ,
                                                 char const   *__s2 )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) size_t strxfrm(char * __restrict  __dest ,
                                                    char const   * __restrict  __src ,
                                                    size_t __n )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) int strcoll_l(char const   *__s1 ,
                                                   char const   *__s2 ,
                                                   __locale_t __l )  __attribute__((__pure__,
__nonnull__(1,2,3))) ;
extern  __attribute__((__nothrow__)) size_t strxfrm_l(char *__dest ,
                                                      char const   *__src ,
                                                      size_t __n ,
                                                      __locale_t __l )  __attribute__((__nonnull__(2,4))) ;
extern  __attribute__((__nothrow__)) char *strdup(char const   *__s )  __attribute__((__nonnull__(1),
__malloc__)) ;
extern  __attribute__((__nothrow__)) char *strndup(char const   *__string ,
                                                   size_t __n )  __attribute__((__nonnull__(1),
__malloc__)) ;
extern  __attribute__((__nothrow__)) char *strchr(char const   *__s , int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *strrchr(char const   *__s , int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *strchrnul(char const   *__s ,
                                                     int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) size_t strcspn(char const   *__s ,
                                                    char const   *__reject )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) size_t strspn(char const   *__s ,
                                                   char const   *__accept )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strpbrk(char const   *__s ,
                                                   char const   *__accept )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strstr(char const   *__haystack ,
                                                  char const   *__needle )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strtok(char * __restrict  __s ,
                                                  char const   * __restrict  __delim )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) char *__strtok_r(char * __restrict  __s ,
                                                      char const   * __restrict  __delim ,
                                                      char ** __restrict  __save_ptr )  __attribute__((__nonnull__(2,3))) ;
extern  __attribute__((__nothrow__)) char *strtok_r(char * __restrict  __s ,
                                                    char const   * __restrict  __delim ,
                                                    char ** __restrict  __save_ptr )  __attribute__((__nonnull__(2,3))) ;
extern  __attribute__((__nothrow__)) char *strcasestr(char const   *__haystack ,
                                                      char const   *__needle )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *memmem(void const   *__haystack ,
                                                  size_t __haystacklen ,
                                                  void const   *__needle ,
                                                  size_t __needlelen )  __attribute__((__pure__,
__nonnull__(1,3))) ;
extern  __attribute__((__nothrow__)) void *__mempcpy(void * __restrict  __dest ,
                                                     void const   * __restrict  __src ,
                                                     size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void *mempcpy(void * __restrict  __dest ,
                                                   void const   * __restrict  __src ,
                                                   size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) size_t strlen(char const   *__s )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) size_t strnlen(char const   *__string ,
                                                    size_t __maxlen )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *strerror(int __errnum ) ;
extern  __attribute__((__nothrow__)) char *strerror_r(int __errnum ,
                                                      char *__buf ,
                                                      size_t __buflen )  __attribute__((__nonnull__(2))) ;
extern  __attribute__((__nothrow__)) char *strerror_l(int __errnum ,
                                                      __locale_t __l ) ;
extern  __attribute__((__nothrow__)) void __bzero(void *__s , size_t __n )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void bcopy(void const   *__src ,
                                                void *__dest , size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) void bzero(void *__s , size_t __n )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int bcmp(void const   *__s1 ,
                                              void const   *__s2 , size_t __n )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *index(char const   *__s , int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *rindex(char const   *__s , int __c )  __attribute__((__pure__,
__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) int ffs(int __i )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int ffsl(long __l )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int ffsll(long long __ll )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int strcasecmp(char const   *__s1 ,
                                                    char const   *__s2 )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strncasecmp(char const   *__s1 ,
                                                     char const   *__s2 ,
                                                     size_t __n )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strcasecmp_l(char const   *__s1 ,
                                                      char const   *__s2 ,
                                                      __locale_t __loc )  __attribute__((__pure__,
__nonnull__(1,2,3))) ;
extern  __attribute__((__nothrow__)) int strncasecmp_l(char const   *__s1 ,
                                                       char const   *__s2 ,
                                                       size_t __n ,
                                                       __locale_t __loc )  __attribute__((__pure__,
__nonnull__(1,2,4))) ;
extern  __attribute__((__nothrow__)) char *strsep(char ** __restrict  __stringp ,
                                                  char const   * __restrict  __delim )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strsignal(int __sig ) ;
extern  __attribute__((__nothrow__)) char *__stpcpy(char * __restrict  __dest ,
                                                    char const   * __restrict  __src )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *stpcpy(char * __restrict  __dest ,
                                                  char const   * __restrict  __src )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *__stpncpy(char * __restrict  __dest ,
                                                     char const   * __restrict  __src ,
                                                     size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *stpncpy(char * __restrict  __dest ,
                                                   char const   * __restrict  __src ,
                                                   size_t __n )  __attribute__((__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) int strverscmp(char const   *__s1 ,
                                                    char const   *__s2 )  __attribute__((__pure__,
__nonnull__(1,2))) ;
extern  __attribute__((__nothrow__)) char *strfry(char *__string )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) void *memfrob(void *__s , size_t __n )  __attribute__((__nonnull__(1))) ;
extern  __attribute__((__nothrow__)) char *basename(char const   *__filename )  __attribute__((__nonnull__(1))) ;
extern void *__rawmemchr(void const   *__s , int __c ) ;
__inline extern size_t __strcspn_c1(char const   *__s , int __reject ) ;
__inline extern size_t __strcspn_c1(char const   *__s , int __reject ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22800\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22801\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22798\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) != 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22797\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*(__s + __result) != (int const   )__reject) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22796\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22799\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22802\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern size_t __strcspn_c2(char const   *__s , int __reject1 ,
                                    int __reject2 ) ;
__inline extern size_t __strcspn_c2(char const   *__s , int __reject1 ,
                                    int __reject2 ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22808\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22809\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22806\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) != 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22805\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*(__s + __result) != (int const   )__reject1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22804\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int const   )*(__s + __result) != (int const   )__reject2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22803\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        } else {
          break;
        }
      } else {
        break;
      }
    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22807\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22810\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern size_t __strcspn_c3(char const   *__s , int __reject1 ,
                                    int __reject2 , int __reject3 ) ;
__inline extern size_t __strcspn_c3(char const   *__s , int __reject1 ,
                                    int __reject2 , int __reject3 ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22817\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22818\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22815\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) != 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22814\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*(__s + __result) != (int const   )__reject1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22813\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int const   )*(__s + __result) != (int const   )__reject2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22812\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if ((int const   )*(__s + __result) != (int const   )__reject3) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22811\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }

          } else {
            break;
          }
        } else {
          break;
        }
      } else {
        break;
      }
    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22816\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22819\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern size_t __strspn_c1(char const   *__s , int __accept ) ;
__inline extern size_t __strspn_c1(char const   *__s , int __accept ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22823\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22824\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22821\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) == (int const   )__accept) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22820\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22822\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22825\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern size_t __strspn_c2(char const   *__s , int __accept1 ,
                                   int __accept2 ) ;
__inline extern size_t __strspn_c2(char const   *__s , int __accept1 ,
                                   int __accept2 ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22831\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22832\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22829\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) == (int const   )__accept1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22826\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22828\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*(__s + __result) == (int const   )__accept2) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22827\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22830\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22833\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern size_t __strspn_c3(char const   *__s , int __accept1 ,
                                   int __accept2 , int __accept3 ) ;
__inline extern size_t __strspn_c3(char const   *__s , int __accept1 ,
                                   int __accept2 , int __accept3 ) 
{ register size_t __result ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22841\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (size_t )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22842\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22839\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*(__s + __result) == (int const   )__accept1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22834\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22838\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*(__s + __result) == (int const   )__accept2) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22835\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22837\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int const   )*(__s + __result) == (int const   )__accept3) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22836\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        } else {
          break;
        }
      }
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22840\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22843\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
__inline extern char *__strpbrk_c2(char const   *__s , int __accept1 ,
                                   int __accept2 ) ;
__inline extern char *__strpbrk_c2(char const   *__s , int __accept1 ,
                                   int __accept2 ) 
{ char *tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22851\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22847\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*__s != 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22846\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*__s != (int const   )__accept1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22845\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int const   )*__s != (int const   )__accept2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22844\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        } else {
          break;
        }
      } else {
        break;
      }
    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22848\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __s ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22852\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((int const   )*__s == 0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22849\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = (char *)((void *)0);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22850\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = (char *)((unsigned int )__s);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22853\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern char *__strpbrk_c3(char const   *__s , int __accept1 ,
                                   int __accept2 , int __accept3 ) ;
__inline extern char *__strpbrk_c3(char const   *__s , int __accept1 ,
                                   int __accept2 , int __accept3 ) 
{ char *tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22862\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22858\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int const   )*__s != 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22857\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int const   )*__s != (int const   )__accept1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22856\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int const   )*__s != (int const   )__accept2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22855\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if ((int const   )*__s != (int const   )__accept3) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22854\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }

          } else {
            break;
          }
        } else {
          break;
        }
      } else {
        break;
      }
    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22859\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __s ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22863\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((int const   )*__s == 0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22860\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = (char *)((void *)0);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22861\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = (char *)((unsigned int )__s);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22864\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern char *__strtok_r_1c(char *__s , char __sep , char **__nextp ) ;
__inline extern char *__strtok_r_1c(char *__s , char __sep , char **__nextp ) 
{ char *__result ;
  char *tmp ;
  char *tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22882\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((unsigned int )__s == (unsigned int )((void *)0)) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22865\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __s = *__nextp;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22866\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22883\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22868\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((int )*__s == (int )__sep) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22867\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22869\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __s ++;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22884\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __result = (char *)((void *)0);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22885\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((int )*__s != 0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22877\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = __s;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22878\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __s ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22879\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __result = tmp;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22880\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22873\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*__s != 0) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22870\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22874\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp___0 = __s;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22875\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      __s ++;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22876\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*tmp___0 == (int )__sep) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22871\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        *(__s + -1) = (char )'\000';
        break;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22872\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22881\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22886\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  *__nextp = __s;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22887\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__result);
}
}
extern char *__strsep_g(char **__stringp , char const   *__delim ) ;
__inline extern char *__strsep_1c(char **__s , char __reject ) ;
__inline extern char *__strsep_1c(char **__s , char __reject ) 
{ register char *__retval ;
  char *tmp ;
  char *tmp___0 ;
  void *tmp___1 ;
  char *tmp___2 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22897\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __retval = *__s;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22898\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((unsigned int )__retval != (unsigned int )((void *)0)) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22892\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = __builtin_strchr(__retval, (int )__reject);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22893\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = tmp___2;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22894\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *__s = tmp___0;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22895\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((unsigned int )tmp___0 != (unsigned int )((void *)0)) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22888\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = *__s;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22889\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      (*__s) ++;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22890\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      *tmp = (char )'\000';
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22891\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22896\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22899\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__retval);
}
}
__inline extern char *__strsep_2c(char **__s , char __reject1 , char __reject2 ) ;
__inline extern char *__strsep_2c(char **__s , char __reject1 , char __reject2 ) 
{ register char *__retval ;
  register char *__cp ;
  char *tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22917\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __retval = *__s;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22918\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((unsigned int )__retval != (unsigned int )((void *)0)) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22913\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __cp = __retval;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22914\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22910\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*__cp == 0) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22900\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        __cp = (char *)((void *)0);
        break;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22901\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22911\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*__cp == (int )__reject1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22902\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp = __cp;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22903\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        __cp ++;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22904\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        *tmp = (char )'\000';
        break;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22909\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int )*__cp == (int )__reject2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22905\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          tmp = __cp;
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22906\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          __cp ++;
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22907\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          *tmp = (char )'\000';
          break;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22908\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        }
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22912\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      __cp ++;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22915\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *__s = __cp;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22916\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22919\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__retval);
}
}
__inline extern char *__strsep_3c(char **__s , char __reject1 , char __reject2 ,
                                  char __reject3 ) ;
__inline extern char *__strsep_3c(char **__s , char __reject1 , char __reject2 ,
                                  char __reject3 ) 
{ register char *__retval ;
  register char *__cp ;
  char *tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22941\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __retval = *__s;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22942\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((unsigned int )__retval != (unsigned int )((void *)0)) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22937\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    __cp = __retval;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22938\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22934\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*__cp == 0) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22920\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        __cp = (char *)((void *)0);
        break;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22921\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22935\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if ((int )*__cp == (int )__reject1) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22922\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp = __cp;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22923\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        __cp ++;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22924\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        *tmp = (char )'\000';
        break;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "22933\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if ((int )*__cp == (int )__reject2) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22925\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          tmp = __cp;
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22926\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          __cp ++;
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22927\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          *tmp = (char )'\000';
          break;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "22932\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if ((int )*__cp == (int )__reject3) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22928\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            tmp = __cp;
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22929\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            __cp ++;
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22930\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            *tmp = (char )'\000';
            break;
          } else {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "22931\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }

          }
        }
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22936\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      __cp ++;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22939\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *__s = __cp;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22940\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22943\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__retval);
}
}
extern  __attribute__((__nothrow__)) char *__strdup(char const   *__string )  __attribute__((__malloc__)) ;
extern  __attribute__((__nothrow__)) char *__strndup(char const   *__string ,
                                                     size_t __n )  __attribute__((__malloc__)) ;
extern  __attribute__((__nothrow__)) double acos(double __x ) ;
extern  __attribute__((__nothrow__)) double __acos(double __x ) ;
extern  __attribute__((__nothrow__)) double asin(double __x ) ;
extern  __attribute__((__nothrow__)) double __asin(double __x ) ;
extern  __attribute__((__nothrow__)) double atan(double __x ) ;
extern  __attribute__((__nothrow__)) double __atan(double __x ) ;
extern  __attribute__((__nothrow__)) double atan2(double __y , double __x ) ;
extern  __attribute__((__nothrow__)) double __atan2(double __y , double __x ) ;
extern  __attribute__((__nothrow__)) double cos(double __x ) ;
extern  __attribute__((__nothrow__)) double __cos(double __x ) ;
extern  __attribute__((__nothrow__)) double sin(double __x ) ;
extern  __attribute__((__nothrow__)) double __sin(double __x ) ;
extern  __attribute__((__nothrow__)) double tan(double __x ) ;
extern  __attribute__((__nothrow__)) double __tan(double __x ) ;
extern  __attribute__((__nothrow__)) double cosh(double __x ) ;
extern  __attribute__((__nothrow__)) double __cosh(double __x ) ;
extern  __attribute__((__nothrow__)) double sinh(double __x ) ;
extern  __attribute__((__nothrow__)) double __sinh(double __x ) ;
extern  __attribute__((__nothrow__)) double tanh(double __x ) ;
extern  __attribute__((__nothrow__)) double __tanh(double __x ) ;
extern  __attribute__((__nothrow__)) void sincos(double __x , double *__sinx ,
                                                 double *__cosx ) ;
extern  __attribute__((__nothrow__)) void __sincos(double __x , double *__sinx ,
                                                   double *__cosx ) ;
extern  __attribute__((__nothrow__)) double acosh(double __x ) ;
extern  __attribute__((__nothrow__)) double __acosh(double __x ) ;
extern  __attribute__((__nothrow__)) double asinh(double __x ) ;
extern  __attribute__((__nothrow__)) double __asinh(double __x ) ;
extern  __attribute__((__nothrow__)) double atanh(double __x ) ;
extern  __attribute__((__nothrow__)) double __atanh(double __x ) ;
extern  __attribute__((__nothrow__)) double exp(double __x ) ;
extern  __attribute__((__nothrow__)) double __exp(double __x ) ;
extern  __attribute__((__nothrow__)) double frexp(double __x , int *__exponent ) ;
extern  __attribute__((__nothrow__)) double __frexp(double __x ,
                                                    int *__exponent ) ;
extern  __attribute__((__nothrow__)) double ldexp(double __x , int __exponent ) ;
extern  __attribute__((__nothrow__)) double __ldexp(double __x , int __exponent ) ;
extern  __attribute__((__nothrow__)) double log(double __x ) ;
extern  __attribute__((__nothrow__)) double __log(double __x ) ;
extern  __attribute__((__nothrow__)) double log10(double __x ) ;
extern  __attribute__((__nothrow__)) double __log10(double __x ) ;
extern  __attribute__((__nothrow__)) double modf(double __x , double *__iptr ) ;
extern  __attribute__((__nothrow__)) double __modf(double __x , double *__iptr ) ;
extern  __attribute__((__nothrow__)) double exp10(double __x ) ;
extern  __attribute__((__nothrow__)) double __exp10(double __x ) ;
extern  __attribute__((__nothrow__)) double pow10(double __x ) ;
extern  __attribute__((__nothrow__)) double __pow10(double __x ) ;
extern  __attribute__((__nothrow__)) double expm1(double __x ) ;
extern  __attribute__((__nothrow__)) double __expm1(double __x ) ;
extern  __attribute__((__nothrow__)) double log1p(double __x ) ;
extern  __attribute__((__nothrow__)) double __log1p(double __x ) ;
extern  __attribute__((__nothrow__)) double logb(double __x ) ;
extern  __attribute__((__nothrow__)) double __logb(double __x ) ;
extern  __attribute__((__nothrow__)) double exp2(double __x ) ;
extern  __attribute__((__nothrow__)) double __exp2(double __x ) ;
extern  __attribute__((__nothrow__)) double log2(double __x ) ;
extern  __attribute__((__nothrow__)) double __log2(double __x ) ;
extern  __attribute__((__nothrow__)) double pow(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __pow(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double sqrt(double __x ) ;
extern  __attribute__((__nothrow__)) double __sqrt(double __x ) ;
extern  __attribute__((__nothrow__)) double hypot(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __hypot(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double cbrt(double __x ) ;
extern  __attribute__((__nothrow__)) double __cbrt(double __x ) ;
__inline extern  __attribute__((__nothrow__)) double ceil(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __ceil(double __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) double fabs(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __fabs(double __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) double floor(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __floor(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double fmod(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __fmod(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) int __isinf(double __value )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) int __finite(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isinf(double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int finite(double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double drem(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __drem(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double significand(double __x ) ;
extern  __attribute__((__nothrow__)) double __significand(double __x ) ;
extern  __attribute__((__nothrow__)) double copysign(double __x , double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __copysign(double __x , double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double nan(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __nan(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int __isnan(double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isnan(double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double j0(double  ) ;
extern  __attribute__((__nothrow__)) double __j0(double  ) ;
extern  __attribute__((__nothrow__)) double j1(double  ) ;
extern  __attribute__((__nothrow__)) double __j1(double  ) ;
extern  __attribute__((__nothrow__)) double jn(int  , double  ) ;
extern  __attribute__((__nothrow__)) double __jn(int  , double  ) ;
extern  __attribute__((__nothrow__)) double y0(double  ) ;
extern  __attribute__((__nothrow__)) double __y0(double  ) ;
extern  __attribute__((__nothrow__)) double y1(double  ) ;
extern  __attribute__((__nothrow__)) double __y1(double  ) ;
extern  __attribute__((__nothrow__)) double yn(int  , double  ) ;
extern  __attribute__((__nothrow__)) double __yn(int  , double  ) ;
extern  __attribute__((__nothrow__)) double erf(double  ) ;
extern  __attribute__((__nothrow__)) double __erf(double  ) ;
extern  __attribute__((__nothrow__)) double erfc(double  ) ;
extern  __attribute__((__nothrow__)) double __erfc(double  ) ;
extern  __attribute__((__nothrow__)) double lgamma(double  ) ;
extern  __attribute__((__nothrow__)) double __lgamma(double  ) ;
extern  __attribute__((__nothrow__)) double tgamma(double  ) ;
extern  __attribute__((__nothrow__)) double __tgamma(double  ) ;
extern  __attribute__((__nothrow__)) double gamma(double  ) ;
extern  __attribute__((__nothrow__)) double __gamma(double  ) ;
extern  __attribute__((__nothrow__)) double lgamma_r(double  , int *__signgamp ) ;
extern  __attribute__((__nothrow__)) double __lgamma_r(double  ,
                                                       int *__signgamp ) ;
extern  __attribute__((__nothrow__)) double rint(double __x ) ;
extern  __attribute__((__nothrow__)) double __rint(double __x ) ;
extern  __attribute__((__nothrow__)) double nextafter(double __x , double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __nextafter(double __x , double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double nexttoward(double __x ,
                                                       long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __nexttoward(double __x ,
                                                         long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double remainder(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __remainder(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double scalbn(double __x , int __n ) ;
extern  __attribute__((__nothrow__)) double __scalbn(double __x , int __n ) ;
extern  __attribute__((__nothrow__)) int ilogb(double __x ) ;
extern  __attribute__((__nothrow__)) int __ilogb(double __x ) ;
extern  __attribute__((__nothrow__)) double scalbln(double __x , long __n ) ;
extern  __attribute__((__nothrow__)) double __scalbln(double __x , long __n ) ;
extern  __attribute__((__nothrow__)) double nearbyint(double __x ) ;
extern  __attribute__((__nothrow__)) double __nearbyint(double __x ) ;
extern  __attribute__((__nothrow__)) double round(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __round(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double trunc(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double __trunc(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double remquo(double __x , double __y ,
                                                   int *__quo ) ;
extern  __attribute__((__nothrow__)) double __remquo(double __x , double __y ,
                                                     int *__quo ) ;
__inline extern  __attribute__((__nothrow__)) long lrint(double __x ) ;
extern  __attribute__((__nothrow__)) long __lrint(double __x ) ;
__inline extern  __attribute__((__nothrow__)) long long llrint(double __x ) ;
extern  __attribute__((__nothrow__)) long long __llrint(double __x ) ;
extern  __attribute__((__nothrow__)) long lround(double __x ) ;
extern  __attribute__((__nothrow__)) long __lround(double __x ) ;
extern  __attribute__((__nothrow__)) long long llround(double __x ) ;
extern  __attribute__((__nothrow__)) long long __llround(double __x ) ;
extern  __attribute__((__nothrow__)) double fdim(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __fdim(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double fmax(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __fmax(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double fmin(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) double __fmin(double __x , double __y ) ;
extern  __attribute__((__nothrow__)) int __fpclassify(double __value )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) int __signbit(double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) double fma(double __x , double __y ,
                                                double __z ) ;
extern  __attribute__((__nothrow__)) double __fma(double __x , double __y ,
                                                  double __z ) ;
extern  __attribute__((__nothrow__)) double scalb(double __x , double __n ) ;
extern  __attribute__((__nothrow__)) double __scalb(double __x , double __n ) ;
extern  __attribute__((__nothrow__)) float acosf(float __x ) ;
extern  __attribute__((__nothrow__)) float __acosf(float __x ) ;
extern  __attribute__((__nothrow__)) float asinf(float __x ) ;
extern  __attribute__((__nothrow__)) float __asinf(float __x ) ;
extern  __attribute__((__nothrow__)) float atanf(float __x ) ;
extern  __attribute__((__nothrow__)) float __atanf(float __x ) ;
extern  __attribute__((__nothrow__)) float atan2f(float __y , float __x ) ;
extern  __attribute__((__nothrow__)) float __atan2f(float __y , float __x ) ;
extern  __attribute__((__nothrow__)) float cosf(float __x ) ;
extern  __attribute__((__nothrow__)) float __cosf(float __x ) ;
extern  __attribute__((__nothrow__)) float sinf(float __x ) ;
extern  __attribute__((__nothrow__)) float __sinf(float __x ) ;
extern  __attribute__((__nothrow__)) float tanf(float __x ) ;
extern  __attribute__((__nothrow__)) float __tanf(float __x ) ;
extern  __attribute__((__nothrow__)) float coshf(float __x ) ;
extern  __attribute__((__nothrow__)) float __coshf(float __x ) ;
extern  __attribute__((__nothrow__)) float sinhf(float __x ) ;
extern  __attribute__((__nothrow__)) float __sinhf(float __x ) ;
extern  __attribute__((__nothrow__)) float tanhf(float __x ) ;
extern  __attribute__((__nothrow__)) float __tanhf(float __x ) ;
extern  __attribute__((__nothrow__)) void sincosf(float __x , float *__sinx ,
                                                  float *__cosx ) ;
extern  __attribute__((__nothrow__)) void __sincosf(float __x , float *__sinx ,
                                                    float *__cosx ) ;
extern  __attribute__((__nothrow__)) float acoshf(float __x ) ;
extern  __attribute__((__nothrow__)) float __acoshf(float __x ) ;
extern  __attribute__((__nothrow__)) float asinhf(float __x ) ;
extern  __attribute__((__nothrow__)) float __asinhf(float __x ) ;
extern  __attribute__((__nothrow__)) float atanhf(float __x ) ;
extern  __attribute__((__nothrow__)) float __atanhf(float __x ) ;
extern  __attribute__((__nothrow__)) float expf(float __x ) ;
extern  __attribute__((__nothrow__)) float __expf(float __x ) ;
extern  __attribute__((__nothrow__)) float frexpf(float __x , int *__exponent ) ;
extern  __attribute__((__nothrow__)) float __frexpf(float __x , int *__exponent ) ;
extern  __attribute__((__nothrow__)) float ldexpf(float __x , int __exponent ) ;
extern  __attribute__((__nothrow__)) float __ldexpf(float __x , int __exponent ) ;
extern  __attribute__((__nothrow__)) float logf(float __x ) ;
extern  __attribute__((__nothrow__)) float __logf(float __x ) ;
extern  __attribute__((__nothrow__)) float log10f(float __x ) ;
extern  __attribute__((__nothrow__)) float __log10f(float __x ) ;
extern  __attribute__((__nothrow__)) float modff(float __x , float *__iptr ) ;
extern  __attribute__((__nothrow__)) float __modff(float __x , float *__iptr ) ;
extern  __attribute__((__nothrow__)) float exp10f(float __x ) ;
extern  __attribute__((__nothrow__)) float __exp10f(float __x ) ;
extern  __attribute__((__nothrow__)) float pow10f(float __x ) ;
extern  __attribute__((__nothrow__)) float __pow10f(float __x ) ;
extern  __attribute__((__nothrow__)) float expm1f(float __x ) ;
extern  __attribute__((__nothrow__)) float __expm1f(float __x ) ;
extern  __attribute__((__nothrow__)) float log1pf(float __x ) ;
extern  __attribute__((__nothrow__)) float __log1pf(float __x ) ;
extern  __attribute__((__nothrow__)) float logbf(float __x ) ;
extern  __attribute__((__nothrow__)) float __logbf(float __x ) ;
extern  __attribute__((__nothrow__)) float exp2f(float __x ) ;
extern  __attribute__((__nothrow__)) float __exp2f(float __x ) ;
extern  __attribute__((__nothrow__)) float log2f(float __x ) ;
extern  __attribute__((__nothrow__)) float __log2f(float __x ) ;
extern  __attribute__((__nothrow__)) float powf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __powf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float sqrtf(float __x ) ;
extern  __attribute__((__nothrow__)) float __sqrtf(float __x ) ;
extern  __attribute__((__nothrow__)) float hypotf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __hypotf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float cbrtf(float __x ) ;
extern  __attribute__((__nothrow__)) float __cbrtf(float __x ) ;
__inline extern  __attribute__((__nothrow__)) float ceilf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __ceilf(float __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) float fabsf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __fabsf(float __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) float floorf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __floorf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float fmodf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __fmodf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) int __isinff(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int __finitef(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isinff(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int finitef(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float dremf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __dremf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float significandf(float __x ) ;
extern  __attribute__((__nothrow__)) float __significandf(float __x ) ;
extern  __attribute__((__nothrow__)) float copysignf(float __x , float __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __copysignf(float __x , float __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float nanf(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __nanf(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int __isnanf(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isnanf(float __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float j0f(float  ) ;
extern  __attribute__((__nothrow__)) float __j0f(float  ) ;
extern  __attribute__((__nothrow__)) float j1f(float  ) ;
extern  __attribute__((__nothrow__)) float __j1f(float  ) ;
extern  __attribute__((__nothrow__)) float jnf(int  , float  ) ;
extern  __attribute__((__nothrow__)) float __jnf(int  , float  ) ;
extern  __attribute__((__nothrow__)) float y0f(float  ) ;
extern  __attribute__((__nothrow__)) float __y0f(float  ) ;
extern  __attribute__((__nothrow__)) float y1f(float  ) ;
extern  __attribute__((__nothrow__)) float __y1f(float  ) ;
extern  __attribute__((__nothrow__)) float ynf(int  , float  ) ;
extern  __attribute__((__nothrow__)) float __ynf(int  , float  ) ;
extern  __attribute__((__nothrow__)) float erff(float  ) ;
extern  __attribute__((__nothrow__)) float __erff(float  ) ;
extern  __attribute__((__nothrow__)) float erfcf(float  ) ;
extern  __attribute__((__nothrow__)) float __erfcf(float  ) ;
extern  __attribute__((__nothrow__)) float lgammaf(float  ) ;
extern  __attribute__((__nothrow__)) float __lgammaf(float  ) ;
extern  __attribute__((__nothrow__)) float tgammaf(float  ) ;
extern  __attribute__((__nothrow__)) float __tgammaf(float  ) ;
extern  __attribute__((__nothrow__)) float gammaf(float  ) ;
extern  __attribute__((__nothrow__)) float __gammaf(float  ) ;
extern  __attribute__((__nothrow__)) float lgammaf_r(float  , int *__signgamp ) ;
extern  __attribute__((__nothrow__)) float __lgammaf_r(float  , int *__signgamp ) ;
extern  __attribute__((__nothrow__)) float rintf(float __x ) ;
extern  __attribute__((__nothrow__)) float __rintf(float __x ) ;
extern  __attribute__((__nothrow__)) float nextafterf(float __x , float __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __nextafterf(float __x , float __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float nexttowardf(float __x ,
                                                       long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __nexttowardf(float __x ,
                                                         long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float remainderf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __remainderf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float scalbnf(float __x , int __n ) ;
extern  __attribute__((__nothrow__)) float __scalbnf(float __x , int __n ) ;
extern  __attribute__((__nothrow__)) int ilogbf(float __x ) ;
extern  __attribute__((__nothrow__)) int __ilogbf(float __x ) ;
extern  __attribute__((__nothrow__)) float scalblnf(float __x , long __n ) ;
extern  __attribute__((__nothrow__)) float __scalblnf(float __x , long __n ) ;
extern  __attribute__((__nothrow__)) float nearbyintf(float __x ) ;
extern  __attribute__((__nothrow__)) float __nearbyintf(float __x ) ;
extern  __attribute__((__nothrow__)) float roundf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __roundf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float truncf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float __truncf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float remquof(float __x , float __y ,
                                                   int *__quo ) ;
extern  __attribute__((__nothrow__)) float __remquof(float __x , float __y ,
                                                     int *__quo ) ;
__inline extern  __attribute__((__nothrow__)) long lrintf(float __x ) ;
extern  __attribute__((__nothrow__)) long __lrintf(float __x ) ;
__inline extern  __attribute__((__nothrow__)) long long llrintf(float __x ) ;
extern  __attribute__((__nothrow__)) long long __llrintf(float __x ) ;
extern  __attribute__((__nothrow__)) long lroundf(float __x ) ;
extern  __attribute__((__nothrow__)) long __lroundf(float __x ) ;
extern  __attribute__((__nothrow__)) long long llroundf(float __x ) ;
extern  __attribute__((__nothrow__)) long long __llroundf(float __x ) ;
extern  __attribute__((__nothrow__)) float fdimf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __fdimf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float fmaxf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __fmaxf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float fminf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) float __fminf(float __x , float __y ) ;
extern  __attribute__((__nothrow__)) int __fpclassifyf(float __value )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) int __signbitf(float __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) float fmaf(float __x , float __y ,
                                                float __z ) ;
extern  __attribute__((__nothrow__)) float __fmaf(float __x , float __y ,
                                                  float __z ) ;
extern  __attribute__((__nothrow__)) float scalbf(float __x , float __n ) ;
extern  __attribute__((__nothrow__)) float __scalbf(float __x , float __n ) ;
extern  __attribute__((__nothrow__)) long double acosl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __acosl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double asinl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __asinl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double atanl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __atanl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double atan2l(long double __y ,
                                                        long double __x ) ;
__inline extern  __attribute__((__nothrow__)) long double __atan2l(long double __y ,
                                                                   long double __x ) ;
extern  __attribute__((__nothrow__)) long double cosl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __cosl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double sinl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __sinl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double tanl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __tanl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double coshl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __coshl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double sinhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __sinhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double tanhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __tanhl(long double __x ) ;
extern  __attribute__((__nothrow__)) void sincosl(long double __x ,
                                                  long double *__sinx ,
                                                  long double *__cosx ) ;
extern  __attribute__((__nothrow__)) void __sincosl(long double __x ,
                                                    long double *__sinx ,
                                                    long double *__cosx ) ;
extern  __attribute__((__nothrow__)) long double acoshl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __acoshl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double asinhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __asinhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double atanhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __atanhl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double expl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __expl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double frexpl(long double __x ,
                                                        int *__exponent ) ;
extern  __attribute__((__nothrow__)) long double __frexpl(long double __x ,
                                                          int *__exponent ) ;
extern  __attribute__((__nothrow__)) long double ldexpl(long double __x ,
                                                        int __exponent ) ;
extern  __attribute__((__nothrow__)) long double __ldexpl(long double __x ,
                                                          int __exponent ) ;
extern  __attribute__((__nothrow__)) long double logl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __logl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double log10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __log10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double modfl(long double __x ,
                                                       long double *__iptr ) ;
extern  __attribute__((__nothrow__)) long double __modfl(long double __x ,
                                                         long double *__iptr ) ;
extern  __attribute__((__nothrow__)) long double exp10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __exp10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double pow10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __pow10l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double expm1l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __expm1l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double log1pl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __log1pl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double logbl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __logbl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double exp2l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __exp2l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double log2l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __log2l(long double __x ) ;
extern  __attribute__((__nothrow__)) long double powl(long double __x ,
                                                      long double __y ) ;
extern  __attribute__((__nothrow__)) long double __powl(long double __x ,
                                                        long double __y ) ;
extern  __attribute__((__nothrow__)) long double sqrtl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __sqrtl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double hypotl(long double __x ,
                                                        long double __y ) ;
extern  __attribute__((__nothrow__)) long double __hypotl(long double __x ,
                                                          long double __y ) ;
extern  __attribute__((__nothrow__)) long double cbrtl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __cbrtl(long double __x ) ;
__inline extern  __attribute__((__nothrow__)) long double ceill(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __ceill(long double __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) long double fabsl(long double __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) long double __fabsl(long double __x )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) long double floorl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __floorl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double fmodl(long double __x ,
                                                       long double __y ) ;
extern  __attribute__((__nothrow__)) long double __fmodl(long double __x ,
                                                         long double __y ) ;
extern  __attribute__((__nothrow__)) int __isinfl(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int __finitel(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isinfl(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int finitel(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double dreml(long double __x ,
                                                       long double __y ) ;
extern  __attribute__((__nothrow__)) long double __dreml(long double __x ,
                                                         long double __y ) ;
extern  __attribute__((__nothrow__)) long double significandl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __significandl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double copysignl(long double __x ,
                                                           long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __copysignl(long double __x ,
                                                             long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double nanl(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __nanl(char const   *__tagb )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int __isnanl(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) int isnanl(long double __value )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double j0l(long double  ) ;
extern  __attribute__((__nothrow__)) long double __j0l(long double  ) ;
extern  __attribute__((__nothrow__)) long double j1l(long double  ) ;
extern  __attribute__((__nothrow__)) long double __j1l(long double  ) ;
extern  __attribute__((__nothrow__)) long double jnl(int  , long double  ) ;
extern  __attribute__((__nothrow__)) long double __jnl(int  , long double  ) ;
extern  __attribute__((__nothrow__)) long double y0l(long double  ) ;
extern  __attribute__((__nothrow__)) long double __y0l(long double  ) ;
extern  __attribute__((__nothrow__)) long double y1l(long double  ) ;
extern  __attribute__((__nothrow__)) long double __y1l(long double  ) ;
extern  __attribute__((__nothrow__)) long double ynl(int  , long double  ) ;
extern  __attribute__((__nothrow__)) long double __ynl(int  , long double  ) ;
extern  __attribute__((__nothrow__)) long double erfl(long double  ) ;
extern  __attribute__((__nothrow__)) long double __erfl(long double  ) ;
extern  __attribute__((__nothrow__)) long double erfcl(long double  ) ;
extern  __attribute__((__nothrow__)) long double __erfcl(long double  ) ;
extern  __attribute__((__nothrow__)) long double lgammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double __lgammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double tgammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double __tgammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double gammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double __gammal(long double  ) ;
extern  __attribute__((__nothrow__)) long double lgammal_r(long double  ,
                                                           int *__signgamp ) ;
extern  __attribute__((__nothrow__)) long double __lgammal_r(long double  ,
                                                             int *__signgamp ) ;
extern  __attribute__((__nothrow__)) long double rintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __rintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double nextafterl(long double __x ,
                                                            long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __nextafterl(long double __x ,
                                                              long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double nexttowardl(long double __x ,
                                                             long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __nexttowardl(long double __x ,
                                                               long double __y )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double remainderl(long double __x ,
                                                            long double __y ) ;
extern  __attribute__((__nothrow__)) long double __remainderl(long double __x ,
                                                              long double __y ) ;
extern  __attribute__((__nothrow__)) long double scalbnl(long double __x ,
                                                         int __n ) ;
extern  __attribute__((__nothrow__)) long double __scalbnl(long double __x ,
                                                           int __n ) ;
extern  __attribute__((__nothrow__)) int ilogbl(long double __x ) ;
extern  __attribute__((__nothrow__)) int __ilogbl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double scalblnl(long double __x ,
                                                          long __n ) ;
extern  __attribute__((__nothrow__)) long double __scalblnl(long double __x ,
                                                            long __n ) ;
extern  __attribute__((__nothrow__)) long double nearbyintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double __nearbyintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double roundl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __roundl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double truncl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double __truncl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double remquol(long double __x ,
                                                         long double __y ,
                                                         int *__quo ) ;
extern  __attribute__((__nothrow__)) long double __remquol(long double __x ,
                                                           long double __y ,
                                                           int *__quo ) ;
__inline extern  __attribute__((__nothrow__)) long lrintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long __lrintl(long double __x ) ;
__inline extern  __attribute__((__nothrow__)) long long llrintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long long __llrintl(long double __x ) ;
extern  __attribute__((__nothrow__)) long lroundl(long double __x ) ;
extern  __attribute__((__nothrow__)) long __lroundl(long double __x ) ;
extern  __attribute__((__nothrow__)) long long llroundl(long double __x ) ;
extern  __attribute__((__nothrow__)) long long __llroundl(long double __x ) ;
extern  __attribute__((__nothrow__)) long double fdiml(long double __x ,
                                                       long double __y ) ;
extern  __attribute__((__nothrow__)) long double __fdiml(long double __x ,
                                                         long double __y ) ;
extern  __attribute__((__nothrow__)) long double fmaxl(long double __x ,
                                                       long double __y ) ;
extern  __attribute__((__nothrow__)) long double __fmaxl(long double __x ,
                                                         long double __y ) ;
extern  __attribute__((__nothrow__)) long double fminl(long double __x ,
                                                       long double __y ) ;
extern  __attribute__((__nothrow__)) long double __fminl(long double __x ,
                                                         long double __y ) ;
extern  __attribute__((__nothrow__)) int __fpclassifyl(long double __value )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) int __signbitl(long double __x )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) long double fmal(long double __x ,
                                                      long double __y ,
                                                      long double __z ) ;
extern  __attribute__((__nothrow__)) long double __fmal(long double __x ,
                                                        long double __y ,
                                                        long double __z ) ;
extern  __attribute__((__nothrow__)) long double scalbl(long double __x ,
                                                        long double __n ) ;
extern  __attribute__((__nothrow__)) long double __scalbl(long double __x ,
                                                          long double __n ) ;
extern int signgam ;
extern _LIB_VERSION_TYPE _LIB_VERSION ;
extern int matherr(struct exception *__exc ) ;
__inline extern  __attribute__((__nothrow__)) int __signbitf(float __x )  __attribute__((__const__)) ;
__inline extern int __signbitf(float __x ) 
{ union __anonunion___u_24 __u ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22944\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __u.__f = __x;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22945\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__u.__i < 0);
}
}
__inline extern  __attribute__((__nothrow__)) int __signbit(double __x )  __attribute__((__const__)) ;
__inline extern int __signbit(double __x ) 
{ union __anonunion___u_25 __u ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22946\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __u.__d = __x;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22947\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__u.__i[1] < 0);
}
}
__inline extern  __attribute__((__nothrow__)) int __signbitl(long double __x )  __attribute__((__const__)) ;
__inline extern int __signbitl(long double __x ) 
{ union __anonunion___u_26 __u ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22948\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __u.__l = __x;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22949\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((__u.__i[2] & 0x8000) != 0);
}
}
__inline extern  __attribute__((__nothrow__)) double __sgn(double __x ) ;
__inline extern  __attribute__((__nothrow__)) double __sgn(double __x ) ;
__inline extern double __sgn(double __x ) 
{ double tmp ;
  double tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22955\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (__x == 0.0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22950\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = 0.0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22953\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (__x > 0.0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22951\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = 1.0;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22952\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = - 1.0;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22954\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = tmp;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22956\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___0);
}
}
__inline extern  __attribute__((__nothrow__)) float __sgnf(float __x ) ;
__inline extern  __attribute__((__nothrow__)) float __sgnf(float __x ) ;
__inline extern float __sgnf(float __x ) 
{ double tmp ;
  double tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22962\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if ((double )__x == 0.0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22957\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = 0.0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22960\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if ((double )__x > 0.0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22958\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = 1.0;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22959\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = - 1.0;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22961\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = tmp;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22963\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((float )tmp___0);
}
}
__inline extern  __attribute__((__nothrow__)) long double __sgnl(long double __x ) ;
__inline extern  __attribute__((__nothrow__)) long double __sgnl(long double __x ) ;
__inline extern long double __sgnl(long double __x ) 
{ double tmp ;
  double tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22969\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (__x == (long double )0.0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22964\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = 0.0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22967\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (__x > (long double )0.0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22965\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = 1.0;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "22966\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = - 1.0;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "22968\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = tmp;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22970\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((long double )tmp___0);
}
}
__inline extern  __attribute__((__nothrow__)) long double __atan2l(long double __y ,
                                                                   long double __x ) ;
__inline extern long double __atan2l(long double __y , long double __x ) 
{ long double tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22971\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __builtin_atan2l(__y, __x);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22972\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) double fabs(double __x )  __attribute__((__const__)) ;
__inline extern double fabs(double __x ) 
{ double tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22973\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __builtin_fabs(__x);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22974\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) float fabsf(float __x )  __attribute__((__const__)) ;
__inline extern float fabsf(float __x ) 
{ float tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22975\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __builtin_fabsf(__x);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22976\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) long double fabsl(long double __x )  __attribute__((__const__)) ;
__inline extern long double fabsl(long double __x ) 
{ long double tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22977\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __builtin_fabsl(__x);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22978\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) long double __fabsl(long double __x )  __attribute__((__const__)) ;
__inline extern long double __fabsl(long double __x ) 
{ long double tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22979\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __builtin_fabsl(__x);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22980\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) long double __sgn1l(long double __x ) ;
__inline extern  __attribute__((__nothrow__)) long double __sgn1l(long double __x ) ;
__inline extern long double __sgn1l(long double __x ) 
{ union __anonunion___n_27 __n ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22981\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __n.__xld = __x;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22982\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __n.__xi[2] = (__n.__xi[2] & 32768U) | 16383U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22983\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __n.__xi[1] = 0x80000000;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22984\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __n.__xi[0] = 0U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22985\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__n.__xld);
}
}
__inline extern  __attribute__((__nothrow__)) double floor(double __x )  __attribute__((__const__)) ;
__inline extern double floor(double __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22986\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0400, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22987\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((double )__value);
}
}
__inline extern  __attribute__((__nothrow__)) float floorf(float __x )  __attribute__((__const__)) ;
__inline extern float floorf(float __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22988\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0400, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22989\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((float )__value);
}
}
__inline extern  __attribute__((__nothrow__)) long double floorl(long double __x )  __attribute__((__const__)) ;
__inline extern long double floorl(long double __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22990\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0400, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22991\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__value);
}
}
__inline extern  __attribute__((__nothrow__)) double ceil(double __x )  __attribute__((__const__)) ;
__inline extern double ceil(double __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22992\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0800, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22993\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((double )__value);
}
}
__inline extern  __attribute__((__nothrow__)) float ceilf(float __x )  __attribute__((__const__)) ;
__inline extern float ceilf(float __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22994\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0800, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22995\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((float )__value);
}
}
__inline extern  __attribute__((__nothrow__)) long double ceill(long double __x )  __attribute__((__const__)) ;
__inline extern long double ceill(long double __x ) 
{ register long double __value ;
  register int __ignore ;
  unsigned short __cw ;
  unsigned short __cwtmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22996\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fnstcw %3\n\t"
                       "movzwl %3, %1\n\t"
                       "andl $0xf3ff, %1\n\t"
                       "orl $0x0800, %1\n\t"
                       "movw %w1, %2\n\t"
                       "fldcw %2\n\t"
                       "frndint\n\t"
                       "fldcw %3": "=t" (__value), "=&q" (__ignore),
                       "=m" (__cwtmp), "=m" (__cw): "0" (__x));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22997\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__value);
}
}
__inline extern  __attribute__((__nothrow__)) long lrintf(float __x ) ;
__inline extern long lrintf(float __x ) 
{ long __lrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22998\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpl %0": "=m" (__lrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "22999\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__lrintres);
}
}
__inline extern  __attribute__((__nothrow__)) long lrint(double __x ) ;
__inline extern long lrint(double __x ) 
{ long __lrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23000\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpl %0": "=m" (__lrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23001\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__lrintres);
}
}
__inline extern  __attribute__((__nothrow__)) long lrintl(long double __x ) ;
__inline extern long lrintl(long double __x ) 
{ long __lrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23002\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpl %0": "=m" (__lrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23003\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__lrintres);
}
}
__inline extern  __attribute__((__nothrow__)) long long llrintf(float __x ) ;
__inline extern long long llrintf(float __x ) 
{ long long __llrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23004\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpll %0": "=m" (__llrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23005\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__llrintres);
}
}
__inline extern  __attribute__((__nothrow__)) long long llrint(double __x ) ;
__inline extern long long llrint(double __x ) 
{ long long __llrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23006\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpll %0": "=m" (__llrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23007\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__llrintres);
}
}
__inline extern  __attribute__((__nothrow__)) long long llrintl(long double __x ) ;
__inline extern long long llrintl(long double __x ) 
{ long long __llrintres ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23008\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __asm__  volatile   ("fistpll %0": "=m" (__llrintres): "t" (__x): "st");
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23009\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (__llrintres);
}
}
__inline extern  __attribute__((__nothrow__)) int __finite(double __x )  __attribute__((__const__)) ;
__inline extern int __finite(double __x ) 
{ union __anonunion_28 __constr_expr_0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23010\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  __constr_expr_0.__d = __x;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23011\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((int )((((unsigned int )__constr_expr_0.__i[1] | 0x800fffffu) + 1U) >> 31));
}
}
extern  __attribute__((__nothrow__)) intmax_t imaxabs(intmax_t __n )  __attribute__((__const__)) ;
extern  __attribute__((__nothrow__)) imaxdiv_t imaxdiv(intmax_t __numer ,
                                                       intmax_t __denom )  __attribute__((__const__)) ;
__inline extern  __attribute__((__nothrow__)) intmax_t strtoimax(char const   * __restrict  nptr ,
                                                                 char ** __restrict  endptr ,
                                                                 int base ) ;
__inline extern  __attribute__((__nothrow__)) uintmax_t strtoumax(char const   * __restrict  nptr ,
                                                                  char ** __restrict  endptr ,
                                                                  int base ) ;
__inline extern  __attribute__((__nothrow__)) intmax_t wcstoimax(__gwchar_t const   * __restrict  nptr ,
                                                                 __gwchar_t ** __restrict  endptr ,
                                                                 int base ) ;
__inline extern  __attribute__((__nothrow__)) uintmax_t wcstoumax(__gwchar_t const   * __restrict  nptr ,
                                                                  __gwchar_t ** __restrict  endptr ,
                                                                  int base ) ;
extern  __attribute__((__nothrow__)) long long __strtoll_internal(char const   * __restrict  __nptr ,
                                                                  char ** __restrict  __endptr ,
                                                                  int __base ,
                                                                  int __group )  __attribute__((__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) intmax_t strtoimax(char const   * __restrict  nptr ,
                                                                 char ** __restrict  endptr ,
                                                                 int base ) ;
__inline extern intmax_t strtoimax(char const   * __restrict  nptr ,
                                   char ** __restrict  endptr , int base ) 
{ long long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23012\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __strtoll_internal(nptr, endptr, base, 0);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23013\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
extern  __attribute__((__nothrow__)) unsigned long long __strtoull_internal(char const   * __restrict  __nptr ,
                                                                            char ** __restrict  __endptr ,
                                                                            int __base ,
                                                                            int __group )  __attribute__((__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) uintmax_t strtoumax(char const   * __restrict  nptr ,
                                                                  char ** __restrict  endptr ,
                                                                  int base ) ;
__inline extern uintmax_t strtoumax(char const   * __restrict  nptr ,
                                    char ** __restrict  endptr , int base ) 
{ unsigned long long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23014\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __strtoull_internal(nptr, endptr, base, 0);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23015\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
extern  __attribute__((__nothrow__)) long long __wcstoll_internal(__gwchar_t const   * __restrict  __nptr ,
                                                                  __gwchar_t ** __restrict  __endptr ,
                                                                  int __base ,
                                                                  int __group )  __attribute__((__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) intmax_t wcstoimax(__gwchar_t const   * __restrict  nptr ,
                                                                 __gwchar_t ** __restrict  endptr ,
                                                                 int base ) ;
__inline extern intmax_t wcstoimax(__gwchar_t const   * __restrict  nptr ,
                                   __gwchar_t ** __restrict  endptr , int base ) 
{ long long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23016\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __wcstoll_internal(nptr, endptr, base, 0);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23017\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
extern  __attribute__((__nothrow__)) unsigned long long __wcstoull_internal(__gwchar_t const   * __restrict  __nptr ,
                                                                            __gwchar_t ** __restrict  __endptr ,
                                                                            int __base ,
                                                                            int __group )  __attribute__((__nonnull__(1))) ;
__inline extern  __attribute__((__nothrow__)) uintmax_t wcstoumax(__gwchar_t const   * __restrict  nptr ,
                                                                  __gwchar_t ** __restrict  endptr ,
                                                                  int base ) ;
__inline extern uintmax_t wcstoumax(__gwchar_t const   * __restrict  nptr ,
                                    __gwchar_t ** __restrict  endptr , int base ) 
{ unsigned long long tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23018\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __wcstoull_internal(nptr, endptr, base, 0);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23019\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
extern struct _IO_FILE_plus _IO_2_1_stdin_ ;
extern struct _IO_FILE_plus _IO_2_1_stdout_ ;
extern struct _IO_FILE_plus _IO_2_1_stderr_ ;
extern void _IO_cookie_init(struct _IO_cookie_file *__cfile , int __read_write ,
                            void *__cookie , _IO_cookie_io_functions_t __fns ) ;
extern int __underflow(_IO_FILE * ) ;
extern int __uflow(_IO_FILE * ) ;
extern int __overflow(_IO_FILE * , int  ) ;
extern int _IO_getc(_IO_FILE *__fp ) ;
extern int _IO_putc(int __c , _IO_FILE *__fp ) ;
extern  __attribute__((__nothrow__)) int _IO_feof(_IO_FILE *__fp ) ;
extern  __attribute__((__nothrow__)) int _IO_ferror(_IO_FILE *__fp ) ;
extern int _IO_peekc_locked(_IO_FILE *__fp ) ;
extern  __attribute__((__nothrow__)) void _IO_flockfile(_IO_FILE * ) ;
extern  __attribute__((__nothrow__)) void _IO_funlockfile(_IO_FILE * ) ;
extern  __attribute__((__nothrow__)) int _IO_ftrylockfile(_IO_FILE * ) ;
extern int _IO_vfscanf(_IO_FILE * __restrict   , char const   * __restrict   ,
                       __gnuc_va_list  , int * __restrict   ) ;
extern int _IO_vfprintf(_IO_FILE * __restrict   , char const   * __restrict   ,
                        __gnuc_va_list  ) ;
extern __ssize_t _IO_padn(_IO_FILE * , int  , __ssize_t  ) ;
extern size_t _IO_sgetn(_IO_FILE * , void * , size_t  ) ;
extern __off64_t _IO_seekoff(_IO_FILE * , __off64_t  , int  , int  ) ;
extern __off64_t _IO_seekpos(_IO_FILE * , __off64_t  , int  ) ;
extern  __attribute__((__nothrow__)) void _IO_free_backup_area(_IO_FILE * ) ;
extern struct _IO_FILE *stdin ;
extern struct _IO_FILE *stdout ;
extern struct _IO_FILE *stderr ;
extern  __attribute__((__nothrow__)) int remove(char const   *__filename ) ;
extern  __attribute__((__nothrow__)) int rename(char const   *__old ,
                                                char const   *__new ) ;
extern  __attribute__((__nothrow__)) int renameat(int __oldfd ,
                                                  char const   *__old ,
                                                  int __newfd ,
                                                  char const   *__new ) ;
extern FILE *tmpfile(void) ;
extern FILE *tmpfile64(void) ;
extern  __attribute__((__nothrow__)) char *tmpnam(char *__s ) ;
extern  __attribute__((__nothrow__)) char *tmpnam_r(char *__s ) ;
extern  __attribute__((__nothrow__)) char *tempnam(char const   *__dir ,
                                                   char const   *__pfx )  __attribute__((__malloc__)) ;
extern int fflush_unlocked(FILE *__stream ) ;
extern int fcloseall(void) ;
extern FILE *freopen(char const   * __restrict  __filename ,
                     char const   * __restrict  __modes ,
                     FILE * __restrict  __stream ) ;
extern FILE *fopen64(char const   * __restrict  __filename ,
                     char const   * __restrict  __modes ) ;
extern FILE *freopen64(char const   * __restrict  __filename ,
                       char const   * __restrict  __modes ,
                       FILE * __restrict  __stream ) ;
extern  __attribute__((__nothrow__)) FILE *fdopen(int __fd ,
                                                  char const   *__modes ) ;
extern  __attribute__((__nothrow__)) FILE *fopencookie(void * __restrict  __magic_cookie ,
                                                       char const   * __restrict  __modes ,
                                                       _IO_cookie_io_functions_t __io_funcs ) ;
extern  __attribute__((__nothrow__)) FILE *fmemopen(void *__s , size_t __len ,
                                                    char const   *__modes ) ;
extern  __attribute__((__nothrow__)) FILE *open_memstream(char **__bufloc ,
                                                          size_t *__sizeloc ) ;
extern  __attribute__((__nothrow__)) void setbuf(FILE * __restrict  __stream ,
                                                 char * __restrict  __buf ) ;
extern  __attribute__((__nothrow__)) int setvbuf(FILE * __restrict  __stream ,
                                                 char * __restrict  __buf ,
                                                 int __modes , size_t __n ) ;
extern  __attribute__((__nothrow__)) void setbuffer(FILE * __restrict  __stream ,
                                                    char * __restrict  __buf ,
                                                    size_t __size ) ;
extern  __attribute__((__nothrow__)) void setlinebuf(FILE *__stream ) ;
extern int printf(char const   * __restrict  __format  , ...) ;
extern  __attribute__((__nothrow__)) int sprintf(char * __restrict  __s ,
                                                 char const   * __restrict  __format 
                                                 , ...) ;
extern int vfprintf(FILE * __restrict  __s ,
                    char const   * __restrict  __format , __gnuc_va_list __arg ) ;
__inline extern int vprintf(char const   * __restrict  __fmt ,
                            __gnuc_va_list __arg ) ;
extern  __attribute__((__nothrow__)) int vsprintf(char * __restrict  __s ,
                                                  char const   * __restrict  __format ,
                                                  __gnuc_va_list __arg ) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  snprintf)(char * __restrict  __s ,
                                                                             size_t __maxlen ,
                                                                             char const   * __restrict  __format 
                                                                             , ...) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  vsnprintf)(char * __restrict  __s ,
                                                                              size_t __maxlen ,
                                                                              char const   * __restrict  __format ,
                                                                              __gnuc_va_list __arg ) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  vasprintf)(char ** __restrict  __ptr ,
                                                                              char const   * __restrict  __f ,
                                                                              __gnuc_va_list __arg ) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  __asprintf)(char ** __restrict  __ptr ,
                                                                               char const   * __restrict  __fmt 
                                                                               , ...) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  asprintf)(char ** __restrict  __ptr ,
                                                                             char const   * __restrict  __fmt 
                                                                             , ...) ;
extern int ( /* format attribute */  vdprintf)(int __fd ,
                                               char const   * __restrict  __fmt ,
                                               __gnuc_va_list __arg ) ;
extern int ( /* format attribute */  dprintf)(int __fd ,
                                              char const   * __restrict  __fmt 
                                              , ...) ;
extern int fscanf(FILE * __restrict  __stream ,
                  char const   * __restrict  __format  , ...) ;
extern int scanf(char const   * __restrict  __format  , ...) ;
extern  __attribute__((__nothrow__)) int sscanf(char const   * __restrict  __s ,
                                                char const   * __restrict  __format 
                                                , ...) ;
extern int ( /* format attribute */  vfscanf)(FILE * __restrict  __s ,
                                              char const   * __restrict  __format ,
                                              __gnuc_va_list __arg ) ;
extern int ( /* format attribute */  vscanf)(char const   * __restrict  __format ,
                                             __gnuc_va_list __arg ) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  vsscanf)(char const   * __restrict  __s ,
                                                                            char const   * __restrict  __format ,
                                                                            __gnuc_va_list __arg ) ;
extern int fgetc(FILE *__stream ) ;
extern int getc(FILE *__stream ) ;
__inline extern int getchar(void) ;
__inline extern int getc_unlocked(FILE *__fp ) ;
__inline extern int getchar_unlocked(void) ;
__inline extern int fgetc_unlocked(FILE *__fp ) ;
extern int fputc(int __c , FILE *__stream ) ;
extern int putc(int __c , FILE *__stream ) ;
__inline extern int putchar(int __c ) ;
__inline extern int fputc_unlocked(int __c , FILE *__stream ) ;
__inline extern int putc_unlocked(int __c , FILE *__stream ) ;
__inline extern int putchar_unlocked(int __c ) ;
extern int getw(FILE *__stream ) ;
extern int putw(int __w , FILE *__stream ) ;
extern char *fgets(char * __restrict  __s , int __n ,
                   FILE * __restrict  __stream ) ;
extern char *gets(char *__s ) ;
extern char *fgets_unlocked(char * __restrict  __s , int __n ,
                            FILE * __restrict  __stream ) ;
extern __ssize_t __getdelim(char ** __restrict  __lineptr ,
                            size_t * __restrict  __n , int __delimiter ,
                            FILE * __restrict  __stream ) ;
extern __ssize_t getdelim(char ** __restrict  __lineptr ,
                          size_t * __restrict  __n , int __delimiter ,
                          FILE * __restrict  __stream ) ;
__inline extern __ssize_t getline(char ** __restrict  __lineptr ,
                                  size_t * __restrict  __n ,
                                  FILE * __restrict  __stream ) ;
extern int fputs(char const   * __restrict  __s , FILE * __restrict  __stream ) ;
extern int puts(char const   *__s ) ;
extern int ungetc(int __c , FILE *__stream ) ;
extern size_t fread(void * __restrict  __ptr , size_t __size , size_t __n ,
                    FILE * __restrict  __stream ) ;
extern size_t fwrite(void const   * __restrict  __ptr , size_t __size ,
                     size_t __n , FILE * __restrict  __s ) ;
extern int fputs_unlocked(char const   * __restrict  __s ,
                          FILE * __restrict  __stream ) ;
extern size_t fread_unlocked(void * __restrict  __ptr , size_t __size ,
                             size_t __n , FILE * __restrict  __stream ) ;
extern size_t fwrite_unlocked(void const   * __restrict  __ptr , size_t __size ,
                              size_t __n , FILE * __restrict  __stream ) ;
extern int fseek(FILE *__stream , long __off , int __whence ) ;
extern long ftell(FILE *__stream ) ;
extern void rewind(FILE *__stream ) ;
extern int fseeko(FILE *__stream , __off_t __off , int __whence ) ;
extern __off_t ftello(FILE *__stream ) ;
extern int fgetpos(FILE * __restrict  __stream , fpos_t * __restrict  __pos ) ;
extern int fsetpos(FILE *__stream , fpos_t const   *__pos ) ;
extern int fseeko64(FILE *__stream , __off64_t __off , int __whence ) ;
extern __off64_t ftello64(FILE *__stream ) ;
extern int fgetpos64(FILE * __restrict  __stream , fpos64_t * __restrict  __pos ) ;
extern int fsetpos64(FILE *__stream , fpos64_t const   *__pos ) ;
extern  __attribute__((__nothrow__)) void clearerr(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) int feof(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) int ferror(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) void clearerr_unlocked(FILE *__stream ) ;
__inline extern  __attribute__((__nothrow__)) int feof_unlocked(FILE *__stream ) ;
__inline extern  __attribute__((__nothrow__)) int ferror_unlocked(FILE *__stream ) ;
extern void perror(char const   *__s ) ;
extern int sys_nerr ;
extern char const   * const  sys_errlist[] ;
extern int _sys_nerr ;
extern char const   * const  _sys_errlist[] ;
extern  __attribute__((__nothrow__)) int fileno(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) int fileno_unlocked(FILE *__stream ) ;
extern FILE *popen(char const   *__command , char const   *__modes ) ;
extern int pclose(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) char *ctermid(char *__s ) ;
extern char *cuserid(char *__s ) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  obstack_printf)(struct obstack * __restrict  __obstack ,
                                                                                   char const   * __restrict  __format 
                                                                                   , ...) ;
extern  __attribute__((__nothrow__)) int ( /* format attribute */  obstack_vprintf)(struct obstack * __restrict  __obstack ,
                                                                                    char const   * __restrict  __format ,
                                                                                    __gnuc_va_list __args ) ;
extern  __attribute__((__nothrow__)) void flockfile(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) int ftrylockfile(FILE *__stream ) ;
extern  __attribute__((__nothrow__)) void funlockfile(FILE *__stream ) ;
__inline extern int vprintf(char const   * __restrict  __fmt ,
                            __gnuc_va_list __arg ) 
{ int tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23020\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = vfprintf((FILE */* __restrict  */)stdout, __fmt, __arg);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23021\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern int getchar(void) 
{ int tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23022\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = _IO_getc(stdin);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23023\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern int fgetc_unlocked(FILE *__fp ) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  int tmp___2 ;
  long tmp___3 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23029\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___3 = __builtin_expect((long )((unsigned int )__fp->_IO_read_ptr >= (unsigned int )__fp->_IO_read_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23030\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___3) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23024\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __uflow(__fp);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23025\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23026\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = __fp->_IO_read_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23027\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (__fp->_IO_read_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23028\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (int )*((unsigned char *)tmp___1);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23031\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___2);
}
}
__inline extern int getc_unlocked(FILE *__fp ) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  int tmp___2 ;
  long tmp___3 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23037\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___3 = __builtin_expect((long )((unsigned int )__fp->_IO_read_ptr >= (unsigned int )__fp->_IO_read_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23038\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___3) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23032\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __uflow(__fp);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23033\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23034\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = __fp->_IO_read_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23035\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (__fp->_IO_read_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23036\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (int )*((unsigned char *)tmp___1);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23039\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___2);
}
}
__inline extern int getchar_unlocked(void) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  int tmp___2 ;
  long tmp___3 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23045\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___3 = __builtin_expect((long )((unsigned int )stdin->_IO_read_ptr >= (unsigned int )stdin->_IO_read_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23046\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___3) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23040\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __uflow(stdin);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23041\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23042\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = stdin->_IO_read_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23043\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (stdin->_IO_read_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23044\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (int )*((unsigned char *)tmp___1);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23047\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___2);
}
}
__inline extern int putchar(int __c ) 
{ int tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23048\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = _IO_putc(__c, stdout);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23049\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern int fputc_unlocked(int __c , FILE *__stream ) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  char tmp___2 ;
  int tmp___3 ;
  long tmp___4 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23057\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___4 = __builtin_expect((long )((unsigned int )__stream->_IO_write_ptr >= (unsigned int )__stream->_IO_write_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23058\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___4) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23050\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __overflow(__stream, (int )((unsigned char )__c));
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23051\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23052\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = __stream->_IO_write_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23053\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (__stream->_IO_write_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23054\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (char )__c;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23055\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *tmp___1 = tmp___2;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23056\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = (int )((unsigned char )tmp___2);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23059\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___3);
}
}
__inline extern int putc_unlocked(int __c , FILE *__stream ) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  char tmp___2 ;
  int tmp___3 ;
  long tmp___4 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23067\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___4 = __builtin_expect((long )((unsigned int )__stream->_IO_write_ptr >= (unsigned int )__stream->_IO_write_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23068\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___4) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23060\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __overflow(__stream, (int )((unsigned char )__c));
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23061\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23062\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = __stream->_IO_write_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23063\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (__stream->_IO_write_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23064\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (char )__c;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23065\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *tmp___1 = tmp___2;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23066\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = (int )((unsigned char )tmp___2);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23069\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___3);
}
}
__inline extern int putchar_unlocked(int __c ) 
{ long tmp ;
  int tmp___0 ;
  char *tmp___1 ;
  char tmp___2 ;
  int tmp___3 ;
  long tmp___4 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23077\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___4 = __builtin_expect((long )((unsigned int )stdout->_IO_write_ptr >= (unsigned int )stdout->_IO_write_end),
                             0L);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23078\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tmp___4) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23070\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = __overflow(stdout, (int )((unsigned char )__c));
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23071\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = tmp___0;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23072\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = stdout->_IO_write_ptr;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23073\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (stdout->_IO_write_ptr) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23074\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = (char )__c;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23075\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *tmp___1 = tmp___2;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23076\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = (int )((unsigned char )tmp___2);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23079\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp___3);
}
}
__inline extern __ssize_t getline(char ** __restrict  __lineptr ,
                                  size_t * __restrict  __n ,
                                  FILE * __restrict  __stream ) 
{ __ssize_t tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23080\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = __getdelim(__lineptr, __n, '\n', __stream);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23081\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
}
}
__inline extern  __attribute__((__nothrow__)) int feof_unlocked(FILE *__stream ) ;
__inline extern int feof_unlocked(FILE *__stream ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23082\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((__stream->_flags & 0x10) != 0);
}
}
__inline extern  __attribute__((__nothrow__)) int ferror_unlocked(FILE *__stream ) ;
__inline extern int ferror_unlocked(FILE *__stream ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23083\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((__stream->_flags & 0x20) != 0);
}
}
extern timelib_sll timelib_day_of_week(timelib_sll y , timelib_sll m ,
                                       timelib_sll d ) ;
extern timelib_sll timelib_iso_day_of_week(timelib_sll y , timelib_sll m ,
                                           timelib_sll d ) ;
extern timelib_sll timelib_day_of_year(timelib_sll y , timelib_sll m ,
                                       timelib_sll d ) ;
extern timelib_sll timelib_daynr_from_weeknr(timelib_sll y , timelib_sll w ,
                                             timelib_sll d ) ;
extern timelib_sll timelib_days_in_month(timelib_sll y , timelib_sll m ) ;
extern void timelib_isoweek_from_date(timelib_sll y , timelib_sll m ,
                                      timelib_sll d , timelib_sll *iw ,
                                      timelib_sll *iy ) ;
extern int timelib_valid_time(timelib_sll h , timelib_sll i , timelib_sll s ) ;
extern int timelib_valid_date(timelib_sll y , timelib_sll m , timelib_sll d ) ;
extern timelib_time *timelib_strtotime(char *s , int len ,
                                       timelib_error_container **errors ,
                                       timelib_tzdb const   *tzdb ) ;
extern timelib_time *timelib_parse_from_format(char *format , char *s ,
                                               int len ,
                                               timelib_error_container **errors ,
                                               timelib_tzdb const   *tzdb ) ;
extern void timelib_fill_holes(timelib_time *parsed , timelib_time *now ,
                               int options ) ;
extern char *timelib_timezone_id_from_abbr(char const   *abbr , long gmtoffset ,
                                           int isdst ) ;
extern timelib_tz_lookup_table const   *timelib_timezone_abbreviations_list(void) ;
extern void timelib_strtointerval(char *s , int len , timelib_time **begin ,
                                  timelib_time **end ,
                                  timelib_rel_time **period , int *recurrences ,
                                  struct timelib_error_container **errors ) ;
void timelib_update_ts(timelib_time *time , timelib_tzinfo *tzi ) ;
void timelib_do_rel_normalize(timelib_time *base , timelib_rel_time *rt ) ;
extern int timelib_apply_localtime(timelib_time *t , unsigned int localtime ) ;
extern void timelib_unixtime2gmt(timelib_time *tm , timelib_sll ts ) ;
extern void timelib_unixtime2local(timelib_time *tm , timelib_sll ts ) ;
extern void timelib_update_from_sse(timelib_time *tm ) ;
extern void timelib_set_timezone(timelib_time *t , timelib_tzinfo *tz ) ;
extern int timelib_timezone_id_is_valid(char *timezone ,
                                        timelib_tzdb const   *tzdb ) ;
extern timelib_tzinfo *timelib_parse_tzfile(char *timezone ,
                                            timelib_tzdb const   *tzdb ) ;
extern int timelib_timestamp_is_in_dst(timelib_sll ts , timelib_tzinfo *tz ) ;
extern timelib_time_offset *timelib_get_time_zone_info(timelib_sll ts ,
                                                       timelib_tzinfo *tz ) ;
extern timelib_sll timelib_get_current_offset(timelib_time *t ) ;
extern void timelib_dump_tzinfo(timelib_tzinfo *tz ) ;
extern timelib_tzdb const   *timelib_builtin_db(void) ;
extern timelib_tzdb_index_entry const   *timelib_timezone_builtin_identifiers_list(int *count ) ;
extern timelib_tzinfo *timelib_tzinfo_ctor(char *name ) ;
extern void timelib_time_tz_abbr_update(timelib_time *tm , char *tz_abbr ) ;
extern void timelib_time_tz_name_update(timelib_time *tm , char *tz_name ) ;
extern void timelib_tzinfo_dtor(timelib_tzinfo *tz ) ;
extern timelib_tzinfo *timelib_tzinfo_clone(timelib_tzinfo *tz ) ;
extern timelib_rel_time *timelib_rel_time_ctor(void) ;
extern void timelib_rel_time_dtor(timelib_rel_time *t ) ;
extern timelib_rel_time *timelib_rel_time_clone(timelib_rel_time *tz ) ;
extern timelib_time *timelib_time_ctor(void) ;
extern void timelib_time_set_option(timelib_time *tm , int option ,
                                    void *option_value ) ;
extern void timelib_time_dtor(timelib_time *t ) ;
extern timelib_time *timelib_time_clone(timelib_time *orig ) ;
extern timelib_time_offset *timelib_time_offset_ctor(void) ;
extern void timelib_time_offset_dtor(timelib_time_offset *t ) ;
extern void timelib_error_container_dtor(timelib_error_container *errors ) ;
extern long timelib_date_to_int(timelib_time *d , int *error ) ;
extern void timelib_dump_date(timelib_time *d , int options ) ;
extern void timelib_dump_rel_time(timelib_rel_time *d ) ;
extern void timelib_decimal_hour_to_hms(double h , int *hour , int *min ,
                                        int *sec ) ;
extern double timelib_ts_to_juliandate(timelib_sll ts ) ;
extern int timelib_astro_rise_set_altitude(timelib_time *time , double lon ,
                                           double lat , double altit ,
                                           int upper_limb , double *h_rise ,
                                           double *h_set ,
                                           timelib_sll *ts_rise ,
                                           timelib_sll *ts_set ,
                                           timelib_sll *ts_transit ) ;
extern timelib_rel_time *timelib_diff(timelib_time *one , timelib_time *two ) ;
static int month_tab_leap[12]  = 
  {      -1,      30,      59,      90, 
        120,      151,      181,      212, 
        243,      273,      304,      334};
static int month_tab[12]  = 
  {      0,      31,      59,      90, 
        120,      151,      181,      212, 
        243,      273,      304,      334};
static int days_in_month_leap[13]  = 
  {      31,      31,      29,      31, 
        30,      31,      30,      31, 
        31,      30,      31,      30, 
        31};
static int days_in_month[13]  = 
  {      31,      31,      28,      31, 
        30,      31,      30,      31, 
        31,      30,      31,      30, 
        31};
static int do_range_limit(timelib_sll start , timelib_sll end ,
                          timelib_sll adj , timelib_sll *a , timelib_sll *b ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23090\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*a < start) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23084\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *b -= ((start - *a) - 1LL) / adj + 1LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23085\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *a += adj * (((start - *a) - 1LL) / adj + 1LL);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23086\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23091\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*a >= end) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23087\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *b += *a / adj;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23088\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *a -= adj * (*a / adj);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23089\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23092\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (0);
}
}
static void inc_month(timelib_sll *y , timelib_sll *m ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23096\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  (*m) ++;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23097\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*m > 12LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23093\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *m -= 12LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23094\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (*y) ++;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23095\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23098\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void dec_month(timelib_sll *y , timelib_sll *m ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23102\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  (*m) --;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23103\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*m < 1LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23099\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *m += 12LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23100\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (*y) --;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23101\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23104\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_range_limit_days_relative(timelib_sll *base_y ,
                                         timelib_sll *base_m , timelib_sll *y ,
                                         timelib_sll *m , timelib_sll *d ,
                                         timelib_sll invert ) 
{ timelib_sll leapyear ;
  timelib_sll month ;
  timelib_sll year ;
  timelib_sll days ;
  int tmp ;
  int tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23139\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_range_limit(1LL, 13LL, 12LL, base_m, base_y);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23140\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  year = *base_y;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23141\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  month = *base_m;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23142\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (! invert) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23121\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23114\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (*d < 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23105\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23115\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      dec_month(& year, & month);
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23116\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (year % 4LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23110\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (year % 100LL != 0LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23106\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          tmp = 1;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23109\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if (year % 400LL == 0LL) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23107\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            tmp = 1;
          } else {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23108\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            tmp = 0;
          }
        }
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23111\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp = 0;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23117\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      leapyear = (long long )tmp;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23118\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (leapyear) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23112\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        days = (long long )days_in_month_leap[month];
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23113\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        days = (long long )days_in_month[month];
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23119\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      *d += days;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23120\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      (*m) --;
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23138\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23131\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (*d < 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23122\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23132\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (year % 4LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23127\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (year % 100LL != 0LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23123\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          tmp___0 = 1;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23126\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if (year % 400LL == 0LL) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23124\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            tmp___0 = 1;
          } else {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23125\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            tmp___0 = 0;
          }
        }
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23128\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___0 = 0;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23133\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      leapyear = (long long )tmp___0;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23134\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (leapyear) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23129\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        days = (long long )days_in_month_leap[month];
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23130\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        days = (long long )days_in_month[month];
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23135\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      *d += days;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23136\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      (*m) --;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23137\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      inc_month(& year, & month);
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23143\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static int do_range_limit_days(timelib_sll *y , timelib_sll *m , timelib_sll *d ) 
{ timelib_sll leapyear ;
  timelib_sll days_this_month ;
  timelib_sll last_month ;
  timelib_sll last_year ;
  timelib_sll days_last_month ;
  int tmp ;
  int tmp___0 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23177\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*d >= 146097LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23144\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *y += 400LL * (*d / 146097LL);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23145\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *d -= 146097LL * (*d / 146097LL);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23149\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (*d <= -146097LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23146\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      *y += 400LL * (*d / 146097LL);
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23147\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      *d -= 146097LL * (*d / 146097LL);
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23148\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23178\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_range_limit(1LL, 13LL, 12LL, m, y);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23179\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*y % 4LL == 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23154\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (*y % 100LL != 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23150\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp = 1;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23153\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (*y % 400LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23151\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp = 1;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23152\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp = 0;
      }
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23155\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = 0;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23180\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  leapyear = (long long )tmp;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23181\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (leapyear) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23156\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    days_this_month = (long long )days_in_month_leap[*m];
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23157\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    days_this_month = (long long )days_in_month[*m];
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23182\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  last_month = *m - 1LL;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23183\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (last_month < 1LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23158\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    last_month += 12LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23159\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    last_year = *y - 1LL;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23160\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    last_year = *y;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23184\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (last_year % 4LL == 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23165\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (last_year % 100LL != 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23161\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp___0 = 1;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23164\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (last_year % 400LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23162\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___0 = 1;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23163\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___0 = 0;
      }
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23166\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = 0;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23185\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  leapyear = (long long )tmp___0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23186\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (leapyear) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23167\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    days_last_month = (long long )days_in_month_leap[last_month];
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23168\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    days_last_month = (long long )days_in_month[last_month];
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23187\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*d <= 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23169\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *d += days_last_month;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23170\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (*m) --;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23171\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    return (1);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23172\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23188\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (*d > days_this_month) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23173\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    *d -= days_this_month;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23174\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    (*m) ++;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23175\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    return (1);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23176\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23189\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (0);
}
}
static void do_adjust_for_weekday(timelib_time *time ) 
{ timelib_sll current_dow ;
  timelib_sll difference ;
  int tmp ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23207\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  current_dow = timelib_day_of_week(time->y, time->m, time->d);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23208\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.weekday_behavior == 2) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23192\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (time->relative.weekday == 0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23190\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      time->relative.weekday = 7;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23191\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23193\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->d -= current_dow;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23194\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->d += (timelib_sll )time->relative.weekday;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23195\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    return;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23196\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23209\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  difference = (timelib_sll )time->relative.weekday - current_dow;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23210\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.d < 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23198\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (difference < 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23197\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      difference += 7LL;
    } else {
      goto _L;
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23203\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    _L: /* CIL Label */ 
    if (time->relative.d >= 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23201\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (difference <= (timelib_sll )(- time->relative.weekday_behavior)) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23199\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        difference += 7LL;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23200\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23202\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23211\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.weekday >= 0) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23204\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->d += difference;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23205\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = abs(time->relative.weekday);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23206\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->d -= 7LL - ((timelib_sll )tmp - current_dow);
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23212\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->relative.have_weekday_relative = 0U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23213\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
void timelib_do_rel_normalize(timelib_time *base , timelib_rel_time *rt ) 
{ int tmp ;
  int tmp___0 ;
  int tmp___1 ;
  int tmp___2 ;
  int tmp___3 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23229\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23215\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = do_range_limit(0LL, 60LL, 60LL, & rt->s, & rt->i);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23216\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23214\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23230\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23218\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = do_range_limit(0LL, 60LL, 60LL, & rt->i, & rt->h);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23219\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23217\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23231\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23221\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = do_range_limit(0LL, 24LL, 24LL, & rt->h, & rt->d);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23222\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23220\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23232\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23224\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = do_range_limit(0LL, 12LL, 12LL, & rt->m, & rt->y);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23225\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___2) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23223\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23233\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_range_limit_days_relative(& base->y, & base->m, & rt->y, & rt->m, & rt->d,
                               (long long )rt->invert);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23234\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23227\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = do_range_limit(0LL, 12LL, 12LL, & rt->m, & rt->y);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23228\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___3) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23226\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23235\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_normalize(timelib_time *time ) 
{ int tmp ;
  int tmp___0 ;
  int tmp___1 ;
  int tmp___2 ;
  int tmp___3 ;
  int tmp___4 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23254\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23237\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp = do_range_limit(0LL, 60LL, 60LL, & time->s, & time->i);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23238\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23236\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23255\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23240\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___0 = do_range_limit(0LL, 60LL, 60LL, & time->i, & time->h);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23241\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___0) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23239\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23256\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23243\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___1 = do_range_limit(0LL, 24LL, 24LL, & time->h, & time->d);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23244\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23242\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23257\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23246\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___2 = do_range_limit(1LL, 13LL, 12LL, & time->m, & time->y);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23247\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___2) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23245\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23258\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23249\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___3 = do_range_limit_days(& time->y, & time->m, & time->d);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23250\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___3) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23248\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23259\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  while (1) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23252\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___4 = do_range_limit(1LL, 13LL, 12LL, & time->m, & time->y);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23253\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tmp___4) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23251\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    } else {
      break;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23260\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_adjust_relative(timelib_time *time ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23273\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.have_weekday_relative) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23261\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    do_adjust_for_weekday(time);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23262\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23274\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_normalize(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23275\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->have_relative) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23263\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->s += time->relative.s;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23264\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->i += time->relative.i;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23265\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->h += time->relative.h;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23266\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->d += time->relative.d;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23267\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->m += time->relative.m;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23268\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->y += time->relative.y;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23269\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  switch (time->relative.first_last_day_of) {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23270\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  case 1: 
  time->d = 1LL;
  break;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23271\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  case 2: 
  time->d = 0LL;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23272\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  (time->m) ++;
  break;
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23276\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_normalize(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23277\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_adjust_special_weekday(timelib_time *time ) 
{ timelib_sll current_dow ;
  timelib_sll count ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23311\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  count = time->relative.special.amount;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23312\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  current_dow = timelib_day_of_week(time->y, time->m, time->d);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23313\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (count == 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23282\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (current_dow == 6LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23278\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      time->d += 2LL;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23279\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23283\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (current_dow == 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23280\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      (time->d) ++;
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23281\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23310\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (count > 0LL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23290\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (current_dow == 5LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23284\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        time->d += 2LL;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23285\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23291\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (current_dow == 6LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23286\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        (time->d) ++;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23287\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23292\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      time->d += (count / 5LL) * 7LL;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23293\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      current_dow = timelib_day_of_week(time->y, time->m, time->d);
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23294\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      time->d += count % 5LL;
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23295\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (count % 5LL + current_dow > 5LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23288\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        time->d += 2LL;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23289\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23309\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (count < 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23302\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (current_dow == 1LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23296\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          time->d -= 2LL;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23297\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        }
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23303\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (current_dow == 0LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23298\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          (time->d) --;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23299\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        }
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23304\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        time->d += (count / 5LL) * 7LL;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23305\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        current_dow = timelib_day_of_week(time->y, time->m, time->d);
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23306\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        time->d += count % 5LL;
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23307\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (count % 5LL + current_dow < 1LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23300\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          time->d -= 2LL;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23301\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }

        }
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23308\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      }
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23314\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_adjust_special(timelib_time *time ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23317\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.have_special_relative) {
    switch ((int )time->relative.special.type) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23315\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    case 0x01: 
    do_adjust_special_weekday(time);
    break;
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23316\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23318\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_normalize(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23319\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  memset((void *)(& time->relative.special), 0, sizeof(time->relative.special));
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23320\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static void do_adjust_special_early(timelib_time *time ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23328\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (time->relative.have_special_relative) {
    switch ((int )time->relative.special.type) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23321\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    case 0x02: 
    time->d = 1LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23322\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->m += time->relative.m;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23323\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->relative.m = 0LL;
    break;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23324\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    case 0x03: 
    time->d = 1LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23325\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->m += time->relative.m + 1LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23326\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    time->relative.m = 0LL;
    break;
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23327\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23329\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_normalize(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23330\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
static timelib_sll do_years(timelib_sll year ) 
{ timelib_sll i ;
  timelib_sll res ;
  timelib_sll eras ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23358\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (timelib_sll )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23359\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  eras = (year - 1970LL) / 40000LL;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23360\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (eras != 0LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23331\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    year -= eras * 40000LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23332\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    res += (12622780800ll * eras) * 100LL;
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23333\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23361\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (year >= 1970LL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23344\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    i = year - 1LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23345\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23341\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (i >= 1970LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23334\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23342\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (i % 4LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23339\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (i % 100LL != 0LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23335\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          res += 31622400LL;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23338\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if (i % 400LL == 0LL) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23336\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            res += 31622400LL;
          } else {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23337\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            res += 31536000LL;
          }
        }
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23340\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        res += 31536000LL;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23343\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      i --;
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23356\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    i = 1969LL;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23357\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    while (1) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23353\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (i >= year) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23346\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }

      } else {
        break;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23354\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (i % 4LL == 0LL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23351\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        if (i % 100LL != 0LL) {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23347\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          res -= 31622400LL;
        } else {
          {
          _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                 "a");
          fprintf(_coverage_fout, "23350\n");
          fflush(_coverage_fout);
          fclose(_coverage_fout);
          }
          if (i % 400LL == 0LL) {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23348\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            res -= 31622400LL;
          } else {
            {
            _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                                   "a");
            fprintf(_coverage_fout, "23349\n");
            fflush(_coverage_fout);
            fclose(_coverage_fout);
            }
            res -= 31536000LL;
          }
        }
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23352\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        res -= 31536000LL;
      }
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23355\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      i --;
    }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23362\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (res);
}
}
static timelib_sll do_months(timelib_ull month , timelib_ull year ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23369\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (year % 4ULL == 0ULL) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23367\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (year % 100ULL != 0ULL) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23363\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      return ((long long )((month_tab_leap[month - 1ULL] + 1) * 86400));
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23366\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (year % 400ULL == 0ULL) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23364\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        return ((long long )((month_tab_leap[month - 1ULL] + 1) * 86400));
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23365\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        return ((long long )(month_tab[month - 1ULL] * 86400));
      }
    }
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23368\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    return ((long long )(month_tab[month - 1ULL] * 86400));
  }
}
}
static timelib_sll do_days(timelib_ull day ) 
{ 

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23370\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((long long )((day - 1ULL) * 86400ULL));
}
}
static timelib_sll do_time(timelib_ull hour , timelib_ull minute ,
                           timelib_ull second ) 
{ timelib_sll res ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23371\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (timelib_sll )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23372\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (long long )((timelib_ull )res + hour * 3600ULL);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23373\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (long long )((timelib_ull )res + minute * 60ULL);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23374\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (long long )((timelib_ull )res + second);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23375\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (res);
}
}
static timelib_sll do_adjust_timezone(timelib_time *tz , timelib_tzinfo *tzi ) 
{ timelib_sll tmp ;
  timelib_time_offset *before ;
  timelib_time_offset *after ;
  timelib_sll tmp___0 ;
  int in_transistion ;
  int tmp___1 ;
  timelib_time_offset *gmt_offset ;
  void *tmp___2 ;
  size_t __len ;
  size_t tmp___3 ;
  char *__retval ;
  void *tmp___4 ;
  void *tmp___5 ;
  char *tmp___6 ;
  char *tmp___7 ;

  {
  switch ((int )tz->zone_type) {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23404\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  case 1: 
  tz->is_localtime = 1U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23405\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return ((long long )(tz->z * 60));
  break;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23406\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  case 2: 
  tz->is_localtime = 1U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23407\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = (long long )tz->z;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23408\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp -= (timelib_sll )(tz->dst * 60);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23409\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp *= 60LL;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23410\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (tmp);
  break;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23411\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  case 3: 
  tzi = tz->tz_info;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23412\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  default: ;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23413\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  if (tzi) {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23386\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tz->is_localtime = 1U;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23387\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    before = timelib_get_time_zone_info(tz->sse, tzi);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23388\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    after = timelib_get_time_zone_info(tz->sse - (timelib_sll )before->offset,
                                       tzi);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23389\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    timelib_set_timezone(tz, tzi);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23390\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tz->sse - (timelib_sll )after->offset >= after->transistion_time + (timelib_sll )(before->offset - after->offset)) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23378\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (tz->sse - (timelib_sll )after->offset < after->transistion_time) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23376\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___1 = 1;
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23377\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___1 = 0;
      }
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23379\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp___1 = 0;
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23391\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    in_transistion = tmp___1;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23392\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (before->offset != after->offset) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23382\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      if (! in_transistion) {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23380\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___0 = (long long )(- after->offset);
      } else {
        {
        _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                               "a");
        fprintf(_coverage_fout, "23381\n");
        fflush(_coverage_fout);
        fclose(_coverage_fout);
        }
        tmp___0 = (long long )(- tz->z);
      }
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23383\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      tmp___0 = (long long )(- tz->z);
    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23393\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    timelib_time_offset_dtor(before);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23394\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    timelib_time_offset_dtor(after);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23395\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    gmt_offset = timelib_get_time_zone_info(tz->sse + tmp___0, tzi);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23396\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tz->z = gmt_offset->offset;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23397\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tz->dst = (int )gmt_offset->is_dst;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23398\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    if (tz->tz_abbr) {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23384\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }
      free((void *)tz->tz_abbr);
    } else {
      {
      _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                             "a");
      fprintf(_coverage_fout, "23385\n");
      fflush(_coverage_fout);
      fclose(_coverage_fout);
      }

    }
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23399\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tmp___7 = __strdup((char const   *)gmt_offset->abbr);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23400\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    tz->tz_abbr = tmp___7;
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23401\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    timelib_time_offset_dtor(gmt_offset);
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23402\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }
    return (tmp___0);
  } else {
    {
    _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                           "a");
    fprintf(_coverage_fout, "23403\n");
    fflush(_coverage_fout);
    fclose(_coverage_fout);
    }

  }
  }
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23414\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return (0LL);
}
}
void timelib_update_ts(timelib_time *time , timelib_tzinfo *tzi ) 
{ timelib_sll res ;
  timelib_sll tmp ;
  timelib_sll tmp___0 ;
  timelib_sll tmp___1 ;
  timelib_sll tmp___2 ;
  timelib_sll tmp___3 ;
  unsigned int tmp___4 ;
  unsigned int tmp___5 ;

  {
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23415\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res = (timelib_sll )0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23416\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_adjust_special_early(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23417\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_adjust_relative(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23418\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  do_adjust_special(time);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23419\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp = do_years(time->y);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23420\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res += tmp;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23421\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___0 = do_months((unsigned long long )time->m, (unsigned long long )time->y);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23422\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res += tmp___0;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23423\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___1 = do_days((unsigned long long )time->d);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23424\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res += tmp___1;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23425\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___2 = do_time((unsigned long long )time->h, (unsigned long long )time->i,
                    (unsigned long long )time->s);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23426\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res += tmp___2;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23427\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->sse = res;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23428\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___3 = do_adjust_timezone(time, tzi);
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23429\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  res += tmp___3;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23430\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->sse = res;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23431\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->sse_uptodate = 1U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23432\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___5 = 0U;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23433\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->relative.have_special_relative = tmp___5;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23434\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  tmp___4 = tmp___5;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23435\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->relative.have_weekday_relative = tmp___4;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23436\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  time->have_relative = tmp___4;
  {
  _coverage_fout = fopen("/root/mountpoint-genprog/genprog-many-bugs/php-bug-2011-12-06-5087b0ee42-ac631dd580/coverage/coverage.path",
                         "a");
  fprintf(_coverage_fout, "23437\n");
  fflush(_coverage_fout);
  fclose(_coverage_fout);
  }
  return;
}
}
